// A message containing doubles, as used by other applications.
syntax = "proto2";

message DoubleMessage {
    required double field1 = 1;
    required double field2 = 2;
}

// A message containing doubles, but redefined using uint64_t.
// For use in AVR code.
message AVRDoubleMessage {
    required fixed64 field1 = 1;
    required fixed64 field2 = 2;
}

