# Copyright (c) 2014, Google Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
# OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#if !defined(OPENSSL_NO_ASM) && defined(__arm__)

.syntax unified
.cpu cortex-a8
.fpu neon
.text
.thumb
.align 2
.global CRYPTO_arm_neon_probe
.hidden CRYPTO_arm_neon_probe
.type CRYPTO_arm_neon_probe, %function
.thumb_func
CRYPTO_arm_neon_probe:
  vorr q1, q1, q1
  bx lr
.section .note.GNU-stack,"",%progbits

#endif  /* !OPENSSL_NO_ASM && __arm__ */
