/* Generated by Cython 0.24.1 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "OPENSSL_NO_ASM",
                1
            ],
            [
                "_WIN32_WINNT",
                1536
            ],
            [
                "GPR_BACKWARDS_COMPATIBILITY_MODE",
                1
            ],
            [
                "PyMODINIT_FUNC",
                "__attribute__((visibility (\"default\"))) PyObject*"
            ]
        ],
        "depends": [
            "include/grpc/byte_buffer_reader.h",
            "include/grpc/compression.h",
            "include/grpc/grpc.h",
            "include/grpc/grpc_security.h",
            "include/grpc/support/alloc.h"
        ],
        "extra_compile_args": [
            "-fno-wrapv",
            "-fvisibility=hidden"
        ],
        "extra_link_args": [
            "-lpthread"
        ],
        "include_dirs": [
            "src/python/grpcio",
            "include",
            ".",
            "third_party/boringssl/include",
            "third_party/zlib"
        ],
        "libraries": [
            "m"
        ]
    },
    "module_name": "grpc._cython.cygrpc"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_24_1"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
  #define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__grpc___cython__cygrpc
#define __PYX_HAVE_API__grpc___cython__cygrpc
#include "stddef.h"
#include "time.h"
#include "grpc/support/alloc.h"
#include "grpc/byte_buffer_reader.h"
#include "grpc/grpc.h"
#include "grpc/grpc_security.h"
#include "grpc/compression.h"
#include "string.h"
#include "stdio.h"
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi",
  "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi",
  "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi",
  "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi",
  "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi",
  "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi",
  "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi",
  "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi",
  "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi",
  "src/python/grpcio/grpc/_cython/cygrpc.pyx",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/* "src/python/grpcio/grpc/_cython/_cygrpc/grpc.pxi":35
 * # Typedef types with approximately the same semantics to provide their names to
 * # Cython
 * ctypedef int int32_t             # <<<<<<<<<<<<<<
 * ctypedef unsigned uint32_t
 * ctypedef long int64_t
 */
typedef int __pyx_t_4grpc_7_cython_6cygrpc_int32_t;

/* "src/python/grpcio/grpc/_cython/_cygrpc/grpc.pxi":36
 * # Cython
 * ctypedef int int32_t
 * ctypedef unsigned uint32_t             # <<<<<<<<<<<<<<
 * ctypedef long int64_t
 * 
 */
typedef unsigned int __pyx_t_4grpc_7_cython_6cygrpc_uint32_t;

/* "src/python/grpcio/grpc/_cython/_cygrpc/grpc.pxi":37
 * ctypedef int int32_t
 * ctypedef unsigned uint32_t
 * ctypedef long int64_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef long __pyx_t_4grpc_7_cython_6cygrpc_int64_t;

/*--- Type declarations ---*/
struct __pyx_obj_4grpc_7_cython_6cygrpc_Call;
struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel;
struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials;
struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials;
struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials;
struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin;
struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext;
struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue;
struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec;
struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails;
struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag;
struct __pyx_obj_4grpc_7_cython_6cygrpc_Event;
struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer;
struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair;
struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg;
struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs;
struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum;
struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata;
struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation;
struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations;
struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions;
struct __pyx_obj_4grpc_7_cython_6cygrpc_Server;
struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator;
struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator;
struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata;

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pxd.pxi":31
 * 
 * 
 * cdef class Call:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_call *c_call
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Call {
  PyObject_HEAD
  grpc_call *c_call;
  PyObject *references;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pxd.pxi":31
 * 
 * 
 * cdef class Channel:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_channel *c_channel
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel {
  PyObject_HEAD
  grpc_channel *c_channel;
  PyObject *references;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pxd.pxi":33
 * 
 * 
 * cdef class ChannelCredentials:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_channel_credentials *c_credentials
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials {
  PyObject_HEAD
  grpc_channel_credentials *c_credentials;
  grpc_ssl_pem_key_cert_pair c_ssl_pem_key_cert_pair;
  PyObject *references;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pxd.pxi":40
 * 
 * 
 * cdef class CallCredentials:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_call_credentials *c_credentials
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials {
  PyObject_HEAD
  grpc_call_credentials *c_credentials;
  PyObject *references;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pxd.pxi":46
 * 
 * 
 * cdef class ServerCredentials:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_server_credentials *c_credentials
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials {
  PyObject_HEAD
  grpc_server_credentials *c_credentials;
  grpc_ssl_pem_key_cert_pair *c_ssl_pem_key_cert_pairs;
  size_t c_ssl_pem_key_cert_pairs_count;
  PyObject *references;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pxd.pxi":54
 * 
 * 
 * cdef class CredentialsMetadataPlugin:             # <<<<<<<<<<<<<<
 * 
 *   cdef object plugin_callback
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin {
  PyObject_HEAD
  struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_vtab;
  PyObject *plugin_callback;
  PyObject *plugin_name;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pxd.pxi":62
 * 
 * 
 * cdef class AuthMetadataContext:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_auth_metadata_context context
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext {
  PyObject_HEAD
  grpc_auth_metadata_context context;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pxd.pxi":31
 * 
 * 
 * cdef class CompletionQueue:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_completion_queue *c_completion_queue
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue {
  PyObject_HEAD
  struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_vtab;
  grpc_completion_queue *c_completion_queue;
  int is_shutting_down;
  int is_shutdown;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":31
 * 
 * 
 * cdef class Timespec:             # <<<<<<<<<<<<<<
 * 
 *   cdef gpr_timespec c_time
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec {
  PyObject_HEAD
  gpr_timespec c_time;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":36
 * 
 * 
 * cdef class CallDetails:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_call_details c_details
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails {
  PyObject_HEAD
  grpc_call_details c_details;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":41
 * 
 * 
 * cdef class OperationTag:             # <<<<<<<<<<<<<<
 * 
 *   cdef object user_tag
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag {
  PyObject_HEAD
  PyObject *user_tag;
  PyObject *references;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *shutting_down_server;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *operation_call;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *request_call_details;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *request_metadata;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *batch_operations;
  int is_new_request;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":55
 * 
 * 
 * cdef class Event:             # <<<<<<<<<<<<<<
 * 
 *   cdef readonly grpc_completion_type type
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Event {
  PyObject_HEAD
  grpc_completion_type type;
  int success;
  PyObject *tag;
  int is_new_request;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *request_call_details;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *request_metadata;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *operation_call;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *batch_operations;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":73
 * 
 * 
 * cdef class ByteBuffer:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_byte_buffer *c_byte_buffer
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer {
  PyObject_HEAD
  grpc_byte_buffer *c_byte_buffer;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":78
 * 
 * 
 * cdef class SslPemKeyCertPair:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_ssl_pem_key_cert_pair c_pair
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair {
  PyObject_HEAD
  grpc_ssl_pem_key_cert_pair c_pair;
  PyObject *private_key;
  PyObject *certificate_chain;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":84
 * 
 * 
 * cdef class ChannelArg:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_arg c_arg
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg {
  PyObject_HEAD
  grpc_arg c_arg;
  PyObject *key;
  PyObject *value;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":90
 * 
 * 
 * cdef class ChannelArgs:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_channel_args c_args
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs {
  PyObject_HEAD
  grpc_channel_args c_args;
  PyObject *args;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":96
 * 
 * 
 * cdef class Metadatum:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_metadata c_metadata
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum {
  PyObject_HEAD
  grpc_metadata c_metadata;
  PyObject *_key;
  PyObject *_value;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":102
 * 
 * 
 * cdef class Metadata:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_metadata_array c_metadata_array
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata {
  PyObject_HEAD
  grpc_metadata_array c_metadata_array;
  PyObject *metadata;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":108
 * 
 * 
 * cdef class Operation:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_op c_op
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation {
  PyObject_HEAD
  grpc_op c_op;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *_received_message;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *_received_metadata;
  grpc_status_code _received_status_code;
  char *_received_status_details;
  size_t _received_status_details_capacity;
  int _received_cancelled;
  int is_valid;
  PyObject *references;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":121
 * 
 * 
 * cdef class Operations:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_op *c_ops
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations {
  PyObject_HEAD
  grpc_op *c_ops;
  size_t c_nops;
  PyObject *operations;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":128
 * 
 * 
 * cdef class CompressionOptions:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_compression_options c_options
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions {
  PyObject_HEAD
  grpc_compression_options c_options;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pxd.pxi":31
 * 
 * 
 * cdef class Server:             # <<<<<<<<<<<<<<
 * 
 *   cdef grpc_server *c_server
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc_Server {
  PyObject_HEAD
  struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_Server *__pyx_vtab;
  grpc_server *c_server;
  int is_started;
  int is_shutting_down;
  int is_shutdown;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *backup_shutdown_queue;
  PyObject *references;
  PyObject *registered_completion_queues;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":384
 * 
 * 
 * cdef class _MetadataIterator:             # <<<<<<<<<<<<<<
 * 
 *   cdef size_t i
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator {
  PyObject_HEAD
  size_t i;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *metadata;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":634
 * 
 * 
 * cdef class _OperationsIterator:             # <<<<<<<<<<<<<<
 * 
 *   cdef size_t i
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator {
  PyObject_HEAD
  size_t i;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *operations;
};


/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":138
 * 
 * 
 * cdef void plugin_get_metadata(             # <<<<<<<<<<<<<<
 *     void *state, grpc_auth_metadata_context context,
 *     grpc_credentials_plugin_metadata_cb cb, void *user_data) with gil:
 */
struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata {
  PyObject_HEAD
  grpc_credentials_plugin_metadata_cb __pyx_v_cb;
  void *__pyx_v_user_data;
};



/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":86
 * 
 * 
 * cdef class CredentialsMetadataPlugin:             # <<<<<<<<<<<<<<
 * 
 *   def __cinit__(self, object plugin_callback, bytes name):
 */

struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin {
  grpc_metadata_credentials_plugin (*make_c_plugin)(struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *);
};
static struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_vtabptr_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin;


/* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":38
 * 
 * 
 * cdef class CompletionQueue:             # <<<<<<<<<<<<<<
 * 
 *   def __cinit__(self):
 */

struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CompletionQueue {
  PyObject *(*_interpret_event)(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *, grpc_event);
};
static struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_vtabptr_4grpc_7_cython_6cygrpc_CompletionQueue;


/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":35
 * 
 * 
 * cdef class Server:             # <<<<<<<<<<<<<<
 * 
 *   def __cinit__(self, ChannelArgs arguments=None):
 */

struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_Server {
  PyObject *(*_c_shutdown)(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *, PyObject *);
  PyObject *(*notify_shutdown_complete)(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *);
};
static struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_Server *__pyx_vtabptr_4grpc_7_cython_6cygrpc_Server;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ListAppend.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* CallableCheck.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#define __Pyx_PyCallable_Check(obj)   ((obj)->ob_type->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* IncludeStringH.proto */
#include <string.h>

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* GetNameInClass.proto */
static PyObject *__Pyx_GetNameInClass(PyObject *nmspace, PyObject *name);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_completion_type(grpc_completion_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_connectivity_state(grpc_connectivity_state value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_status_code(grpc_status_code value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_call_error(grpc_call_error value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_op_type(grpc_op_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_compression_algorithm(grpc_compression_algorithm value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_compression_level(grpc_compression_level value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntFromPy.proto */
static CYTHON_INLINE grpc_status_code __Pyx_PyInt_As_grpc_status_code(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE grpc_connectivity_state __Pyx_PyInt_As_grpc_connectivity_state(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE grpc_completion_type __Pyx_PyInt_As_grpc_completion_type(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE grpc_compression_algorithm __Pyx_PyInt_As_grpc_compression_algorithm(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static grpc_metadata_credentials_plugin __pyx_f_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_make_c_plugin(struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_4grpc_7_cython_6cygrpc_15CompletionQueue__interpret_event(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self, grpc_event __pyx_v_event); /* proto*/
static PyObject *__pyx_f_4grpc_7_cython_6cygrpc_6Server__c_shutdown(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue, PyObject *__pyx_v_tag); /* proto*/
static PyObject *__pyx_f_4grpc_7_cython_6cygrpc_6Server_notify_shutdown_complete(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self); /* proto*/

/* Module declarations from 'libc.stddef' */

/* Module declarations from 'libc.time' */

/* Module declarations from 'libc' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'grpc._cython.cygrpc' */
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Call = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Channel = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_ChannelCredentials = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_ServerCredentials = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_AuthMetadataContext = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_CallDetails = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_OperationTag = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Event = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_ByteBuffer = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_SslPemKeyCertPair = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_ChannelArg = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_ChannelArgs = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Metadatum = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Metadata = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Operation = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Operations = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_CompressionOptions = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc_Server = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc__MetadataIterator = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc__OperationsIterator = 0;
static PyTypeObject *__pyx_ptype_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata = 0;
static int __pyx_v_4grpc_7_cython_6cygrpc__INTERRUPT_CHECK_PERIOD_MS;
static void __pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata(void *, grpc_auth_metadata_context, grpc_credentials_plugin_metadata_cb, void *); /*proto*/
static void __pyx_f_4grpc_7_cython_6cygrpc_plugin_destroy_c_plugin_state(void *); /*proto*/
static grpc_ssl_roots_override_result __pyx_f_4grpc_7_cython_6cygrpc_ssl_roots_override_callback(char **); /*proto*/
static PyObject *__pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "grpc._cython.cygrpc"
int __pyx_module_is_main_grpc___cython__cygrpc = 0;

/* Implementation of 'grpc._cython.cygrpc' */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_RuntimeError;
static const char __pyx_k_i[] = "i";
static const char __pyx_k_ok[] = "ok";
static const char __pyx_k_op[] = "op";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_inf[] = "+inf";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_low[] = "low";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_tag[] = "tag";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_code[] = "code";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_gzip[] = "gzip";
static const char __pyx_k_high[] = "high";
static const char __pyx_k_host[] = "host";
static const char __pyx_k_idle[] = "idle";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_none[] = "none";
static const char __pyx_k_pair[] = "pair";
static const char __pyx_k_poll[] = "poll";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_inf_2[] = "-inf";
static const char __pyx_k_queue[] = "queue";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ready[] = "ready";
static const char __pyx_k_sleep[] = "sleep";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_medium[] = "medium";
static const char __pyx_k_method[] = "method";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_plugin[] = "plugin";
static const char __pyx_k_status[] = "status";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_aborted[] = "aborted";
static const char __pyx_k_address[] = "address";
static const char __pyx_k_deflate[] = "deflate";
static const char __pyx_k_details[] = "details";
static const char __pyx_k_os_path[] = "os.path";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_success[] = "success";
static const char __pyx_k_unknown[] = "unknown";
static const char __pyx_k_c_plugin[] = "c_plugin";
static const char __pyx_k_deadline[] = "deadline";
static const char __pyx_k_internal[] = "internal";
static const char __pyx_k_is_valid[] = "is_valid";
static const char __pyx_k_json_key[] = "json_key";
static const char __pyx_k_metadata[] = "metadata";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_shutdown[] = "shutdown";
static const char __pyx_k_user_tag[] = "user_tag";
static const char __pyx_k_CallError[] = "CallError";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_WriteFlag[] = "WriteFlag";
static const char __pyx_k_algorithm[] = "algorithm";
static const char __pyx_k_arguments[] = "arguments";
static const char __pyx_k_cancelled[] = "cancelled";
static const char __pyx_k_data_loss[] = "data_loss";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_not_found[] = "not_found";
static const char __pyx_k_threading[] = "threading";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_StatusCode[] = "StatusCode";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_call_queue[] = "call_queue";
static const char __pyx_k_connecting[] = "connecting";
static const char __pyx_k_error_code[] = "error_code";
static const char __pyx_k_initialize[] = "_initialize";
static const char __pyx_k_operations[] = "operations";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_buffer_hint[] = "buffer_hint";
static const char __pyx_k_byte_buffer[] = "byte_buffer";
static const char __pyx_k_credentials[] = "credentials";
static const char __pyx_k_no_compress[] = "no_compress";
static const char __pyx_k_not_invoked[] = "not_invoked";
static const char __pyx_k_private_key[] = "private_key";
static const char __pyx_k_retain_self[] = "retain_self";
static const char __pyx_k_start_batch[] = "_start_batch";
static const char __pyx_k_unavailable[] = "unavailable";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_grpc__cython[] = "grpc._cython";
static const char __pyx_k_out_of_range[] = "out_of_range";
static const char __pyx_k_send_message[] = "send_message";
static const char __pyx_k_server_queue[] = "server_queue";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_ChannelArgKey[] = "ChannelArgKey";
static const char __pyx_k_OperationType[] = "OperationType";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_credentials_1[] = "credentials_1";
static const char __pyx_k_credentials_2[] = "credentials_2";
static const char __pyx_k_enable_census[] = "enable_census";
static const char __pyx_k_error_details[] = "error_details";
static const char __pyx_k_infinite_past[] = "infinite_past";
static const char __pyx_k_invalid_flags[] = "invalid_flags";
static const char __pyx_k_not_on_client[] = "not_on_client";
static const char __pyx_k_not_on_server[] = "not_on_server";
static const char __pyx_k_pkg_resources[] = "pkg_resources";
static const char __pyx_k_queue_timeout[] = "queue_timeout";
static const char __pyx_k_unimplemented[] = "unimplemented";
static const char __pyx_k_CompletionType[] = "CompletionType";
static const char __pyx_k_already_exists[] = "already_exists";
static const char __pyx_k_is_new_request[] = "is_new_request";
static const char __pyx_k_operation_call[] = "operation_call";
static const char __pyx_k_pem_root_certs[] = "pem_root_certs";
static const char __pyx_k_queue_shutdown[] = "queue_shutdown";
static const char __pyx_k_token_lifetime[] = "token_lifetime";
static const char __pyx_k_already_invoked[] = "already_invoked";
static const char __pyx_k_infinite_future[] = "infinite_future";
static const char __pyx_k_plugin_callback[] = "plugin_callback";
static const char __pyx_k_python_callback[] = "python_callback";
static const char __pyx_k_receive_message[] = "receive_message";
static const char __pyx_k_resource_string[] = "resource_string";
static const char __pyx_k_unauthenticated[] = "unauthenticated";
static const char __pyx_k_CompressionLevel[] = "CompressionLevel";
static const char __pyx_k_already_accepted[] = "already_accepted";
static const char __pyx_k_already_finished[] = "already_finished";
static const char __pyx_k_batch_operations[] = "batch_operations";
static const char __pyx_k_c_pem_root_certs[] = "c_pem_root_certs";
static const char __pyx_k_invalid_argument[] = "invalid_argument";
static const char __pyx_k_invalid_metadata[] = "invalid_metadata";
static const char __pyx_k_request_metadata[] = "request_metadata";
static const char __pyx_k_ConnectivityState[] = "ConnectivityState";
static const char __pyx_k_certificate_chain[] = "certificate_chain";
static const char __pyx_k_deadline_exceeded[] = "deadline_exceeded";
static const char __pyx_k_default_authority[] = "default_authority";
static const char __pyx_k_force_client_auth[] = "force_client_auth";
static const char __pyx_k_json_key_c_string[] = "json_key_c_string";
static const char __pyx_k_permission_denied[] = "permission_denied";
static const char __pyx_k_transient_failure[] = "transient_failure";
static const char __pyx_k_authority_selector[] = "authority_selector";
static const char __pyx_k_json_refresh_token[] = "json_refresh_token";
static const char __pyx_k_max_message_length[] = "max_message_length";
static const char __pyx_k_operation_complete[] = "operation_complete";
static const char __pyx_k_pem_key_cert_pairs[] = "pem_key_cert_pairs";
static const char __pyx_k_queue_must_be_live[] = "queue must be live";
static const char __pyx_k_resource_exhausted[] = "resource_exhausted";
static const char __pyx_k_server_credentials[] = "server_credentials";
static const char __pyx_k_authorization_token[] = "authorization_token";
static const char __pyx_k_channel_credentials[] = "channel_credentials";
static const char __pyx_k_failed_precondition[] = "failed_precondition";
static const char __pyx_k_grpc__cython_cygrpc[] = "grpc._cython.cygrpc";
static const char __pyx_k_last_observed_state[] = "last_observed_state";
static const char __pyx_k_too_many_operations[] = "too_many_operations";
static const char __pyx_k_CompressionAlgorithm[] = "CompressionAlgorithm";
static const char __pyx_k_request_call_details[] = "request_call_details";
static const char __pyx_k_credentials_roots_pem[] = "_credentials/roots.pem";
static const char __pyx_k_pem_root_certificates[] = "pem_root_certificates";
static const char __pyx_k_send_initial_metadata[] = "send_initial_metadata";
static const char __pyx_k_ssl_pem_key_cert_pair[] = "ssl_pem_key_cert_pair";
static const char __pyx_k_max_concurrent_streams[] = "max_concurrent_streams";
static const char __pyx_k_operation_send_message[] = "operation_send_message";
static const char __pyx_k_send_close_from_client[] = "send_close_from_client";
static const char __pyx_k_server_credentials_ssl[] = "server_credentials_ssl";
static const char __pyx_k_c_pem_root_certificates[] = "c_pem_root_certificates";
static const char __pyx_k_channel_credentials_ssl[] = "channel_credentials_ssl";
static const char __pyx_k_receive_close_on_server[] = "receive_close_on_server";
static const char __pyx_k_send_status_from_server[] = "send_status_from_server";
static const char __pyx_k_receive_initial_metadata[] = "receive_initial_metadata";
static const char __pyx_k_receive_status_on_client[] = "receive_status_on_client";
static const char __pyx_k_ssl_target_name_override[] = "ssl_target_name_override";
static const char __pyx_k_Expected_int_or_bytes_got[] = "Expected int or bytes, got {}";
static const char __pyx_k_operation_receive_message[] = "operation_receive_message";
static const char __pyx_k_primary_user_agent_string[] = "primary_user_agent_string";
static const char __pyx_k_call_credentials_composite[] = "call_credentials_composite";
static const char __pyx_k_compression_algorithm_name[] = "compression_algorithm_name";
static const char __pyx_k_expected_list_of_Metadatum[] = "expected list of Metadatum";
static const char __pyx_k_authority_selector_c_string[] = "authority_selector_c_string";
static const char __pyx_k_call_credentials_google_iam[] = "call_credentials_google_iam";
static const char __pyx_k_expected_list_of_ChannelArg[] = "expected list of ChannelArg";
static const char __pyx_k_json_refresh_token_c_string[] = "json_refresh_token_c_string";
static const char __pyx_k_secondary_user_agent_string[] = "secondary_user_agent_string";
static const char __pyx_k_authorization_token_c_string[] = "authorization_token_c_string";
static const char __pyx_k_channel_credentials_composite[] = "channel_credentials_composite";
static const char __pyx_k_http2_initial_sequence_number[] = "http2_initial_sequence_number";
static const char __pyx_k_the_server_hasn_t_started_yet[] = "the server hasn't started yet";
static const char __pyx_k_index_must_be_0_key_or_1_value[] = "index must be 0 (key) or 1 (value)";
static const char __pyx_k_src_python_grpcio_grpc__cython[] = "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi";
static const char __pyx_k_the_server_has_already_started[] = "the server has already started";
static const char __pyx_k_call_credentials_google_compute[] = "call_credentials_google_compute_engine";
static const char __pyx_k_call_credentials_google_refresh[] = "call_credentials_google_refresh_token";
static const char __pyx_k_operation_send_initial_metadata[] = "operation_send_initial_metadata";
static const char __pyx_k_passed_credentials_must_both_be[] = "passed credentials must both be valid";
static const char __pyx_k_private_var_folders_g4_3ppgfd8x[] = "/private/var/folders/g4/3ppgfd8x4274zrnf_tf60_8h0000gn/T/pygrpc.10KrrE/src/python/grpcio/grpc/_cython/cygrpc.pyx";
static const char __pyx_k_queue_must_not_be_shutting_down[] = "queue must not be shutting down or shutdown";
static const char __pyx_k_Expected_bytes_str_or_unicode_no[] = "Expected bytes, str, or unicode, not {}";
static const char __pyx_k_call_credentials_metadata_plugin[] = "call_credentials_metadata_plugin";
static const char __pyx_k_call_credentials_service_account[] = "call_credentials_service_account_jwt_access";
static const char __pyx_k_cannot_register_completion_queue[] = "cannot register completion queues after start";
static const char __pyx_k_channel_credentials_google_defau[] = "channel_credentials_google_default";
static const char __pyx_k_expected_callable_plugin_callbac[] = "expected callable plugin_callback";
static const char __pyx_k_expected_operations_to_be_iterab[] = "expected operations to be iterable of Operation";
static const char __pyx_k_expected_pem_key_cert_pairs_to_b[] = "expected pem_key_cert_pairs to be sequence of SslPemKeyCertPair";
static const char __pyx_k_expected_registered_completion_q[] = "expected registered completion queue";
static const char __pyx_k_expected_time_to_be_float_int_or[] = "expected time to be float, int, or Timespec, not {}";
static const char __pyx_k_if_error_code_is_specified_so_mu[] = "if error_code is specified, so must details (and vice-versa)";
static const char __pyx_k_invalid_call_object_cannot_be_us[] = "invalid call object cannot be used from Python";
static const char __pyx_k_operation_receive_close_on_serve[] = "operation_receive_close_on_server";
static const char __pyx_k_operation_receive_initial_metada[] = "operation_receive_initial_metadata";
static const char __pyx_k_operation_receive_status_on_clie[] = "operation_receive_status_on_client";
static const char __pyx_k_operation_send_close_from_client[] = "operation_send_close_from_client";
static const char __pyx_k_operation_send_status_from_serve[] = "operation_send_status_from_server";
static const char __pyx_k_plugin_get_metadata_locals_pytho[] = "plugin_get_metadata.<locals>.python_callback";
static const char __pyx_k_queue_must_be_shutting_down_to_b[] = "queue must be shutting down to be cleared";
static const char __pyx_k_register_non_listening_completio[] = "register_non_listening_completion_queue";
static const char __pyx_k_self_must_be_an_operation_receiv[] = "self must be an operation receiving a message";
static const char __pyx_k_server_must_be_started_and_not_s[] = "server must be started and not shutting down";
static const char __pyx_k_server_queue_must_be_a_registere[] = "server_queue must be a registered completion queue";
static const char __pyx_k_src_python_grpcio_grpc__cython_2[] = "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi";
static const char __pyx_k_the_server_must_be_shutting_down[] = "the server must be shutting down to cancel all calls";
static const char __pyx_k_self_must_be_an_operation_receiv_2[] = "self must be an operation receiving metadata";
static const char __pyx_k_self_must_be_an_operation_receiv_3[] = "self must be an operation receiving a status code";
static const char __pyx_k_self_must_be_an_operation_receiv_4[] = "self must be an operation receiving status details";
static const char __pyx_k_self_must_be_an_operation_receiv_5[] = "self must be an operation receiving cancellation information";
static PyObject *__pyx_n_s_CallError;
static PyObject *__pyx_n_s_ChannelArgKey;
static PyObject *__pyx_n_s_CompletionType;
static PyObject *__pyx_n_s_CompressionAlgorithm;
static PyObject *__pyx_n_s_CompressionLevel;
static PyObject *__pyx_n_s_ConnectivityState;
static PyObject *__pyx_kp_s_Expected_bytes_str_or_unicode_no;
static PyObject *__pyx_kp_s_Expected_int_or_bytes_got;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_OperationType;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_StatusCode;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_WriteFlag;
static PyObject *__pyx_n_s_aborted;
static PyObject *__pyx_n_s_address;
static PyObject *__pyx_n_s_algorithm;
static PyObject *__pyx_n_s_already_accepted;
static PyObject *__pyx_n_s_already_exists;
static PyObject *__pyx_n_s_already_finished;
static PyObject *__pyx_n_s_already_invoked;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_arguments;
static PyObject *__pyx_n_s_ascii;
static PyObject *__pyx_n_s_authority_selector;
static PyObject *__pyx_n_s_authority_selector_c_string;
static PyObject *__pyx_n_s_authorization_token;
static PyObject *__pyx_n_s_authorization_token_c_string;
static PyObject *__pyx_n_s_batch_operations;
static PyObject *__pyx_n_s_buffer_hint;
static PyObject *__pyx_n_s_byte_buffer;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_n_s_c_pem_root_certificates;
static PyObject *__pyx_n_s_c_pem_root_certs;
static PyObject *__pyx_n_s_c_plugin;
static PyObject *__pyx_n_s_call_credentials_composite;
static PyObject *__pyx_n_s_call_credentials_google_compute;
static PyObject *__pyx_n_s_call_credentials_google_iam;
static PyObject *__pyx_n_s_call_credentials_google_refresh;
static PyObject *__pyx_n_s_call_credentials_metadata_plugin;
static PyObject *__pyx_n_s_call_credentials_service_account;
static PyObject *__pyx_n_s_call_queue;
static PyObject *__pyx_n_s_cancelled;
static PyObject *__pyx_kp_s_cannot_register_completion_queue;
static PyObject *__pyx_n_s_certificate_chain;
static PyObject *__pyx_n_s_channel_credentials;
static PyObject *__pyx_n_s_channel_credentials_composite;
static PyObject *__pyx_n_s_channel_credentials_google_defau;
static PyObject *__pyx_n_s_channel_credentials_ssl;
static PyObject *__pyx_n_s_code;
static PyObject *__pyx_n_s_compression_algorithm_name;
static PyObject *__pyx_n_s_connecting;
static PyObject *__pyx_n_s_credentials;
static PyObject *__pyx_n_s_credentials_1;
static PyObject *__pyx_n_s_credentials_2;
static PyObject *__pyx_kp_s_credentials_roots_pem;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_data_loss;
static PyObject *__pyx_n_s_deadline;
static PyObject *__pyx_n_s_deadline_exceeded;
static PyObject *__pyx_n_s_default_authority;
static PyObject *__pyx_n_s_deflate;
static PyObject *__pyx_n_s_details;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_enable_census;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_error_code;
static PyObject *__pyx_n_s_error_details;
static PyObject *__pyx_kp_s_expected_callable_plugin_callbac;
static PyObject *__pyx_kp_s_expected_list_of_ChannelArg;
static PyObject *__pyx_kp_s_expected_list_of_Metadatum;
static PyObject *__pyx_kp_s_expected_operations_to_be_iterab;
static PyObject *__pyx_kp_s_expected_pem_key_cert_pairs_to_b;
static PyObject *__pyx_kp_s_expected_registered_completion_q;
static PyObject *__pyx_kp_s_expected_time_to_be_float_int_or;
static PyObject *__pyx_n_s_failed_precondition;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_force_client_auth;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_kp_s_grpc__cython;
static PyObject *__pyx_n_s_grpc__cython_cygrpc;
static PyObject *__pyx_n_s_gzip;
static PyObject *__pyx_n_s_high;
static PyObject *__pyx_n_s_host;
static PyObject *__pyx_n_s_http2_initial_sequence_number;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_idle;
static PyObject *__pyx_kp_s_if_error_code_is_specified_so_mu;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_kp_s_index_must_be_0_key_or_1_value;
static PyObject *__pyx_kp_s_inf;
static PyObject *__pyx_kp_s_inf_2;
static PyObject *__pyx_n_s_infinite_future;
static PyObject *__pyx_n_s_infinite_past;
static PyObject *__pyx_n_s_initialize;
static PyObject *__pyx_n_s_internal;
static PyObject *__pyx_n_s_invalid_argument;
static PyObject *__pyx_kp_s_invalid_call_object_cannot_be_us;
static PyObject *__pyx_n_s_invalid_flags;
static PyObject *__pyx_n_s_invalid_metadata;
static PyObject *__pyx_n_s_is_new_request;
static PyObject *__pyx_n_s_is_valid;
static PyObject *__pyx_n_s_json_key;
static PyObject *__pyx_n_s_json_key_c_string;
static PyObject *__pyx_n_s_json_refresh_token;
static PyObject *__pyx_n_s_json_refresh_token_c_string;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_last_observed_state;
static PyObject *__pyx_n_s_low;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max_concurrent_streams;
static PyObject *__pyx_n_s_max_message_length;
static PyObject *__pyx_n_s_medium;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_metadata;
static PyObject *__pyx_n_s_method;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_no_compress;
static PyObject *__pyx_n_s_none;
static PyObject *__pyx_n_s_not_found;
static PyObject *__pyx_n_s_not_invoked;
static PyObject *__pyx_n_s_not_on_client;
static PyObject *__pyx_n_s_not_on_server;
static PyObject *__pyx_n_s_ok;
static PyObject *__pyx_n_s_op;
static PyObject *__pyx_n_s_operation_call;
static PyObject *__pyx_n_s_operation_complete;
static PyObject *__pyx_n_s_operation_receive_close_on_serve;
static PyObject *__pyx_n_s_operation_receive_initial_metada;
static PyObject *__pyx_n_s_operation_receive_message;
static PyObject *__pyx_n_s_operation_receive_status_on_clie;
static PyObject *__pyx_n_s_operation_send_close_from_client;
static PyObject *__pyx_n_s_operation_send_initial_metadata;
static PyObject *__pyx_n_s_operation_send_message;
static PyObject *__pyx_n_s_operation_send_status_from_serve;
static PyObject *__pyx_n_s_operations;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_os_path;
static PyObject *__pyx_n_s_out_of_range;
static PyObject *__pyx_n_s_pair;
static PyObject *__pyx_n_s_parent;
static PyObject *__pyx_kp_s_passed_credentials_must_both_be;
static PyObject *__pyx_n_s_pem_key_cert_pairs;
static PyObject *__pyx_n_s_pem_root_certificates;
static PyObject *__pyx_n_s_pem_root_certs;
static PyObject *__pyx_n_s_permission_denied;
static PyObject *__pyx_n_s_pkg_resources;
static PyObject *__pyx_n_s_plugin;
static PyObject *__pyx_n_s_plugin_callback;
static PyObject *__pyx_n_s_plugin_get_metadata_locals_pytho;
static PyObject *__pyx_n_s_poll;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_primary_user_agent_string;
static PyObject *__pyx_n_s_private_key;
static PyObject *__pyx_kp_s_private_var_folders_g4_3ppgfd8x;
static PyObject *__pyx_n_s_python_callback;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_queue;
static PyObject *__pyx_kp_s_queue_must_be_live;
static PyObject *__pyx_kp_s_queue_must_be_shutting_down_to_b;
static PyObject *__pyx_kp_s_queue_must_not_be_shutting_down;
static PyObject *__pyx_n_s_queue_shutdown;
static PyObject *__pyx_n_s_queue_timeout;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_ready;
static PyObject *__pyx_n_s_receive_close_on_server;
static PyObject *__pyx_n_s_receive_initial_metadata;
static PyObject *__pyx_n_s_receive_message;
static PyObject *__pyx_n_s_receive_status_on_client;
static PyObject *__pyx_n_s_register_non_listening_completio;
static PyObject *__pyx_n_s_request_call_details;
static PyObject *__pyx_n_s_request_metadata;
static PyObject *__pyx_n_s_resource_exhausted;
static PyObject *__pyx_n_s_resource_string;
static PyObject *__pyx_n_s_retain_self;
static PyObject *__pyx_n_s_secondary_user_agent_string;
static PyObject *__pyx_kp_s_self_must_be_an_operation_receiv;
static PyObject *__pyx_kp_s_self_must_be_an_operation_receiv_2;
static PyObject *__pyx_kp_s_self_must_be_an_operation_receiv_3;
static PyObject *__pyx_kp_s_self_must_be_an_operation_receiv_4;
static PyObject *__pyx_kp_s_self_must_be_an_operation_receiv_5;
static PyObject *__pyx_n_s_send_close_from_client;
static PyObject *__pyx_n_s_send_initial_metadata;
static PyObject *__pyx_n_s_send_message;
static PyObject *__pyx_n_s_send_status_from_server;
static PyObject *__pyx_n_s_server_credentials;
static PyObject *__pyx_n_s_server_credentials_ssl;
static PyObject *__pyx_kp_s_server_must_be_started_and_not_s;
static PyObject *__pyx_n_s_server_queue;
static PyObject *__pyx_kp_s_server_queue_must_be_a_registere;
static PyObject *__pyx_n_s_shutdown;
static PyObject *__pyx_n_s_sleep;
static PyObject *__pyx_kp_s_src_python_grpcio_grpc__cython;
static PyObject *__pyx_kp_s_src_python_grpcio_grpc__cython_2;
static PyObject *__pyx_n_s_ssl_pem_key_cert_pair;
static PyObject *__pyx_n_s_ssl_target_name_override;
static PyObject *__pyx_n_s_start_batch;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_status;
static PyObject *__pyx_n_s_success;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_tag;
static PyObject *__pyx_n_s_target;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_the_server_has_already_started;
static PyObject *__pyx_kp_s_the_server_hasn_t_started_yet;
static PyObject *__pyx_kp_s_the_server_must_be_shutting_down;
static PyObject *__pyx_n_s_threading;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_token_lifetime;
static PyObject *__pyx_n_s_too_many_operations;
static PyObject *__pyx_n_s_transient_failure;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_unauthenticated;
static PyObject *__pyx_n_s_unavailable;
static PyObject *__pyx_n_s_unimplemented;
static PyObject *__pyx_n_s_unknown;
static PyObject *__pyx_n_s_user_tag;
static PyObject *__pyx_n_s_value;
static int __pyx_pf_4grpc_7_cython_6cygrpc_4Call___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_2_start_batch(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, PyObject *__pyx_v_operations, PyObject *__pyx_v_tag, PyObject *__pyx_v_retain_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_4start_client_batch(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, PyObject *__pyx_v_operations, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_6start_server_batch(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, PyObject *__pyx_v_operations, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_8cancel(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, grpc_status_code __pyx_v_error_code, PyObject *__pyx_v_details); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_10set_credentials(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_call_credentials); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_12peer(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_4Call_14__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_8is_valid___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_7Channel___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self, PyObject *__pyx_v_target, struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_arguments, struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_channel_credentials); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_7Channel_2create_call(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_parent, int __pyx_v_flags, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue, PyObject *__pyx_v_method, PyObject *__pyx_v_host, struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_7Channel_4check_connectivity_state(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self, int __pyx_v_try_to_connect); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_7Channel_6watch_connectivity_state(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self, grpc_connectivity_state __pyx_v_last_observed_state, struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_7Channel_8target(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_7Channel_10__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials_8is_valid___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials_8is_valid___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_17ServerCredentials___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_17ServerCredentials_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_self, PyObject *__pyx_v_plugin_callback, PyObject *__pyx_v_name); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11service_url___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11method_name___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19plugin_get_metadata_python_callback(PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata, grpc_status_code __pyx_v_status, PyObject *__pyx_v_error_details); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_channel_credentials_google_default(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_2channel_credentials_ssl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pem_root_certificates, struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_ssl_pem_key_cert_pair); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4channel_credentials_composite(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_credentials_1, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_2); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6call_credentials_composite(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_1, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_2); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8call_credentials_google_compute_engine(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10call_credentials_service_account_jwt_access(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_json_key, struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_token_lifetime); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_12call_credentials_google_refresh_token(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_json_refresh_token); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_14call_credentials_google_iam(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_authorization_token, PyObject *__pyx_v_authority_selector); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_16call_credentials_metadata_plugin(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_plugin); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18server_credentials_ssl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pem_root_certs, PyObject *__pyx_v_pem_key_cert_pairs, int __pyx_v_force_client_auth); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_2poll(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_4shutdown(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_6clear(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_8__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_8Timespec___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_self, PyObject *__pyx_v_time); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_7seconds___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_11nanoseconds___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_2__float__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_4infinite_future(); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_6infinite_past(); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_6method___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_4host___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_8deadline___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_12OperationTag___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *__pyx_v_self, PyObject *__pyx_v_user_tag); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_5Event___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self, grpc_completion_type __pyx_v_type, int __pyx_v_success, PyObject *__pyx_v_tag, struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_operation_call, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_request_call_details, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_request_metadata, int __pyx_v_is_new_request, struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_batch_operations); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_4type___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_7success___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_3tag___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_14is_new_request___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_20request_call_details___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_16request_metadata___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_14operation_call___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_16batch_operations___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_2bytes(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_4__len__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_6__str__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_8__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_self, PyObject *__pyx_v_private_key, PyObject *__pyx_v_certificate_chain); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_11private_key___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_17certificate_chain___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg_3key___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg_5value___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_self, PyObject *__pyx_v_args); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_4__len__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_6__getitem__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_self, size_t __pyx_v_i); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_3key___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_5value___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_2__len__(CYTHON_UNUSED struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_4__getitem__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self, size_t __pyx_v_i); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_6__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator_2__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator_4__next__(struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self, PyObject *__pyx_v_metadata); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_4__len__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_6__getitem__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self, size_t __pyx_v_i); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_8__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_9Operation___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_4type___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_5flags___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_10has_status___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_16received_message___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_24received_message_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_17received_metadata___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_25received_metadata_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_20received_status_code___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_28received_status_code_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_23received_status_details___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_31received_status_details_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_18received_cancelled___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_26received_cancelled_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_8is_valid___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_20operation_send_initial_metadata(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_22operation_send_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_24operation_send_close_from_client(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_26operation_send_status_from_server(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata, grpc_status_code __pyx_v_code, PyObject *__pyx_v_details, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_28operation_receive_initial_metadata(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_30operation_receive_message(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_32operation_receive_status_on_client(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_34operation_receive_close_on_server(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_operations); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator_2__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator_4__next__(struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_10Operations___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self, PyObject *__pyx_v_operations); /* proto */
static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_10Operations_2__len__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10Operations_4__getitem__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self, size_t __pyx_v_i); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_10Operations_6__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10Operations_8__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_2enable_algorithm(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self, grpc_compression_algorithm __pyx_v_algorithm); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_4disable_algorithm(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self, grpc_compression_algorithm __pyx_v_algorithm); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_6is_algorithm_enabled(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self, grpc_compression_algorithm __pyx_v_algorithm); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_8to_channel_arg(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_36compression_algorithm_name(CYTHON_UNUSED PyObject *__pyx_self, grpc_compression_algorithm __pyx_v_algorithm); /* proto */
static int __pyx_pf_4grpc_7_cython_6cygrpc_6Server___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_arguments); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_2request_call(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_call_queue, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_server_queue, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_4register_completion_queue(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_6register_non_listening_completion_queue(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_8start(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_10add_http2_port(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, PyObject *__pyx_v_address, struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *__pyx_v_server_credentials); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_12shutdown(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_14cancel_all_calls(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self); /* proto */
static void __pyx_pf_4grpc_7_cython_6cygrpc_6Server_16__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_38_initialize(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Call(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Channel(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelCredentials(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CallCredentials(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ServerCredentials(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_AuthMetadataContext(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CompletionQueue(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Timespec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CallDetails(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_OperationTag(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Event(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ByteBuffer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_SslPemKeyCertPair(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelArg(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelArgs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Metadatum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Metadata(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Operation(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Operations(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CompressionOptions(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Server(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc__MetadataIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc__OperationsIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_1e9;
static PyObject *__pyx_int_0;
static grpc_status_code __pyx_k__3;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_codeobj__9;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__59;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__65;
static PyObject *__pyx_codeobj__67;
static PyObject *__pyx_codeobj__69;
static PyObject *__pyx_codeobj__71;
static PyObject *__pyx_codeobj__73;
static PyObject *__pyx_codeobj__75;
static PyObject *__pyx_codeobj__76;

/* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":33
 * # This function will ascii encode unicode string inputs if neccesary.
 * # In Python3, unicode strings are the default str type.
 * cdef bytes str_to_bytes(object s):             # <<<<<<<<<<<<<<
 *   if s is None or isinstance(s, bytes):
 *     return s
 */

static PyObject *__pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("str_to_bytes", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":34
 * # In Python3, unicode strings are the default str type.
 * cdef bytes str_to_bytes(object s):
 *   if s is None or isinstance(s, bytes):             # <<<<<<<<<<<<<<
 *     return s
 *   elif isinstance(s, unicode):
 */
  __pyx_t_2 = (__pyx_v_s == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyBytes_Check(__pyx_v_s); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":35
 * cdef bytes str_to_bytes(object s):
 *   if s is None or isinstance(s, bytes):
 *     return s             # <<<<<<<<<<<<<<
 *   elif isinstance(s, unicode):
 *     return s.encode('ascii')
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyBytes_CheckExact(__pyx_v_s))||((__pyx_v_s) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_s)->tp_name), 0))) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_s);
    __pyx_r = ((PyObject*)__pyx_v_s);
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":34
 * # In Python3, unicode strings are the default str type.
 * cdef bytes str_to_bytes(object s):
 *   if s is None or isinstance(s, bytes):             # <<<<<<<<<<<<<<
 *     return s
 *   elif isinstance(s, unicode):
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":36
 *   if s is None or isinstance(s, bytes):
 *     return s
 *   elif isinstance(s, unicode):             # <<<<<<<<<<<<<<
 *     return s.encode('ascii')
 *   else:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_s); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":37
 *     return s
 *   elif isinstance(s, unicode):
 *     return s.encode('ascii')             # <<<<<<<<<<<<<<
 *   else:
 *     raise TypeError('Expected bytes, str, or unicode, not {}'.format(type(s)))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(1, 37, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":36
 *   if s is None or isinstance(s, bytes):
 *     return s
 *   elif isinstance(s, unicode):             # <<<<<<<<<<<<<<
 *     return s.encode('ascii')
 *   else:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":39
 *     return s.encode('ascii')
 *   else:
 *     raise TypeError('Expected bytes, str, or unicode, not {}'.format(type(s)))             # <<<<<<<<<<<<<<
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Expected_bytes_str_or_unicode_no, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)Py_TYPE(__pyx_v_s))); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_s)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_s)));
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, ((PyObject *)Py_TYPE(__pyx_v_s)));
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(1, 39, __pyx_L1_error)
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":33
 * # This function will ascii encode unicode string inputs if neccesary.
 * # In Python3, unicode strings are the default str type.
 * cdef bytes str_to_bytes(object s):             # <<<<<<<<<<<<<<
 *   if s is None or isinstance(s, bytes):
 *     return s
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("grpc._cython.cygrpc.str_to_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":35
 * cdef class Call:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     # Create an *empty* call
 *     grpc_init()
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_4Call_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_4Call_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4Call___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_4Call___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":37
 *   def __cinit__(self):
 *     # Create an *empty* call
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.c_call = NULL
 *     self.references = []
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":38
 *     # Create an *empty* call
 *     grpc_init()
 *     self.c_call = NULL             # <<<<<<<<<<<<<<
 *     self.references = []
 * 
 */
  __pyx_v_self->c_call = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":39
 *     grpc_init()
 *     self.c_call = NULL
 *     self.references = []             # <<<<<<<<<<<<<<
 * 
 *   def _start_batch(self, operations, tag, retain_self):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->references);
  __Pyx_DECREF(__pyx_v_self->references);
  __pyx_v_self->references = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":35
 * cdef class Call:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     # Create an *empty* call
 *     grpc_init()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":41
 *     self.references = []
 * 
 *   def _start_batch(self, operations, tag, retain_self):             # <<<<<<<<<<<<<<
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_3_start_batch(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_3_start_batch(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_operations = 0;
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_v_retain_self = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_start_batch (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_operations,&__pyx_n_s_tag,&__pyx_n_s_retain_self,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_operations)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tag)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_start_batch", 1, 3, 3, 1); __PYX_ERR(2, 41, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_retain_self)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_start_batch", 1, 3, 3, 2); __PYX_ERR(2, 41, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_start_batch") < 0)) __PYX_ERR(2, 41, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_operations = values[0];
    __pyx_v_tag = values[1];
    __pyx_v_retain_self = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_start_batch", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 41, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call._start_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4Call_2_start_batch(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self), __pyx_v_operations, __pyx_v_tag, __pyx_v_retain_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_2_start_batch(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, PyObject *__pyx_v_operations, PyObject *__pyx_v_tag, PyObject *__pyx_v_retain_self) {
  grpc_call_error __pyx_v_result;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_cy_operations = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *__pyx_v_operation_tag = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_start_batch", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":42
 * 
 *   def _start_batch(self, operations, tag, retain_self):
 *     if not self.is_valid:             # <<<<<<<<<<<<<<
 *       raise ValueError("invalid call object cannot be used from Python")
 *     cdef grpc_call_error result
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_valid); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":43
 *   def _start_batch(self, operations, tag, retain_self):
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")             # <<<<<<<<<<<<<<
 *     cdef grpc_call_error result
 *     cdef Operations cy_operations = Operations(operations)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 43, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":42
 * 
 *   def _start_batch(self, operations, tag, retain_self):
 *     if not self.is_valid:             # <<<<<<<<<<<<<<
 *       raise ValueError("invalid call object cannot be used from Python")
 *     cdef grpc_call_error result
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":45
 *       raise ValueError("invalid call object cannot be used from Python")
 *     cdef grpc_call_error result
 *     cdef Operations cy_operations = Operations(operations)             # <<<<<<<<<<<<<<
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     if retain_self:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_operations);
  __Pyx_GIVEREF(__pyx_v_operations);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_operations);
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operations), __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_cy_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":46
 *     cdef grpc_call_error result
 *     cdef Operations cy_operations = Operations(operations)
 *     cdef OperationTag operation_tag = OperationTag(tag)             # <<<<<<<<<<<<<<
 *     if retain_self:
 *       operation_tag.operation_call = self
 */
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_tag);
  __Pyx_GIVEREF(__pyx_v_tag);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_tag);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_OperationTag), __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_operation_tag = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":47
 *     cdef Operations cy_operations = Operations(operations)
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     if retain_self:             # <<<<<<<<<<<<<<
 *       operation_tag.operation_call = self
 *     else:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_retain_self); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":48
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     if retain_self:
 *       operation_tag.operation_call = self             # <<<<<<<<<<<<<<
 *     else:
 *       operation_tag.operation_call = None
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_operation_tag->operation_call);
    __Pyx_DECREF(((PyObject *)__pyx_v_operation_tag->operation_call));
    __pyx_v_operation_tag->operation_call = __pyx_v_self;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":47
 *     cdef Operations cy_operations = Operations(operations)
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     if retain_self:             # <<<<<<<<<<<<<<
 *       operation_tag.operation_call = self
 *     else:
 */
    goto __pyx_L4;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":50
 *       operation_tag.operation_call = self
 *     else:
 *       operation_tag.operation_call = None             # <<<<<<<<<<<<<<
 *     operation_tag.batch_operations = cy_operations
 *     cpython.Py_INCREF(operation_tag)
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_operation_tag->operation_call);
    __Pyx_DECREF(((PyObject *)__pyx_v_operation_tag->operation_call));
    __pyx_v_operation_tag->operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)Py_None);
  }
  __pyx_L4:;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":51
 *     else:
 *       operation_tag.operation_call = None
 *     operation_tag.batch_operations = cy_operations             # <<<<<<<<<<<<<<
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_cy_operations));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_cy_operations));
  __Pyx_GOTREF(__pyx_v_operation_tag->batch_operations);
  __Pyx_DECREF(((PyObject *)__pyx_v_operation_tag->batch_operations));
  __pyx_v_operation_tag->batch_operations = __pyx_v_cy_operations;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":52
 *       operation_tag.operation_call = None
 *     operation_tag.batch_operations = cy_operations
 *     cpython.Py_INCREF(operation_tag)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       result = grpc_call_start_batch(
 */
  Py_INCREF(((PyObject *)__pyx_v_operation_tag));

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":53
 *     operation_tag.batch_operations = cy_operations
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_call_start_batch(
 *           self.c_call, cy_operations.c_ops, cy_operations.c_nops,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":54
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:
 *       result = grpc_call_start_batch(             # <<<<<<<<<<<<<<
 *           self.c_call, cy_operations.c_ops, cy_operations.c_nops,
 *           <cpython.PyObject *>operation_tag, NULL)
 */
        __pyx_v_result = grpc_call_start_batch(__pyx_v_self->c_call, __pyx_v_cy_operations->c_ops, __pyx_v_cy_operations->c_nops, ((PyObject *)__pyx_v_operation_tag), NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":53
 *     operation_tag.batch_operations = cy_operations
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_call_start_batch(
 *           self.c_call, cy_operations.c_ops, cy_operations.c_nops,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":57
 *           self.c_call, cy_operations.c_ops, cy_operations.c_nops,
 *           <cpython.PyObject *>operation_tag, NULL)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   def start_client_batch(self, operations, tag):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_grpc_call_error(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":41
 *     self.references = []
 * 
 *   def _start_batch(self, operations, tag, retain_self):             # <<<<<<<<<<<<<<
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call._start_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cy_operations);
  __Pyx_XDECREF((PyObject *)__pyx_v_operation_tag);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":59
 *     return result
 * 
 *   def start_client_batch(self, operations, tag):             # <<<<<<<<<<<<<<
 *     # We don't reference this call in the operations tag because
 *     # it should be cancelled when it goes out of scope
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_5start_client_batch(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_5start_client_batch(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_operations = 0;
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_client_batch (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_operations,&__pyx_n_s_tag,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_operations)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tag)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("start_client_batch", 1, 2, 2, 1); __PYX_ERR(2, 59, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "start_client_batch") < 0)) __PYX_ERR(2, 59, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_operations = values[0];
    __pyx_v_tag = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start_client_batch", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 59, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.start_client_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4Call_4start_client_batch(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self), __pyx_v_operations, __pyx_v_tag);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_4start_client_batch(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, PyObject *__pyx_v_operations, PyObject *__pyx_v_tag) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("start_client_batch", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":62
 *     # We don't reference this call in the operations tag because
 *     # it should be cancelled when it goes out of scope
 *     return self._start_batch(operations, tag, False)             # <<<<<<<<<<<<<<
 * 
 *   def start_server_batch(self, operations, tag):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start_batch); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  __pyx_t_5 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_3) {
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_operations);
  __Pyx_GIVEREF(__pyx_v_operations);
  PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_operations);
  __Pyx_INCREF(__pyx_v_tag);
  __Pyx_GIVEREF(__pyx_v_tag);
  PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_tag);
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, Py_False);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":59
 *     return result
 * 
 *   def start_client_batch(self, operations, tag):             # <<<<<<<<<<<<<<
 *     # We don't reference this call in the operations tag because
 *     # it should be cancelled when it goes out of scope
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.start_client_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":64
 *     return self._start_batch(operations, tag, False)
 * 
 *   def start_server_batch(self, operations, tag):             # <<<<<<<<<<<<<<
 *     return self._start_batch(operations, tag, True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_7start_server_batch(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_7start_server_batch(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_operations = 0;
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_server_batch (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_operations,&__pyx_n_s_tag,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_operations)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tag)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("start_server_batch", 1, 2, 2, 1); __PYX_ERR(2, 64, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "start_server_batch") < 0)) __PYX_ERR(2, 64, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_operations = values[0];
    __pyx_v_tag = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start_server_batch", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 64, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.start_server_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4Call_6start_server_batch(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self), __pyx_v_operations, __pyx_v_tag);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_6start_server_batch(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, PyObject *__pyx_v_operations, PyObject *__pyx_v_tag) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("start_server_batch", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":65
 * 
 *   def start_server_batch(self, operations, tag):
 *     return self._start_batch(operations, tag, True)             # <<<<<<<<<<<<<<
 * 
 *   def cancel(
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start_batch); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  __pyx_t_5 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_3) {
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_operations);
  __Pyx_GIVEREF(__pyx_v_operations);
  PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_operations);
  __Pyx_INCREF(__pyx_v_tag);
  __Pyx_GIVEREF(__pyx_v_tag);
  PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_tag);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, Py_True);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":64
 *     return self._start_batch(operations, tag, False)
 * 
 *   def start_server_batch(self, operations, tag):             # <<<<<<<<<<<<<<
 *     return self._start_batch(operations, tag, True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.start_server_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":67
 *     return self._start_batch(operations, tag, True)
 * 
 *   def cancel(             # <<<<<<<<<<<<<<
 *       self, grpc_status_code error_code=GRPC_STATUS__DO_NOT_USE,
 *       details=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_9cancel(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_9cancel(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  grpc_status_code __pyx_v_error_code;
  PyObject *__pyx_v_details = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_error_code,&__pyx_n_s_details,0};
    PyObject* values[2] = {0,0};

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":69
 *   def cancel(
 *       self, grpc_status_code error_code=GRPC_STATUS__DO_NOT_USE,
 *       details=None):             # <<<<<<<<<<<<<<
 *     details = str_to_bytes(details)
 *     if not self.is_valid:
 */
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_error_code);
          if (value) { values[0] = value; kw_args--; }
        }
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_details);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cancel") < 0)) __PYX_ERR(2, 67, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_error_code = ((grpc_status_code)__Pyx_PyInt_As_grpc_status_code(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 68, __pyx_L3_error)
    } else {
      __pyx_v_error_code = __pyx_k__3;
    }
    __pyx_v_details = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4Call_8cancel(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self), __pyx_v_error_code, __pyx_v_details);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":67
 *     return self._start_batch(operations, tag, True)
 * 
 *   def cancel(             # <<<<<<<<<<<<<<
 *       self, grpc_status_code error_code=GRPC_STATUS__DO_NOT_USE,
 *       details=None):
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_8cancel(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, grpc_status_code __pyx_v_error_code, PyObject *__pyx_v_details) {
  grpc_call_error __pyx_v_result;
  char *__pyx_v_c_details;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  char *__pyx_t_5;
  __Pyx_RefNannySetupContext("cancel", 0);
  __Pyx_INCREF(__pyx_v_details);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":70
 *       self, grpc_status_code error_code=GRPC_STATUS__DO_NOT_USE,
 *       details=None):
 *     details = str_to_bytes(details)             # <<<<<<<<<<<<<<
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")
 */
  __pyx_t_1 = __pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(__pyx_v_details); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_details, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":71
 *       details=None):
 *     details = str_to_bytes(details)
 *     if not self.is_valid:             # <<<<<<<<<<<<<<
 *       raise ValueError("invalid call object cannot be used from Python")
 *     if (details is None) != (error_code == GRPC_STATUS__DO_NOT_USE):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_valid); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":72
 *     details = str_to_bytes(details)
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")             # <<<<<<<<<<<<<<
 *     if (details is None) != (error_code == GRPC_STATUS__DO_NOT_USE):
 *       raise ValueError("if error_code is specified, so must details "
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 72, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":71
 *       details=None):
 *     details = str_to_bytes(details)
 *     if not self.is_valid:             # <<<<<<<<<<<<<<
 *       raise ValueError("invalid call object cannot be used from Python")
 *     if (details is None) != (error_code == GRPC_STATUS__DO_NOT_USE):
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":73
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")
 *     if (details is None) != (error_code == GRPC_STATUS__DO_NOT_USE):             # <<<<<<<<<<<<<<
 *       raise ValueError("if error_code is specified, so must details "
 *                        "(and vice-versa)")
 */
  __pyx_t_3 = (__pyx_v_details == Py_None);
  __pyx_t_2 = ((__pyx_t_3 != (__pyx_v_error_code == GRPC_STATUS__DO_NOT_USE)) != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":74
 *       raise ValueError("invalid call object cannot be used from Python")
 *     if (details is None) != (error_code == GRPC_STATUS__DO_NOT_USE):
 *       raise ValueError("if error_code is specified, so must details "             # <<<<<<<<<<<<<<
 *                        "(and vice-versa)")
 *     cdef grpc_call_error result
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 74, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":73
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")
 *     if (details is None) != (error_code == GRPC_STATUS__DO_NOT_USE):             # <<<<<<<<<<<<<<
 *       raise ValueError("if error_code is specified, so must details "
 *                        "(and vice-versa)")
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":77
 *                        "(and vice-versa)")
 *     cdef grpc_call_error result
 *     cdef char *c_details = NULL             # <<<<<<<<<<<<<<
 *     if error_code != GRPC_STATUS__DO_NOT_USE:
 *       self.references.append(details)
 */
  __pyx_v_c_details = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":78
 *     cdef grpc_call_error result
 *     cdef char *c_details = NULL
 *     if error_code != GRPC_STATUS__DO_NOT_USE:             # <<<<<<<<<<<<<<
 *       self.references.append(details)
 *       c_details = details
 */
  __pyx_t_2 = ((__pyx_v_error_code != GRPC_STATUS__DO_NOT_USE) != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":79
 *     cdef char *c_details = NULL
 *     if error_code != GRPC_STATUS__DO_NOT_USE:
 *       self.references.append(details)             # <<<<<<<<<<<<<<
 *       c_details = details
 *       with nogil:
 */
    if (unlikely(__pyx_v_self->references == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
      __PYX_ERR(2, 79, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_self->references, __pyx_v_details); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(2, 79, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":80
 *     if error_code != GRPC_STATUS__DO_NOT_USE:
 *       self.references.append(details)
 *       c_details = details             # <<<<<<<<<<<<<<
 *       with nogil:
 *         result = grpc_call_cancel_with_status(
 */
    __pyx_t_5 = __Pyx_PyObject_AsString(__pyx_v_details); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(2, 80, __pyx_L1_error)
    __pyx_v_c_details = __pyx_t_5;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":81
 *       self.references.append(details)
 *       c_details = details
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_call_cancel_with_status(
 *             self.c_call, error_code, c_details, NULL)
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":82
 *       c_details = details
 *       with nogil:
 *         result = grpc_call_cancel_with_status(             # <<<<<<<<<<<<<<
 *             self.c_call, error_code, c_details, NULL)
 *       return result
 */
          __pyx_v_result = grpc_call_cancel_with_status(__pyx_v_self->c_call, __pyx_v_error_code, __pyx_v_c_details, NULL);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":81
 *       self.references.append(details)
 *       c_details = details
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_call_cancel_with_status(
 *             self.c_call, error_code, c_details, NULL)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L8;
          }
          __pyx_L8:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":84
 *         result = grpc_call_cancel_with_status(
 *             self.c_call, error_code, c_details, NULL)
 *       return result             # <<<<<<<<<<<<<<
 *     else:
 *       with nogil:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyInt_From_grpc_call_error(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":78
 *     cdef grpc_call_error result
 *     cdef char *c_details = NULL
 *     if error_code != GRPC_STATUS__DO_NOT_USE:             # <<<<<<<<<<<<<<
 *       self.references.append(details)
 *       c_details = details
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":86
 *       return result
 *     else:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_call_cancel(self.c_call, NULL)
 *       return result
 */
  /*else*/ {
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":87
 *     else:
 *       with nogil:
 *         result = grpc_call_cancel(self.c_call, NULL)             # <<<<<<<<<<<<<<
 *       return result
 * 
 */
          __pyx_v_result = grpc_call_cancel(__pyx_v_self->c_call, NULL);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":86
 *       return result
 *     else:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_call_cancel(self.c_call, NULL)
 *       return result
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L11;
          }
          __pyx_L11:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":88
 *       with nogil:
 *         result = grpc_call_cancel(self.c_call, NULL)
 *       return result             # <<<<<<<<<<<<<<
 * 
 *   def set_credentials(
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyInt_From_grpc_call_error(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":67
 *     return self._start_batch(operations, tag, True)
 * 
 *   def cancel(             # <<<<<<<<<<<<<<
 *       self, grpc_status_code error_code=GRPC_STATUS__DO_NOT_USE,
 *       details=None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_details);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":90
 *       return result
 * 
 *   def set_credentials(             # <<<<<<<<<<<<<<
 *       self, CallCredentials call_credentials not None):
 *     cdef grpc_call_error result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_11set_credentials(PyObject *__pyx_v_self, PyObject *__pyx_v_call_credentials); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_11set_credentials(PyObject *__pyx_v_self, PyObject *__pyx_v_call_credentials) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_credentials (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_call_credentials), __pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials, 0, "call_credentials", 0))) __PYX_ERR(2, 91, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4Call_10set_credentials(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self), ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_v_call_credentials));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_10set_credentials(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_call_credentials) {
  grpc_call_error __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("set_credentials", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":93
 *       self, CallCredentials call_credentials not None):
 *     cdef grpc_call_error result
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_call_set_credentials(
 *           self.c_call, call_credentials.c_credentials)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":94
 *     cdef grpc_call_error result
 *     with nogil:
 *       result = grpc_call_set_credentials(             # <<<<<<<<<<<<<<
 *           self.c_call, call_credentials.c_credentials)
 *     return result
 */
        __pyx_v_result = grpc_call_set_credentials(__pyx_v_self->c_call, __pyx_v_call_credentials->c_credentials);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":93
 *       self, CallCredentials call_credentials not None):
 *     cdef grpc_call_error result
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_call_set_credentials(
 *           self.c_call, call_credentials.c_credentials)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":96
 *       result = grpc_call_set_credentials(
 *           self.c_call, call_credentials.c_credentials)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   def peer(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_grpc_call_error(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":90
 *       return result
 * 
 *   def set_credentials(             # <<<<<<<<<<<<<<
 *       self, CallCredentials call_credentials not None):
 *     cdef grpc_call_error result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.set_credentials", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":98
 *     return result
 * 
 *   def peer(self):             # <<<<<<<<<<<<<<
 *     cdef char *peer = NULL
 *     with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_13peer(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_13peer(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("peer (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4Call_12peer(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_12peer(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self) {
  char *__pyx_v_peer;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("peer", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":99
 * 
 *   def peer(self):
 *     cdef char *peer = NULL             # <<<<<<<<<<<<<<
 *     with nogil:
 *       peer = grpc_call_get_peer(self.c_call)
 */
  __pyx_v_peer = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":100
 *   def peer(self):
 *     cdef char *peer = NULL
 *     with nogil:             # <<<<<<<<<<<<<<
 *       peer = grpc_call_get_peer(self.c_call)
 *     result = <bytes>peer
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":101
 *     cdef char *peer = NULL
 *     with nogil:
 *       peer = grpc_call_get_peer(self.c_call)             # <<<<<<<<<<<<<<
 *     result = <bytes>peer
 *     with nogil:
 */
        __pyx_v_peer = grpc_call_get_peer(__pyx_v_self->c_call);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":100
 *   def peer(self):
 *     cdef char *peer = NULL
 *     with nogil:             # <<<<<<<<<<<<<<
 *       peer = grpc_call_get_peer(self.c_call)
 *     result = <bytes>peer
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":102
 *     with nogil:
 *       peer = grpc_call_get_peer(self.c_call)
 *     result = <bytes>peer             # <<<<<<<<<<<<<<
 *     with nogil:
 *       gpr_free(peer)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_peer); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_result = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":103
 *       peer = grpc_call_get_peer(self.c_call)
 *     result = <bytes>peer
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_free(peer)
 *     return result
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":104
 *     result = <bytes>peer
 *     with nogil:
 *       gpr_free(peer)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
        gpr_free(__pyx_v_peer);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":103
 *       peer = grpc_call_get_peer(self.c_call)
 *     result = <bytes>peer
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_free(peer)
 *     return result
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":105
 *     with nogil:
 *       gpr_free(peer)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":98
 *     return result
 * 
 *   def peer(self):             # <<<<<<<<<<<<<<
 *     cdef char *peer = NULL
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.peer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":107
 *     return result
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_call != NULL:
 *       grpc_call_destroy(self.c_call)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_4Call_15__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_4Call_15__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_4Call_14__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_4Call_14__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":108
 * 
 *   def __dealloc__(self):
 *     if self.c_call != NULL:             # <<<<<<<<<<<<<<
 *       grpc_call_destroy(self.c_call)
 *     grpc_shutdown()
 */
  __pyx_t_1 = ((__pyx_v_self->c_call != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":109
 *   def __dealloc__(self):
 *     if self.c_call != NULL:
 *       grpc_call_destroy(self.c_call)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
    grpc_call_destroy(__pyx_v_self->c_call);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":108
 * 
 *   def __dealloc__(self):
 *     if self.c_call != NULL:             # <<<<<<<<<<<<<<
 *       grpc_call_destroy(self.c_call)
 *     grpc_shutdown()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":110
 *     if self.c_call != NULL:
 *       grpc_call_destroy(self.c_call)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 *   # The object *should* always be valid from Python. Used for debugging.
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":107
 *     return result
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_call != NULL:
 *       grpc_call_destroy(self.c_call)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":114
 *   # The object *should* always be valid from Python. Used for debugging.
 *   @property
 *   def is_valid(self):             # <<<<<<<<<<<<<<
 *     return self.c_call != NULL
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_8is_valid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_4Call_8is_valid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4Call_8is_valid___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4Call_8is_valid___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":115
 *   @property
 *   def is_valid(self):
 *     return self.c_call != NULL             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->c_call != NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":114
 *   # The object *should* always be valid from Python. Used for debugging.
 *   @property
 *   def is_valid(self):             # <<<<<<<<<<<<<<
 *     return self.c_call != NULL
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Call.is_valid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":35
 * cdef class Channel:
 * 
 *   def __cinit__(self, bytes target, ChannelArgs arguments=None,             # <<<<<<<<<<<<<<
 *                 ChannelCredentials channel_credentials=None):
 *     grpc_init()
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_7Channel_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_7Channel_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_target = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_arguments = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_channel_credentials = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_target,&__pyx_n_s_arguments,&__pyx_n_s_channel_credentials,0};
    PyObject* values[3] = {0,0,0};
    values[1] = (PyObject *)((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)Py_None);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":36
 * 
 *   def __cinit__(self, bytes target, ChannelArgs arguments=None,
 *                 ChannelCredentials channel_credentials=None):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     cdef grpc_channel_args *c_arguments = NULL
 */
    values[2] = (PyObject *)((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_target)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_arguments);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_channel_credentials);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(5, 35, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_target = ((PyObject*)values[0]);
    __pyx_v_arguments = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)values[1]);
    __pyx_v_channel_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(5, 35, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_target), (&PyBytes_Type), 1, "target", 1))) __PYX_ERR(5, 35, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_arguments), __pyx_ptype_4grpc_7_cython_6cygrpc_ChannelArgs, 1, "arguments", 0))) __PYX_ERR(5, 35, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_channel_credentials), __pyx_ptype_4grpc_7_cython_6cygrpc_ChannelCredentials, 1, "channel_credentials", 0))) __PYX_ERR(5, 36, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_7Channel___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)__pyx_v_self), __pyx_v_target, __pyx_v_arguments, __pyx_v_channel_credentials);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":35
 * cdef class Channel:
 * 
 *   def __cinit__(self, bytes target, ChannelArgs arguments=None,             # <<<<<<<<<<<<<<
 *                 ChannelCredentials channel_credentials=None):
 *     grpc_init()
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_7Channel___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self, PyObject *__pyx_v_target, struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_arguments, struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_channel_credentials) {
  grpc_channel_args *__pyx_v_c_arguments;
  char *__pyx_v_c_target;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  char *__pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":37
 *   def __cinit__(self, bytes target, ChannelArgs arguments=None,
 *                 ChannelCredentials channel_credentials=None):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     cdef grpc_channel_args *c_arguments = NULL
 *     cdef char *c_target = NULL
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":38
 *                 ChannelCredentials channel_credentials=None):
 *     grpc_init()
 *     cdef grpc_channel_args *c_arguments = NULL             # <<<<<<<<<<<<<<
 *     cdef char *c_target = NULL
 *     self.c_channel = NULL
 */
  __pyx_v_c_arguments = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":39
 *     grpc_init()
 *     cdef grpc_channel_args *c_arguments = NULL
 *     cdef char *c_target = NULL             # <<<<<<<<<<<<<<
 *     self.c_channel = NULL
 *     self.references = []
 */
  __pyx_v_c_target = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":40
 *     cdef grpc_channel_args *c_arguments = NULL
 *     cdef char *c_target = NULL
 *     self.c_channel = NULL             # <<<<<<<<<<<<<<
 *     self.references = []
 *     if arguments is not None:
 */
  __pyx_v_self->c_channel = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":41
 *     cdef char *c_target = NULL
 *     self.c_channel = NULL
 *     self.references = []             # <<<<<<<<<<<<<<
 *     if arguments is not None:
 *       c_arguments = &arguments.c_args
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->references);
  __Pyx_DECREF(__pyx_v_self->references);
  __pyx_v_self->references = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":42
 *     self.c_channel = NULL
 *     self.references = []
 *     if arguments is not None:             # <<<<<<<<<<<<<<
 *       c_arguments = &arguments.c_args
 *     c_target = target
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_arguments) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":43
 *     self.references = []
 *     if arguments is not None:
 *       c_arguments = &arguments.c_args             # <<<<<<<<<<<<<<
 *     c_target = target
 *     if channel_credentials is None:
 */
    __pyx_v_c_arguments = (&__pyx_v_arguments->c_args);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":42
 *     self.c_channel = NULL
 *     self.references = []
 *     if arguments is not None:             # <<<<<<<<<<<<<<
 *       c_arguments = &arguments.c_args
 *     c_target = target
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":44
 *     if arguments is not None:
 *       c_arguments = &arguments.c_args
 *     c_target = target             # <<<<<<<<<<<<<<
 *     if channel_credentials is None:
 *       with nogil:
 */
  __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_target); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(5, 44, __pyx_L1_error)
  __pyx_v_c_target = __pyx_t_4;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":45
 *       c_arguments = &arguments.c_args
 *     c_target = target
 *     if channel_credentials is None:             # <<<<<<<<<<<<<<
 *       with nogil:
 *         self.c_channel = grpc_insecure_channel_create(c_target, c_arguments,
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_channel_credentials) == Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":46
 *     c_target = target
 *     if channel_credentials is None:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         self.c_channel = grpc_insecure_channel_create(c_target, c_arguments,
 *                                                       NULL)
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":47
 *     if channel_credentials is None:
 *       with nogil:
 *         self.c_channel = grpc_insecure_channel_create(c_target, c_arguments,             # <<<<<<<<<<<<<<
 *                                                       NULL)
 *     else:
 */
          __pyx_v_self->c_channel = grpc_insecure_channel_create(__pyx_v_c_target, __pyx_v_c_arguments, NULL);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":46
 *     c_target = target
 *     if channel_credentials is None:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         self.c_channel = grpc_insecure_channel_create(c_target, c_arguments,
 *                                                       NULL)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L7;
          }
          __pyx_L7:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":45
 *       c_arguments = &arguments.c_args
 *     c_target = target
 *     if channel_credentials is None:             # <<<<<<<<<<<<<<
 *       with nogil:
 *         self.c_channel = grpc_insecure_channel_create(c_target, c_arguments,
 */
    goto __pyx_L4;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":50
 *                                                       NULL)
 *     else:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         self.c_channel = grpc_secure_channel_create(
 *             channel_credentials.c_credentials, c_target, c_arguments, NULL)
 */
  /*else*/ {
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":51
 *     else:
 *       with nogil:
 *         self.c_channel = grpc_secure_channel_create(             # <<<<<<<<<<<<<<
 *             channel_credentials.c_credentials, c_target, c_arguments, NULL)
 *       self.references.append(channel_credentials)
 */
          __pyx_v_self->c_channel = grpc_secure_channel_create(__pyx_v_channel_credentials->c_credentials, __pyx_v_c_target, __pyx_v_c_arguments, NULL);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":50
 *                                                       NULL)
 *     else:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         self.c_channel = grpc_secure_channel_create(
 *             channel_credentials.c_credentials, c_target, c_arguments, NULL)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L10;
          }
          __pyx_L10:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":53
 *         self.c_channel = grpc_secure_channel_create(
 *             channel_credentials.c_credentials, c_target, c_arguments, NULL)
 *       self.references.append(channel_credentials)             # <<<<<<<<<<<<<<
 *     self.references.append(target)
 *     self.references.append(arguments)
 */
    if (unlikely(__pyx_v_self->references == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
      __PYX_ERR(5, 53, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->references, ((PyObject *)__pyx_v_channel_credentials)); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(5, 53, __pyx_L1_error)
  }
  __pyx_L4:;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":54
 *             channel_credentials.c_credentials, c_target, c_arguments, NULL)
 *       self.references.append(channel_credentials)
 *     self.references.append(target)             # <<<<<<<<<<<<<<
 *     self.references.append(arguments)
 * 
 */
  if (unlikely(__pyx_v_self->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(5, 54, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->references, __pyx_v_target); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(5, 54, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":55
 *       self.references.append(channel_credentials)
 *     self.references.append(target)
 *     self.references.append(arguments)             # <<<<<<<<<<<<<<
 * 
 *   def create_call(self, Call parent, int flags,
 */
  if (unlikely(__pyx_v_self->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(5, 55, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->references, ((PyObject *)__pyx_v_arguments)); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(5, 55, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":35
 * cdef class Channel:
 * 
 *   def __cinit__(self, bytes target, ChannelArgs arguments=None,             # <<<<<<<<<<<<<<
 *                 ChannelCredentials channel_credentials=None):
 *     grpc_init()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":57
 *     self.references.append(arguments)
 * 
 *   def create_call(self, Call parent, int flags,             # <<<<<<<<<<<<<<
 *                   CompletionQueue queue not None,
 *                   method, host, Timespec deadline not None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_3create_call(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_3create_call(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_parent = 0;
  int __pyx_v_flags;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue = 0;
  PyObject *__pyx_v_method = 0;
  PyObject *__pyx_v_host = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_call (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_parent,&__pyx_n_s_flags,&__pyx_n_s_queue,&__pyx_n_s_method,&__pyx_n_s_host,&__pyx_n_s_deadline,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_call", 1, 6, 6, 1); __PYX_ERR(5, 57, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_queue)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_call", 1, 6, 6, 2); __PYX_ERR(5, 57, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_method)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_call", 1, 6, 6, 3); __PYX_ERR(5, 57, __pyx_L3_error)
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_host)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_call", 1, 6, 6, 4); __PYX_ERR(5, 57, __pyx_L3_error)
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_deadline)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_call", 1, 6, 6, 5); __PYX_ERR(5, 57, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_call") < 0)) __PYX_ERR(5, 57, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_parent = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)values[0]);
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 57, __pyx_L3_error)
    __pyx_v_queue = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)values[2]);
    __pyx_v_method = values[3];
    __pyx_v_host = values[4];
    __pyx_v_deadline = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)values[5]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_call", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(5, 57, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.create_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parent), __pyx_ptype_4grpc_7_cython_6cygrpc_Call, 1, "parent", 0))) __PYX_ERR(5, 57, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_queue), __pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue, 0, "queue", 0))) __PYX_ERR(5, 58, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_deadline), __pyx_ptype_4grpc_7_cython_6cygrpc_Timespec, 0, "deadline", 0))) __PYX_ERR(5, 59, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_7Channel_2create_call(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)__pyx_v_self), __pyx_v_parent, __pyx_v_flags, __pyx_v_queue, __pyx_v_method, __pyx_v_host, __pyx_v_deadline);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_7Channel_2create_call(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_parent, int __pyx_v_flags, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue, PyObject *__pyx_v_method, PyObject *__pyx_v_host, struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline) {
  char *__pyx_v_method_c_string;
  char *__pyx_v_host_c_string;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_operation_call = 0;
  grpc_call *__pyx_v_parent_call;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  int __pyx_t_4;
  grpc_call *__pyx_t_5;
  __Pyx_RefNannySetupContext("create_call", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":60
 *                   CompletionQueue queue not None,
 *                   method, host, Timespec deadline not None):
 *     if queue.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise ValueError("queue must not be shutting down or shutdown")
 *     cdef char *method_c_string = method
 */
  __pyx_t_1 = (__pyx_v_queue->is_shutting_down != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":61
 *                   method, host, Timespec deadline not None):
 *     if queue.is_shutting_down:
 *       raise ValueError("queue must not be shutting down or shutdown")             # <<<<<<<<<<<<<<
 *     cdef char *method_c_string = method
 *     cdef char *host_c_string = NULL
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(5, 61, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":60
 *                   CompletionQueue queue not None,
 *                   method, host, Timespec deadline not None):
 *     if queue.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise ValueError("queue must not be shutting down or shutdown")
 *     cdef char *method_c_string = method
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":62
 *     if queue.is_shutting_down:
 *       raise ValueError("queue must not be shutting down or shutdown")
 *     cdef char *method_c_string = method             # <<<<<<<<<<<<<<
 *     cdef char *host_c_string = NULL
 *     if host is not None:
 */
  __pyx_t_3 = __Pyx_PyObject_AsString(__pyx_v_method); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(5, 62, __pyx_L1_error)
  __pyx_v_method_c_string = __pyx_t_3;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":63
 *       raise ValueError("queue must not be shutting down or shutdown")
 *     cdef char *method_c_string = method
 *     cdef char *host_c_string = NULL             # <<<<<<<<<<<<<<
 *     if host is not None:
 *       host_c_string = host
 */
  __pyx_v_host_c_string = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":64
 *     cdef char *method_c_string = method
 *     cdef char *host_c_string = NULL
 *     if host is not None:             # <<<<<<<<<<<<<<
 *       host_c_string = host
 *     cdef Call operation_call = Call()
 */
  __pyx_t_1 = (__pyx_v_host != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":65
 *     cdef char *host_c_string = NULL
 *     if host is not None:
 *       host_c_string = host             # <<<<<<<<<<<<<<
 *     cdef Call operation_call = Call()
 *     operation_call.references = [self, method, host, queue]
 */
    __pyx_t_3 = __Pyx_PyObject_AsString(__pyx_v_host); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(5, 65, __pyx_L1_error)
    __pyx_v_host_c_string = __pyx_t_3;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":64
 *     cdef char *method_c_string = method
 *     cdef char *host_c_string = NULL
 *     if host is not None:             # <<<<<<<<<<<<<<
 *       host_c_string = host
 *     cdef Call operation_call = Call()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":66
 *     if host is not None:
 *       host_c_string = host
 *     cdef Call operation_call = Call()             # <<<<<<<<<<<<<<
 *     operation_call.references = [self, method, host, queue]
 *     cdef grpc_call *parent_call = NULL
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Call), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":67
 *       host_c_string = host
 *     cdef Call operation_call = Call()
 *     operation_call.references = [self, method, host, queue]             # <<<<<<<<<<<<<<
 *     cdef grpc_call *parent_call = NULL
 *     if parent is not None:
 */
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_method);
  __Pyx_GIVEREF(__pyx_v_method);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_method);
  __Pyx_INCREF(__pyx_v_host);
  __Pyx_GIVEREF(__pyx_v_host);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_v_host);
  __Pyx_INCREF(((PyObject *)__pyx_v_queue));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_queue));
  PyList_SET_ITEM(__pyx_t_2, 3, ((PyObject *)__pyx_v_queue));
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_operation_call->references);
  __Pyx_DECREF(__pyx_v_operation_call->references);
  __pyx_v_operation_call->references = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":68
 *     cdef Call operation_call = Call()
 *     operation_call.references = [self, method, host, queue]
 *     cdef grpc_call *parent_call = NULL             # <<<<<<<<<<<<<<
 *     if parent is not None:
 *       parent_call = parent.c_call
 */
  __pyx_v_parent_call = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":69
 *     operation_call.references = [self, method, host, queue]
 *     cdef grpc_call *parent_call = NULL
 *     if parent is not None:             # <<<<<<<<<<<<<<
 *       parent_call = parent.c_call
 *     with nogil:
 */
  __pyx_t_4 = (((PyObject *)__pyx_v_parent) != Py_None);
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":70
 *     cdef grpc_call *parent_call = NULL
 *     if parent is not None:
 *       parent_call = parent.c_call             # <<<<<<<<<<<<<<
 *     with nogil:
 *       operation_call.c_call = grpc_channel_create_call(
 */
    __pyx_t_5 = __pyx_v_parent->c_call;
    __pyx_v_parent_call = __pyx_t_5;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":69
 *     operation_call.references = [self, method, host, queue]
 *     cdef grpc_call *parent_call = NULL
 *     if parent is not None:             # <<<<<<<<<<<<<<
 *       parent_call = parent.c_call
 *     with nogil:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":71
 *     if parent is not None:
 *       parent_call = parent.c_call
 *     with nogil:             # <<<<<<<<<<<<<<
 *       operation_call.c_call = grpc_channel_create_call(
 *           self.c_channel, parent_call, flags,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":72
 *       parent_call = parent.c_call
 *     with nogil:
 *       operation_call.c_call = grpc_channel_create_call(             # <<<<<<<<<<<<<<
 *           self.c_channel, parent_call, flags,
 *           queue.c_completion_queue, method_c_string, host_c_string,
 */
        __pyx_v_operation_call->c_call = grpc_channel_create_call(__pyx_v_self->c_channel, __pyx_v_parent_call, __pyx_v_flags, __pyx_v_queue->c_completion_queue, __pyx_v_method_c_string, __pyx_v_host_c_string, __pyx_v_deadline->c_time, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":71
 *     if parent is not None:
 *       parent_call = parent.c_call
 *     with nogil:             # <<<<<<<<<<<<<<
 *       operation_call.c_call = grpc_channel_create_call(
 *           self.c_channel, parent_call, flags,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":76
 *           queue.c_completion_queue, method_c_string, host_c_string,
 *           deadline.c_time, NULL)
 *     return operation_call             # <<<<<<<<<<<<<<
 * 
 *   def check_connectivity_state(self, bint try_to_connect):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_operation_call));
  __pyx_r = ((PyObject *)__pyx_v_operation_call);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":57
 *     self.references.append(arguments)
 * 
 *   def create_call(self, Call parent, int flags,             # <<<<<<<<<<<<<<
 *                   CompletionQueue queue not None,
 *                   method, host, Timespec deadline not None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.create_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_operation_call);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":78
 *     return operation_call
 * 
 *   def check_connectivity_state(self, bint try_to_connect):             # <<<<<<<<<<<<<<
 *     cdef grpc_connectivity_state result
 *     with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_5check_connectivity_state(PyObject *__pyx_v_self, PyObject *__pyx_arg_try_to_connect); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_5check_connectivity_state(PyObject *__pyx_v_self, PyObject *__pyx_arg_try_to_connect) {
  int __pyx_v_try_to_connect;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_connectivity_state (wrapper)", 0);
  assert(__pyx_arg_try_to_connect); {
    __pyx_v_try_to_connect = __Pyx_PyObject_IsTrue(__pyx_arg_try_to_connect); if (unlikely((__pyx_v_try_to_connect == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 78, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.check_connectivity_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_7Channel_4check_connectivity_state(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)__pyx_v_self), ((int)__pyx_v_try_to_connect));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_7Channel_4check_connectivity_state(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self, int __pyx_v_try_to_connect) {
  grpc_connectivity_state __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("check_connectivity_state", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":80
 *   def check_connectivity_state(self, bint try_to_connect):
 *     cdef grpc_connectivity_state result
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_channel_check_connectivity_state(self.c_channel,
 *                                                      try_to_connect)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":81
 *     cdef grpc_connectivity_state result
 *     with nogil:
 *       result = grpc_channel_check_connectivity_state(self.c_channel,             # <<<<<<<<<<<<<<
 *                                                      try_to_connect)
 *     return result
 */
        __pyx_v_result = grpc_channel_check_connectivity_state(__pyx_v_self->c_channel, __pyx_v_try_to_connect);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":80
 *   def check_connectivity_state(self, bint try_to_connect):
 *     cdef grpc_connectivity_state result
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_channel_check_connectivity_state(self.c_channel,
 *                                                      try_to_connect)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":83
 *       result = grpc_channel_check_connectivity_state(self.c_channel,
 *                                                      try_to_connect)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   def watch_connectivity_state(
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_grpc_connectivity_state(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":78
 *     return operation_call
 * 
 *   def check_connectivity_state(self, bint try_to_connect):             # <<<<<<<<<<<<<<
 *     cdef grpc_connectivity_state result
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.check_connectivity_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":85
 *     return result
 * 
 *   def watch_connectivity_state(             # <<<<<<<<<<<<<<
 *       self, grpc_connectivity_state last_observed_state,
 *       Timespec deadline not None, CompletionQueue queue not None, tag):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_7watch_connectivity_state(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_7watch_connectivity_state(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  grpc_connectivity_state __pyx_v_last_observed_state;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue = 0;
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("watch_connectivity_state (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_last_observed_state,&__pyx_n_s_deadline,&__pyx_n_s_queue,&__pyx_n_s_tag,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_last_observed_state)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_deadline)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("watch_connectivity_state", 1, 4, 4, 1); __PYX_ERR(5, 85, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_queue)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("watch_connectivity_state", 1, 4, 4, 2); __PYX_ERR(5, 85, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tag)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("watch_connectivity_state", 1, 4, 4, 3); __PYX_ERR(5, 85, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "watch_connectivity_state") < 0)) __PYX_ERR(5, 85, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_last_observed_state = ((grpc_connectivity_state)__Pyx_PyInt_As_grpc_connectivity_state(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 86, __pyx_L3_error)
    __pyx_v_deadline = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)values[1]);
    __pyx_v_queue = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)values[2]);
    __pyx_v_tag = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("watch_connectivity_state", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(5, 85, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.watch_connectivity_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_deadline), __pyx_ptype_4grpc_7_cython_6cygrpc_Timespec, 0, "deadline", 0))) __PYX_ERR(5, 87, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_queue), __pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue, 0, "queue", 0))) __PYX_ERR(5, 87, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_7Channel_6watch_connectivity_state(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)__pyx_v_self), __pyx_v_last_observed_state, __pyx_v_deadline, __pyx_v_queue, __pyx_v_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_7Channel_6watch_connectivity_state(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self, grpc_connectivity_state __pyx_v_last_observed_state, struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue, PyObject *__pyx_v_tag) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *__pyx_v_operation_tag = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("watch_connectivity_state", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":88
 *       self, grpc_connectivity_state last_observed_state,
 *       Timespec deadline not None, CompletionQueue queue not None, tag):
 *     cdef OperationTag operation_tag = OperationTag(tag)             # <<<<<<<<<<<<<<
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_tag);
  __Pyx_GIVEREF(__pyx_v_tag);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_tag);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_OperationTag), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_operation_tag = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":89
 *       Timespec deadline not None, CompletionQueue queue not None, tag):
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     cpython.Py_INCREF(operation_tag)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_channel_watch_connectivity_state(
 */
  Py_INCREF(((PyObject *)__pyx_v_operation_tag));

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":90
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_channel_watch_connectivity_state(
 *           self.c_channel, last_observed_state, deadline.c_time,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":91
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:
 *       grpc_channel_watch_connectivity_state(             # <<<<<<<<<<<<<<
 *           self.c_channel, last_observed_state, deadline.c_time,
 *           queue.c_completion_queue, <cpython.PyObject *>operation_tag)
 */
        grpc_channel_watch_connectivity_state(__pyx_v_self->c_channel, __pyx_v_last_observed_state, __pyx_v_deadline->c_time, __pyx_v_queue->c_completion_queue, ((PyObject *)__pyx_v_operation_tag));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":90
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_channel_watch_connectivity_state(
 *           self.c_channel, last_observed_state, deadline.c_time,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":85
 *     return result
 * 
 *   def watch_connectivity_state(             # <<<<<<<<<<<<<<
 *       self, grpc_connectivity_state last_observed_state,
 *       Timespec deadline not None, CompletionQueue queue not None, tag):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.watch_connectivity_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_operation_tag);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":95
 *           queue.c_completion_queue, <cpython.PyObject *>operation_tag)
 * 
 *   def target(self):             # <<<<<<<<<<<<<<
 *     cdef char *target = NULL
 *     with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_9target(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_9target(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("target (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_7Channel_8target(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_7Channel_8target(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self) {
  char *__pyx_v_target;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("target", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":96
 * 
 *   def target(self):
 *     cdef char *target = NULL             # <<<<<<<<<<<<<<
 *     with nogil:
 *       target = grpc_channel_get_target(self.c_channel)
 */
  __pyx_v_target = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":97
 *   def target(self):
 *     cdef char *target = NULL
 *     with nogil:             # <<<<<<<<<<<<<<
 *       target = grpc_channel_get_target(self.c_channel)
 *     result = <bytes>target
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":98
 *     cdef char *target = NULL
 *     with nogil:
 *       target = grpc_channel_get_target(self.c_channel)             # <<<<<<<<<<<<<<
 *     result = <bytes>target
 *     with nogil:
 */
        __pyx_v_target = grpc_channel_get_target(__pyx_v_self->c_channel);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":97
 *   def target(self):
 *     cdef char *target = NULL
 *     with nogil:             # <<<<<<<<<<<<<<
 *       target = grpc_channel_get_target(self.c_channel)
 *     result = <bytes>target
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":99
 *     with nogil:
 *       target = grpc_channel_get_target(self.c_channel)
 *     result = <bytes>target             # <<<<<<<<<<<<<<
 *     with nogil:
 *       gpr_free(target)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_result = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":100
 *       target = grpc_channel_get_target(self.c_channel)
 *     result = <bytes>target
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_free(target)
 *     return result
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":101
 *     result = <bytes>target
 *     with nogil:
 *       gpr_free(target)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
        gpr_free(__pyx_v_target);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":100
 *       target = grpc_channel_get_target(self.c_channel)
 *     result = <bytes>target
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_free(target)
 *     return result
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":102
 *     with nogil:
 *       gpr_free(target)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":95
 *           queue.c_completion_queue, <cpython.PyObject *>operation_tag)
 * 
 *   def target(self):             # <<<<<<<<<<<<<<
 *     cdef char *target = NULL
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Channel.target", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":104
 *     return result
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_channel != NULL:
 *       grpc_channel_destroy(self.c_channel)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_7Channel_11__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_7Channel_11__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_7Channel_10__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_7Channel_10__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":105
 * 
 *   def __dealloc__(self):
 *     if self.c_channel != NULL:             # <<<<<<<<<<<<<<
 *       grpc_channel_destroy(self.c_channel)
 *     grpc_shutdown()
 */
  __pyx_t_1 = ((__pyx_v_self->c_channel != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":106
 *   def __dealloc__(self):
 *     if self.c_channel != NULL:
 *       grpc_channel_destroy(self.c_channel)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 */
    grpc_channel_destroy(__pyx_v_self->c_channel);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":105
 * 
 *   def __dealloc__(self):
 *     if self.c_channel != NULL:             # <<<<<<<<<<<<<<
 *       grpc_channel_destroy(self.c_channel)
 *     grpc_shutdown()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":107
 *     if self.c_channel != NULL:
 *       grpc_channel_destroy(self.c_channel)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":104
 *     return result
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_channel != NULL:
 *       grpc_channel_destroy(self.c_channel)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":35
 * cdef class ChannelCredentials:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.c_credentials = NULL
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":36
 * 
 *   def __cinit__(self):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.c_credentials = NULL
 *     self.c_ssl_pem_key_cert_pair.private_key = NULL
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":37
 *   def __cinit__(self):
 *     grpc_init()
 *     self.c_credentials = NULL             # <<<<<<<<<<<<<<
 *     self.c_ssl_pem_key_cert_pair.private_key = NULL
 *     self.c_ssl_pem_key_cert_pair.certificate_chain = NULL
 */
  __pyx_v_self->c_credentials = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":38
 *     grpc_init()
 *     self.c_credentials = NULL
 *     self.c_ssl_pem_key_cert_pair.private_key = NULL             # <<<<<<<<<<<<<<
 *     self.c_ssl_pem_key_cert_pair.certificate_chain = NULL
 *     self.references = []
 */
  __pyx_v_self->c_ssl_pem_key_cert_pair.private_key = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":39
 *     self.c_credentials = NULL
 *     self.c_ssl_pem_key_cert_pair.private_key = NULL
 *     self.c_ssl_pem_key_cert_pair.certificate_chain = NULL             # <<<<<<<<<<<<<<
 *     self.references = []
 * 
 */
  __pyx_v_self->c_ssl_pem_key_cert_pair.cert_chain = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":40
 *     self.c_ssl_pem_key_cert_pair.private_key = NULL
 *     self.c_ssl_pem_key_cert_pair.certificate_chain = NULL
 *     self.references = []             # <<<<<<<<<<<<<<
 * 
 *   # The object *can* be invalid in Python if we fail to make the credentials
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->references);
  __Pyx_DECREF(__pyx_v_self->references);
  __pyx_v_self->references = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":35
 * cdef class ChannelCredentials:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.c_credentials = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelCredentials.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":45
 *   # (and the core thus returns NULL credentials). Used primarily for debugging.
 *   @property
 *   def is_valid(self):             # <<<<<<<<<<<<<<
 *     return self.c_credentials != NULL
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_8is_valid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_8is_valid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials_8is_valid___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials_8is_valid___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":46
 *   @property
 *   def is_valid(self):
 *     return self.c_credentials != NULL             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->c_credentials != NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":45
 *   # (and the core thus returns NULL credentials). Used primarily for debugging.
 *   @property
 *   def is_valid(self):             # <<<<<<<<<<<<<<
 *     return self.c_credentials != NULL
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelCredentials.is_valid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":48
 *     return self.c_credentials != NULL
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_credentials != NULL:
 *       grpc_channel_credentials_release(self.c_credentials)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_18ChannelCredentials_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":49
 * 
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:             # <<<<<<<<<<<<<<
 *       grpc_channel_credentials_release(self.c_credentials)
 *     grpc_shutdown()
 */
  __pyx_t_1 = ((__pyx_v_self->c_credentials != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":50
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:
 *       grpc_channel_credentials_release(self.c_credentials)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
    grpc_channel_credentials_release(__pyx_v_self->c_credentials);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":49
 * 
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:             # <<<<<<<<<<<<<<
 *       grpc_channel_credentials_release(self.c_credentials)
 *     grpc_shutdown()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":51
 *     if self.c_credentials != NULL:
 *       grpc_channel_credentials_release(self.c_credentials)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":48
 *     return self.c_credentials != NULL
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_credentials != NULL:
 *       grpc_channel_credentials_release(self.c_credentials)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":56
 * cdef class CallCredentials:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.c_credentials = NULL
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":57
 * 
 *   def __cinit__(self):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.c_credentials = NULL
 *     self.references = []
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":58
 *   def __cinit__(self):
 *     grpc_init()
 *     self.c_credentials = NULL             # <<<<<<<<<<<<<<
 *     self.references = []
 * 
 */
  __pyx_v_self->c_credentials = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":59
 *     grpc_init()
 *     self.c_credentials = NULL
 *     self.references = []             # <<<<<<<<<<<<<<
 * 
 *   # The object *can* be invalid in Python if we fail to make the credentials
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->references);
  __Pyx_DECREF(__pyx_v_self->references);
  __pyx_v_self->references = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":56
 * cdef class CallCredentials:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.c_credentials = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CallCredentials.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":64
 *   # (and the core thus returns NULL credentials). Used primarily for debugging.
 *   @property
 *   def is_valid(self):             # <<<<<<<<<<<<<<
 *     return self.c_credentials != NULL
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_8is_valid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_8is_valid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials_8is_valid___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials_8is_valid___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":65
 *   @property
 *   def is_valid(self):
 *     return self.c_credentials != NULL             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->c_credentials != NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":64
 *   # (and the core thus returns NULL credentials). Used primarily for debugging.
 *   @property
 *   def is_valid(self):             # <<<<<<<<<<<<<<
 *     return self.c_credentials != NULL
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CallCredentials.is_valid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":67
 *     return self.c_credentials != NULL
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_credentials != NULL:
 *       grpc_call_credentials_release(self.c_credentials)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_15CallCredentials_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":68
 * 
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:             # <<<<<<<<<<<<<<
 *       grpc_call_credentials_release(self.c_credentials)
 *     grpc_shutdown()
 */
  __pyx_t_1 = ((__pyx_v_self->c_credentials != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":69
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:
 *       grpc_call_credentials_release(self.c_credentials)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
    grpc_call_credentials_release(__pyx_v_self->c_credentials);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":68
 * 
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:             # <<<<<<<<<<<<<<
 *       grpc_call_credentials_release(self.c_credentials)
 *     grpc_shutdown()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":70
 *     if self.c_credentials != NULL:
 *       grpc_call_credentials_release(self.c_credentials)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":67
 *     return self.c_credentials != NULL
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_credentials != NULL:
 *       grpc_call_credentials_release(self.c_credentials)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":75
 * cdef class ServerCredentials:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.c_credentials = NULL
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_17ServerCredentials_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_17ServerCredentials_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_17ServerCredentials___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_17ServerCredentials___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":76
 * 
 *   def __cinit__(self):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.c_credentials = NULL
 *     self.references = []
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":77
 *   def __cinit__(self):
 *     grpc_init()
 *     self.c_credentials = NULL             # <<<<<<<<<<<<<<
 *     self.references = []
 * 
 */
  __pyx_v_self->c_credentials = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":78
 *     grpc_init()
 *     self.c_credentials = NULL
 *     self.references = []             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->references);
  __Pyx_DECREF(__pyx_v_self->references);
  __pyx_v_self->references = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":75
 * cdef class ServerCredentials:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.c_credentials = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ServerCredentials.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":80
 *     self.references = []
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_credentials != NULL:
 *       grpc_server_credentials_release(self.c_credentials)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_17ServerCredentials_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_17ServerCredentials_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_17ServerCredentials_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_17ServerCredentials_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":81
 * 
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:             # <<<<<<<<<<<<<<
 *       grpc_server_credentials_release(self.c_credentials)
 *     grpc_shutdown()
 */
  __pyx_t_1 = ((__pyx_v_self->c_credentials != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":82
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:
 *       grpc_server_credentials_release(self.c_credentials)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
    grpc_server_credentials_release(__pyx_v_self->c_credentials);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":81
 * 
 *   def __dealloc__(self):
 *     if self.c_credentials != NULL:             # <<<<<<<<<<<<<<
 *       grpc_server_credentials_release(self.c_credentials)
 *     grpc_shutdown()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":83
 *     if self.c_credentials != NULL:
 *       grpc_server_credentials_release(self.c_credentials)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":80
 *     self.references = []
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_credentials != NULL:
 *       grpc_server_credentials_release(self.c_credentials)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":88
 * cdef class CredentialsMetadataPlugin:
 * 
 *   def __cinit__(self, object plugin_callback, bytes name):             # <<<<<<<<<<<<<<
 *     """
 *     Args:
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_plugin_callback = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_plugin_callback,&__pyx_n_s_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_plugin_callback)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(3, 88, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(3, 88, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_plugin_callback = values[0];
    __pyx_v_name = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 88, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.CredentialsMetadataPlugin.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyBytes_Type), 1, "name", 1))) __PYX_ERR(3, 88, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)__pyx_v_self), __pyx_v_plugin_callback, __pyx_v_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_self, PyObject *__pyx_v_plugin_callback, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":99
 *       name (bytes): Plugin name.
 *     """
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     if not callable(plugin_callback):
 *       raise ValueError('expected callable plugin_callback')
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":100
 *     """
 *     grpc_init()
 *     if not callable(plugin_callback):             # <<<<<<<<<<<<<<
 *       raise ValueError('expected callable plugin_callback')
 *     self.plugin_callback = plugin_callback
 */
  __pyx_t_1 = __Pyx_PyCallable_Check(__pyx_v_plugin_callback); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(3, 100, __pyx_L1_error)
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":101
 *     grpc_init()
 *     if not callable(plugin_callback):
 *       raise ValueError('expected callable plugin_callback')             # <<<<<<<<<<<<<<
 *     self.plugin_callback = plugin_callback
 *     self.plugin_name = name
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 101, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":100
 *     """
 *     grpc_init()
 *     if not callable(plugin_callback):             # <<<<<<<<<<<<<<
 *       raise ValueError('expected callable plugin_callback')
 *     self.plugin_callback = plugin_callback
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":102
 *     if not callable(plugin_callback):
 *       raise ValueError('expected callable plugin_callback')
 *     self.plugin_callback = plugin_callback             # <<<<<<<<<<<<<<
 *     self.plugin_name = name
 * 
 */
  __Pyx_INCREF(__pyx_v_plugin_callback);
  __Pyx_GIVEREF(__pyx_v_plugin_callback);
  __Pyx_GOTREF(__pyx_v_self->plugin_callback);
  __Pyx_DECREF(__pyx_v_self->plugin_callback);
  __pyx_v_self->plugin_callback = __pyx_v_plugin_callback;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":103
 *       raise ValueError('expected callable plugin_callback')
 *     self.plugin_callback = plugin_callback
 *     self.plugin_name = name             # <<<<<<<<<<<<<<
 * 
 *   @staticmethod
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->plugin_name);
  __Pyx_DECREF(__pyx_v_self->plugin_name);
  __pyx_v_self->plugin_name = __pyx_v_name;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":88
 * cdef class CredentialsMetadataPlugin:
 * 
 *   def __cinit__(self, object plugin_callback, bytes name):             # <<<<<<<<<<<<<<
 *     """
 *     Args:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CredentialsMetadataPlugin.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":106
 * 
 *   @staticmethod
 *   cdef grpc_metadata_credentials_plugin make_c_plugin(self):             # <<<<<<<<<<<<<<
 *     cdef grpc_metadata_credentials_plugin result
 *     result.get_metadata = plugin_get_metadata
 */

static grpc_metadata_credentials_plugin __pyx_f_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_make_c_plugin(struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_self) {
  grpc_metadata_credentials_plugin __pyx_v_result;
  grpc_metadata_credentials_plugin __pyx_r;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  __Pyx_RefNannySetupContext("make_c_plugin", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":108
 *   cdef grpc_metadata_credentials_plugin make_c_plugin(self):
 *     cdef grpc_metadata_credentials_plugin result
 *     result.get_metadata = plugin_get_metadata             # <<<<<<<<<<<<<<
 *     result.destroy = plugin_destroy_c_plugin_state
 *     result.state = <void *>self
 */
  __pyx_v_result.get_metadata = __pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":109
 *     cdef grpc_metadata_credentials_plugin result
 *     result.get_metadata = plugin_get_metadata
 *     result.destroy = plugin_destroy_c_plugin_state             # <<<<<<<<<<<<<<
 *     result.state = <void *>self
 *     result.type = self.plugin_name
 */
  __pyx_v_result.destroy = __pyx_f_4grpc_7_cython_6cygrpc_plugin_destroy_c_plugin_state;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":110
 *     result.get_metadata = plugin_get_metadata
 *     result.destroy = plugin_destroy_c_plugin_state
 *     result.state = <void *>self             # <<<<<<<<<<<<<<
 *     result.type = self.plugin_name
 *     cpython.Py_INCREF(self)
 */
  __pyx_v_result.state = ((void *)__pyx_v_self);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":111
 *     result.destroy = plugin_destroy_c_plugin_state
 *     result.state = <void *>self
 *     result.type = self.plugin_name             # <<<<<<<<<<<<<<
 *     cpython.Py_INCREF(self)
 *     return result
 */
  __pyx_t_1 = __Pyx_PyObject_AsString(__pyx_v_self->plugin_name); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(3, 111, __pyx_L1_error)
  __pyx_v_result.type = __pyx_t_1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":112
 *     result.state = <void *>self
 *     result.type = self.plugin_name
 *     cpython.Py_INCREF(self)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":113
 *     result.type = self.plugin_name
 *     cpython.Py_INCREF(self)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":106
 * 
 *   @staticmethod
 *   cdef grpc_metadata_credentials_plugin make_c_plugin(self):             # <<<<<<<<<<<<<<
 *     cdef grpc_metadata_credentials_plugin result
 *     result.get_metadata = plugin_get_metadata
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("grpc._cython.cygrpc.CredentialsMetadataPlugin.make_c_plugin", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":115
 *     return result
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":116
 * 
 *   def __dealloc__(self):
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":115
 *     return result
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":121
 * cdef class AuthMetadataContext:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.context.service_url = NULL
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":122
 * 
 *   def __cinit__(self):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.context.service_url = NULL
 *     self.context.method_name = NULL
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":123
 *   def __cinit__(self):
 *     grpc_init()
 *     self.context.service_url = NULL             # <<<<<<<<<<<<<<
 *     self.context.method_name = NULL
 * 
 */
  __pyx_v_self->context.service_url = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":124
 *     grpc_init()
 *     self.context.service_url = NULL
 *     self.context.method_name = NULL             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __pyx_v_self->context.method_name = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":121
 * cdef class AuthMetadataContext:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.context.service_url = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":127
 * 
 *   @property
 *   def service_url(self):             # <<<<<<<<<<<<<<
 *     return self.context.service_url
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11service_url_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11service_url_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11service_url___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11service_url___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":128
 *   @property
 *   def service_url(self):
 *     return self.context.service_url             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->context.service_url); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":127
 * 
 *   @property
 *   def service_url(self):             # <<<<<<<<<<<<<<
 *     return self.context.service_url
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.AuthMetadataContext.service_url.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":131
 * 
 *   @property
 *   def method_name(self):             # <<<<<<<<<<<<<<
 *     return self.context.method_name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11method_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11method_name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11method_name___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11method_name___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":132
 *   @property
 *   def method_name(self):
 *     return self.context.method_name             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->context.method_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":131
 * 
 *   @property
 *   def method_name(self):             # <<<<<<<<<<<<<<
 *     return self.context.method_name
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.AuthMetadataContext.method_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":134
 *     return self.context.method_name
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_19AuthMetadataContext_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":135
 * 
 *   def __dealloc__(self):
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":134
 *     return self.context.method_name
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":141
 *     void *state, grpc_auth_metadata_context context,
 *     grpc_credentials_plugin_metadata_cb cb, void *user_data) with gil:
 *   def python_callback(             # <<<<<<<<<<<<<<
 *       Metadata metadata, grpc_status_code status,
 *       bytes error_details):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19plugin_get_metadata_1python_callback(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_19plugin_get_metadata_1python_callback = {"python_callback", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_19plugin_get_metadata_1python_callback, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19plugin_get_metadata_1python_callback(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata = 0;
  grpc_status_code __pyx_v_status;
  PyObject *__pyx_v_error_details = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("python_callback (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_metadata,&__pyx_n_s_status,&__pyx_n_s_error_details,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_metadata)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_status)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("python_callback", 1, 3, 3, 1); __PYX_ERR(3, 141, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_error_details)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("python_callback", 1, 3, 3, 2); __PYX_ERR(3, 141, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "python_callback") < 0)) __PYX_ERR(3, 141, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)values[0]);
    __pyx_v_status = ((grpc_status_code)__Pyx_PyInt_As_grpc_status_code(values[1])); if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 142, __pyx_L3_error)
    __pyx_v_error_details = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("python_callback", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 141, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.plugin_get_metadata.python_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_metadata), __pyx_ptype_4grpc_7_cython_6cygrpc_Metadata, 1, "metadata", 0))) __PYX_ERR(3, 142, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_error_details), (&PyBytes_Type), 1, "error_details", 1))) __PYX_ERR(3, 143, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_19plugin_get_metadata_python_callback(__pyx_self, __pyx_v_metadata, __pyx_v_status, __pyx_v_error_details);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19plugin_get_metadata_python_callback(PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata, grpc_status_code __pyx_v_status, PyObject *__pyx_v_error_details) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata *__pyx_cur_scope;
  struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  __Pyx_RefNannySetupContext("python_callback", 0);
  __pyx_outer_scope = (struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":145
 *       bytes error_details):
 *     cb(user_data, metadata.c_metadata_array.metadata,
 *        metadata.c_metadata_array.count, status, error_details)             # <<<<<<<<<<<<<<
 *   cdef CredentialsMetadataPlugin self = <CredentialsMetadataPlugin>state
 *   cdef AuthMetadataContext cy_context = AuthMetadataContext()
 */
  __pyx_t_1 = __Pyx_PyObject_AsString(__pyx_v_error_details); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(3, 145, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":144
 *       Metadata metadata, grpc_status_code status,
 *       bytes error_details):
 *     cb(user_data, metadata.c_metadata_array.metadata,             # <<<<<<<<<<<<<<
 *        metadata.c_metadata_array.count, status, error_details)
 *   cdef CredentialsMetadataPlugin self = <CredentialsMetadataPlugin>state
 */
  __pyx_cur_scope->__pyx_v_cb(__pyx_cur_scope->__pyx_v_user_data, __pyx_v_metadata->c_metadata_array.metadata, __pyx_v_metadata->c_metadata_array.count, __pyx_v_status, __pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":141
 *     void *state, grpc_auth_metadata_context context,
 *     grpc_credentials_plugin_metadata_cb cb, void *user_data) with gil:
 *   def python_callback(             # <<<<<<<<<<<<<<
 *       Metadata metadata, grpc_status_code status,
 *       bytes error_details):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.plugin_get_metadata.python_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":138
 * 
 * 
 * cdef void plugin_get_metadata(             # <<<<<<<<<<<<<<
 *     void *state, grpc_auth_metadata_context context,
 *     grpc_credentials_plugin_metadata_cb cb, void *user_data) with gil:
 */

static void __pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata(void *__pyx_v_state, grpc_auth_metadata_context __pyx_v_context, grpc_credentials_plugin_metadata_cb __pyx_v_cb, void *__pyx_v_user_data) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata *__pyx_cur_scope;
  PyObject *__pyx_v_python_callback = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_self = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *__pyx_v_cy_context = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("plugin_get_metadata", 0);
  __pyx_cur_scope = (struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata *)__pyx_tp_new_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata(__pyx_ptype_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    #ifdef WITH_THREAD
    PyGILState_Release(__pyx_gilstate_save);
    #endif
    __Pyx_WriteUnraisable("grpc._cython.cygrpc.plugin_get_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
    return;
  }
  __Pyx_GOTREF(__pyx_cur_scope);
  __pyx_cur_scope->__pyx_v_cb = __pyx_v_cb;
  __pyx_cur_scope->__pyx_v_user_data = __pyx_v_user_data;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":141
 *     void *state, grpc_auth_metadata_context context,
 *     grpc_credentials_plugin_metadata_cb cb, void *user_data) with gil:
 *   def python_callback(             # <<<<<<<<<<<<<<
 *       Metadata metadata, grpc_status_code status,
 *       bytes error_details):
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_19plugin_get_metadata_1python_callback, 0, __pyx_n_s_plugin_get_metadata_locals_pytho, ((PyObject*)__pyx_cur_scope), __pyx_n_s_grpc__cython_cygrpc, __pyx_d, ((PyObject *)__pyx_codeobj__9)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_python_callback = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":146
 *     cb(user_data, metadata.c_metadata_array.metadata,
 *        metadata.c_metadata_array.count, status, error_details)
 *   cdef CredentialsMetadataPlugin self = <CredentialsMetadataPlugin>state             # <<<<<<<<<<<<<<
 *   cdef AuthMetadataContext cy_context = AuthMetadataContext()
 *   cy_context.context = context
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_state);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_self = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":147
 *        metadata.c_metadata_array.count, status, error_details)
 *   cdef CredentialsMetadataPlugin self = <CredentialsMetadataPlugin>state
 *   cdef AuthMetadataContext cy_context = AuthMetadataContext()             # <<<<<<<<<<<<<<
 *   cy_context.context = context
 *   self.plugin_callback(cy_context, python_callback)
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_AuthMetadataContext), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cy_context = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":148
 *   cdef CredentialsMetadataPlugin self = <CredentialsMetadataPlugin>state
 *   cdef AuthMetadataContext cy_context = AuthMetadataContext()
 *   cy_context.context = context             # <<<<<<<<<<<<<<
 *   self.plugin_callback(cy_context, python_callback)
 * 
 */
  __pyx_v_cy_context->context = __pyx_v_context;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":149
 *   cdef AuthMetadataContext cy_context = AuthMetadataContext()
 *   cy_context.context = context
 *   self.plugin_callback(cy_context, python_callback)             # <<<<<<<<<<<<<<
 * 
 * cdef void plugin_destroy_c_plugin_state(void *state) with gil:
 */
  __Pyx_INCREF(__pyx_v_self->plugin_callback);
  __pyx_t_2 = __pyx_v_self->plugin_callback; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_3) {
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(((PyObject *)__pyx_v_cy_context));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_cy_context));
  PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, ((PyObject *)__pyx_v_cy_context));
  __Pyx_INCREF(__pyx_v_python_callback);
  __Pyx_GIVEREF(__pyx_v_python_callback);
  PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_python_callback);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":138
 * 
 * 
 * cdef void plugin_get_metadata(             # <<<<<<<<<<<<<<
 *     void *state, grpc_auth_metadata_context context,
 *     grpc_credentials_plugin_metadata_cb cb, void *user_data) with gil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("grpc._cython.cygrpc.plugin_get_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_python_callback);
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XDECREF((PyObject *)__pyx_v_cy_context);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":151
 *   self.plugin_callback(cy_context, python_callback)
 * 
 * cdef void plugin_destroy_c_plugin_state(void *state) with gil:             # <<<<<<<<<<<<<<
 *   cpython.Py_DECREF(<CredentialsMetadataPlugin>state)
 * 
 */

static void __pyx_f_4grpc_7_cython_6cygrpc_plugin_destroy_c_plugin_state(void *__pyx_v_state) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("plugin_destroy_c_plugin_state", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":152
 * 
 * cdef void plugin_destroy_c_plugin_state(void *state) with gil:
 *   cpython.Py_DECREF(<CredentialsMetadataPlugin>state)             # <<<<<<<<<<<<<<
 * 
 * def channel_credentials_google_default():
 */
  Py_DECREF(((PyObject *)__pyx_v_state));

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":151
 *   self.plugin_callback(cy_context, python_callback)
 * 
 * cdef void plugin_destroy_c_plugin_state(void *state) with gil:             # <<<<<<<<<<<<<<
 *   cpython.Py_DECREF(<CredentialsMetadataPlugin>state)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":154
 *   cpython.Py_DECREF(<CredentialsMetadataPlugin>state)
 * 
 * def channel_credentials_google_default():             # <<<<<<<<<<<<<<
 *   cdef ChannelCredentials credentials = ChannelCredentials();
 *   with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_1channel_credentials_google_default(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_1channel_credentials_google_default = {"channel_credentials_google_default", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_1channel_credentials_google_default, METH_NOARGS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_1channel_credentials_google_default(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("channel_credentials_google_default (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_channel_credentials_google_default(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_channel_credentials_google_default(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_credentials = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("channel_credentials_google_default", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":155
 * 
 * def channel_credentials_google_default():
 *   cdef ChannelCredentials credentials = ChannelCredentials();             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_credentials = grpc_google_default_credentials_create()
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_ChannelCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":156
 * def channel_credentials_google_default():
 *   cdef ChannelCredentials credentials = ChannelCredentials();
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_google_default_credentials_create()
 *   return credentials
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":157
 *   cdef ChannelCredentials credentials = ChannelCredentials();
 *   with nogil:
 *     credentials.c_credentials = grpc_google_default_credentials_create()             # <<<<<<<<<<<<<<
 *   return credentials
 * 
 */
        __pyx_v_credentials->c_credentials = grpc_google_default_credentials_create();
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":156
 * def channel_credentials_google_default():
 *   cdef ChannelCredentials credentials = ChannelCredentials();
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_google_default_credentials_create()
 *   return credentials
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":158
 *   with nogil:
 *     credentials.c_credentials = grpc_google_default_credentials_create()
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def channel_credentials_ssl(pem_root_certificates,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":154
 *   cpython.Py_DECREF(<CredentialsMetadataPlugin>state)
 * 
 * def channel_credentials_google_default():             # <<<<<<<<<<<<<<
 *   cdef ChannelCredentials credentials = ChannelCredentials();
 *   with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.channel_credentials_google_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":160
 *   return credentials
 * 
 * def channel_credentials_ssl(pem_root_certificates,             # <<<<<<<<<<<<<<
 *                             SslPemKeyCertPair ssl_pem_key_cert_pair):
 *   pem_root_certificates = str_to_bytes(pem_root_certificates)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_3channel_credentials_ssl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_3channel_credentials_ssl = {"channel_credentials_ssl", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_3channel_credentials_ssl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_3channel_credentials_ssl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pem_root_certificates = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_ssl_pem_key_cert_pair = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("channel_credentials_ssl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pem_root_certificates,&__pyx_n_s_ssl_pem_key_cert_pair,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pem_root_certificates)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ssl_pem_key_cert_pair)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("channel_credentials_ssl", 1, 2, 2, 1); __PYX_ERR(3, 160, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "channel_credentials_ssl") < 0)) __PYX_ERR(3, 160, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_pem_root_certificates = values[0];
    __pyx_v_ssl_pem_key_cert_pair = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("channel_credentials_ssl", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 160, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.channel_credentials_ssl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ssl_pem_key_cert_pair), __pyx_ptype_4grpc_7_cython_6cygrpc_SslPemKeyCertPair, 1, "ssl_pem_key_cert_pair", 0))) __PYX_ERR(3, 161, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_2channel_credentials_ssl(__pyx_self, __pyx_v_pem_root_certificates, __pyx_v_ssl_pem_key_cert_pair);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_2channel_credentials_ssl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pem_root_certificates, struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_ssl_pem_key_cert_pair) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_credentials = 0;
  char const *__pyx_v_c_pem_root_certificates;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  char const *__pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("channel_credentials_ssl", 0);
  __Pyx_INCREF(__pyx_v_pem_root_certificates);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":162
 * def channel_credentials_ssl(pem_root_certificates,
 *                             SslPemKeyCertPair ssl_pem_key_cert_pair):
 *   pem_root_certificates = str_to_bytes(pem_root_certificates)             # <<<<<<<<<<<<<<
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   cdef const char *c_pem_root_certificates = NULL
 */
  __pyx_t_1 = __pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(__pyx_v_pem_root_certificates); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_pem_root_certificates, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":163
 *                             SslPemKeyCertPair ssl_pem_key_cert_pair):
 *   pem_root_certificates = str_to_bytes(pem_root_certificates)
 *   cdef ChannelCredentials credentials = ChannelCredentials()             # <<<<<<<<<<<<<<
 *   cdef const char *c_pem_root_certificates = NULL
 *   if pem_root_certificates is not None:
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_ChannelCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":164
 *   pem_root_certificates = str_to_bytes(pem_root_certificates)
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   cdef const char *c_pem_root_certificates = NULL             # <<<<<<<<<<<<<<
 *   if pem_root_certificates is not None:
 *     c_pem_root_certificates = pem_root_certificates
 */
  __pyx_v_c_pem_root_certificates = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":165
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   cdef const char *c_pem_root_certificates = NULL
 *   if pem_root_certificates is not None:             # <<<<<<<<<<<<<<
 *     c_pem_root_certificates = pem_root_certificates
 *     credentials.references.append(pem_root_certificates)
 */
  __pyx_t_2 = (__pyx_v_pem_root_certificates != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":166
 *   cdef const char *c_pem_root_certificates = NULL
 *   if pem_root_certificates is not None:
 *     c_pem_root_certificates = pem_root_certificates             # <<<<<<<<<<<<<<
 *     credentials.references.append(pem_root_certificates)
 *   if ssl_pem_key_cert_pair is not None:
 */
    __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_pem_root_certificates); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(3, 166, __pyx_L1_error)
    __pyx_v_c_pem_root_certificates = __pyx_t_4;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":167
 *   if pem_root_certificates is not None:
 *     c_pem_root_certificates = pem_root_certificates
 *     credentials.references.append(pem_root_certificates)             # <<<<<<<<<<<<<<
 *   if ssl_pem_key_cert_pair is not None:
 *     with nogil:
 */
    if (unlikely(__pyx_v_credentials->references == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
      __PYX_ERR(3, 167, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_credentials->references, __pyx_v_pem_root_certificates); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(3, 167, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":165
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   cdef const char *c_pem_root_certificates = NULL
 *   if pem_root_certificates is not None:             # <<<<<<<<<<<<<<
 *     c_pem_root_certificates = pem_root_certificates
 *     credentials.references.append(pem_root_certificates)
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":168
 *     c_pem_root_certificates = pem_root_certificates
 *     credentials.references.append(pem_root_certificates)
 *   if ssl_pem_key_cert_pair is not None:             # <<<<<<<<<<<<<<
 *     with nogil:
 *       credentials.c_credentials = grpc_ssl_credentials_create(
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_ssl_pem_key_cert_pair) != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":169
 *     credentials.references.append(pem_root_certificates)
 *   if ssl_pem_key_cert_pair is not None:
 *     with nogil:             # <<<<<<<<<<<<<<
 *       credentials.c_credentials = grpc_ssl_credentials_create(
 *           c_pem_root_certificates, &ssl_pem_key_cert_pair.c_pair, NULL)
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":170
 *   if ssl_pem_key_cert_pair is not None:
 *     with nogil:
 *       credentials.c_credentials = grpc_ssl_credentials_create(             # <<<<<<<<<<<<<<
 *           c_pem_root_certificates, &ssl_pem_key_cert_pair.c_pair, NULL)
 *     credentials.references.append(ssl_pem_key_cert_pair)
 */
          __pyx_v_credentials->c_credentials = grpc_ssl_credentials_create(__pyx_v_c_pem_root_certificates, (&__pyx_v_ssl_pem_key_cert_pair->c_pair), NULL);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":169
 *     credentials.references.append(pem_root_certificates)
 *   if ssl_pem_key_cert_pair is not None:
 *     with nogil:             # <<<<<<<<<<<<<<
 *       credentials.c_credentials = grpc_ssl_credentials_create(
 *           c_pem_root_certificates, &ssl_pem_key_cert_pair.c_pair, NULL)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L7;
          }
          __pyx_L7:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":172
 *       credentials.c_credentials = grpc_ssl_credentials_create(
 *           c_pem_root_certificates, &ssl_pem_key_cert_pair.c_pair, NULL)
 *     credentials.references.append(ssl_pem_key_cert_pair)             # <<<<<<<<<<<<<<
 *   else:
 *     with nogil:
 */
    if (unlikely(__pyx_v_credentials->references == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
      __PYX_ERR(3, 172, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_credentials->references, ((PyObject *)__pyx_v_ssl_pem_key_cert_pair)); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(3, 172, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":168
 *     c_pem_root_certificates = pem_root_certificates
 *     credentials.references.append(pem_root_certificates)
 *   if ssl_pem_key_cert_pair is not None:             # <<<<<<<<<<<<<<
 *     with nogil:
 *       credentials.c_credentials = grpc_ssl_credentials_create(
 */
    goto __pyx_L4;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":174
 *     credentials.references.append(ssl_pem_key_cert_pair)
 *   else:
 *     with nogil:             # <<<<<<<<<<<<<<
 *       credentials.c_credentials = grpc_ssl_credentials_create(
 *         c_pem_root_certificates, NULL, NULL)
 */
  /*else*/ {
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":175
 *   else:
 *     with nogil:
 *       credentials.c_credentials = grpc_ssl_credentials_create(             # <<<<<<<<<<<<<<
 *         c_pem_root_certificates, NULL, NULL)
 *   return credentials
 */
          __pyx_v_credentials->c_credentials = grpc_ssl_credentials_create(__pyx_v_c_pem_root_certificates, NULL, NULL);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":174
 *     credentials.references.append(ssl_pem_key_cert_pair)
 *   else:
 *     with nogil:             # <<<<<<<<<<<<<<
 *       credentials.c_credentials = grpc_ssl_credentials_create(
 *         c_pem_root_certificates, NULL, NULL)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L10;
          }
          __pyx_L10:;
        }
    }
  }
  __pyx_L4:;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":177
 *       credentials.c_credentials = grpc_ssl_credentials_create(
 *         c_pem_root_certificates, NULL, NULL)
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def channel_credentials_composite(
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":160
 *   return credentials
 * 
 * def channel_credentials_ssl(pem_root_certificates,             # <<<<<<<<<<<<<<
 *                             SslPemKeyCertPair ssl_pem_key_cert_pair):
 *   pem_root_certificates = str_to_bytes(pem_root_certificates)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.channel_credentials_ssl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XDECREF(__pyx_v_pem_root_certificates);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":179
 *   return credentials
 * 
 * def channel_credentials_composite(             # <<<<<<<<<<<<<<
 *     ChannelCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5channel_credentials_composite(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_5channel_credentials_composite = {"channel_credentials_composite", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_5channel_credentials_composite, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5channel_credentials_composite(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_credentials_1 = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_2 = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("channel_credentials_composite (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_credentials_1,&__pyx_n_s_credentials_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_credentials_1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_credentials_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("channel_credentials_composite", 1, 2, 2, 1); __PYX_ERR(3, 179, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "channel_credentials_composite") < 0)) __PYX_ERR(3, 179, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_credentials_1 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)values[0]);
    __pyx_v_credentials_2 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("channel_credentials_composite", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 179, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.channel_credentials_composite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_credentials_1), __pyx_ptype_4grpc_7_cython_6cygrpc_ChannelCredentials, 0, "credentials_1", 0))) __PYX_ERR(3, 180, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_credentials_2), __pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials, 0, "credentials_2", 0))) __PYX_ERR(3, 181, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_4channel_credentials_composite(__pyx_self, __pyx_v_credentials_1, __pyx_v_credentials_2);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_4channel_credentials_composite(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_credentials_1, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_2) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *__pyx_v_credentials = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("channel_credentials_composite", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":182
 *     ChannelCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 *   if not credentials_1.is_valid or not credentials_2.is_valid:             # <<<<<<<<<<<<<<
 *     raise ValueError("passed credentials must both be valid")
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_credentials_1), __pyx_n_s_is_valid); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_credentials_2), __pyx_n_s_is_valid); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":183
 *     CallCredentials credentials_2 not None):
 *   if not credentials_1.is_valid or not credentials_2.is_valid:
 *     raise ValueError("passed credentials must both be valid")             # <<<<<<<<<<<<<<
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   with nogil:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(3, 183, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":182
 *     ChannelCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 *   if not credentials_1.is_valid or not credentials_2.is_valid:             # <<<<<<<<<<<<<<
 *     raise ValueError("passed credentials must both be valid")
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":184
 *   if not credentials_1.is_valid or not credentials_2.is_valid:
 *     raise ValueError("passed credentials must both be valid")
 *   cdef ChannelCredentials credentials = ChannelCredentials()             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_credentials = grpc_composite_channel_credentials_create(
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_ChannelCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":185
 *     raise ValueError("passed credentials must both be valid")
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_composite_channel_credentials_create(
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":186
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   with nogil:
 *     credentials.c_credentials = grpc_composite_channel_credentials_create(             # <<<<<<<<<<<<<<
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 *   credentials.references.append(credentials_1)
 */
        __pyx_v_credentials->c_credentials = grpc_composite_channel_credentials_create(__pyx_v_credentials_1->c_credentials, __pyx_v_credentials_2->c_credentials, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":185
 *     raise ValueError("passed credentials must both be valid")
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_composite_channel_credentials_create(
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":188
 *     credentials.c_credentials = grpc_composite_channel_credentials_create(
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 *   credentials.references.append(credentials_1)             # <<<<<<<<<<<<<<
 *   credentials.references.append(credentials_2)
 *   return credentials
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 188, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_credentials->references, ((PyObject *)__pyx_v_credentials_1)); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(3, 188, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":189
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 *   credentials.references.append(credentials_1)
 *   credentials.references.append(credentials_2)             # <<<<<<<<<<<<<<
 *   return credentials
 * 
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 189, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_credentials->references, ((PyObject *)__pyx_v_credentials_2)); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(3, 189, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":190
 *   credentials.references.append(credentials_1)
 *   credentials.references.append(credentials_2)
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def call_credentials_composite(
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":179
 *   return credentials
 * 
 * def channel_credentials_composite(             # <<<<<<<<<<<<<<
 *     ChannelCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.channel_credentials_composite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":192
 *   return credentials
 * 
 * def call_credentials_composite(             # <<<<<<<<<<<<<<
 *     CallCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7call_credentials_composite(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_7call_credentials_composite = {"call_credentials_composite", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_7call_credentials_composite, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_7call_credentials_composite(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_1 = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_2 = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_credentials_composite (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_credentials_1,&__pyx_n_s_credentials_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_credentials_1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_credentials_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("call_credentials_composite", 1, 2, 2, 1); __PYX_ERR(3, 192, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "call_credentials_composite") < 0)) __PYX_ERR(3, 192, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_credentials_1 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)values[0]);
    __pyx_v_credentials_2 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_credentials_composite", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 192, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_composite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_credentials_1), __pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials, 0, "credentials_1", 0))) __PYX_ERR(3, 193, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_credentials_2), __pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials, 0, "credentials_2", 0))) __PYX_ERR(3, 194, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6call_credentials_composite(__pyx_self, __pyx_v_credentials_1, __pyx_v_credentials_2);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6call_credentials_composite(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_1, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials_2) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("call_credentials_composite", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":195
 *     CallCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 *   if not credentials_1.is_valid or not credentials_2.is_valid:             # <<<<<<<<<<<<<<
 *     raise ValueError("passed credentials must both be valid")
 *   cdef CallCredentials credentials = CallCredentials()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_credentials_1), __pyx_n_s_is_valid); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(3, 195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_credentials_2), __pyx_n_s_is_valid); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(3, 195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":196
 *     CallCredentials credentials_2 not None):
 *   if not credentials_1.is_valid or not credentials_2.is_valid:
 *     raise ValueError("passed credentials must both be valid")             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(3, 196, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":195
 *     CallCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 *   if not credentials_1.is_valid or not credentials_2.is_valid:             # <<<<<<<<<<<<<<
 *     raise ValueError("passed credentials must both be valid")
 *   cdef CallCredentials credentials = CallCredentials()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":197
 *   if not credentials_1.is_valid or not credentials_2.is_valid:
 *     raise ValueError("passed credentials must both be valid")
 *   cdef CallCredentials credentials = CallCredentials()             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_credentials = grpc_composite_call_credentials_create(
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":198
 *     raise ValueError("passed credentials must both be valid")
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_composite_call_credentials_create(
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":199
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:
 *     credentials.c_credentials = grpc_composite_call_credentials_create(             # <<<<<<<<<<<<<<
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 *   credentials.references.append(credentials_1)
 */
        __pyx_v_credentials->c_credentials = grpc_composite_call_credentials_create(__pyx_v_credentials_1->c_credentials, __pyx_v_credentials_2->c_credentials, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":198
 *     raise ValueError("passed credentials must both be valid")
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_composite_call_credentials_create(
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":201
 *     credentials.c_credentials = grpc_composite_call_credentials_create(
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 *   credentials.references.append(credentials_1)             # <<<<<<<<<<<<<<
 *   credentials.references.append(credentials_2)
 *   return credentials
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 201, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_credentials->references, ((PyObject *)__pyx_v_credentials_1)); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(3, 201, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":202
 *         credentials_1.c_credentials, credentials_2.c_credentials, NULL)
 *   credentials.references.append(credentials_1)
 *   credentials.references.append(credentials_2)             # <<<<<<<<<<<<<<
 *   return credentials
 * 
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 202, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_credentials->references, ((PyObject *)__pyx_v_credentials_2)); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(3, 202, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":203
 *   credentials.references.append(credentials_1)
 *   credentials.references.append(credentials_2)
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def call_credentials_google_compute_engine():
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":192
 *   return credentials
 * 
 * def call_credentials_composite(             # <<<<<<<<<<<<<<
 *     CallCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_composite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":205
 *   return credentials
 * 
 * def call_credentials_google_compute_engine():             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9call_credentials_google_compute_engine(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_9call_credentials_google_compute_engine = {"call_credentials_google_compute_engine", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_9call_credentials_google_compute_engine, METH_NOARGS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9call_credentials_google_compute_engine(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_credentials_google_compute_engine (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8call_credentials_google_compute_engine(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8call_credentials_google_compute_engine(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("call_credentials_google_compute_engine", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":206
 * 
 * def call_credentials_google_compute_engine():
 *   cdef CallCredentials credentials = CallCredentials()             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_credentials = (
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":207
 * def call_credentials_google_compute_engine():
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = (
 *         grpc_google_compute_engine_credentials_create(NULL))
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":208
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:
 *     credentials.c_credentials = (             # <<<<<<<<<<<<<<
 *         grpc_google_compute_engine_credentials_create(NULL))
 *   return credentials
 */
        __pyx_v_credentials->c_credentials = grpc_google_compute_engine_credentials_create(NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":207
 * def call_credentials_google_compute_engine():
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = (
 *         grpc_google_compute_engine_credentials_create(NULL))
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":210
 *     credentials.c_credentials = (
 *         grpc_google_compute_engine_credentials_create(NULL))
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def call_credentials_service_account_jwt_access(
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":205
 *   return credentials
 * 
 * def call_credentials_google_compute_engine():             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_google_compute_engine", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":212
 *   return credentials
 * 
 * def call_credentials_service_account_jwt_access(             # <<<<<<<<<<<<<<
 *     json_key, Timespec token_lifetime not None):
 *   json_key = str_to_bytes(json_key)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11call_credentials_service_account_jwt_access(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_11call_credentials_service_account_jwt_access = {"call_credentials_service_account_jwt_access", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_11call_credentials_service_account_jwt_access, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11call_credentials_service_account_jwt_access(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_json_key = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_token_lifetime = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_credentials_service_account_jwt_access (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_json_key,&__pyx_n_s_token_lifetime,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_json_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_token_lifetime)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("call_credentials_service_account_jwt_access", 1, 2, 2, 1); __PYX_ERR(3, 212, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "call_credentials_service_account_jwt_access") < 0)) __PYX_ERR(3, 212, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_json_key = values[0];
    __pyx_v_token_lifetime = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_credentials_service_account_jwt_access", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 212, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_service_account_jwt_access", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_token_lifetime), __pyx_ptype_4grpc_7_cython_6cygrpc_Timespec, 0, "token_lifetime", 0))) __PYX_ERR(3, 213, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10call_credentials_service_account_jwt_access(__pyx_self, __pyx_v_json_key, __pyx_v_token_lifetime);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10call_credentials_service_account_jwt_access(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_json_key, struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_token_lifetime) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials = 0;
  char *__pyx_v_json_key_c_string;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("call_credentials_service_account_jwt_access", 0);
  __Pyx_INCREF(__pyx_v_json_key);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":214
 * def call_credentials_service_account_jwt_access(
 *     json_key, Timespec token_lifetime not None):
 *   json_key = str_to_bytes(json_key)             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *json_key_c_string = json_key
 */
  __pyx_t_1 = __pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(__pyx_v_json_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_json_key, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":215
 *     json_key, Timespec token_lifetime not None):
 *   json_key = str_to_bytes(json_key)
 *   cdef CallCredentials credentials = CallCredentials()             # <<<<<<<<<<<<<<
 *   cdef char *json_key_c_string = json_key
 *   with nogil:
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":216
 *   json_key = str_to_bytes(json_key)
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *json_key_c_string = json_key             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_credentials = (
 */
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_v_json_key); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(3, 216, __pyx_L1_error)
  __pyx_v_json_key_c_string = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":217
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *json_key_c_string = json_key
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = (
 *         grpc_service_account_jwt_access_credentials_create(
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":218
 *   cdef char *json_key_c_string = json_key
 *   with nogil:
 *     credentials.c_credentials = (             # <<<<<<<<<<<<<<
 *         grpc_service_account_jwt_access_credentials_create(
 *             json_key_c_string, token_lifetime.c_time, NULL))
 */
        __pyx_v_credentials->c_credentials = grpc_service_account_jwt_access_credentials_create(__pyx_v_json_key_c_string, __pyx_v_token_lifetime->c_time, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":217
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *json_key_c_string = json_key
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = (
 *         grpc_service_account_jwt_access_credentials_create(
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":221
 *         grpc_service_account_jwt_access_credentials_create(
 *             json_key_c_string, token_lifetime.c_time, NULL))
 *   credentials.references.append(json_key)             # <<<<<<<<<<<<<<
 *   return credentials
 * 
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 221, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_credentials->references, __pyx_v_json_key); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(3, 221, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":222
 *             json_key_c_string, token_lifetime.c_time, NULL))
 *   credentials.references.append(json_key)
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def call_credentials_google_refresh_token(json_refresh_token):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":212
 *   return credentials
 * 
 * def call_credentials_service_account_jwt_access(             # <<<<<<<<<<<<<<
 *     json_key, Timespec token_lifetime not None):
 *   json_key = str_to_bytes(json_key)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_service_account_jwt_access", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XDECREF(__pyx_v_json_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":224
 *   return credentials
 * 
 * def call_credentials_google_refresh_token(json_refresh_token):             # <<<<<<<<<<<<<<
 *   json_refresh_token = str_to_bytes(json_refresh_token)
 *   cdef CallCredentials credentials = CallCredentials()
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_13call_credentials_google_refresh_token(PyObject *__pyx_self, PyObject *__pyx_v_json_refresh_token); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_13call_credentials_google_refresh_token = {"call_credentials_google_refresh_token", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_13call_credentials_google_refresh_token, METH_O, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_13call_credentials_google_refresh_token(PyObject *__pyx_self, PyObject *__pyx_v_json_refresh_token) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_credentials_google_refresh_token (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_12call_credentials_google_refresh_token(__pyx_self, ((PyObject *)__pyx_v_json_refresh_token));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_12call_credentials_google_refresh_token(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_json_refresh_token) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials = 0;
  char *__pyx_v_json_refresh_token_c_string;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("call_credentials_google_refresh_token", 0);
  __Pyx_INCREF(__pyx_v_json_refresh_token);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":225
 * 
 * def call_credentials_google_refresh_token(json_refresh_token):
 *   json_refresh_token = str_to_bytes(json_refresh_token)             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *json_refresh_token_c_string = json_refresh_token
 */
  __pyx_t_1 = __pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(__pyx_v_json_refresh_token); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_json_refresh_token, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":226
 * def call_credentials_google_refresh_token(json_refresh_token):
 *   json_refresh_token = str_to_bytes(json_refresh_token)
 *   cdef CallCredentials credentials = CallCredentials()             # <<<<<<<<<<<<<<
 *   cdef char *json_refresh_token_c_string = json_refresh_token
 *   with nogil:
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":227
 *   json_refresh_token = str_to_bytes(json_refresh_token)
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *json_refresh_token_c_string = json_refresh_token             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_credentials = grpc_google_refresh_token_credentials_create(
 */
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_v_json_refresh_token); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(3, 227, __pyx_L1_error)
  __pyx_v_json_refresh_token_c_string = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":228
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *json_refresh_token_c_string = json_refresh_token
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_google_refresh_token_credentials_create(
 *         json_refresh_token_c_string, NULL)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":229
 *   cdef char *json_refresh_token_c_string = json_refresh_token
 *   with nogil:
 *     credentials.c_credentials = grpc_google_refresh_token_credentials_create(             # <<<<<<<<<<<<<<
 *         json_refresh_token_c_string, NULL)
 *   credentials.references.append(json_refresh_token)
 */
        __pyx_v_credentials->c_credentials = grpc_google_refresh_token_credentials_create(__pyx_v_json_refresh_token_c_string, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":228
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *json_refresh_token_c_string = json_refresh_token
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_google_refresh_token_credentials_create(
 *         json_refresh_token_c_string, NULL)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":231
 *     credentials.c_credentials = grpc_google_refresh_token_credentials_create(
 *         json_refresh_token_c_string, NULL)
 *   credentials.references.append(json_refresh_token)             # <<<<<<<<<<<<<<
 *   return credentials
 * 
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 231, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_credentials->references, __pyx_v_json_refresh_token); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(3, 231, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":232
 *         json_refresh_token_c_string, NULL)
 *   credentials.references.append(json_refresh_token)
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def call_credentials_google_iam(authorization_token, authority_selector):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":224
 *   return credentials
 * 
 * def call_credentials_google_refresh_token(json_refresh_token):             # <<<<<<<<<<<<<<
 *   json_refresh_token = str_to_bytes(json_refresh_token)
 *   cdef CallCredentials credentials = CallCredentials()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_google_refresh_token", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XDECREF(__pyx_v_json_refresh_token);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":234
 *   return credentials
 * 
 * def call_credentials_google_iam(authorization_token, authority_selector):             # <<<<<<<<<<<<<<
 *   authorization_token = str_to_bytes(authorization_token)
 *   authority_selector = str_to_bytes(authority_selector)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15call_credentials_google_iam(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_15call_credentials_google_iam = {"call_credentials_google_iam", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_15call_credentials_google_iam, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15call_credentials_google_iam(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_authorization_token = 0;
  PyObject *__pyx_v_authority_selector = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_credentials_google_iam (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_authorization_token,&__pyx_n_s_authority_selector,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_authorization_token)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_authority_selector)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("call_credentials_google_iam", 1, 2, 2, 1); __PYX_ERR(3, 234, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "call_credentials_google_iam") < 0)) __PYX_ERR(3, 234, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_authorization_token = values[0];
    __pyx_v_authority_selector = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_credentials_google_iam", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 234, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_google_iam", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_14call_credentials_google_iam(__pyx_self, __pyx_v_authorization_token, __pyx_v_authority_selector);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_14call_credentials_google_iam(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_authorization_token, PyObject *__pyx_v_authority_selector) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials = 0;
  char *__pyx_v_authorization_token_c_string;
  char *__pyx_v_authority_selector_c_string;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("call_credentials_google_iam", 0);
  __Pyx_INCREF(__pyx_v_authorization_token);
  __Pyx_INCREF(__pyx_v_authority_selector);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":235
 * 
 * def call_credentials_google_iam(authorization_token, authority_selector):
 *   authorization_token = str_to_bytes(authorization_token)             # <<<<<<<<<<<<<<
 *   authority_selector = str_to_bytes(authority_selector)
 *   cdef CallCredentials credentials = CallCredentials()
 */
  __pyx_t_1 = __pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(__pyx_v_authorization_token); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_authorization_token, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":236
 * def call_credentials_google_iam(authorization_token, authority_selector):
 *   authorization_token = str_to_bytes(authorization_token)
 *   authority_selector = str_to_bytes(authority_selector)             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *authorization_token_c_string = authorization_token
 */
  __pyx_t_1 = __pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(__pyx_v_authority_selector); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_authority_selector, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":237
 *   authorization_token = str_to_bytes(authorization_token)
 *   authority_selector = str_to_bytes(authority_selector)
 *   cdef CallCredentials credentials = CallCredentials()             # <<<<<<<<<<<<<<
 *   cdef char *authorization_token_c_string = authorization_token
 *   cdef char *authority_selector_c_string = authority_selector
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":238
 *   authority_selector = str_to_bytes(authority_selector)
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *authorization_token_c_string = authorization_token             # <<<<<<<<<<<<<<
 *   cdef char *authority_selector_c_string = authority_selector
 *   with nogil:
 */
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_v_authorization_token); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(3, 238, __pyx_L1_error)
  __pyx_v_authorization_token_c_string = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":239
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef char *authorization_token_c_string = authorization_token
 *   cdef char *authority_selector_c_string = authority_selector             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_credentials = grpc_google_iam_credentials_create(
 */
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_v_authority_selector); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(3, 239, __pyx_L1_error)
  __pyx_v_authority_selector_c_string = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":240
 *   cdef char *authorization_token_c_string = authorization_token
 *   cdef char *authority_selector_c_string = authority_selector
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_google_iam_credentials_create(
 *         authorization_token_c_string, authority_selector_c_string, NULL)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":241
 *   cdef char *authority_selector_c_string = authority_selector
 *   with nogil:
 *     credentials.c_credentials = grpc_google_iam_credentials_create(             # <<<<<<<<<<<<<<
 *         authorization_token_c_string, authority_selector_c_string, NULL)
 *   credentials.references.append(authorization_token)
 */
        __pyx_v_credentials->c_credentials = grpc_google_iam_credentials_create(__pyx_v_authorization_token_c_string, __pyx_v_authority_selector_c_string, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":240
 *   cdef char *authorization_token_c_string = authorization_token
 *   cdef char *authority_selector_c_string = authority_selector
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = grpc_google_iam_credentials_create(
 *         authorization_token_c_string, authority_selector_c_string, NULL)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":243
 *     credentials.c_credentials = grpc_google_iam_credentials_create(
 *         authorization_token_c_string, authority_selector_c_string, NULL)
 *   credentials.references.append(authorization_token)             # <<<<<<<<<<<<<<
 *   credentials.references.append(authority_selector)
 *   return credentials
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 243, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_credentials->references, __pyx_v_authorization_token); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(3, 243, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":244
 *         authorization_token_c_string, authority_selector_c_string, NULL)
 *   credentials.references.append(authorization_token)
 *   credentials.references.append(authority_selector)             # <<<<<<<<<<<<<<
 *   return credentials
 * 
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 244, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_credentials->references, __pyx_v_authority_selector); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(3, 244, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":245
 *   credentials.references.append(authorization_token)
 *   credentials.references.append(authority_selector)
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def call_credentials_metadata_plugin(CredentialsMetadataPlugin plugin):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":234
 *   return credentials
 * 
 * def call_credentials_google_iam(authorization_token, authority_selector):             # <<<<<<<<<<<<<<
 *   authorization_token = str_to_bytes(authorization_token)
 *   authority_selector = str_to_bytes(authority_selector)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_google_iam", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XDECREF(__pyx_v_authorization_token);
  __Pyx_XDECREF(__pyx_v_authority_selector);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":247
 *   return credentials
 * 
 * def call_credentials_metadata_plugin(CredentialsMetadataPlugin plugin):             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17call_credentials_metadata_plugin(PyObject *__pyx_self, PyObject *__pyx_v_plugin); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_17call_credentials_metadata_plugin = {"call_credentials_metadata_plugin", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_17call_credentials_metadata_plugin, METH_O, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17call_credentials_metadata_plugin(PyObject *__pyx_self, PyObject *__pyx_v_plugin) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_credentials_metadata_plugin (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_plugin), __pyx_ptype_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin, 1, "plugin", 0))) __PYX_ERR(3, 247, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_16call_credentials_metadata_plugin(__pyx_self, ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)__pyx_v_plugin));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_16call_credentials_metadata_plugin(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *__pyx_v_plugin) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *__pyx_v_credentials = 0;
  grpc_metadata_credentials_plugin __pyx_v_c_plugin;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("call_credentials_metadata_plugin", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":248
 * 
 * def call_credentials_metadata_plugin(CredentialsMetadataPlugin plugin):
 *   cdef CallCredentials credentials = CallCredentials()             # <<<<<<<<<<<<<<
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()
 *   with nogil:
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":249
 * def call_credentials_metadata_plugin(CredentialsMetadataPlugin plugin):
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_credentials = (
 */
  __pyx_v_c_plugin = ((struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)__pyx_v_plugin->__pyx_vtab)->make_c_plugin(__pyx_v_plugin);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":250
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = (
 *         grpc_metadata_credentials_create_from_plugin(c_plugin, NULL))
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":251
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()
 *   with nogil:
 *     credentials.c_credentials = (             # <<<<<<<<<<<<<<
 *         grpc_metadata_credentials_create_from_plugin(c_plugin, NULL))
 *   # TODO(atash): the following held reference is *probably* never necessary
 */
        __pyx_v_credentials->c_credentials = grpc_metadata_credentials_create_from_plugin(__pyx_v_c_plugin, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":250
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_credentials = (
 *         grpc_metadata_credentials_create_from_plugin(c_plugin, NULL))
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":254
 *         grpc_metadata_credentials_create_from_plugin(c_plugin, NULL))
 *   # TODO(atash): the following held reference is *probably* never necessary
 *   credentials.references.append(plugin)             # <<<<<<<<<<<<<<
 *   return credentials
 * 
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 254, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_Append(__pyx_v_credentials->references, ((PyObject *)__pyx_v_plugin)); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(3, 254, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":255
 *   # TODO(atash): the following held reference is *probably* never necessary
 *   credentials.references.append(plugin)
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 * def server_credentials_ssl(pem_root_certs, pem_key_cert_pairs,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":247
 *   return credentials
 * 
 * def call_credentials_metadata_plugin(CredentialsMetadataPlugin plugin):             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.call_credentials_metadata_plugin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":257
 *   return credentials
 * 
 * def server_credentials_ssl(pem_root_certs, pem_key_cert_pairs,             # <<<<<<<<<<<<<<
 *                            bint force_client_auth):
 *   pem_root_certs = str_to_bytes(pem_root_certs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19server_credentials_ssl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_19server_credentials_ssl = {"server_credentials_ssl", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_19server_credentials_ssl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19server_credentials_ssl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pem_root_certs = 0;
  PyObject *__pyx_v_pem_key_cert_pairs = 0;
  int __pyx_v_force_client_auth;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("server_credentials_ssl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pem_root_certs,&__pyx_n_s_pem_key_cert_pairs,&__pyx_n_s_force_client_auth,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pem_root_certs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pem_key_cert_pairs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("server_credentials_ssl", 1, 3, 3, 1); __PYX_ERR(3, 257, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_force_client_auth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("server_credentials_ssl", 1, 3, 3, 2); __PYX_ERR(3, 257, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "server_credentials_ssl") < 0)) __PYX_ERR(3, 257, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_pem_root_certs = values[0];
    __pyx_v_pem_key_cert_pairs = values[1];
    __pyx_v_force_client_auth = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_force_client_auth == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 258, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("server_credentials_ssl", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 257, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.server_credentials_ssl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_18server_credentials_ssl(__pyx_self, __pyx_v_pem_root_certs, __pyx_v_pem_key_cert_pairs, __pyx_v_force_client_auth);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18server_credentials_ssl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pem_root_certs, PyObject *__pyx_v_pem_key_cert_pairs, int __pyx_v_force_client_auth) {
  char *__pyx_v_c_pem_root_certs;
  PyObject *__pyx_v_pair = NULL;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *__pyx_v_credentials = 0;
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  grpc_ssl_pem_key_cert_pair __pyx_t_11;
  grpc_ssl_client_certificate_request_type __pyx_t_12;
  __Pyx_RefNannySetupContext("server_credentials_ssl", 0);
  __Pyx_INCREF(__pyx_v_pem_root_certs);
  __Pyx_INCREF(__pyx_v_pem_key_cert_pairs);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":259
 * def server_credentials_ssl(pem_root_certs, pem_key_cert_pairs,
 *                            bint force_client_auth):
 *   pem_root_certs = str_to_bytes(pem_root_certs)             # <<<<<<<<<<<<<<
 *   cdef char *c_pem_root_certs = NULL
 *   if pem_root_certs is not None:
 */
  __pyx_t_1 = __pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(__pyx_v_pem_root_certs); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_pem_root_certs, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":260
 *                            bint force_client_auth):
 *   pem_root_certs = str_to_bytes(pem_root_certs)
 *   cdef char *c_pem_root_certs = NULL             # <<<<<<<<<<<<<<
 *   if pem_root_certs is not None:
 *     c_pem_root_certs = pem_root_certs
 */
  __pyx_v_c_pem_root_certs = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":261
 *   pem_root_certs = str_to_bytes(pem_root_certs)
 *   cdef char *c_pem_root_certs = NULL
 *   if pem_root_certs is not None:             # <<<<<<<<<<<<<<
 *     c_pem_root_certs = pem_root_certs
 *   pem_key_cert_pairs = list(pem_key_cert_pairs)
 */
  __pyx_t_2 = (__pyx_v_pem_root_certs != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":262
 *   cdef char *c_pem_root_certs = NULL
 *   if pem_root_certs is not None:
 *     c_pem_root_certs = pem_root_certs             # <<<<<<<<<<<<<<
 *   pem_key_cert_pairs = list(pem_key_cert_pairs)
 *   for pair in pem_key_cert_pairs:
 */
    __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_pem_root_certs); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(3, 262, __pyx_L1_error)
    __pyx_v_c_pem_root_certs = __pyx_t_4;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":261
 *   pem_root_certs = str_to_bytes(pem_root_certs)
 *   cdef char *c_pem_root_certs = NULL
 *   if pem_root_certs is not None:             # <<<<<<<<<<<<<<
 *     c_pem_root_certs = pem_root_certs
 *   pem_key_cert_pairs = list(pem_key_cert_pairs)
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":263
 *   if pem_root_certs is not None:
 *     c_pem_root_certs = pem_root_certs
 *   pem_key_cert_pairs = list(pem_key_cert_pairs)             # <<<<<<<<<<<<<<
 *   for pair in pem_key_cert_pairs:
 *     if not isinstance(pair, SslPemKeyCertPair):
 */
  __pyx_t_1 = PySequence_List(__pyx_v_pem_key_cert_pairs); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_pem_key_cert_pairs, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":264
 *     c_pem_root_certs = pem_root_certs
 *   pem_key_cert_pairs = list(pem_key_cert_pairs)
 *   for pair in pem_key_cert_pairs:             # <<<<<<<<<<<<<<
 *     if not isinstance(pair, SslPemKeyCertPair):
 *       raise TypeError("expected pem_key_cert_pairs to be sequence of "
 */
  if (likely(PyList_CheckExact(__pyx_v_pem_key_cert_pairs)) || PyTuple_CheckExact(__pyx_v_pem_key_cert_pairs)) {
    __pyx_t_1 = __pyx_v_pem_key_cert_pairs; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_pem_key_cert_pairs); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 264, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(3, 264, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 264, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(3, 264, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 264, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_1);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(3, 264, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_pair, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":265
 *   pem_key_cert_pairs = list(pem_key_cert_pairs)
 *   for pair in pem_key_cert_pairs:
 *     if not isinstance(pair, SslPemKeyCertPair):             # <<<<<<<<<<<<<<
 *       raise TypeError("expected pem_key_cert_pairs to be sequence of "
 *                       "SslPemKeyCertPair")
 */
    __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_pair, __pyx_ptype_4grpc_7_cython_6cygrpc_SslPemKeyCertPair); 
    __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
    if (__pyx_t_2) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":266
 *   for pair in pem_key_cert_pairs:
 *     if not isinstance(pair, SslPemKeyCertPair):
 *       raise TypeError("expected pem_key_cert_pairs to be sequence of "             # <<<<<<<<<<<<<<
 *                       "SslPemKeyCertPair")
 *   cdef ServerCredentials credentials = ServerCredentials()
 */
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(3, 266, __pyx_L1_error)

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":265
 *   pem_key_cert_pairs = list(pem_key_cert_pairs)
 *   for pair in pem_key_cert_pairs:
 *     if not isinstance(pair, SslPemKeyCertPair):             # <<<<<<<<<<<<<<
 *       raise TypeError("expected pem_key_cert_pairs to be sequence of "
 *                       "SslPemKeyCertPair")
 */
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":264
 *     c_pem_root_certs = pem_root_certs
 *   pem_key_cert_pairs = list(pem_key_cert_pairs)
 *   for pair in pem_key_cert_pairs:             # <<<<<<<<<<<<<<
 *     if not isinstance(pair, SslPemKeyCertPair):
 *       raise TypeError("expected pem_key_cert_pairs to be sequence of "
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":268
 *       raise TypeError("expected pem_key_cert_pairs to be sequence of "
 *                       "SslPemKeyCertPair")
 *   cdef ServerCredentials credentials = ServerCredentials()             # <<<<<<<<<<<<<<
 *   credentials.references.append(pem_key_cert_pairs)
 *   credentials.references.append(pem_root_certs)
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_ServerCredentials), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":269
 *                       "SslPemKeyCertPair")
 *   cdef ServerCredentials credentials = ServerCredentials()
 *   credentials.references.append(pem_key_cert_pairs)             # <<<<<<<<<<<<<<
 *   credentials.references.append(pem_root_certs)
 *   credentials.c_ssl_pem_key_cert_pairs_count = len(pem_key_cert_pairs)
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 269, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_credentials->references, __pyx_v_pem_key_cert_pairs); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(3, 269, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":270
 *   cdef ServerCredentials credentials = ServerCredentials()
 *   credentials.references.append(pem_key_cert_pairs)
 *   credentials.references.append(pem_root_certs)             # <<<<<<<<<<<<<<
 *   credentials.c_ssl_pem_key_cert_pairs_count = len(pem_key_cert_pairs)
 *   with nogil:
 */
  if (unlikely(__pyx_v_credentials->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(3, 270, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_credentials->references, __pyx_v_pem_root_certs); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(3, 270, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":271
 *   credentials.references.append(pem_key_cert_pairs)
 *   credentials.references.append(pem_root_certs)
 *   credentials.c_ssl_pem_key_cert_pairs_count = len(pem_key_cert_pairs)             # <<<<<<<<<<<<<<
 *   with nogil:
 *     credentials.c_ssl_pem_key_cert_pairs = (
 */
  __pyx_t_5 = PyObject_Length(__pyx_v_pem_key_cert_pairs); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(3, 271, __pyx_L1_error)
  __pyx_v_credentials->c_ssl_pem_key_cert_pairs_count = __pyx_t_5;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":272
 *   credentials.references.append(pem_root_certs)
 *   credentials.c_ssl_pem_key_cert_pairs_count = len(pem_key_cert_pairs)
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_ssl_pem_key_cert_pairs = (
 *         <grpc_ssl_pem_key_cert_pair *>gpr_malloc(
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":273
 *   credentials.c_ssl_pem_key_cert_pairs_count = len(pem_key_cert_pairs)
 *   with nogil:
 *     credentials.c_ssl_pem_key_cert_pairs = (             # <<<<<<<<<<<<<<
 *         <grpc_ssl_pem_key_cert_pair *>gpr_malloc(
 *             sizeof(grpc_ssl_pem_key_cert_pair) *
 */
        __pyx_v_credentials->c_ssl_pem_key_cert_pairs = ((grpc_ssl_pem_key_cert_pair *)gpr_malloc(((sizeof(grpc_ssl_pem_key_cert_pair)) * __pyx_v_credentials->c_ssl_pem_key_cert_pairs_count)));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":272
 *   credentials.references.append(pem_root_certs)
 *   credentials.c_ssl_pem_key_cert_pairs_count = len(pem_key_cert_pairs)
 *   with nogil:             # <<<<<<<<<<<<<<
 *     credentials.c_ssl_pem_key_cert_pairs = (
 *         <grpc_ssl_pem_key_cert_pair *>gpr_malloc(
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":278
 *                 credentials.c_ssl_pem_key_cert_pairs_count
 *         ))
 *   for i in range(credentials.c_ssl_pem_key_cert_pairs_count):             # <<<<<<<<<<<<<<
 *     credentials.c_ssl_pem_key_cert_pairs[i] = (
 *         (<SslPemKeyCertPair>pem_key_cert_pairs[i]).c_pair)
 */
  __pyx_t_9 = __pyx_v_credentials->c_ssl_pem_key_cert_pairs_count;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":280
 *   for i in range(credentials.c_ssl_pem_key_cert_pairs_count):
 *     credentials.c_ssl_pem_key_cert_pairs[i] = (
 *         (<SslPemKeyCertPair>pem_key_cert_pairs[i]).c_pair)             # <<<<<<<<<<<<<<
 *   credentials.c_credentials = grpc_ssl_server_credentials_create(
 *       c_pem_root_certs, credentials.c_ssl_pem_key_cert_pairs,
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_pem_key_cert_pairs, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)__pyx_t_1)->c_pair;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":279
 *         ))
 *   for i in range(credentials.c_ssl_pem_key_cert_pairs_count):
 *     credentials.c_ssl_pem_key_cert_pairs[i] = (             # <<<<<<<<<<<<<<
 *         (<SslPemKeyCertPair>pem_key_cert_pairs[i]).c_pair)
 *   credentials.c_credentials = grpc_ssl_server_credentials_create(
 */
    (__pyx_v_credentials->c_ssl_pem_key_cert_pairs[__pyx_v_i]) = __pyx_t_11;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":284
 *       c_pem_root_certs, credentials.c_ssl_pem_key_cert_pairs,
 *       credentials.c_ssl_pem_key_cert_pairs_count,
 *       GRPC_SSL_REQUEST_AND_REQUIRE_CLIENT_CERTIFICATE_AND_VERIFY if force_client_auth else GRPC_SSL_DONT_REQUEST_CLIENT_CERTIFICATE,             # <<<<<<<<<<<<<<
 *       NULL)
 *   return credentials
 */
  if ((__pyx_v_force_client_auth != 0)) {
    __pyx_t_12 = GRPC_SSL_REQUEST_AND_REQUIRE_CLIENT_CERTIFICATE_AND_VERIFY;
  } else {
    __pyx_t_12 = GRPC_SSL_DONT_REQUEST_CLIENT_CERTIFICATE;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":281
 *     credentials.c_ssl_pem_key_cert_pairs[i] = (
 *         (<SslPemKeyCertPair>pem_key_cert_pairs[i]).c_pair)
 *   credentials.c_credentials = grpc_ssl_server_credentials_create(             # <<<<<<<<<<<<<<
 *       c_pem_root_certs, credentials.c_ssl_pem_key_cert_pairs,
 *       credentials.c_ssl_pem_key_cert_pairs_count,
 */
  __pyx_v_credentials->c_credentials = grpc_ssl_server_credentials_create(__pyx_v_c_pem_root_certs, __pyx_v_credentials->c_ssl_pem_key_cert_pairs, __pyx_v_credentials->c_ssl_pem_key_cert_pairs_count, __pyx_t_12, NULL);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":286
 *       GRPC_SSL_REQUEST_AND_REQUIRE_CLIENT_CERTIFICATE_AND_VERIFY if force_client_auth else GRPC_SSL_DONT_REQUEST_CLIENT_CERTIFICATE,
 *       NULL)
 *   return credentials             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_credentials));
  __pyx_r = ((PyObject *)__pyx_v_credentials);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":257
 *   return credentials
 * 
 * def server_credentials_ssl(pem_root_certs, pem_key_cert_pairs,             # <<<<<<<<<<<<<<
 *                            bint force_client_auth):
 *   pem_root_certs = str_to_bytes(pem_root_certs)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("grpc._cython.cygrpc.server_credentials_ssl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pair);
  __Pyx_XDECREF((PyObject *)__pyx_v_credentials);
  __Pyx_XDECREF(__pyx_v_pem_root_certs);
  __Pyx_XDECREF(__pyx_v_pem_key_cert_pairs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":40
 * cdef class CompletionQueue:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     with nogil:
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":41
 * 
 *   def __cinit__(self):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     with nogil:
 *       self.c_completion_queue = grpc_completion_queue_create(NULL)
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":42
 *   def __cinit__(self):
 *     grpc_init()
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_completion_queue = grpc_completion_queue_create(NULL)
 *     self.is_shutting_down = False
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":43
 *     grpc_init()
 *     with nogil:
 *       self.c_completion_queue = grpc_completion_queue_create(NULL)             # <<<<<<<<<<<<<<
 *     self.is_shutting_down = False
 *     self.is_shutdown = False
 */
        __pyx_v_self->c_completion_queue = grpc_completion_queue_create(NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":42
 *   def __cinit__(self):
 *     grpc_init()
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_completion_queue = grpc_completion_queue_create(NULL)
 *     self.is_shutting_down = False
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":44
 *     with nogil:
 *       self.c_completion_queue = grpc_completion_queue_create(NULL)
 *     self.is_shutting_down = False             # <<<<<<<<<<<<<<
 *     self.is_shutdown = False
 * 
 */
  __pyx_v_self->is_shutting_down = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":45
 *       self.c_completion_queue = grpc_completion_queue_create(NULL)
 *     self.is_shutting_down = False
 *     self.is_shutdown = False             # <<<<<<<<<<<<<<
 * 
 *   cdef _interpret_event(self, grpc_event event):
 */
  __pyx_v_self->is_shutdown = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":40
 * cdef class CompletionQueue:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     with nogil:
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":47
 *     self.is_shutdown = False
 * 
 *   cdef _interpret_event(self, grpc_event event):             # <<<<<<<<<<<<<<
 *     cdef OperationTag tag = None
 *     cdef object user_tag = None
 */

static PyObject *__pyx_f_4grpc_7_cython_6cygrpc_15CompletionQueue__interpret_event(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self, grpc_event __pyx_v_event) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *__pyx_v_tag = 0;
  PyObject *__pyx_v_user_tag = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_operation_call = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_request_call_details = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_request_metadata = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_batch_operations = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_interpret_event", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":48
 * 
 *   cdef _interpret_event(self, grpc_event event):
 *     cdef OperationTag tag = None             # <<<<<<<<<<<<<<
 *     cdef object user_tag = None
 *     cdef Call operation_call = None
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_tag = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)Py_None);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":49
 *   cdef _interpret_event(self, grpc_event event):
 *     cdef OperationTag tag = None
 *     cdef object user_tag = None             # <<<<<<<<<<<<<<
 *     cdef Call operation_call = None
 *     cdef CallDetails request_call_details = None
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_user_tag = Py_None;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":50
 *     cdef OperationTag tag = None
 *     cdef object user_tag = None
 *     cdef Call operation_call = None             # <<<<<<<<<<<<<<
 *     cdef CallDetails request_call_details = None
 *     cdef Metadata request_metadata = None
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)Py_None);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":51
 *     cdef object user_tag = None
 *     cdef Call operation_call = None
 *     cdef CallDetails request_call_details = None             # <<<<<<<<<<<<<<
 *     cdef Metadata request_metadata = None
 *     cdef Operations batch_operations = None
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_request_call_details = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)Py_None);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":52
 *     cdef Call operation_call = None
 *     cdef CallDetails request_call_details = None
 *     cdef Metadata request_metadata = None             # <<<<<<<<<<<<<<
 *     cdef Operations batch_operations = None
 *     if event.type == GRPC_QUEUE_TIMEOUT:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_request_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":53
 *     cdef CallDetails request_call_details = None
 *     cdef Metadata request_metadata = None
 *     cdef Operations batch_operations = None             # <<<<<<<<<<<<<<
 *     if event.type == GRPC_QUEUE_TIMEOUT:
 *       return Event(
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_batch_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)Py_None);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":54
 *     cdef Metadata request_metadata = None
 *     cdef Operations batch_operations = None
 *     if event.type == GRPC_QUEUE_TIMEOUT:             # <<<<<<<<<<<<<<
 *       return Event(
 *           event.type, False, None, None, None, None, False, None)
 */
  switch (__pyx_v_event.type) {
    case GRPC_QUEUE_TIMEOUT:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":55
 *     cdef Operations batch_operations = None
 *     if event.type == GRPC_QUEUE_TIMEOUT:
 *       return Event(             # <<<<<<<<<<<<<<
 *           event.type, False, None, None, None, None, False, None)
 *     elif event.type == GRPC_QUEUE_SHUTDOWN:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":56
 *     if event.type == GRPC_QUEUE_TIMEOUT:
 *       return Event(
 *           event.type, False, None, None, None, None, False, None)             # <<<<<<<<<<<<<<
 *     elif event.type == GRPC_QUEUE_SHUTDOWN:
 *       self.is_shutdown = True
 */
    __pyx_t_1 = __Pyx_PyInt_From_grpc_completion_type(__pyx_v_event.type); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":55
 *     cdef Operations batch_operations = None
 *     if event.type == GRPC_QUEUE_TIMEOUT:
 *       return Event(             # <<<<<<<<<<<<<<
 *           event.type, False, None, None, None, None, False, None)
 *     elif event.type == GRPC_QUEUE_SHUTDOWN:
 */
    __pyx_t_2 = PyTuple_New(8); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    PyTuple_SET_ITEM(__pyx_t_2, 1, Py_False);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 3, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 4, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 5, Py_None);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    PyTuple_SET_ITEM(__pyx_t_2, 6, Py_False);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 7, Py_None);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Event), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":54
 *     cdef Metadata request_metadata = None
 *     cdef Operations batch_operations = None
 *     if event.type == GRPC_QUEUE_TIMEOUT:             # <<<<<<<<<<<<<<
 *       return Event(
 *           event.type, False, None, None, None, None, False, None)
 */
    break;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":57
 *       return Event(
 *           event.type, False, None, None, None, None, False, None)
 *     elif event.type == GRPC_QUEUE_SHUTDOWN:             # <<<<<<<<<<<<<<
 *       self.is_shutdown = True
 *       return Event(
 */
    case GRPC_QUEUE_SHUTDOWN:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":58
 *           event.type, False, None, None, None, None, False, None)
 *     elif event.type == GRPC_QUEUE_SHUTDOWN:
 *       self.is_shutdown = True             # <<<<<<<<<<<<<<
 *       return Event(
 *           event.type, True, None, None, None, None, False, None)
 */
    __pyx_v_self->is_shutdown = 1;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":59
 *     elif event.type == GRPC_QUEUE_SHUTDOWN:
 *       self.is_shutdown = True
 *       return Event(             # <<<<<<<<<<<<<<
 *           event.type, True, None, None, None, None, False, None)
 *     else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":60
 *       self.is_shutdown = True
 *       return Event(
 *           event.type, True, None, None, None, None, False, None)             # <<<<<<<<<<<<<<
 *     else:
 *       if event.tag != NULL:
 */
    __pyx_t_1 = __Pyx_PyInt_From_grpc_completion_type(__pyx_v_event.type); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":59
 *     elif event.type == GRPC_QUEUE_SHUTDOWN:
 *       self.is_shutdown = True
 *       return Event(             # <<<<<<<<<<<<<<
 *           event.type, True, None, None, None, None, False, None)
 *     else:
 */
    __pyx_t_2 = PyTuple_New(8); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_INCREF(Py_True);
    __Pyx_GIVEREF(Py_True);
    PyTuple_SET_ITEM(__pyx_t_2, 1, Py_True);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 3, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 4, Py_None);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 5, Py_None);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    PyTuple_SET_ITEM(__pyx_t_2, 6, Py_False);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 7, Py_None);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Event), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":57
 *       return Event(
 *           event.type, False, None, None, None, None, False, None)
 *     elif event.type == GRPC_QUEUE_SHUTDOWN:             # <<<<<<<<<<<<<<
 *       self.is_shutdown = True
 *       return Event(
 */
    break;
    default:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":62
 *           event.type, True, None, None, None, None, False, None)
 *     else:
 *       if event.tag != NULL:             # <<<<<<<<<<<<<<
 *         tag = <OperationTag>event.tag
 *         # We receive event tags only after they've been inc-ref'd elsewhere in
 */
    __pyx_t_3 = ((__pyx_v_event.tag != NULL) != 0);
    if (__pyx_t_3) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":63
 *     else:
 *       if event.tag != NULL:
 *         tag = <OperationTag>event.tag             # <<<<<<<<<<<<<<
 *         # We receive event tags only after they've been inc-ref'd elsewhere in
 *         # the code.
 */
      __pyx_t_1 = ((PyObject *)__pyx_v_event.tag);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_tag, ((struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":66
 *         # We receive event tags only after they've been inc-ref'd elsewhere in
 *         # the code.
 *         cpython.Py_DECREF(tag)             # <<<<<<<<<<<<<<
 *         if tag.shutting_down_server is not None:
 *           tag.shutting_down_server.notify_shutdown_complete()
 */
      Py_DECREF(((PyObject *)__pyx_v_tag));

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":67
 *         # the code.
 *         cpython.Py_DECREF(tag)
 *         if tag.shutting_down_server is not None:             # <<<<<<<<<<<<<<
 *           tag.shutting_down_server.notify_shutdown_complete()
 *         user_tag = tag.user_tag
 */
      __pyx_t_3 = (((PyObject *)__pyx_v_tag->shutting_down_server) != Py_None);
      __pyx_t_4 = (__pyx_t_3 != 0);
      if (__pyx_t_4) {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":68
 *         cpython.Py_DECREF(tag)
 *         if tag.shutting_down_server is not None:
 *           tag.shutting_down_server.notify_shutdown_complete()             # <<<<<<<<<<<<<<
 *         user_tag = tag.user_tag
 *         operation_call = tag.operation_call
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_Server *)__pyx_v_tag->shutting_down_server->__pyx_vtab)->notify_shutdown_complete(__pyx_v_tag->shutting_down_server); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 68, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":67
 *         # the code.
 *         cpython.Py_DECREF(tag)
 *         if tag.shutting_down_server is not None:             # <<<<<<<<<<<<<<
 *           tag.shutting_down_server.notify_shutdown_complete()
 *         user_tag = tag.user_tag
 */
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":69
 *         if tag.shutting_down_server is not None:
 *           tag.shutting_down_server.notify_shutdown_complete()
 *         user_tag = tag.user_tag             # <<<<<<<<<<<<<<
 *         operation_call = tag.operation_call
 *         request_call_details = tag.request_call_details
 */
      __pyx_t_1 = __pyx_v_tag->user_tag;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_user_tag, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":70
 *           tag.shutting_down_server.notify_shutdown_complete()
 *         user_tag = tag.user_tag
 *         operation_call = tag.operation_call             # <<<<<<<<<<<<<<
 *         request_call_details = tag.request_call_details
 *         request_metadata = tag.request_metadata
 */
      __pyx_t_1 = ((PyObject *)__pyx_v_tag->operation_call);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_operation_call, ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":71
 *         user_tag = tag.user_tag
 *         operation_call = tag.operation_call
 *         request_call_details = tag.request_call_details             # <<<<<<<<<<<<<<
 *         request_metadata = tag.request_metadata
 *         batch_operations = tag.batch_operations
 */
      __pyx_t_1 = ((PyObject *)__pyx_v_tag->request_call_details);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_request_call_details, ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":72
 *         operation_call = tag.operation_call
 *         request_call_details = tag.request_call_details
 *         request_metadata = tag.request_metadata             # <<<<<<<<<<<<<<
 *         batch_operations = tag.batch_operations
 *         if tag.is_new_request:
 */
      __pyx_t_1 = ((PyObject *)__pyx_v_tag->request_metadata);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_request_metadata, ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":73
 *         request_call_details = tag.request_call_details
 *         request_metadata = tag.request_metadata
 *         batch_operations = tag.batch_operations             # <<<<<<<<<<<<<<
 *         if tag.is_new_request:
 *           # Stuff in the tag not explicitly handled by us needs to live through
 */
      __pyx_t_1 = ((PyObject *)__pyx_v_tag->batch_operations);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_batch_operations, ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":74
 *         request_metadata = tag.request_metadata
 *         batch_operations = tag.batch_operations
 *         if tag.is_new_request:             # <<<<<<<<<<<<<<
 *           # Stuff in the tag not explicitly handled by us needs to live through
 *           # the life of the call
 */
      __pyx_t_4 = (__pyx_v_tag->is_new_request != 0);
      if (__pyx_t_4) {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":77
 *           # Stuff in the tag not explicitly handled by us needs to live through
 *           # the life of the call
 *           operation_call.references.extend(tag.references)             # <<<<<<<<<<<<<<
 *       return Event(
 *           event.type, event.success, user_tag, operation_call,
 */
        if (unlikely(__pyx_v_operation_call->references == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "extend");
          __PYX_ERR(6, 77, __pyx_L1_error)
        }
        __pyx_t_1 = __pyx_v_tag->references;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyList_Extend(__pyx_v_operation_call->references, __pyx_t_1); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(6, 77, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":74
 *         request_metadata = tag.request_metadata
 *         batch_operations = tag.batch_operations
 *         if tag.is_new_request:             # <<<<<<<<<<<<<<
 *           # Stuff in the tag not explicitly handled by us needs to live through
 *           # the life of the call
 */
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":62
 *           event.type, True, None, None, None, None, False, None)
 *     else:
 *       if event.tag != NULL:             # <<<<<<<<<<<<<<
 *         tag = <OperationTag>event.tag
 *         # We receive event tags only after they've been inc-ref'd elsewhere in
 */
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":78
 *           # the life of the call
 *           operation_call.references.extend(tag.references)
 *       return Event(             # <<<<<<<<<<<<<<
 *           event.type, event.success, user_tag, operation_call,
 *           request_call_details, request_metadata, tag.is_new_request,
 */
    __Pyx_XDECREF(__pyx_r);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":79
 *           operation_call.references.extend(tag.references)
 *       return Event(
 *           event.type, event.success, user_tag, operation_call,             # <<<<<<<<<<<<<<
 *           request_call_details, request_metadata, tag.is_new_request,
 *           batch_operations)
 */
    __pyx_t_1 = __Pyx_PyInt_From_grpc_completion_type(__pyx_v_event.type); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_event.success); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":80
 *       return Event(
 *           event.type, event.success, user_tag, operation_call,
 *           request_call_details, request_metadata, tag.is_new_request,             # <<<<<<<<<<<<<<
 *           batch_operations)
 * 
 */
    __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_tag->is_new_request); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":78
 *           # the life of the call
 *           operation_call.references.extend(tag.references)
 *       return Event(             # <<<<<<<<<<<<<<
 *           event.type, event.success, user_tag, operation_call,
 *           request_call_details, request_metadata, tag.is_new_request,
 */
    __pyx_t_7 = PyTuple_New(8); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_user_tag);
    __Pyx_GIVEREF(__pyx_v_user_tag);
    PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_user_tag);
    __Pyx_INCREF(((PyObject *)__pyx_v_operation_call));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_operation_call));
    PyTuple_SET_ITEM(__pyx_t_7, 3, ((PyObject *)__pyx_v_operation_call));
    __Pyx_INCREF(((PyObject *)__pyx_v_request_call_details));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_request_call_details));
    PyTuple_SET_ITEM(__pyx_t_7, 4, ((PyObject *)__pyx_v_request_call_details));
    __Pyx_INCREF(((PyObject *)__pyx_v_request_metadata));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_request_metadata));
    PyTuple_SET_ITEM(__pyx_t_7, 5, ((PyObject *)__pyx_v_request_metadata));
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_t_6);
    __Pyx_INCREF(((PyObject *)__pyx_v_batch_operations));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_batch_operations));
    PyTuple_SET_ITEM(__pyx_t_7, 7, ((PyObject *)__pyx_v_batch_operations));
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Event), __pyx_t_7, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
    break;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":47
 *     self.is_shutdown = False
 * 
 *   cdef _interpret_event(self, grpc_event event):             # <<<<<<<<<<<<<<
 *     cdef OperationTag tag = None
 *     cdef object user_tag = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompletionQueue._interpret_event", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_tag);
  __Pyx_XDECREF(__pyx_v_user_tag);
  __Pyx_XDECREF((PyObject *)__pyx_v_operation_call);
  __Pyx_XDECREF((PyObject *)__pyx_v_request_call_details);
  __Pyx_XDECREF((PyObject *)__pyx_v_request_metadata);
  __Pyx_XDECREF((PyObject *)__pyx_v_batch_operations);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":83
 *           batch_operations)
 * 
 *   def poll(self, Timespec deadline=None):             # <<<<<<<<<<<<<<
 *     # We name this 'poll' to avoid problems with CPython's expectations for
 *     # 'special' methods (like next and __next__).
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_3poll(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_3poll(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("poll (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deadline,0};
    PyObject* values[1] = {0};
    values[0] = (PyObject *)((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_deadline);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "poll") < 0)) __PYX_ERR(6, 83, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_deadline = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("poll", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 83, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompletionQueue.poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_deadline), __pyx_ptype_4grpc_7_cython_6cygrpc_Timespec, 1, "deadline", 0))) __PYX_ERR(6, 83, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_2poll(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_self), __pyx_v_deadline);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_2poll(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_deadline) {
  gpr_timespec __pyx_v_c_increment;
  gpr_timespec __pyx_v_c_timeout;
  gpr_timespec __pyx_v_c_deadline;
  grpc_event __pyx_v_event;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  gpr_timespec __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("poll", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":89
 *     cdef gpr_timespec c_timeout
 *     cdef gpr_timespec c_deadline
 *     with nogil:             # <<<<<<<<<<<<<<
 *       c_increment = gpr_time_from_millis(_INTERRUPT_CHECK_PERIOD_MS, GPR_TIMESPAN)
 *       c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":90
 *     cdef gpr_timespec c_deadline
 *     with nogil:
 *       c_increment = gpr_time_from_millis(_INTERRUPT_CHECK_PERIOD_MS, GPR_TIMESPAN)             # <<<<<<<<<<<<<<
 *       c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       if deadline is not None:
 */
        __pyx_v_c_increment = gpr_time_from_millis(__pyx_v_4grpc_7_cython_6cygrpc__INTERRUPT_CHECK_PERIOD_MS, GPR_TIMESPAN);

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":91
 *     with nogil:
 *       c_increment = gpr_time_from_millis(_INTERRUPT_CHECK_PERIOD_MS, GPR_TIMESPAN)
 *       c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)             # <<<<<<<<<<<<<<
 *       if deadline is not None:
 *         c_deadline = deadline.c_time
 */
        __pyx_v_c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME);

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":92
 *       c_increment = gpr_time_from_millis(_INTERRUPT_CHECK_PERIOD_MS, GPR_TIMESPAN)
 *       c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       if deadline is not None:             # <<<<<<<<<<<<<<
 *         c_deadline = deadline.c_time
 * 
 */
        __pyx_t_1 = (((PyObject *)__pyx_v_deadline) != Py_None);
        __pyx_t_2 = (__pyx_t_1 != 0);
        if (__pyx_t_2) {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":93
 *       c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       if deadline is not None:
 *         c_deadline = deadline.c_time             # <<<<<<<<<<<<<<
 * 
 *       while True:
 */
          __pyx_t_3 = __pyx_v_deadline->c_time;
          __pyx_v_c_deadline = __pyx_t_3;

          /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":92
 *       c_increment = gpr_time_from_millis(_INTERRUPT_CHECK_PERIOD_MS, GPR_TIMESPAN)
 *       c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       if deadline is not None:             # <<<<<<<<<<<<<<
 *         c_deadline = deadline.c_time
 * 
 */
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":95
 *         c_deadline = deadline.c_time
 * 
 *       while True:             # <<<<<<<<<<<<<<
 *         c_timeout = gpr_time_add(gpr_now(GPR_CLOCK_REALTIME), c_increment)
 *         if gpr_time_cmp(c_timeout, c_deadline) > 0:
 */
        while (1) {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":96
 * 
 *       while True:
 *         c_timeout = gpr_time_add(gpr_now(GPR_CLOCK_REALTIME), c_increment)             # <<<<<<<<<<<<<<
 *         if gpr_time_cmp(c_timeout, c_deadline) > 0:
 *           c_timeout = c_deadline
 */
          __pyx_v_c_timeout = gpr_time_add(gpr_now(GPR_CLOCK_REALTIME), __pyx_v_c_increment);

          /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":97
 *       while True:
 *         c_timeout = gpr_time_add(gpr_now(GPR_CLOCK_REALTIME), c_increment)
 *         if gpr_time_cmp(c_timeout, c_deadline) > 0:             # <<<<<<<<<<<<<<
 *           c_timeout = c_deadline
 *         event = grpc_completion_queue_next(
 */
          __pyx_t_2 = ((gpr_time_cmp(__pyx_v_c_timeout, __pyx_v_c_deadline) > 0) != 0);
          if (__pyx_t_2) {

            /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":98
 *         c_timeout = gpr_time_add(gpr_now(GPR_CLOCK_REALTIME), c_increment)
 *         if gpr_time_cmp(c_timeout, c_deadline) > 0:
 *           c_timeout = c_deadline             # <<<<<<<<<<<<<<
 *         event = grpc_completion_queue_next(
 *           self.c_completion_queue, c_timeout, NULL)
 */
            __pyx_v_c_timeout = __pyx_v_c_deadline;

            /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":97
 *       while True:
 *         c_timeout = gpr_time_add(gpr_now(GPR_CLOCK_REALTIME), c_increment)
 *         if gpr_time_cmp(c_timeout, c_deadline) > 0:             # <<<<<<<<<<<<<<
 *           c_timeout = c_deadline
 *         event = grpc_completion_queue_next(
 */
          }

          /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":99
 *         if gpr_time_cmp(c_timeout, c_deadline) > 0:
 *           c_timeout = c_deadline
 *         event = grpc_completion_queue_next(             # <<<<<<<<<<<<<<
 *           self.c_completion_queue, c_timeout, NULL)
 *         if event.type != GRPC_QUEUE_TIMEOUT or gpr_time_cmp(c_timeout, c_deadline) == 0:
 */
          __pyx_v_event = grpc_completion_queue_next(__pyx_v_self->c_completion_queue, __pyx_v_c_timeout, NULL);

          /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":101
 *         event = grpc_completion_queue_next(
 *           self.c_completion_queue, c_timeout, NULL)
 *         if event.type != GRPC_QUEUE_TIMEOUT or gpr_time_cmp(c_timeout, c_deadline) == 0:             # <<<<<<<<<<<<<<
 *           break;
 * 
 */
          __pyx_t_1 = ((__pyx_v_event.type != GRPC_QUEUE_TIMEOUT) != 0);
          if (!__pyx_t_1) {
          } else {
            __pyx_t_2 = __pyx_t_1;
            goto __pyx_L11_bool_binop_done;
          }
          __pyx_t_1 = ((gpr_time_cmp(__pyx_v_c_timeout, __pyx_v_c_deadline) == 0) != 0);
          __pyx_t_2 = __pyx_t_1;
          __pyx_L11_bool_binop_done:;
          if (__pyx_t_2) {

            /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":102
 *           self.c_completion_queue, c_timeout, NULL)
 *         if event.type != GRPC_QUEUE_TIMEOUT or gpr_time_cmp(c_timeout, c_deadline) == 0:
 *           break;             # <<<<<<<<<<<<<<
 * 
 *         # Handle any signals
 */
            goto __pyx_L8_break;

            /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":101
 *         event = grpc_completion_queue_next(
 *           self.c_completion_queue, c_timeout, NULL)
 *         if event.type != GRPC_QUEUE_TIMEOUT or gpr_time_cmp(c_timeout, c_deadline) == 0:             # <<<<<<<<<<<<<<
 *           break;
 * 
 */
          }

          /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":105
 * 
 *         # Handle any signals
 *         with gil:             # <<<<<<<<<<<<<<
 *           cpython.PyErr_CheckSignals()
 *     return self._interpret_event(event)
 */
          {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();
              #endif
              /*try:*/ {

                /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":106
 *         # Handle any signals
 *         with gil:
 *           cpython.PyErr_CheckSignals()             # <<<<<<<<<<<<<<
 *     return self._interpret_event(event)
 * 
 */
                __pyx_t_4 = PyErr_CheckSignals(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(6, 106, __pyx_L16_error)
              }

              /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":105
 * 
 *         # Handle any signals
 *         with gil:             # <<<<<<<<<<<<<<
 *           cpython.PyErr_CheckSignals()
 *     return self._interpret_event(event)
 */
              /*finally:*/ {
                /*normal exit:*/{
                  #ifdef WITH_THREAD
                  PyGILState_Release(__pyx_gilstate_save);
                  #endif
                  goto __pyx_L17;
                }
                __pyx_L16_error: {
                  #ifdef WITH_THREAD
                  PyGILState_Release(__pyx_gilstate_save);
                  #endif
                  goto __pyx_L4_error;
                }
                __pyx_L17:;
              }
          }
        }
        __pyx_L8_break:;
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":89
 *     cdef gpr_timespec c_timeout
 *     cdef gpr_timespec c_deadline
 *     with nogil:             # <<<<<<<<<<<<<<
 *       c_increment = gpr_time_from_millis(_INTERRUPT_CHECK_PERIOD_MS, GPR_TIMESPAN)
 *       c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":107
 *         with gil:
 *           cpython.PyErr_CheckSignals()
 *     return self._interpret_event(event)             # <<<<<<<<<<<<<<
 * 
 *   def shutdown(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_self->__pyx_vtab)->_interpret_event(__pyx_v_self, __pyx_v_event); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":83
 *           batch_operations)
 * 
 *   def poll(self, Timespec deadline=None):             # <<<<<<<<<<<<<<
 *     # We name this 'poll' to avoid problems with CPython's expectations for
 *     # 'special' methods (like next and __next__).
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompletionQueue.poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":109
 *     return self._interpret_event(event)
 * 
 *   def shutdown(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_completion_queue_shutdown(self.c_completion_queue)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_5shutdown(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_5shutdown(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shutdown (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_4shutdown(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_4shutdown(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shutdown", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":110
 * 
 *   def shutdown(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_completion_queue_shutdown(self.c_completion_queue)
 *     self.is_shutting_down = True
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":111
 *   def shutdown(self):
 *     with nogil:
 *       grpc_completion_queue_shutdown(self.c_completion_queue)             # <<<<<<<<<<<<<<
 *     self.is_shutting_down = True
 * 
 */
        grpc_completion_queue_shutdown(__pyx_v_self->c_completion_queue);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":110
 * 
 *   def shutdown(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_completion_queue_shutdown(self.c_completion_queue)
 *     self.is_shutting_down = True
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":112
 *     with nogil:
 *       grpc_completion_queue_shutdown(self.c_completion_queue)
 *     self.is_shutting_down = True             # <<<<<<<<<<<<<<
 * 
 *   def clear(self):
 */
  __pyx_v_self->is_shutting_down = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":109
 *     return self._interpret_event(event)
 * 
 *   def shutdown(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_completion_queue_shutdown(self.c_completion_queue)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":114
 *     self.is_shutting_down = True
 * 
 *   def clear(self):             # <<<<<<<<<<<<<<
 *     if not self.is_shutting_down:
 *       raise ValueError('queue must be shutting down to be cleared')
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_7clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_7clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_6clear(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_6clear(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("clear", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":115
 * 
 *   def clear(self):
 *     if not self.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise ValueError('queue must be shutting down to be cleared')
 *     while self.poll().type != GRPC_QUEUE_SHUTDOWN:
 */
  __pyx_t_1 = ((!(__pyx_v_self->is_shutting_down != 0)) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":116
 *   def clear(self):
 *     if not self.is_shutting_down:
 *       raise ValueError('queue must be shutting down to be cleared')             # <<<<<<<<<<<<<<
 *     while self.poll().type != GRPC_QUEUE_SHUTDOWN:
 *       pass
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(6, 116, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":115
 * 
 *   def clear(self):
 *     if not self.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise ValueError('queue must be shutting down to be cleared')
 *     while self.poll().type != GRPC_QUEUE_SHUTDOWN:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":117
 *     if not self.is_shutting_down:
 *       raise ValueError('queue must be shutting down to be cleared')
 *     while self.poll().type != GRPC_QUEUE_SHUTDOWN:             # <<<<<<<<<<<<<<
 *       pass
 * 
 */
  while (1) {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_poll); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 117, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 117, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_grpc_completion_type(GRPC_QUEUE_SHUTDOWN); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(6, 117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!__pyx_t_1) break;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":114
 *     self.is_shutting_down = True
 * 
 *   def clear(self):             # <<<<<<<<<<<<<<
 *     if not self.is_shutting_down:
 *       raise ValueError('queue must be shutting down to be cleared')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompletionQueue.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":120
 *       pass
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     cdef gpr_timespec c_deadline
 *     c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_9__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_9__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_8__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_15CompletionQueue_8__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_self) {
  gpr_timespec __pyx_v_c_deadline;
  grpc_event __pyx_v_event;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":122
 *   def __dealloc__(self):
 *     cdef gpr_timespec c_deadline
 *     c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)             # <<<<<<<<<<<<<<
 *     if self.c_completion_queue != NULL:
 *       # Ensure shutdown
 */
  __pyx_v_c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":123
 *     cdef gpr_timespec c_deadline
 *     c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 *     if self.c_completion_queue != NULL:             # <<<<<<<<<<<<<<
 *       # Ensure shutdown
 *       if not self.is_shutting_down:
 */
  __pyx_t_1 = ((__pyx_v_self->c_completion_queue != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":125
 *     if self.c_completion_queue != NULL:
 *       # Ensure shutdown
 *       if not self.is_shutting_down:             # <<<<<<<<<<<<<<
 *         grpc_completion_queue_shutdown(self.c_completion_queue)
 *       # Pump the queue (All outstanding calls should have been cancelled)
 */
    __pyx_t_1 = ((!(__pyx_v_self->is_shutting_down != 0)) != 0);
    if (__pyx_t_1) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":126
 *       # Ensure shutdown
 *       if not self.is_shutting_down:
 *         grpc_completion_queue_shutdown(self.c_completion_queue)             # <<<<<<<<<<<<<<
 *       # Pump the queue (All outstanding calls should have been cancelled)
 *       while not self.is_shutdown:
 */
      grpc_completion_queue_shutdown(__pyx_v_self->c_completion_queue);

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":125
 *     if self.c_completion_queue != NULL:
 *       # Ensure shutdown
 *       if not self.is_shutting_down:             # <<<<<<<<<<<<<<
 *         grpc_completion_queue_shutdown(self.c_completion_queue)
 *       # Pump the queue (All outstanding calls should have been cancelled)
 */
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":128
 *         grpc_completion_queue_shutdown(self.c_completion_queue)
 *       # Pump the queue (All outstanding calls should have been cancelled)
 *       while not self.is_shutdown:             # <<<<<<<<<<<<<<
 *         event = grpc_completion_queue_next(
 *             self.c_completion_queue, c_deadline, NULL)
 */
    while (1) {
      __pyx_t_1 = ((!(__pyx_v_self->is_shutdown != 0)) != 0);
      if (!__pyx_t_1) break;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":129
 *       # Pump the queue (All outstanding calls should have been cancelled)
 *       while not self.is_shutdown:
 *         event = grpc_completion_queue_next(             # <<<<<<<<<<<<<<
 *             self.c_completion_queue, c_deadline, NULL)
 *         self._interpret_event(event)
 */
      __pyx_v_event = grpc_completion_queue_next(__pyx_v_self->c_completion_queue, __pyx_v_c_deadline, NULL);

      /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":131
 *         event = grpc_completion_queue_next(
 *             self.c_completion_queue, c_deadline, NULL)
 *         self._interpret_event(event)             # <<<<<<<<<<<<<<
 *       grpc_completion_queue_destroy(self.c_completion_queue)
 *     grpc_shutdown()
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_self->__pyx_vtab)->_interpret_event(__pyx_v_self, __pyx_v_event); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":132
 *             self.c_completion_queue, c_deadline, NULL)
 *         self._interpret_event(event)
 *       grpc_completion_queue_destroy(self.c_completion_queue)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 */
    grpc_completion_queue_destroy(__pyx_v_self->c_completion_queue);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":123
 *     cdef gpr_timespec c_deadline
 *     c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 *     if self.c_completion_queue != NULL:             # <<<<<<<<<<<<<<
 *       # Ensure shutdown
 *       if not self.is_shutting_down:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":133
 *         self._interpret_event(event)
 *       grpc_completion_queue_destroy(self.c_completion_queue)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":120
 *       pass
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     cdef gpr_timespec c_deadline
 *     c_deadline = gpr_inf_future(GPR_CLOCK_REALTIME)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("grpc._cython.cygrpc.CompletionQueue.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":121
 * cdef class Timespec:
 * 
 *   def __cinit__(self, time):             # <<<<<<<<<<<<<<
 *     if time is None:
 *       with nogil:
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_time = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_time,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 121, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_time = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 121, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Timespec.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Timespec___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)__pyx_v_self), __pyx_v_time);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_8Timespec___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_self, PyObject *__pyx_v_time) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __pyx_t_4grpc_7_cython_6cygrpc_int64_t __pyx_t_5;
  __pyx_t_4grpc_7_cython_6cygrpc_int32_t __pyx_t_6;
  gpr_timespec __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_time);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":122
 * 
 *   def __cinit__(self, time):
 *     if time is None:             # <<<<<<<<<<<<<<
 *       with nogil:
 *         self.c_time = gpr_now(GPR_CLOCK_REALTIME)
 */
  __pyx_t_1 = (__pyx_v_time == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":123
 *   def __cinit__(self, time):
 *     if time is None:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         self.c_time = gpr_now(GPR_CLOCK_REALTIME)
 *       return
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":124
 *     if time is None:
 *       with nogil:
 *         self.c_time = gpr_now(GPR_CLOCK_REALTIME)             # <<<<<<<<<<<<<<
 *       return
 *     if isinstance(time, int):
 */
          __pyx_v_self->c_time = gpr_now(GPR_CLOCK_REALTIME);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":123
 *   def __cinit__(self, time):
 *     if time is None:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         self.c_time = gpr_now(GPR_CLOCK_REALTIME)
 *       return
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L6;
          }
          __pyx_L6:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":125
 *       with nogil:
 *         self.c_time = gpr_now(GPR_CLOCK_REALTIME)
 *       return             # <<<<<<<<<<<<<<
 *     if isinstance(time, int):
 *       time = float(time)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":122
 * 
 *   def __cinit__(self, time):
 *     if time is None:             # <<<<<<<<<<<<<<
 *       with nogil:
 *         self.c_time = gpr_now(GPR_CLOCK_REALTIME)
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":126
 *         self.c_time = gpr_now(GPR_CLOCK_REALTIME)
 *       return
 *     if isinstance(time, int):             # <<<<<<<<<<<<<<
 *       time = float(time)
 *     if isinstance(time, float):
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_time); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":127
 *       return
 *     if isinstance(time, int):
 *       time = float(time)             # <<<<<<<<<<<<<<
 *     if isinstance(time, float):
 *       if time == float("+inf"):
 */
    __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_v_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_time, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":126
 *         self.c_time = gpr_now(GPR_CLOCK_REALTIME)
 *       return
 *     if isinstance(time, int):             # <<<<<<<<<<<<<<
 *       time = float(time)
 *     if isinstance(time, float):
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":128
 *     if isinstance(time, int):
 *       time = float(time)
 *     if isinstance(time, float):             # <<<<<<<<<<<<<<
 *       if time == float("+inf"):
 *         with nogil:
 */
  __pyx_t_1 = PyFloat_Check(__pyx_v_time); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":129
 *       time = float(time)
 *     if isinstance(time, float):
 *       if time == float("+inf"):             # <<<<<<<<<<<<<<
 *         with nogil:
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)
 */
    __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_kp_s_inf); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_time, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":130
 *     if isinstance(time, float):
 *       if time == float("+inf"):
 *         with nogil:             # <<<<<<<<<<<<<<
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       elif time == float("-inf"):
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          #endif
          /*try:*/ {

            /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":131
 *       if time == float("+inf"):
 *         with nogil:
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)             # <<<<<<<<<<<<<<
 *       elif time == float("-inf"):
 *         with nogil:
 */
            __pyx_v_self->c_time = gpr_inf_future(GPR_CLOCK_REALTIME);
          }

          /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":130
 *     if isinstance(time, float):
 *       if time == float("+inf"):
 *         with nogil:             # <<<<<<<<<<<<<<
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       elif time == float("-inf"):
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L12;
            }
            __pyx_L12:;
          }
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":129
 *       time = float(time)
 *     if isinstance(time, float):
 *       if time == float("+inf"):             # <<<<<<<<<<<<<<
 *         with nogil:
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)
 */
      goto __pyx_L9;
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":132
 *         with nogil:
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       elif time == float("-inf"):             # <<<<<<<<<<<<<<
 *         with nogil:
 *           self.c_time = gpr_inf_past(GPR_CLOCK_REALTIME)
 */
    __pyx_t_4 = __Pyx_PyNumber_Float(__pyx_kp_s_inf_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_time, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":133
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       elif time == float("-inf"):
 *         with nogil:             # <<<<<<<<<<<<<<
 *           self.c_time = gpr_inf_past(GPR_CLOCK_REALTIME)
 *       else:
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          #endif
          /*try:*/ {

            /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":134
 *       elif time == float("-inf"):
 *         with nogil:
 *           self.c_time = gpr_inf_past(GPR_CLOCK_REALTIME)             # <<<<<<<<<<<<<<
 *       else:
 *         self.c_time.seconds = time
 */
            __pyx_v_self->c_time = gpr_inf_past(GPR_CLOCK_REALTIME);
          }

          /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":133
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       elif time == float("-inf"):
 *         with nogil:             # <<<<<<<<<<<<<<
 *           self.c_time = gpr_inf_past(GPR_CLOCK_REALTIME)
 *       else:
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L15;
            }
            __pyx_L15:;
          }
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":132
 *         with nogil:
 *           self.c_time = gpr_inf_future(GPR_CLOCK_REALTIME)
 *       elif time == float("-inf"):             # <<<<<<<<<<<<<<
 *         with nogil:
 *           self.c_time = gpr_inf_past(GPR_CLOCK_REALTIME)
 */
      goto __pyx_L9;
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":136
 *           self.c_time = gpr_inf_past(GPR_CLOCK_REALTIME)
 *       else:
 *         self.c_time.seconds = time             # <<<<<<<<<<<<<<
 *         self.c_time.nanoseconds = (time - float(self.c_time.seconds)) * 1e9
 *         self.c_time.clock_type = GPR_CLOCK_REALTIME
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_v_time); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 136, __pyx_L1_error)
      __pyx_v_self->c_time.tv_sec = __pyx_t_5;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":137
 *       else:
 *         self.c_time.seconds = time
 *         self.c_time.nanoseconds = (time - float(self.c_time.seconds)) * 1e9             # <<<<<<<<<<<<<<
 *         self.c_time.clock_type = GPR_CLOCK_REALTIME
 *     elif isinstance(time, Timespec):
 */
      __pyx_t_3 = PyFloat_FromDouble(((double)__pyx_v_self->c_time.tv_sec)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyNumber_Subtract(__pyx_v_time, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_Multiply(__pyx_t_4, __pyx_float_1e9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_self->c_time.tv_nsec = __pyx_t_6;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":138
 *         self.c_time.seconds = time
 *         self.c_time.nanoseconds = (time - float(self.c_time.seconds)) * 1e9
 *         self.c_time.clock_type = GPR_CLOCK_REALTIME             # <<<<<<<<<<<<<<
 *     elif isinstance(time, Timespec):
 *       self.c_time = (<Timespec>time).c_time
 */
      __pyx_v_self->c_time.clock_type = GPR_CLOCK_REALTIME;
    }
    __pyx_L9:;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":128
 *     if isinstance(time, int):
 *       time = float(time)
 *     if isinstance(time, float):             # <<<<<<<<<<<<<<
 *       if time == float("+inf"):
 *         with nogil:
 */
    goto __pyx_L8;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":139
 *         self.c_time.nanoseconds = (time - float(self.c_time.seconds)) * 1e9
 *         self.c_time.clock_type = GPR_CLOCK_REALTIME
 *     elif isinstance(time, Timespec):             # <<<<<<<<<<<<<<
 *       self.c_time = (<Timespec>time).c_time
 *     else:
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_time, __pyx_ptype_4grpc_7_cython_6cygrpc_Timespec); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":140
 *         self.c_time.clock_type = GPR_CLOCK_REALTIME
 *     elif isinstance(time, Timespec):
 *       self.c_time = (<Timespec>time).c_time             # <<<<<<<<<<<<<<
 *     else:
 *       raise TypeError("expected time to be float, int, or Timespec, not {}"
 */
    __pyx_t_7 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)__pyx_v_time)->c_time;
    __pyx_v_self->c_time = __pyx_t_7;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":139
 *         self.c_time.nanoseconds = (time - float(self.c_time.seconds)) * 1e9
 *         self.c_time.clock_type = GPR_CLOCK_REALTIME
 *     elif isinstance(time, Timespec):             # <<<<<<<<<<<<<<
 *       self.c_time = (<Timespec>time).c_time
 *     else:
 */
    goto __pyx_L8;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":142
 *       self.c_time = (<Timespec>time).c_time
 *     else:
 *       raise TypeError("expected time to be float, int, or Timespec, not {}"             # <<<<<<<<<<<<<<
 *                           .format(type(time)))
 * 
 */
  /*else*/ {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":143
 *     else:
 *       raise TypeError("expected time to be float, int, or Timespec, not {}"
 *                           .format(type(time)))             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_expected_time_to_be_float_int_or, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)Py_TYPE(__pyx_v_time))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_time)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_time)));
      PyTuple_SET_ITEM(__pyx_t_9, 0+1, ((PyObject *)Py_TYPE(__pyx_v_time)));
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":142
 *       self.c_time = (<Timespec>time).c_time
 *     else:
 *       raise TypeError("expected time to be float, int, or Timespec, not {}"             # <<<<<<<<<<<<<<
 *                           .format(type(time)))
 * 
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 142, __pyx_L1_error)
  }
  __pyx_L8:;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":121
 * cdef class Timespec:
 * 
 *   def __cinit__(self, time):             # <<<<<<<<<<<<<<
 *     if time is None:
 *       with nogil:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Timespec.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_time);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":146
 * 
 *   @property
 *   def seconds(self):             # <<<<<<<<<<<<<<
 *     # TODO(atash) ensure that everywhere a Timespec is created that it's
 *     # converted to GPR_CLOCK_REALTIME then and not every time someone wants to
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_7seconds_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_7seconds_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_7seconds___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_7seconds___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_self) {
  gpr_timespec __pyx_v_real_time;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":151
 *     # read values off in Python.
 *     cdef gpr_timespec real_time
 *     with nogil:             # <<<<<<<<<<<<<<
 *       real_time = (
 *           gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":153
 *     with nogil:
 *       real_time = (
 *           gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))             # <<<<<<<<<<<<<<
 *     return real_time.seconds
 * 
 */
        __pyx_v_real_time = gpr_convert_clock_type(__pyx_v_self->c_time, GPR_CLOCK_REALTIME);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":151
 *     # read values off in Python.
 *     cdef gpr_timespec real_time
 *     with nogil:             # <<<<<<<<<<<<<<
 *       real_time = (
 *           gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":154
 *       real_time = (
 *           gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 *     return real_time.seconds             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_real_time.tv_sec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":146
 * 
 *   @property
 *   def seconds(self):             # <<<<<<<<<<<<<<
 *     # TODO(atash) ensure that everywhere a Timespec is created that it's
 *     # converted to GPR_CLOCK_REALTIME then and not every time someone wants to
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Timespec.seconds.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":157
 * 
 *   @property
 *   def nanoseconds(self):             # <<<<<<<<<<<<<<
 *     cdef gpr_timespec real_time = (
 *         gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_11nanoseconds_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_11nanoseconds_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_11nanoseconds___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_11nanoseconds___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_self) {
  gpr_timespec __pyx_v_real_time;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":159
 *   def nanoseconds(self):
 *     cdef gpr_timespec real_time = (
 *         gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))             # <<<<<<<<<<<<<<
 *     return real_time.nanoseconds
 * 
 */
  __pyx_v_real_time = gpr_convert_clock_type(__pyx_v_self->c_time, GPR_CLOCK_REALTIME);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":160
 *     cdef gpr_timespec real_time = (
 *         gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 *     return real_time.nanoseconds             # <<<<<<<<<<<<<<
 * 
 *   def __float__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_real_time.tv_nsec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":157
 * 
 *   @property
 *   def nanoseconds(self):             # <<<<<<<<<<<<<<
 *     cdef gpr_timespec real_time = (
 *         gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Timespec.nanoseconds.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":162
 *     return real_time.nanoseconds
 * 
 *   def __float__(self):             # <<<<<<<<<<<<<<
 *     cdef gpr_timespec real_time = (
 *         gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_3__float__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_3__float__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__float__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_2__float__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_2__float__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_self) {
  gpr_timespec __pyx_v_real_time;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__float__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":164
 *   def __float__(self):
 *     cdef gpr_timespec real_time = (
 *         gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))             # <<<<<<<<<<<<<<
 *     return <double>real_time.seconds + <double>real_time.nanoseconds / 1e9
 * 
 */
  __pyx_v_real_time = gpr_convert_clock_type(__pyx_v_self->c_time, GPR_CLOCK_REALTIME);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":165
 *     cdef gpr_timespec real_time = (
 *         gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 *     return <double>real_time.seconds + <double>real_time.nanoseconds / 1e9             # <<<<<<<<<<<<<<
 * 
 *   @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((((double)__pyx_v_real_time.tv_sec) + (((double)__pyx_v_real_time.tv_nsec) / 1e9))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":162
 *     return real_time.nanoseconds
 * 
 *   def __float__(self):             # <<<<<<<<<<<<<<
 *     cdef gpr_timespec real_time = (
 *         gpr_convert_clock_type(self.c_time, GPR_CLOCK_REALTIME))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Timespec.__float__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":168
 * 
 *   @staticmethod
 *   def infinite_future():             # <<<<<<<<<<<<<<
 *     return Timespec(float("+inf"))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_5infinite_future(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_8Timespec_5infinite_future = {"infinite_future", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_5infinite_future, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_5infinite_future(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("infinite_future (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("infinite_future", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "infinite_future", 0))) return NULL;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_4infinite_future();

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_4infinite_future() {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("infinite_future", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":169
 *   @staticmethod
 *   def infinite_future():
 *     return Timespec(float("+inf"))             # <<<<<<<<<<<<<<
 * 
 *   @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_kp_s_inf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":168
 * 
 *   @staticmethod
 *   def infinite_future():             # <<<<<<<<<<<<<<
 *     return Timespec(float("+inf"))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Timespec.infinite_future", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":172
 * 
 *   @staticmethod
 *   def infinite_past():             # <<<<<<<<<<<<<<
 *     return Timespec(float("-inf"))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_7infinite_past(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_8Timespec_7infinite_past = {"infinite_past", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_7infinite_past, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_7infinite_past(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("infinite_past (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("infinite_past", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "infinite_past", 0))) return NULL;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_6infinite_past();

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Timespec_6infinite_past() {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("infinite_past", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":173
 *   @staticmethod
 *   def infinite_past():
 *     return Timespec(float("-inf"))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_kp_s_inf_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":172
 * 
 *   @staticmethod
 *   def infinite_past():             # <<<<<<<<<<<<<<
 *     return Timespec(float("-inf"))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Timespec.infinite_past", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":178
 * cdef class CallDetails:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     with nogil:
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":179
 * 
 *   def __cinit__(self):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_call_details_init(&self.c_details)
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":180
 *   def __cinit__(self):
 *     grpc_init()
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_call_details_init(&self.c_details)
 * 
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":181
 *     grpc_init()
 *     with nogil:
 *       grpc_call_details_init(&self.c_details)             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
        grpc_call_details_init((&__pyx_v_self->c_details));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":180
 *   def __cinit__(self):
 *     grpc_init()
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_call_details_init(&self.c_details)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":178
 * cdef class CallDetails:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     with nogil:
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":183
 *       grpc_call_details_init(&self.c_details)
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_call_details_destroy(&self.c_details)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":184
 * 
 *   def __dealloc__(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_call_details_destroy(&self.c_details)
 *     grpc_shutdown()
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":185
 *   def __dealloc__(self):
 *     with nogil:
 *       grpc_call_details_destroy(&self.c_details)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
        grpc_call_details_destroy((&__pyx_v_self->c_details));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":184
 * 
 *   def __dealloc__(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_call_details_destroy(&self.c_details)
 *     grpc_shutdown()
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":186
 *     with nogil:
 *       grpc_call_details_destroy(&self.c_details)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":183
 *       grpc_call_details_init(&self.c_details)
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_call_details_destroy(&self.c_details)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":189
 * 
 *   @property
 *   def method(self):             # <<<<<<<<<<<<<<
 *     if self.c_details.method != NULL:
 *       return <bytes>self.c_details.method
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_6method_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_6method_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_6method___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_6method___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":190
 *   @property
 *   def method(self):
 *     if self.c_details.method != NULL:             # <<<<<<<<<<<<<<
 *       return <bytes>self.c_details.method
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_self->c_details.method != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":191
 *   def method(self):
 *     if self.c_details.method != NULL:
 *       return <bytes>self.c_details.method             # <<<<<<<<<<<<<<
 *     else:
 *       return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->c_details.method); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject*)__pyx_t_2));
    __pyx_r = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":190
 *   @property
 *   def method(self):
 *     if self.c_details.method != NULL:             # <<<<<<<<<<<<<<
 *       return <bytes>self.c_details.method
 *     else:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":193
 *       return <bytes>self.c_details.method
 *     else:
 *       return None             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":189
 * 
 *   @property
 *   def method(self):             # <<<<<<<<<<<<<<
 *     if self.c_details.method != NULL:
 *       return <bytes>self.c_details.method
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CallDetails.method.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":196
 * 
 *   @property
 *   def host(self):             # <<<<<<<<<<<<<<
 *     if self.c_details.host != NULL:
 *       return <bytes>self.c_details.host
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_4host_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_4host_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_4host___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_4host___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":197
 *   @property
 *   def host(self):
 *     if self.c_details.host != NULL:             # <<<<<<<<<<<<<<
 *       return <bytes>self.c_details.host
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_self->c_details.host != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":198
 *   def host(self):
 *     if self.c_details.host != NULL:
 *       return <bytes>self.c_details.host             # <<<<<<<<<<<<<<
 *     else:
 *       return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->c_details.host); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject*)__pyx_t_2));
    __pyx_r = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":197
 *   @property
 *   def host(self):
 *     if self.c_details.host != NULL:             # <<<<<<<<<<<<<<
 *       return <bytes>self.c_details.host
 *     else:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":200
 *       return <bytes>self.c_details.host
 *     else:
 *       return None             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":196
 * 
 *   @property
 *   def host(self):             # <<<<<<<<<<<<<<
 *     if self.c_details.host != NULL:
 *       return <bytes>self.c_details.host
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CallDetails.host.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":203
 * 
 *   @property
 *   def deadline(self):             # <<<<<<<<<<<<<<
 *     timespec = Timespec(float("-inf"))
 *     timespec.c_time = self.c_details.deadline
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_8deadline_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_8deadline_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_8deadline___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_11CallDetails_8deadline___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_self) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *__pyx_v_timespec = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  gpr_timespec __pyx_t_3;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":204
 *   @property
 *   def deadline(self):
 *     timespec = Timespec(float("-inf"))             # <<<<<<<<<<<<<<
 *     timespec.c_time = self.c_details.deadline
 *     return timespec
 */
  __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_kp_s_inf_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_timespec = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":205
 *   def deadline(self):
 *     timespec = Timespec(float("-inf"))
 *     timespec.c_time = self.c_details.deadline             # <<<<<<<<<<<<<<
 *     return timespec
 * 
 */
  __pyx_t_3 = __pyx_v_self->c_details.deadline;
  __pyx_v_timespec->c_time = __pyx_t_3;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":206
 *     timespec = Timespec(float("-inf"))
 *     timespec.c_time = self.c_details.deadline
 *     return timespec             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_timespec));
  __pyx_r = ((PyObject *)__pyx_v_timespec);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":203
 * 
 *   @property
 *   def deadline(self):             # <<<<<<<<<<<<<<
 *     timespec = Timespec(float("-inf"))
 *     timespec.c_time = self.c_details.deadline
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CallDetails.deadline.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_timespec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":211
 * cdef class OperationTag:
 * 
 *   def __cinit__(self, user_tag):             # <<<<<<<<<<<<<<
 *     self.user_tag = user_tag
 *     self.references = []
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_12OperationTag_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_12OperationTag_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_user_tag = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_user_tag,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_user_tag)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 211, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_user_tag = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 211, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.OperationTag.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_12OperationTag___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)__pyx_v_self), __pyx_v_user_tag);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_12OperationTag___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *__pyx_v_self, PyObject *__pyx_v_user_tag) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":212
 * 
 *   def __cinit__(self, user_tag):
 *     self.user_tag = user_tag             # <<<<<<<<<<<<<<
 *     self.references = []
 * 
 */
  __Pyx_INCREF(__pyx_v_user_tag);
  __Pyx_GIVEREF(__pyx_v_user_tag);
  __Pyx_GOTREF(__pyx_v_self->user_tag);
  __Pyx_DECREF(__pyx_v_self->user_tag);
  __pyx_v_self->user_tag = __pyx_v_user_tag;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":213
 *   def __cinit__(self, user_tag):
 *     self.user_tag = user_tag
 *     self.references = []             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->references);
  __Pyx_DECREF(__pyx_v_self->references);
  __pyx_v_self->references = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":211
 * cdef class OperationTag:
 * 
 *   def __cinit__(self, user_tag):             # <<<<<<<<<<<<<<
 *     self.user_tag = user_tag
 *     self.references = []
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.OperationTag.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":218
 * cdef class Event:
 * 
 *   def __cinit__(self, grpc_completion_type type, bint success,             # <<<<<<<<<<<<<<
 *                 object tag, Call operation_call,
 *                 CallDetails request_call_details,
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_5Event_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_5Event_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  grpc_completion_type __pyx_v_type;
  int __pyx_v_success;
  PyObject *__pyx_v_tag = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_operation_call = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_request_call_details = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_request_metadata = 0;
  int __pyx_v_is_new_request;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_batch_operations = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_type,&__pyx_n_s_success,&__pyx_n_s_tag,&__pyx_n_s_operation_call,&__pyx_n_s_request_call_details,&__pyx_n_s_request_metadata,&__pyx_n_s_is_new_request,&__pyx_n_s_batch_operations,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_success)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 1); __PYX_ERR(0, 218, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tag)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 2); __PYX_ERR(0, 218, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_operation_call)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 3); __PYX_ERR(0, 218, __pyx_L3_error)
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_request_call_details)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 4); __PYX_ERR(0, 218, __pyx_L3_error)
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_request_metadata)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 5); __PYX_ERR(0, 218, __pyx_L3_error)
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_is_new_request)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 6); __PYX_ERR(0, 218, __pyx_L3_error)
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_batch_operations)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 7); __PYX_ERR(0, 218, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 218, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_type = ((grpc_completion_type)__Pyx_PyInt_As_grpc_completion_type(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L3_error)
    __pyx_v_success = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_success == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L3_error)
    __pyx_v_tag = values[2];
    __pyx_v_operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)values[3]);
    __pyx_v_request_call_details = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)values[4]);
    __pyx_v_request_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)values[5]);
    __pyx_v_is_new_request = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_is_new_request == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 222, __pyx_L3_error)
    __pyx_v_batch_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)values[7]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 218, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Event.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_operation_call), __pyx_ptype_4grpc_7_cython_6cygrpc_Call, 1, "operation_call", 0))) __PYX_ERR(0, 219, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_request_call_details), __pyx_ptype_4grpc_7_cython_6cygrpc_CallDetails, 1, "request_call_details", 0))) __PYX_ERR(0, 220, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_request_metadata), __pyx_ptype_4grpc_7_cython_6cygrpc_Metadata, 1, "request_metadata", 0))) __PYX_ERR(0, 221, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_batch_operations), __pyx_ptype_4grpc_7_cython_6cygrpc_Operations, 1, "batch_operations", 0))) __PYX_ERR(0, 223, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self), __pyx_v_type, __pyx_v_success, __pyx_v_tag, __pyx_v_operation_call, __pyx_v_request_call_details, __pyx_v_request_metadata, __pyx_v_is_new_request, __pyx_v_batch_operations);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_5Event___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self, grpc_completion_type __pyx_v_type, int __pyx_v_success, PyObject *__pyx_v_tag, struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *__pyx_v_operation_call, struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *__pyx_v_request_call_details, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_request_metadata, int __pyx_v_is_new_request, struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_batch_operations) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":224
 *                 bint is_new_request,
 *                 Operations batch_operations):
 *     self.type = type             # <<<<<<<<<<<<<<
 *     self.success = success
 *     self.tag = tag
 */
  __pyx_v_self->type = __pyx_v_type;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":225
 *                 Operations batch_operations):
 *     self.type = type
 *     self.success = success             # <<<<<<<<<<<<<<
 *     self.tag = tag
 *     self.operation_call = operation_call
 */
  __pyx_v_self->success = __pyx_v_success;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":226
 *     self.type = type
 *     self.success = success
 *     self.tag = tag             # <<<<<<<<<<<<<<
 *     self.operation_call = operation_call
 *     self.request_call_details = request_call_details
 */
  __Pyx_INCREF(__pyx_v_tag);
  __Pyx_GIVEREF(__pyx_v_tag);
  __Pyx_GOTREF(__pyx_v_self->tag);
  __Pyx_DECREF(__pyx_v_self->tag);
  __pyx_v_self->tag = __pyx_v_tag;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":227
 *     self.success = success
 *     self.tag = tag
 *     self.operation_call = operation_call             # <<<<<<<<<<<<<<
 *     self.request_call_details = request_call_details
 *     self.request_metadata = request_metadata
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_operation_call));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_operation_call));
  __Pyx_GOTREF(__pyx_v_self->operation_call);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->operation_call));
  __pyx_v_self->operation_call = __pyx_v_operation_call;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":228
 *     self.tag = tag
 *     self.operation_call = operation_call
 *     self.request_call_details = request_call_details             # <<<<<<<<<<<<<<
 *     self.request_metadata = request_metadata
 *     self.batch_operations = batch_operations
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_request_call_details));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_request_call_details));
  __Pyx_GOTREF(__pyx_v_self->request_call_details);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->request_call_details));
  __pyx_v_self->request_call_details = __pyx_v_request_call_details;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":229
 *     self.operation_call = operation_call
 *     self.request_call_details = request_call_details
 *     self.request_metadata = request_metadata             # <<<<<<<<<<<<<<
 *     self.batch_operations = batch_operations
 *     self.is_new_request = is_new_request
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_request_metadata));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_request_metadata));
  __Pyx_GOTREF(__pyx_v_self->request_metadata);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->request_metadata));
  __pyx_v_self->request_metadata = __pyx_v_request_metadata;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":230
 *     self.request_call_details = request_call_details
 *     self.request_metadata = request_metadata
 *     self.batch_operations = batch_operations             # <<<<<<<<<<<<<<
 *     self.is_new_request = is_new_request
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_batch_operations));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_batch_operations));
  __Pyx_GOTREF(__pyx_v_self->batch_operations);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->batch_operations));
  __pyx_v_self->batch_operations = __pyx_v_batch_operations;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":231
 *     self.request_metadata = request_metadata
 *     self.batch_operations = batch_operations
 *     self.is_new_request = is_new_request             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->is_new_request = __pyx_v_is_new_request;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":218
 * cdef class Event:
 * 
 *   def __cinit__(self, grpc_completion_type type, bint success,             # <<<<<<<<<<<<<<
 *                 object tag, Call operation_call,
 *                 CallDetails request_call_details,
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":57
 * cdef class Event:
 * 
 *   cdef readonly grpc_completion_type type             # <<<<<<<<<<<<<<
 *   cdef readonly bint success
 *   cdef readonly object tag
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_4type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event_4type___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_4type___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_grpc_completion_type(__pyx_v_self->type); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Event.type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":58
 * 
 *   cdef readonly grpc_completion_type type
 *   cdef readonly bint success             # <<<<<<<<<<<<<<
 *   cdef readonly object tag
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_7success_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_7success_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event_7success___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_7success___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->success); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Event.success.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":59
 *   cdef readonly grpc_completion_type type
 *   cdef readonly bint success
 *   cdef readonly object tag             # <<<<<<<<<<<<<<
 * 
 *   # For Server.request_call
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_3tag_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_3tag_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event_3tag___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_3tag___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->tag);
  __pyx_r = __pyx_v_self->tag;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":62
 * 
 *   # For Server.request_call
 *   cdef readonly bint is_new_request             # <<<<<<<<<<<<<<
 *   cdef readonly CallDetails request_call_details
 *   cdef readonly Metadata request_metadata
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_14is_new_request_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_14is_new_request_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event_14is_new_request___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_14is_new_request___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_new_request); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Event.is_new_request.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":63
 *   # For Server.request_call
 *   cdef readonly bint is_new_request
 *   cdef readonly CallDetails request_call_details             # <<<<<<<<<<<<<<
 *   cdef readonly Metadata request_metadata
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_20request_call_details_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_20request_call_details_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event_20request_call_details___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_20request_call_details___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->request_call_details));
  __pyx_r = ((PyObject *)__pyx_v_self->request_call_details);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":64
 *   cdef readonly bint is_new_request
 *   cdef readonly CallDetails request_call_details
 *   cdef readonly Metadata request_metadata             # <<<<<<<<<<<<<<
 * 
 *   # For server calls
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_16request_metadata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_16request_metadata_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event_16request_metadata___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_16request_metadata___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->request_metadata));
  __pyx_r = ((PyObject *)__pyx_v_self->request_metadata);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":67
 * 
 *   # For server calls
 *   cdef readonly Call operation_call             # <<<<<<<<<<<<<<
 * 
 *   # For Call.start_batch
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_14operation_call_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_14operation_call_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event_14operation_call___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_14operation_call___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->operation_call));
  __pyx_r = ((PyObject *)__pyx_v_self->operation_call);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":70
 * 
 *   # For Call.start_batch
 *   cdef readonly Operations batch_operations             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_16batch_operations_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_5Event_16batch_operations_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_5Event_16batch_operations___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_5Event_16batch_operations___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->batch_operations));
  __pyx_r = ((PyObject *)__pyx_v_self->batch_operations);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":236
 * cdef class ByteBuffer:
 * 
 *   def __cinit__(self, bytes data):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     if data is None:
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 236, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_data = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 236, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.ByteBuffer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 1, "data", 1))) __PYX_ERR(0, 236, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self, PyObject *__pyx_v_data) {
  char *__pyx_v_c_data;
  gpr_slice __pyx_v_data_slice;
  size_t __pyx_v_data_length;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  char *__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":237
 * 
 *   def __cinit__(self, bytes data):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     if data is None:
 *       self.c_byte_buffer = NULL
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":238
 *   def __cinit__(self, bytes data):
 *     grpc_init()
 *     if data is None:             # <<<<<<<<<<<<<<
 *       self.c_byte_buffer = NULL
 *       return
 */
  __pyx_t_1 = (__pyx_v_data == ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":239
 *     grpc_init()
 *     if data is None:
 *       self.c_byte_buffer = NULL             # <<<<<<<<<<<<<<
 *       return
 * 
 */
    __pyx_v_self->c_byte_buffer = NULL;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":240
 *     if data is None:
 *       self.c_byte_buffer = NULL
 *       return             # <<<<<<<<<<<<<<
 * 
 *     cdef char *c_data = data
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":238
 *   def __cinit__(self, bytes data):
 *     grpc_init()
 *     if data is None:             # <<<<<<<<<<<<<<
 *       self.c_byte_buffer = NULL
 *       return
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":242
 *       return
 * 
 *     cdef char *c_data = data             # <<<<<<<<<<<<<<
 *     cdef gpr_slice data_slice
 *     cdef size_t data_length = len(data)
 */
  __pyx_t_3 = __Pyx_PyObject_AsString(__pyx_v_data); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 242, __pyx_L1_error)
  __pyx_v_c_data = __pyx_t_3;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":244
 *     cdef char *c_data = data
 *     cdef gpr_slice data_slice
 *     cdef size_t data_length = len(data)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       data_slice = gpr_slice_from_copied_buffer(c_data, data_length)
 */
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 244, __pyx_L1_error)
  }
  __pyx_t_4 = PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_v_data_length = __pyx_t_4;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":245
 *     cdef gpr_slice data_slice
 *     cdef size_t data_length = len(data)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       data_slice = gpr_slice_from_copied_buffer(c_data, data_length)
 *     with nogil:
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":246
 *     cdef size_t data_length = len(data)
 *     with nogil:
 *       data_slice = gpr_slice_from_copied_buffer(c_data, data_length)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       self.c_byte_buffer = grpc_raw_byte_buffer_create(
 */
        __pyx_v_data_slice = gpr_slice_from_copied_buffer(__pyx_v_c_data, __pyx_v_data_length);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":245
 *     cdef gpr_slice data_slice
 *     cdef size_t data_length = len(data)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       data_slice = gpr_slice_from_copied_buffer(c_data, data_length)
 *     with nogil:
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":247
 *     with nogil:
 *       data_slice = gpr_slice_from_copied_buffer(c_data, data_length)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_byte_buffer = grpc_raw_byte_buffer_create(
 *           &data_slice, 1)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":248
 *       data_slice = gpr_slice_from_copied_buffer(c_data, data_length)
 *     with nogil:
 *       self.c_byte_buffer = grpc_raw_byte_buffer_create(             # <<<<<<<<<<<<<<
 *           &data_slice, 1)
 *     with nogil:
 */
        __pyx_v_self->c_byte_buffer = grpc_raw_byte_buffer_create((&__pyx_v_data_slice), 1);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":247
 *     with nogil:
 *       data_slice = gpr_slice_from_copied_buffer(c_data, data_length)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_byte_buffer = grpc_raw_byte_buffer_create(
 *           &data_slice, 1)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":250
 *       self.c_byte_buffer = grpc_raw_byte_buffer_create(
 *           &data_slice, 1)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_slice_unref(data_slice)
 * 
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":251
 *           &data_slice, 1)
 *     with nogil:
 *       gpr_slice_unref(data_slice)             # <<<<<<<<<<<<<<
 * 
 *   def bytes(self):
 */
        gpr_slice_unref(__pyx_v_data_slice);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":250
 *       self.c_byte_buffer = grpc_raw_byte_buffer_create(
 *           &data_slice, 1)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_slice_unref(data_slice)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L12;
        }
        __pyx_L12:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":236
 * cdef class ByteBuffer:
 * 
 *   def __cinit__(self, bytes data):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     if data is None:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.ByteBuffer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":253
 *       gpr_slice_unref(data_slice)
 * 
 *   def bytes(self):             # <<<<<<<<<<<<<<
 *     cdef grpc_byte_buffer_reader reader
 *     cdef gpr_slice data_slice
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_3bytes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_3bytes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bytes (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_2bytes(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_2bytes(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self) {
  struct grpc_byte_buffer_reader __pyx_v_reader;
  gpr_slice __pyx_v_data_slice;
  size_t __pyx_v_data_slice_length;
  void *__pyx_v_data_slice_pointer;
  int __pyx_v_reader_status;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("bytes", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":259
 *     cdef void *data_slice_pointer
 *     cdef bint reader_status
 *     if self.c_byte_buffer != NULL:             # <<<<<<<<<<<<<<
 *       with nogil:
 *         reader_status = grpc_byte_buffer_reader_init(
 */
  __pyx_t_1 = ((__pyx_v_self->c_byte_buffer != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":260
 *     cdef bint reader_status
 *     if self.c_byte_buffer != NULL:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         reader_status = grpc_byte_buffer_reader_init(
 *             &reader, self.c_byte_buffer)
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":261
 *     if self.c_byte_buffer != NULL:
 *       with nogil:
 *         reader_status = grpc_byte_buffer_reader_init(             # <<<<<<<<<<<<<<
 *             &reader, self.c_byte_buffer)
 *       if not reader_status:
 */
          __pyx_v_reader_status = grpc_byte_buffer_reader_init((&__pyx_v_reader), __pyx_v_self->c_byte_buffer);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":260
 *     cdef bint reader_status
 *     if self.c_byte_buffer != NULL:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         reader_status = grpc_byte_buffer_reader_init(
 *             &reader, self.c_byte_buffer)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L6;
          }
          __pyx_L6:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":263
 *         reader_status = grpc_byte_buffer_reader_init(
 *             &reader, self.c_byte_buffer)
 *       if not reader_status:             # <<<<<<<<<<<<<<
 *         return None
 *       result = bytearray()
 */
    __pyx_t_1 = ((!(__pyx_v_reader_status != 0)) != 0);
    if (__pyx_t_1) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":264
 *             &reader, self.c_byte_buffer)
 *       if not reader_status:
 *         return None             # <<<<<<<<<<<<<<
 *       result = bytearray()
 *       with nogil:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_None);
      __pyx_r = Py_None;
      goto __pyx_L0;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":263
 *         reader_status = grpc_byte_buffer_reader_init(
 *             &reader, self.c_byte_buffer)
 *       if not reader_status:             # <<<<<<<<<<<<<<
 *         return None
 *       result = bytearray()
 */
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":265
 *       if not reader_status:
 *         return None
 *       result = bytearray()             # <<<<<<<<<<<<<<
 *       with nogil:
 *         while grpc_byte_buffer_reader_next(&reader, &data_slice):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyByteArray_Type)), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_result = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":266
 *         return None
 *       result = bytearray()
 *       with nogil:             # <<<<<<<<<<<<<<
 *         while grpc_byte_buffer_reader_next(&reader, &data_slice):
 *           data_slice_pointer = gpr_slice_start_ptr(data_slice)
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":267
 *       result = bytearray()
 *       with nogil:
 *         while grpc_byte_buffer_reader_next(&reader, &data_slice):             # <<<<<<<<<<<<<<
 *           data_slice_pointer = gpr_slice_start_ptr(data_slice)
 *           data_slice_length = gpr_slice_length(data_slice)
 */
          while (1) {
            __pyx_t_1 = (grpc_byte_buffer_reader_next((&__pyx_v_reader), (&__pyx_v_data_slice)) != 0);
            if (!__pyx_t_1) break;

            /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":268
 *       with nogil:
 *         while grpc_byte_buffer_reader_next(&reader, &data_slice):
 *           data_slice_pointer = gpr_slice_start_ptr(data_slice)             # <<<<<<<<<<<<<<
 *           data_slice_length = gpr_slice_length(data_slice)
 *           with gil:
 */
            __pyx_v_data_slice_pointer = GPR_SLICE_START_PTR(__pyx_v_data_slice);

            /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":269
 *         while grpc_byte_buffer_reader_next(&reader, &data_slice):
 *           data_slice_pointer = gpr_slice_start_ptr(data_slice)
 *           data_slice_length = gpr_slice_length(data_slice)             # <<<<<<<<<<<<<<
 *           with gil:
 *             result += (<char *>data_slice_pointer)[:data_slice_length]
 */
            __pyx_v_data_slice_length = GPR_SLICE_LENGTH(__pyx_v_data_slice);

            /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":270
 *           data_slice_pointer = gpr_slice_start_ptr(data_slice)
 *           data_slice_length = gpr_slice_length(data_slice)
 *           with gil:             # <<<<<<<<<<<<<<
 *             result += (<char *>data_slice_pointer)[:data_slice_length]
 *           gpr_slice_unref(data_slice)
 */
            {
                #ifdef WITH_THREAD
                PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();
                #endif
                /*try:*/ {

                  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":271
 *           data_slice_length = gpr_slice_length(data_slice)
 *           with gil:
 *             result += (<char *>data_slice_pointer)[:data_slice_length]             # <<<<<<<<<<<<<<
 *           gpr_slice_unref(data_slice)
 *       with nogil:
 */
                  __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(((char *)__pyx_v_data_slice_pointer) + 0, __pyx_v_data_slice_length - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_result, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 271, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_3);
                  __pyx_t_3 = 0;
                }

                /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":270
 *           data_slice_pointer = gpr_slice_start_ptr(data_slice)
 *           data_slice_length = gpr_slice_length(data_slice)
 *           with gil:             # <<<<<<<<<<<<<<
 *             result += (<char *>data_slice_pointer)[:data_slice_length]
 *           gpr_slice_unref(data_slice)
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    #ifdef WITH_THREAD
                    PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    goto __pyx_L17;
                  }
                  __pyx_L16_error: {
                    #ifdef WITH_THREAD
                    PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    goto __pyx_L9_error;
                  }
                  __pyx_L17:;
                }
            }

            /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":272
 *           with gil:
 *             result += (<char *>data_slice_pointer)[:data_slice_length]
 *           gpr_slice_unref(data_slice)             # <<<<<<<<<<<<<<
 *       with nogil:
 *         grpc_byte_buffer_reader_destroy(&reader)
 */
            gpr_slice_unref(__pyx_v_data_slice);
          }
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":266
 *         return None
 *       result = bytearray()
 *       with nogil:             # <<<<<<<<<<<<<<
 *         while grpc_byte_buffer_reader_next(&reader, &data_slice):
 *           data_slice_pointer = gpr_slice_start_ptr(data_slice)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L10;
          }
          __pyx_L9_error: {
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L10:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":273
 *             result += (<char *>data_slice_pointer)[:data_slice_length]
 *           gpr_slice_unref(data_slice)
 *       with nogil:             # <<<<<<<<<<<<<<
 *         grpc_byte_buffer_reader_destroy(&reader)
 *       return bytes(result)
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":274
 *           gpr_slice_unref(data_slice)
 *       with nogil:
 *         grpc_byte_buffer_reader_destroy(&reader)             # <<<<<<<<<<<<<<
 *       return bytes(result)
 *     else:
 */
          grpc_byte_buffer_reader_destroy((&__pyx_v_reader));
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":273
 *             result += (<char *>data_slice_pointer)[:data_slice_length]
 *           gpr_slice_unref(data_slice)
 *       with nogil:             # <<<<<<<<<<<<<<
 *         grpc_byte_buffer_reader_destroy(&reader)
 *       return bytes(result)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L20;
          }
          __pyx_L20:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":275
 *       with nogil:
 *         grpc_byte_buffer_reader_destroy(&reader)
 *       return bytes(result)             # <<<<<<<<<<<<<<
 *     else:
 *       return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_result);
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyBytes_Type)), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":259
 *     cdef void *data_slice_pointer
 *     cdef bint reader_status
 *     if self.c_byte_buffer != NULL:             # <<<<<<<<<<<<<<
 *       with nogil:
 *         reader_status = grpc_byte_buffer_reader_init(
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":277
 *       return bytes(result)
 *     else:
 *       return None             # <<<<<<<<<<<<<<
 * 
 *   def __len__(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":253
 *       gpr_slice_unref(data_slice)
 * 
 *   def bytes(self):             # <<<<<<<<<<<<<<
 *     cdef grpc_byte_buffer_reader reader
 *     cdef gpr_slice data_slice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ByteBuffer.bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":279
 *       return None
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     cdef size_t result
 *     if self.c_byte_buffer != NULL:
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_4__len__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_4__len__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self) {
  size_t __pyx_v_result;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":281
 *   def __len__(self):
 *     cdef size_t result
 *     if self.c_byte_buffer != NULL:             # <<<<<<<<<<<<<<
 *       with nogil:
 *         result = grpc_byte_buffer_length(self.c_byte_buffer)
 */
  __pyx_t_1 = ((__pyx_v_self->c_byte_buffer != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":282
 *     cdef size_t result
 *     if self.c_byte_buffer != NULL:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_byte_buffer_length(self.c_byte_buffer)
 *       return result
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":283
 *     if self.c_byte_buffer != NULL:
 *       with nogil:
 *         result = grpc_byte_buffer_length(self.c_byte_buffer)             # <<<<<<<<<<<<<<
 *       return result
 *     else:
 */
          __pyx_v_result = grpc_byte_buffer_length(__pyx_v_self->c_byte_buffer);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":282
 *     cdef size_t result
 *     if self.c_byte_buffer != NULL:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_byte_buffer_length(self.c_byte_buffer)
 *       return result
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L6;
          }
          __pyx_L6:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":284
 *       with nogil:
 *         result = grpc_byte_buffer_length(self.c_byte_buffer)
 *       return result             # <<<<<<<<<<<<<<
 *     else:
 *       return 0
 */
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":281
 *   def __len__(self):
 *     cdef size_t result
 *     if self.c_byte_buffer != NULL:             # <<<<<<<<<<<<<<
 *       with nogil:
 *         result = grpc_byte_buffer_length(self.c_byte_buffer)
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":286
 *       return result
 *     else:
 *       return 0             # <<<<<<<<<<<<<<
 * 
 *   def __str__(self):
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":279
 *       return None
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     cdef size_t result
 *     if self.c_byte_buffer != NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":288
 *       return 0
 * 
 *   def __str__(self):             # <<<<<<<<<<<<<<
 *     return self.bytes()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_7__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_7__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_6__str__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_6__str__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":289
 * 
 *   def __str__(self):
 *     return self.bytes()             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":288
 *       return 0
 * 
 *   def __str__(self):             # <<<<<<<<<<<<<<
 *     return self.bytes()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ByteBuffer.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":291
 *     return self.bytes()
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_byte_buffer != NULL:
 *       grpc_byte_buffer_destroy(self.c_byte_buffer)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_9__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_9__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_8__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_10ByteBuffer_8__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":292
 * 
 *   def __dealloc__(self):
 *     if self.c_byte_buffer != NULL:             # <<<<<<<<<<<<<<
 *       grpc_byte_buffer_destroy(self.c_byte_buffer)
 *     grpc_shutdown()
 */
  __pyx_t_1 = ((__pyx_v_self->c_byte_buffer != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":293
 *   def __dealloc__(self):
 *     if self.c_byte_buffer != NULL:
 *       grpc_byte_buffer_destroy(self.c_byte_buffer)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
    grpc_byte_buffer_destroy(__pyx_v_self->c_byte_buffer);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":292
 * 
 *   def __dealloc__(self):
 *     if self.c_byte_buffer != NULL:             # <<<<<<<<<<<<<<
 *       grpc_byte_buffer_destroy(self.c_byte_buffer)
 *     grpc_shutdown()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":294
 *     if self.c_byte_buffer != NULL:
 *       grpc_byte_buffer_destroy(self.c_byte_buffer)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":291
 *     return self.bytes()
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_byte_buffer != NULL:
 *       grpc_byte_buffer_destroy(self.c_byte_buffer)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":299
 * cdef class SslPemKeyCertPair:
 * 
 *   def __cinit__(self, bytes private_key, bytes certificate_chain):             # <<<<<<<<<<<<<<
 *     self.private_key = private_key
 *     self.certificate_chain = certificate_chain
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_private_key = 0;
  PyObject *__pyx_v_certificate_chain = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_private_key,&__pyx_n_s_certificate_chain,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_private_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_certificate_chain)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(0, 299, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 299, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_private_key = ((PyObject*)values[0]);
    __pyx_v_certificate_chain = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 299, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.SslPemKeyCertPair.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_private_key), (&PyBytes_Type), 1, "private_key", 1))) __PYX_ERR(0, 299, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_certificate_chain), (&PyBytes_Type), 1, "certificate_chain", 1))) __PYX_ERR(0, 299, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)__pyx_v_self), __pyx_v_private_key, __pyx_v_certificate_chain);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_self, PyObject *__pyx_v_private_key, PyObject *__pyx_v_certificate_chain) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  char const *__pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":300
 * 
 *   def __cinit__(self, bytes private_key, bytes certificate_chain):
 *     self.private_key = private_key             # <<<<<<<<<<<<<<
 *     self.certificate_chain = certificate_chain
 *     self.c_pair.private_key = self.private_key
 */
  __Pyx_INCREF(__pyx_v_private_key);
  __Pyx_GIVEREF(__pyx_v_private_key);
  __Pyx_GOTREF(__pyx_v_self->private_key);
  __Pyx_DECREF(__pyx_v_self->private_key);
  __pyx_v_self->private_key = __pyx_v_private_key;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":301
 *   def __cinit__(self, bytes private_key, bytes certificate_chain):
 *     self.private_key = private_key
 *     self.certificate_chain = certificate_chain             # <<<<<<<<<<<<<<
 *     self.c_pair.private_key = self.private_key
 *     self.c_pair.certificate_chain = self.certificate_chain
 */
  __Pyx_INCREF(__pyx_v_certificate_chain);
  __Pyx_GIVEREF(__pyx_v_certificate_chain);
  __Pyx_GOTREF(__pyx_v_self->certificate_chain);
  __Pyx_DECREF(__pyx_v_self->certificate_chain);
  __pyx_v_self->certificate_chain = __pyx_v_certificate_chain;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":302
 *     self.private_key = private_key
 *     self.certificate_chain = certificate_chain
 *     self.c_pair.private_key = self.private_key             # <<<<<<<<<<<<<<
 *     self.c_pair.certificate_chain = self.certificate_chain
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_AsString(__pyx_v_self->private_key); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L1_error)
  __pyx_v_self->c_pair.private_key = __pyx_t_1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":303
 *     self.certificate_chain = certificate_chain
 *     self.c_pair.private_key = self.private_key
 *     self.c_pair.certificate_chain = self.certificate_chain             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_v_self->certificate_chain); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 303, __pyx_L1_error)
  __pyx_v_self->c_pair.cert_chain = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":299
 * cdef class SslPemKeyCertPair:
 * 
 *   def __cinit__(self, bytes private_key, bytes certificate_chain):             # <<<<<<<<<<<<<<
 *     self.private_key = private_key
 *     self.certificate_chain = certificate_chain
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.SslPemKeyCertPair.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":81
 * 
 *   cdef grpc_ssl_pem_key_cert_pair c_pair
 *   cdef readonly object private_key, certificate_chain             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_11private_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_11private_key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_11private_key___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_11private_key___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->private_key);
  __pyx_r = __pyx_v_self->private_key;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_17certificate_chain_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_17certificate_chain_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_17certificate_chain___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_17certificate_chain___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->certificate_chain);
  __pyx_r = __pyx_v_self->certificate_chain;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":308
 * cdef class ChannelArg:
 * 
 *   def __cinit__(self, bytes key, value):             # <<<<<<<<<<<<<<
 *     self.key = key
 *     self.c_arg.key = self.key
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(0, 308, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 308, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_key = ((PyObject*)values[0]);
    __pyx_v_value = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 308, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelArg.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyBytes_Type), 1, "key", 1))) __PYX_ERR(0, 308, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *)__pyx_v_self), __pyx_v_key, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":309
 * 
 *   def __cinit__(self, bytes key, value):
 *     self.key = key             # <<<<<<<<<<<<<<
 *     self.c_arg.key = self.key
 *     if isinstance(value, int):
 */
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_key;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":310
 *   def __cinit__(self, bytes key, value):
 *     self.key = key
 *     self.c_arg.key = self.key             # <<<<<<<<<<<<<<
 *     if isinstance(value, int):
 *       self.value = value
 */
  __pyx_t_1 = __Pyx_PyObject_AsString(__pyx_v_self->key); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 310, __pyx_L1_error)
  __pyx_v_self->c_arg.key = __pyx_t_1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":311
 *     self.key = key
 *     self.c_arg.key = self.key
 *     if isinstance(value, int):             # <<<<<<<<<<<<<<
 *       self.value = value
 *       self.c_arg.type = GRPC_ARG_INTEGER
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":312
 *     self.c_arg.key = self.key
 *     if isinstance(value, int):
 *       self.value = value             # <<<<<<<<<<<<<<
 *       self.c_arg.type = GRPC_ARG_INTEGER
 *       self.c_arg.value.integer = self.value
 */
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    __Pyx_GOTREF(__pyx_v_self->value);
    __Pyx_DECREF(__pyx_v_self->value);
    __pyx_v_self->value = __pyx_v_value;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":313
 *     if isinstance(value, int):
 *       self.value = value
 *       self.c_arg.type = GRPC_ARG_INTEGER             # <<<<<<<<<<<<<<
 *       self.c_arg.value.integer = self.value
 *     elif isinstance(value, bytes):
 */
    __pyx_v_self->c_arg.type = GRPC_ARG_INTEGER;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":314
 *       self.value = value
 *       self.c_arg.type = GRPC_ARG_INTEGER
 *       self.c_arg.value.integer = self.value             # <<<<<<<<<<<<<<
 *     elif isinstance(value, bytes):
 *       self.value = value
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_self->value); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 314, __pyx_L1_error)
    __pyx_v_self->c_arg.value.integer = __pyx_t_4;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":311
 *     self.key = key
 *     self.c_arg.key = self.key
 *     if isinstance(value, int):             # <<<<<<<<<<<<<<
 *       self.value = value
 *       self.c_arg.type = GRPC_ARG_INTEGER
 */
    goto __pyx_L3;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":315
 *       self.c_arg.type = GRPC_ARG_INTEGER
 *       self.c_arg.value.integer = self.value
 *     elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *       self.value = value
 *       self.c_arg.type = GRPC_ARG_STRING
 */
  __pyx_t_3 = PyBytes_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":316
 *       self.c_arg.value.integer = self.value
 *     elif isinstance(value, bytes):
 *       self.value = value             # <<<<<<<<<<<<<<
 *       self.c_arg.type = GRPC_ARG_STRING
 *       self.c_arg.value.string = self.value
 */
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    __Pyx_GOTREF(__pyx_v_self->value);
    __Pyx_DECREF(__pyx_v_self->value);
    __pyx_v_self->value = __pyx_v_value;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":317
 *     elif isinstance(value, bytes):
 *       self.value = value
 *       self.c_arg.type = GRPC_ARG_STRING             # <<<<<<<<<<<<<<
 *       self.c_arg.value.string = self.value
 *     else:
 */
    __pyx_v_self->c_arg.type = GRPC_ARG_STRING;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":318
 *       self.value = value
 *       self.c_arg.type = GRPC_ARG_STRING
 *       self.c_arg.value.string = self.value             # <<<<<<<<<<<<<<
 *     else:
 *       raise TypeError('Expected int or bytes, got {}'.format(type(value)))
 */
    __pyx_t_1 = __Pyx_PyObject_AsString(__pyx_v_self->value); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
    __pyx_v_self->c_arg.value.string = __pyx_t_1;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":315
 *       self.c_arg.type = GRPC_ARG_INTEGER
 *       self.c_arg.value.integer = self.value
 *     elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *       self.value = value
 *       self.c_arg.type = GRPC_ARG_STRING
 */
    goto __pyx_L3;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":320
 *       self.c_arg.value.string = self.value
 *     else:
 *       raise TypeError('Expected int or bytes, got {}'.format(type(value)))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Expected_int_or_bytes_got, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)Py_TYPE(__pyx_v_value))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    } else {
      __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_value)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_value)));
      PyTuple_SET_ITEM(__pyx_t_8, 0+1, ((PyObject *)Py_TYPE(__pyx_v_value)));
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 320, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":308
 * cdef class ChannelArg:
 * 
 *   def __cinit__(self, bytes key, value):             # <<<<<<<<<<<<<<
 *     self.key = key
 *     self.c_arg.key = self.key
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelArg.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":87
 * 
 *   cdef grpc_arg c_arg
 *   cdef readonly object key, value             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_3key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_3key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg_3key___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg_3key___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->key);
  __pyx_r = __pyx_v_self->key;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_5value_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg_5value___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10ChannelArg_5value___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->value);
  __pyx_r = __pyx_v_self->value;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":325
 * cdef class ChannelArgs:
 * 
 *   def __cinit__(self, args):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.args = list(args)
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 325, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_args = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 325, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelArgs.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)__pyx_v_self), __pyx_v_args);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_v_arg = NULL;
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  grpc_arg __pyx_t_8;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":326
 * 
 *   def __cinit__(self, args):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.args = list(args)
 *     for arg in self.args:
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":327
 *   def __cinit__(self, args):
 *     grpc_init()
 *     self.args = list(args)             # <<<<<<<<<<<<<<
 *     for arg in self.args:
 *       if not isinstance(arg, ChannelArg):
 */
  __pyx_t_1 = PySequence_List(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":328
 *     grpc_init()
 *     self.args = list(args)
 *     for arg in self.args:             # <<<<<<<<<<<<<<
 *       if not isinstance(arg, ChannelArg):
 *         raise TypeError("expected list of ChannelArg")
 */
  if (unlikely(__pyx_v_self->args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 328, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->args; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 328, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":329
 *     self.args = list(args)
 *     for arg in self.args:
 *       if not isinstance(arg, ChannelArg):             # <<<<<<<<<<<<<<
 *         raise TypeError("expected list of ChannelArg")
 *     self.c_args.arguments_length = len(self.args)
 */
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_ptype_4grpc_7_cython_6cygrpc_ChannelArg); 
    __pyx_t_5 = ((!(__pyx_t_4 != 0)) != 0);
    if (__pyx_t_5) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":330
 *     for arg in self.args:
 *       if not isinstance(arg, ChannelArg):
 *         raise TypeError("expected list of ChannelArg")             # <<<<<<<<<<<<<<
 *     self.c_args.arguments_length = len(self.args)
 *     with nogil:
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 330, __pyx_L1_error)

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":329
 *     self.args = list(args)
 *     for arg in self.args:
 *       if not isinstance(arg, ChannelArg):             # <<<<<<<<<<<<<<
 *         raise TypeError("expected list of ChannelArg")
 *     self.c_args.arguments_length = len(self.args)
 */
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":328
 *     grpc_init()
 *     self.args = list(args)
 *     for arg in self.args:             # <<<<<<<<<<<<<<
 *       if not isinstance(arg, ChannelArg):
 *         raise TypeError("expected list of ChannelArg")
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":331
 *       if not isinstance(arg, ChannelArg):
 *         raise TypeError("expected list of ChannelArg")
 *     self.c_args.arguments_length = len(self.args)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       self.c_args.arguments = <grpc_arg *>gpr_malloc(
 */
  __pyx_t_1 = __pyx_v_self->args;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 331, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->c_args.num_args = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":332
 *         raise TypeError("expected list of ChannelArg")
 *     self.c_args.arguments_length = len(self.args)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_args.arguments = <grpc_arg *>gpr_malloc(
 *           self.c_args.arguments_length*sizeof(grpc_arg))
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":333
 *     self.c_args.arguments_length = len(self.args)
 *     with nogil:
 *       self.c_args.arguments = <grpc_arg *>gpr_malloc(             # <<<<<<<<<<<<<<
 *           self.c_args.arguments_length*sizeof(grpc_arg))
 *     for i in range(self.c_args.arguments_length):
 */
        __pyx_v_self->c_args.args = ((grpc_arg *)gpr_malloc((__pyx_v_self->c_args.num_args * (sizeof(grpc_arg)))));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":332
 *         raise TypeError("expected list of ChannelArg")
 *     self.c_args.arguments_length = len(self.args)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_args.arguments = <grpc_arg *>gpr_malloc(
 *           self.c_args.arguments_length*sizeof(grpc_arg))
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":335
 *       self.c_args.arguments = <grpc_arg *>gpr_malloc(
 *           self.c_args.arguments_length*sizeof(grpc_arg))
 *     for i in range(self.c_args.arguments_length):             # <<<<<<<<<<<<<<
 *       self.c_args.arguments[i] = (<ChannelArg>self.args[i]).c_arg
 * 
 */
  __pyx_t_6 = __pyx_v_self->c_args.num_args;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":336
 *           self.c_args.arguments_length*sizeof(grpc_arg))
 *     for i in range(self.c_args.arguments_length):
 *       self.c_args.arguments[i] = (<ChannelArg>self.args[i]).c_arg             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
    if (unlikely(__pyx_v_self->args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 336, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_self->args, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *)__pyx_t_1)->c_arg;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->c_args.args[__pyx_v_i]) = __pyx_t_8;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":325
 * cdef class ChannelArgs:
 * 
 *   def __cinit__(self, args):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.args = list(args)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelArgs.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":338
 *       self.c_args.arguments[i] = (<ChannelArg>self.args[i]).c_arg
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       gpr_free(self.c_args.arguments)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":339
 * 
 *   def __dealloc__(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_free(self.c_args.arguments)
 *     grpc_shutdown()
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":340
 *   def __dealloc__(self):
 *     with nogil:
 *       gpr_free(self.c_args.arguments)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
        gpr_free(__pyx_v_self->c_args.args);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":339
 * 
 *   def __dealloc__(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_free(self.c_args.arguments)
 *     grpc_shutdown()
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":341
 *     with nogil:
 *       gpr_free(self.c_args.arguments)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 *   def __len__(self):
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":338
 *       self.c_args.arguments[i] = (<ChannelArg>self.args[i]).c_arg
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       gpr_free(self.c_args.arguments)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":343
 *     grpc_shutdown()
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     # self.args is never stale; it's only updated from this file
 *     return len(self.args)
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_4__len__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_4__len__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":345
 *   def __len__(self):
 *     # self.args is never stale; it's only updated from this file
 *     return len(self.args)             # <<<<<<<<<<<<<<
 * 
 *   def __getitem__(self, size_t i):
 */
  __pyx_t_1 = __pyx_v_self->args;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 345, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":343
 *     grpc_shutdown()
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     # self.args is never stale; it's only updated from this file
 *     return len(self.args)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelArgs.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":347
 *     return len(self.args)
 * 
 *   def __getitem__(self, size_t i):             # <<<<<<<<<<<<<<
 *     # self.args is never stale; it's only updated from this file
 *     return self.args[i]
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_size_t(__pyx_arg_i); if (unlikely((__pyx_v_i == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 347, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelArgs.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_6__getitem__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)__pyx_v_self), ((size_t)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_11ChannelArgs_6__getitem__(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_self, size_t __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":349
 *   def __getitem__(self, size_t i):
 *     # self.args is never stale; it's only updated from this file
 *     return self.args[i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 349, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_self->args, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":347
 *     return len(self.args)
 * 
 *   def __getitem__(self, size_t i):             # <<<<<<<<<<<<<<
 *     # self.args is never stale; it's only updated from this file
 *     return self.args[i]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.ChannelArgs.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":354
 * cdef class Metadatum:
 * 
 *   def __cinit__(self, bytes key, bytes value):             # <<<<<<<<<<<<<<
 *     self._key = key
 *     self._value = value
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(0, 354, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 354, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_key = ((PyObject*)values[0]);
    __pyx_v_value = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 354, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadatum.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyBytes_Type), 1, "key", 1))) __PYX_ERR(0, 354, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyBytes_Type), 1, "value", 1))) __PYX_ERR(0, 354, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)__pyx_v_self), __pyx_v_key, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  char const *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":355
 * 
 *   def __cinit__(self, bytes key, bytes value):
 *     self._key = key             # <<<<<<<<<<<<<<
 *     self._value = value
 *     self.c_metadata.key = self._key
 */
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_self->_key);
  __Pyx_DECREF(__pyx_v_self->_key);
  __pyx_v_self->_key = __pyx_v_key;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":356
 *   def __cinit__(self, bytes key, bytes value):
 *     self._key = key
 *     self._value = value             # <<<<<<<<<<<<<<
 *     self.c_metadata.key = self._key
 *     self.c_metadata.value = self._value
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_value);
  __Pyx_DECREF(__pyx_v_self->_value);
  __pyx_v_self->_value = __pyx_v_value;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":357
 *     self._key = key
 *     self._value = value
 *     self.c_metadata.key = self._key             # <<<<<<<<<<<<<<
 *     self.c_metadata.value = self._value
 *     self.c_metadata.value_length = len(self._value)
 */
  __pyx_t_1 = __Pyx_PyObject_AsString(__pyx_v_self->_key); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L1_error)
  __pyx_v_self->c_metadata.key = __pyx_t_1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":358
 *     self._value = value
 *     self.c_metadata.key = self._key
 *     self.c_metadata.value = self._value             # <<<<<<<<<<<<<<
 *     self.c_metadata.value_length = len(self._value)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_v_self->_value); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)
  __pyx_v_self->c_metadata.value = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":359
 *     self.c_metadata.key = self._key
 *     self.c_metadata.value = self._value
 *     self.c_metadata.value_length = len(self._value)             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __pyx_t_3 = __pyx_v_self->_value;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->c_metadata.value_length = __pyx_t_4;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":354
 * cdef class Metadatum:
 * 
 *   def __cinit__(self, bytes key, bytes value):             # <<<<<<<<<<<<<<
 *     self._key = key
 *     self._value = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadatum.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":362
 * 
 *   @property
 *   def key(self):             # <<<<<<<<<<<<<<
 *     return <bytes>self.c_metadata.key
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_3key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_3key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_3key___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_3key___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":363
 *   @property
 *   def key(self):
 *     return <bytes>self.c_metadata.key             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->c_metadata.key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":362
 * 
 *   @property
 *   def key(self):             # <<<<<<<<<<<<<<
 *     return <bytes>self.c_metadata.key
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadatum.key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":366
 * 
 *   @property
 *   def value(self):             # <<<<<<<<<<<<<<
 *     return <bytes>self.c_metadata.value[:self.c_metadata.value_length]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_5value_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_5value___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_5value___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":367
 *   @property
 *   def value(self):
 *     return <bytes>self.c_metadata.value[:self.c_metadata.value_length]             # <<<<<<<<<<<<<<
 * 
 *   def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_self->c_metadata.value + 0, __pyx_v_self->c_metadata.value_length - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":366
 * 
 *   @property
 *   def value(self):             # <<<<<<<<<<<<<<
 *     return <bytes>self.c_metadata.value[:self.c_metadata.value_length]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadatum.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":369
 *     return <bytes>self.c_metadata.value[:self.c_metadata.value_length]
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     return 2
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_3__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_2__len__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_2__len__(CYTHON_UNUSED struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":370
 * 
 *   def __len__(self):
 *     return 2             # <<<<<<<<<<<<<<
 * 
 *   def __getitem__(self, size_t i):
 */
  __pyx_r = 2;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":369
 *     return <bytes>self.c_metadata.value[:self.c_metadata.value_length]
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     return 2
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":372
 *     return 2
 * 
 *   def __getitem__(self, size_t i):             # <<<<<<<<<<<<<<
 *     if i == 0:
 *       return self.key
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_size_t(__pyx_arg_i); if (unlikely((__pyx_v_i == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 372, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadatum.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_4__getitem__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)__pyx_v_self), ((size_t)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_4__getitem__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self, size_t __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":373
 * 
 *   def __getitem__(self, size_t i):
 *     if i == 0:             # <<<<<<<<<<<<<<
 *       return self.key
 *     elif i == 1:
 */
  switch (__pyx_v_i) {
    case 0:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":374
 *   def __getitem__(self, size_t i):
 *     if i == 0:
 *       return self.key             # <<<<<<<<<<<<<<
 *     elif i == 1:
 *       return self.value
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":373
 * 
 *   def __getitem__(self, size_t i):
 *     if i == 0:             # <<<<<<<<<<<<<<
 *       return self.key
 *     elif i == 1:
 */
    break;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":375
 *     if i == 0:
 *       return self.key
 *     elif i == 1:             # <<<<<<<<<<<<<<
 *       return self.value
 *     else:
 */
    case 1:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":376
 *       return self.key
 *     elif i == 1:
 *       return self.value             # <<<<<<<<<<<<<<
 *     else:
 *       raise IndexError("index must be 0 (key) or 1 (value)")
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":375
 *     if i == 0:
 *       return self.key
 *     elif i == 1:             # <<<<<<<<<<<<<<
 *       return self.value
 *     else:
 */
    break;
    default:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":378
 *       return self.value
 *     else:
 *       raise IndexError("index must be 0 (key) or 1 (value)")             # <<<<<<<<<<<<<<
 * 
 *   def __iter__(self):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 378, __pyx_L1_error)
    break;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":372
 *     return 2
 * 
 *   def __getitem__(self, size_t i):             # <<<<<<<<<<<<<<
 *     if i == 0:
 *       return self.key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadatum.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":380
 *       raise IndexError("index must be 0 (key) or 1 (value)")
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return iter((self.key, self.value))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_7__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_7__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_6__iter__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Metadatum_6__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":381
 * 
 *   def __iter__(self):
 *     return iter((self.key, self.value))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":380
 *       raise IndexError("index must be 0 (key) or 1 (value)")
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return iter((self.key, self.value))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadatum.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":389
 *   cdef Metadata metadata
 * 
 *   def __cinit__(self, Metadata metadata not None):             # <<<<<<<<<<<<<<
 *     self.i = 0
 *     self.metadata = metadata
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_metadata,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_metadata)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 389, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 389, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc._MetadataIterator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_metadata), __pyx_ptype_4grpc_7_cython_6cygrpc_Metadata, 0, "metadata", 0))) __PYX_ERR(0, 389, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *)__pyx_v_self), __pyx_v_metadata);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":390
 * 
 *   def __cinit__(self, Metadata metadata not None):
 *     self.i = 0             # <<<<<<<<<<<<<<
 *     self.metadata = metadata
 * 
 */
  __pyx_v_self->i = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":391
 *   def __cinit__(self, Metadata metadata not None):
 *     self.i = 0
 *     self.metadata = metadata             # <<<<<<<<<<<<<<
 * 
 *   def __iter__(self):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_metadata));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_metadata));
  __Pyx_GOTREF(__pyx_v_self->metadata);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->metadata));
  __pyx_v_self->metadata = __pyx_v_metadata;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":389
 *   cdef Metadata metadata
 * 
 *   def __cinit__(self, Metadata metadata not None):             # <<<<<<<<<<<<<<
 *     self.i = 0
 *     self.metadata = metadata
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":393
 *     self.metadata = metadata
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_3__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator_2__iter__(((struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator_2__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":394
 * 
 *   def __iter__(self):
 *     return self             # <<<<<<<<<<<<<<
 * 
 *   def __next__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":393
 *     self.metadata = metadata
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":396
 *     return self
 * 
 *   def __next__(self):             # <<<<<<<<<<<<<<
 *     if self.i < len(self.metadata):
 *       result = self.metadata[self.i]
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_5__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_5__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator_4__next__(((struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_17_MetadataIterator_4__next__(struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":397
 * 
 *   def __next__(self):
 *     if self.i < len(self.metadata):             # <<<<<<<<<<<<<<
 *       result = self.metadata[self.i]
 *       self.i = self.i + 1
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->metadata);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((__pyx_v_self->i < __pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":398
 *   def __next__(self):
 *     if self.i < len(self.metadata):
 *       result = self.metadata[self.i]             # <<<<<<<<<<<<<<
 *       self.i = self.i + 1
 *       return result
 */
    __pyx_t_1 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->metadata), __pyx_v_self->i, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_result = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":399
 *     if self.i < len(self.metadata):
 *       result = self.metadata[self.i]
 *       self.i = self.i + 1             # <<<<<<<<<<<<<<
 *       return result
 *     else:
 */
    __pyx_v_self->i = (__pyx_v_self->i + 1);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":400
 *       result = self.metadata[self.i]
 *       self.i = self.i + 1
 *       return result             # <<<<<<<<<<<<<<
 *     else:
 *       raise StopIteration
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":397
 * 
 *   def __next__(self):
 *     if self.i < len(self.metadata):             # <<<<<<<<<<<<<<
 *       result = self.metadata[self.i]
 *       self.i = self.i + 1
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":402
 *       return result
 *     else:
 *       raise StopIteration             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
    __PYX_ERR(0, 402, __pyx_L1_error)
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":396
 *     return self
 * 
 *   def __next__(self):             # <<<<<<<<<<<<<<
 *     if self.i < len(self.metadata):
 *       result = self.metadata[self.i]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc._MetadataIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":407
 * cdef class Metadata:
 * 
 *   def __cinit__(self, metadata):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.metadata = list(metadata)
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_metadata = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_metadata,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_metadata)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 407, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_metadata = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 407, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadata.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_v_self), __pyx_v_metadata);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self, PyObject *__pyx_v_metadata) {
  PyObject *__pyx_v_metadatum = NULL;
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  grpc_metadata __pyx_t_9;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":408
 * 
 *   def __cinit__(self, metadata):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.metadata = list(metadata)
 *     for metadatum in metadata:
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":409
 *   def __cinit__(self, metadata):
 *     grpc_init()
 *     self.metadata = list(metadata)             # <<<<<<<<<<<<<<
 *     for metadatum in metadata:
 *       if not isinstance(metadatum, Metadatum):
 */
  __pyx_t_1 = PySequence_List(__pyx_v_metadata); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->metadata);
  __Pyx_DECREF(__pyx_v_self->metadata);
  __pyx_v_self->metadata = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":410
 *     grpc_init()
 *     self.metadata = list(metadata)
 *     for metadatum in metadata:             # <<<<<<<<<<<<<<
 *       if not isinstance(metadatum, Metadatum):
 *         raise TypeError("expected list of Metadatum")
 */
  if (likely(PyList_CheckExact(__pyx_v_metadata)) || PyTuple_CheckExact(__pyx_v_metadata)) {
    __pyx_t_1 = __pyx_v_metadata; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_metadata); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 410, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 410, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 410, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 410, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 410, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 410, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_metadatum, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":411
 *     self.metadata = list(metadata)
 *     for metadatum in metadata:
 *       if not isinstance(metadatum, Metadatum):             # <<<<<<<<<<<<<<
 *         raise TypeError("expected list of Metadatum")
 *     with nogil:
 */
    __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_metadatum, __pyx_ptype_4grpc_7_cython_6cygrpc_Metadatum); 
    __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
    if (__pyx_t_6) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":412
 *     for metadatum in metadata:
 *       if not isinstance(metadatum, Metadatum):
 *         raise TypeError("expected list of Metadatum")             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_metadata_array_init(&self.c_metadata_array)
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 412, __pyx_L1_error)

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":411
 *     self.metadata = list(metadata)
 *     for metadatum in metadata:
 *       if not isinstance(metadatum, Metadatum):             # <<<<<<<<<<<<<<
 *         raise TypeError("expected list of Metadatum")
 *     with nogil:
 */
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":410
 *     grpc_init()
 *     self.metadata = list(metadata)
 *     for metadatum in metadata:             # <<<<<<<<<<<<<<
 *       if not isinstance(metadatum, Metadatum):
 *         raise TypeError("expected list of Metadatum")
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":413
 *       if not isinstance(metadatum, Metadatum):
 *         raise TypeError("expected list of Metadatum")
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_metadata_array_init(&self.c_metadata_array)
 *     self.c_metadata_array.count = len(self.metadata)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":414
 *         raise TypeError("expected list of Metadatum")
 *     with nogil:
 *       grpc_metadata_array_init(&self.c_metadata_array)             # <<<<<<<<<<<<<<
 *     self.c_metadata_array.count = len(self.metadata)
 *     self.c_metadata_array.capacity = len(self.metadata)
 */
        grpc_metadata_array_init((&__pyx_v_self->c_metadata_array));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":413
 *       if not isinstance(metadatum, Metadatum):
 *         raise TypeError("expected list of Metadatum")
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_metadata_array_init(&self.c_metadata_array)
 *     self.c_metadata_array.count = len(self.metadata)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":415
 *     with nogil:
 *       grpc_metadata_array_init(&self.c_metadata_array)
 *     self.c_metadata_array.count = len(self.metadata)             # <<<<<<<<<<<<<<
 *     self.c_metadata_array.capacity = len(self.metadata)
 *     with nogil:
 */
  __pyx_t_1 = __pyx_v_self->metadata;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->c_metadata_array.count = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":416
 *       grpc_metadata_array_init(&self.c_metadata_array)
 *     self.c_metadata_array.count = len(self.metadata)
 *     self.c_metadata_array.capacity = len(self.metadata)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       self.c_metadata_array.metadata = <grpc_metadata *>gpr_malloc(
 */
  __pyx_t_1 = __pyx_v_self->metadata;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->c_metadata_array.capacity = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":417
 *     self.c_metadata_array.count = len(self.metadata)
 *     self.c_metadata_array.capacity = len(self.metadata)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_metadata_array.metadata = <grpc_metadata *>gpr_malloc(
 *           self.c_metadata_array.count*sizeof(grpc_metadata)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":418
 *     self.c_metadata_array.capacity = len(self.metadata)
 *     with nogil:
 *       self.c_metadata_array.metadata = <grpc_metadata *>gpr_malloc(             # <<<<<<<<<<<<<<
 *           self.c_metadata_array.count*sizeof(grpc_metadata)
 *       )
 */
        __pyx_v_self->c_metadata_array.metadata = ((grpc_metadata *)gpr_malloc((__pyx_v_self->c_metadata_array.count * (sizeof(grpc_metadata)))));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":417
 *     self.c_metadata_array.count = len(self.metadata)
 *     self.c_metadata_array.capacity = len(self.metadata)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_metadata_array.metadata = <grpc_metadata *>gpr_malloc(
 *           self.c_metadata_array.count*sizeof(grpc_metadata)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L11;
        }
        __pyx_L11:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":421
 *           self.c_metadata_array.count*sizeof(grpc_metadata)
 *       )
 *     for i in range(self.c_metadata_array.count):             # <<<<<<<<<<<<<<
 *       self.c_metadata_array.metadata[i] = (
 *           (<Metadatum>self.metadata[i]).c_metadata)
 */
  __pyx_t_7 = __pyx_v_self->c_metadata_array.count;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":423
 *     for i in range(self.c_metadata_array.count):
 *       self.c_metadata_array.metadata[i] = (
 *           (<Metadatum>self.metadata[i]).c_metadata)             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->metadata, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)__pyx_t_1)->c_metadata;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":422
 *       )
 *     for i in range(self.c_metadata_array.count):
 *       self.c_metadata_array.metadata[i] = (             # <<<<<<<<<<<<<<
 *           (<Metadatum>self.metadata[i]).c_metadata)
 * 
 */
    (__pyx_v_self->c_metadata_array.metadata[__pyx_v_i]) = __pyx_t_9;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":407
 * cdef class Metadata:
 * 
 *   def __cinit__(self, metadata):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.metadata = list(metadata)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadata.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_metadatum);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":425
 *           (<Metadatum>self.metadata[i]).c_metadata)
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     # this frees the allocated memory for the grpc_metadata_array (although
 *     # it'd be nice if that were documented somewhere...)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":429
 *     # it'd be nice if that were documented somewhere...)
 *     # TODO(atash): document this in the C core
 *     grpc_metadata_array_destroy(&self.c_metadata_array)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
  grpc_metadata_array_destroy((&__pyx_v_self->c_metadata_array));

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":430
 *     # TODO(atash): document this in the C core
 *     grpc_metadata_array_destroy(&self.c_metadata_array)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 *   def __len__(self):
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":425
 *           (<Metadatum>self.metadata[i]).c_metadata)
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     # this frees the allocated memory for the grpc_metadata_array (although
 *     # it'd be nice if that were documented somewhere...)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":432
 *     grpc_shutdown()
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     return self.c_metadata_array.count
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_4__len__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_4__len__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":433
 * 
 *   def __len__(self):
 *     return self.c_metadata_array.count             # <<<<<<<<<<<<<<
 * 
 *   def __getitem__(self, size_t i):
 */
  __pyx_r = __pyx_v_self->c_metadata_array.count;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":432
 *     grpc_shutdown()
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     return self.c_metadata_array.count
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":435
 *     return self.c_metadata_array.count
 * 
 *   def __getitem__(self, size_t i):             # <<<<<<<<<<<<<<
 *     return Metadatum(
 *         key=<bytes>self.c_metadata_array.metadata[i].key,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_size_t(__pyx_arg_i); if (unlikely((__pyx_v_i == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 435, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadata.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_6__getitem__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_v_self), ((size_t)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_6__getitem__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self, size_t __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":436
 * 
 *   def __getitem__(self, size_t i):
 *     return Metadatum(             # <<<<<<<<<<<<<<
 *         key=<bytes>self.c_metadata_array.metadata[i].key,
 *         value=<bytes>self.c_metadata_array.metadata[i].value[
 */
  __Pyx_XDECREF(__pyx_r);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":437
 *   def __getitem__(self, size_t i):
 *     return Metadatum(
 *         key=<bytes>self.c_metadata_array.metadata[i].key,             # <<<<<<<<<<<<<<
 *         value=<bytes>self.c_metadata_array.metadata[i].value[
 *             :self.c_metadata_array.metadata[i].value_length])
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_FromString((__pyx_v_self->c_metadata_array.metadata[__pyx_v_i]).key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_key, __pyx_t_2) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":438
 *     return Metadatum(
 *         key=<bytes>self.c_metadata_array.metadata[i].key,
 *         value=<bytes>self.c_metadata_array.metadata[i].value[             # <<<<<<<<<<<<<<
 *             :self.c_metadata_array.metadata[i].value_length])
 * 
 */
  __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize((__pyx_v_self->c_metadata_array.metadata[__pyx_v_i]).value + 0, (__pyx_v_self->c_metadata_array.metadata[__pyx_v_i]).value_length - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_value, __pyx_t_2) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":436
 * 
 *   def __getitem__(self, size_t i):
 *     return Metadatum(             # <<<<<<<<<<<<<<
 *         key=<bytes>self.c_metadata_array.metadata[i].key,
 *         value=<bytes>self.c_metadata_array.metadata[i].value[
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Metadatum), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":435
 *     return self.c_metadata_array.count
 * 
 *   def __getitem__(self, size_t i):             # <<<<<<<<<<<<<<
 *     return Metadatum(
 *         key=<bytes>self.c_metadata_array.metadata[i].key,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadata.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":441
 *             :self.c_metadata_array.metadata[i].value_length])
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return _MetadataIterator(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_9__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_9__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_8__iter__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_8Metadata_8__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":442
 * 
 *   def __iter__(self):
 *     return _MetadataIterator(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc__MetadataIterator), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":441
 *             :self.c_metadata_array.metadata[i].value_length])
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return _MetadataIterator(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Metadata.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":447
 * cdef class Operation:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.references = []
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_9Operation___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":448
 * 
 *   def __cinit__(self):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.references = []
 *     self._received_status_details = NULL
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":449
 *   def __cinit__(self):
 *     grpc_init()
 *     self.references = []             # <<<<<<<<<<<<<<
 *     self._received_status_details = NULL
 *     self._received_status_details_capacity = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->references);
  __Pyx_DECREF(__pyx_v_self->references);
  __pyx_v_self->references = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":450
 *     grpc_init()
 *     self.references = []
 *     self._received_status_details = NULL             # <<<<<<<<<<<<<<
 *     self._received_status_details_capacity = 0
 *     self.is_valid = False
 */
  __pyx_v_self->_received_status_details = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":451
 *     self.references = []
 *     self._received_status_details = NULL
 *     self._received_status_details_capacity = 0             # <<<<<<<<<<<<<<
 *     self.is_valid = False
 * 
 */
  __pyx_v_self->_received_status_details_capacity = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":452
 *     self._received_status_details = NULL
 *     self._received_status_details_capacity = 0
 *     self.is_valid = False             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __pyx_v_self->is_valid = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":447
 * cdef class Operation:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.references = []
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":455
 * 
 *   @property
 *   def type(self):             # <<<<<<<<<<<<<<
 *     return self.c_op.type
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_4type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_4type___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_4type___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":456
 *   @property
 *   def type(self):
 *     return self.c_op.type             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_grpc_op_type(__pyx_v_self->c_op.op); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":455
 * 
 *   @property
 *   def type(self):             # <<<<<<<<<<<<<<
 *     return self.c_op.type
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":459
 * 
 *   @property
 *   def flags(self):             # <<<<<<<<<<<<<<
 *     return self.c_op.flags
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_5flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_5flags_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_5flags___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_5flags___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":460
 *   @property
 *   def flags(self):
 *     return self.c_op.flags             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->c_op.flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":459
 * 
 *   @property
 *   def flags(self):             # <<<<<<<<<<<<<<
 *     return self.c_op.flags
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":463
 * 
 *   @property
 *   def has_status(self):             # <<<<<<<<<<<<<<
 *     return self.c_op.type == GRPC_OP_RECV_STATUS_ON_CLIENT
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_10has_status_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_10has_status_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_10has_status___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_10has_status___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":464
 *   @property
 *   def has_status(self):
 *     return self.c_op.type == GRPC_OP_RECV_STATUS_ON_CLIENT             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->c_op.op == GRPC_OP_RECV_STATUS_ON_CLIENT)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":463
 * 
 *   @property
 *   def has_status(self):             # <<<<<<<<<<<<<<
 *     return self.c_op.type == GRPC_OP_RECV_STATUS_ON_CLIENT
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.has_status.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":467
 * 
 *   @property
 *   def received_message(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       raise TypeError("self must be an operation receiving a message")
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_16received_message_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_16received_message_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_16received_message___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_16received_message___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":468
 *   @property
 *   def received_message(self):
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving a message")
 *     return self._received_message
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op != GRPC_OP_RECV_MESSAGE) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":469
 *   def received_message(self):
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       raise TypeError("self must be an operation receiving a message")             # <<<<<<<<<<<<<<
 *     return self._received_message
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 469, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":468
 *   @property
 *   def received_message(self):
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving a message")
 *     return self._received_message
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":470
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       raise TypeError("self must be an operation receiving a message")
 *     return self._received_message             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_received_message));
  __pyx_r = ((PyObject *)__pyx_v_self->_received_message);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":467
 * 
 *   @property
 *   def received_message(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       raise TypeError("self must be an operation receiving a message")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.received_message.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":473
 * 
 *   @property
 *   def received_message_or_none(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       return None
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_24received_message_or_none_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_24received_message_or_none_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_24received_message_or_none___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_24received_message_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":474
 *   @property
 *   def received_message_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:             # <<<<<<<<<<<<<<
 *       return None
 *     return self._received_message
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op != GRPC_OP_RECV_MESSAGE) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":475
 *   def received_message_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       return None             # <<<<<<<<<<<<<<
 *     return self._received_message
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":474
 *   @property
 *   def received_message_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:             # <<<<<<<<<<<<<<
 *       return None
 *     return self._received_message
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":476
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       return None
 *     return self._received_message             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_received_message));
  __pyx_r = ((PyObject *)__pyx_v_self->_received_message);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":473
 * 
 *   @property
 *   def received_message_or_none(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       return None
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":479
 * 
 *   @property
 *   def received_metadata(self):             # <<<<<<<<<<<<<<
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_17received_metadata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_17received_metadata_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_17received_metadata___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_17received_metadata___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":480
 *   @property
 *   def received_metadata(self):
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and             # <<<<<<<<<<<<<<
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       raise TypeError("self must be an operation receiving metadata")
 */
  switch (__pyx_v_self->c_op.op) {
    case GRPC_OP_RECV_INITIAL_METADATA:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":481
 *   def received_metadata(self):
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving metadata")
 *     return self._received_metadata
 */
    case GRPC_OP_RECV_STATUS_ON_CLIENT:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":480
 *   @property
 *   def received_metadata(self):
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and             # <<<<<<<<<<<<<<
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       raise TypeError("self must be an operation receiving metadata")
 */
    __pyx_t_1 = 0;
    break;
    default:
    __pyx_t_1 = 1;
    break;
  }
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":482
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       raise TypeError("self must be an operation receiving metadata")             # <<<<<<<<<<<<<<
 *     return self._received_metadata
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 482, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":480
 *   @property
 *   def received_metadata(self):
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and             # <<<<<<<<<<<<<<
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       raise TypeError("self must be an operation receiving metadata")
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":483
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       raise TypeError("self must be an operation receiving metadata")
 *     return self._received_metadata             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_received_metadata));
  __pyx_r = ((PyObject *)__pyx_v_self->_received_metadata);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":479
 * 
 *   @property
 *   def received_metadata(self):             # <<<<<<<<<<<<<<
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.received_metadata.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":486
 * 
 *   @property
 *   def received_metadata_or_none(self):             # <<<<<<<<<<<<<<
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_25received_metadata_or_none_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_25received_metadata_or_none_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_25received_metadata_or_none___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_25received_metadata_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":487
 *   @property
 *   def received_metadata_or_none(self):
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and             # <<<<<<<<<<<<<<
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       return None
 */
  switch (__pyx_v_self->c_op.op) {
    case GRPC_OP_RECV_INITIAL_METADATA:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":488
 *   def received_metadata_or_none(self):
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):             # <<<<<<<<<<<<<<
 *       return None
 *     return self._received_metadata
 */
    case GRPC_OP_RECV_STATUS_ON_CLIENT:

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":487
 *   @property
 *   def received_metadata_or_none(self):
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and             # <<<<<<<<<<<<<<
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       return None
 */
    __pyx_t_1 = 0;
    break;
    default:
    __pyx_t_1 = 1;
    break;
  }
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":489
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       return None             # <<<<<<<<<<<<<<
 *     return self._received_metadata
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":487
 *   @property
 *   def received_metadata_or_none(self):
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and             # <<<<<<<<<<<<<<
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       return None
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":490
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       return None
 *     return self._received_metadata             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_received_metadata));
  __pyx_r = ((PyObject *)__pyx_v_self->_received_metadata);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":486
 * 
 *   @property
 *   def received_metadata_or_none(self):             # <<<<<<<<<<<<<<
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":493
 * 
 *   @property
 *   def received_status_code(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving a status code")
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_20received_status_code_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_20received_status_code_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_20received_status_code___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_20received_status_code___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":494
 *   @property
 *   def received_status_code(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving a status code")
 *     return self._received_status_code
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op != GRPC_OP_RECV_STATUS_ON_CLIENT) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":495
 *   def received_status_code(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving a status code")             # <<<<<<<<<<<<<<
 *     return self._received_status_code
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 495, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":494
 *   @property
 *   def received_status_code(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving a status code")
 *     return self._received_status_code
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":496
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving a status code")
 *     return self._received_status_code             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(__pyx_v_self->_received_status_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":493
 * 
 *   @property
 *   def received_status_code(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving a status code")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.received_status_code.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":499
 * 
 *   @property
 *   def received_status_code_or_none(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_28received_status_code_or_none_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_28received_status_code_or_none_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_28received_status_code_or_none___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_28received_status_code_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":500
 *   @property
 *   def received_status_code_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       return None
 *     return self._received_status_code
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op != GRPC_OP_RECV_STATUS_ON_CLIENT) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":501
 *   def received_status_code_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None             # <<<<<<<<<<<<<<
 *     return self._received_status_code
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":500
 *   @property
 *   def received_status_code_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       return None
 *     return self._received_status_code
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":502
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None
 *     return self._received_status_code             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(__pyx_v_self->_received_status_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":499
 * 
 *   @property
 *   def received_status_code_or_none(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.received_status_code_or_none.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":505
 * 
 *   @property
 *   def received_status_details(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving status details")
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_23received_status_details_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_23received_status_details_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_23received_status_details___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_23received_status_details___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":506
 *   @property
 *   def received_status_details(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving status details")
 *     if self._received_status_details:
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op != GRPC_OP_RECV_STATUS_ON_CLIENT) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":507
 *   def received_status_details(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving status details")             # <<<<<<<<<<<<<<
 *     if self._received_status_details:
 *       return self._received_status_details
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 507, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":506
 *   @property
 *   def received_status_details(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving status details")
 *     if self._received_status_details:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":508
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving status details")
 *     if self._received_status_details:             # <<<<<<<<<<<<<<
 *       return self._received_status_details
 *     else:
 */
  __pyx_t_1 = (__pyx_v_self->_received_status_details != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":509
 *       raise TypeError("self must be an operation receiving status details")
 *     if self._received_status_details:
 *       return self._received_status_details             # <<<<<<<<<<<<<<
 *     else:
 *       return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->_received_status_details); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":508
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving status details")
 *     if self._received_status_details:             # <<<<<<<<<<<<<<
 *       return self._received_status_details
 *     else:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":511
 *       return self._received_status_details
 *     else:
 *       return None             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":505
 * 
 *   @property
 *   def received_status_details(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving status details")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.received_status_details.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":514
 * 
 *   @property
 *   def received_status_details_or_none(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_31received_status_details_or_none_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_31received_status_details_or_none_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_31received_status_details_or_none___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_31received_status_details_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":515
 *   @property
 *   def received_status_details_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       return None
 *     if self._received_status_details:
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op != GRPC_OP_RECV_STATUS_ON_CLIENT) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":516
 *   def received_status_details_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None             # <<<<<<<<<<<<<<
 *     if self._received_status_details:
 *       return self._received_status_details
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":515
 *   @property
 *   def received_status_details_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       return None
 *     if self._received_status_details:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":517
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None
 *     if self._received_status_details:             # <<<<<<<<<<<<<<
 *       return self._received_status_details
 *     else:
 */
  __pyx_t_1 = (__pyx_v_self->_received_status_details != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":518
 *       return None
 *     if self._received_status_details:
 *       return self._received_status_details             # <<<<<<<<<<<<<<
 *     else:
 *       return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->_received_status_details); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":517
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None
 *     if self._received_status_details:             # <<<<<<<<<<<<<<
 *       return self._received_status_details
 *     else:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":520
 *       return self._received_status_details
 *     else:
 *       return None             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":514
 * 
 *   @property
 *   def received_status_details_or_none(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       return None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.received_status_details_or_none.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":523
 * 
 *   @property
 *   def received_cancelled(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:
 *       raise TypeError("self must be an operation receiving cancellation "
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_18received_cancelled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_18received_cancelled_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_18received_cancelled___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_18received_cancelled___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":524
 *   @property
 *   def received_cancelled(self):
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving cancellation "
 *                       "information")
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op != GRPC_OP_RECV_CLOSE_ON_SERVER) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":525
 *   def received_cancelled(self):
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:
 *       raise TypeError("self must be an operation receiving cancellation "             # <<<<<<<<<<<<<<
 *                       "information")
 *     return False if self._received_cancelled == 0 else True
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 525, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":524
 *   @property
 *   def received_cancelled(self):
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:             # <<<<<<<<<<<<<<
 *       raise TypeError("self must be an operation receiving cancellation "
 *                       "information")
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":527
 *       raise TypeError("self must be an operation receiving cancellation "
 *                       "information")
 *     return False if self._received_cancelled == 0 else True             # <<<<<<<<<<<<<<
 * 
 *   @property
 */
  __Pyx_XDECREF(__pyx_r);
  if (((__pyx_v_self->_received_cancelled == 0) != 0)) {
    __Pyx_INCREF(Py_False);
    __pyx_t_2 = Py_False;
  } else {
    __Pyx_INCREF(Py_True);
    __pyx_t_2 = Py_True;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":523
 * 
 *   @property
 *   def received_cancelled(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:
 *       raise TypeError("self must be an operation receiving cancellation "
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.received_cancelled.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":530
 * 
 *   @property
 *   def received_cancelled_or_none(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:
 *       return None
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_26received_cancelled_or_none_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_26received_cancelled_or_none_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_26received_cancelled_or_none___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_26received_cancelled_or_none___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":531
 *   @property
 *   def received_cancelled_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:             # <<<<<<<<<<<<<<
 *       return None
 *     return False if self._received_cancelled == 0 else True
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op != GRPC_OP_RECV_CLOSE_ON_SERVER) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":532
 *   def received_cancelled_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:
 *       return None             # <<<<<<<<<<<<<<
 *     return False if self._received_cancelled == 0 else True
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":531
 *   @property
 *   def received_cancelled_or_none(self):
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:             # <<<<<<<<<<<<<<
 *       return None
 *     return False if self._received_cancelled == 0 else True
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":533
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:
 *       return None
 *     return False if self._received_cancelled == 0 else True             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (((__pyx_v_self->_received_cancelled == 0) != 0)) {
    __Pyx_INCREF(Py_False);
    __pyx_t_2 = Py_False;
  } else {
    __Pyx_INCREF(Py_True);
    __pyx_t_2 = Py_True;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":530
 * 
 *   @property
 *   def received_cancelled_or_none(self):             # <<<<<<<<<<<<<<
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:
 *       return None
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":535
 *     return False if self._received_cancelled == 0 else True
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     # We *almost* don't need to do anything; most of the objects are handled by
 *     # Python. The remaining one(s) are primitive fields filled in by GRPC core.
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_2__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_2__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":539
 *     # Python. The remaining one(s) are primitive fields filled in by GRPC core.
 *     # This means that we need to clean up after receive_status_on_client.
 *     if self.c_op.type == GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       gpr_free(self._received_status_details)
 *     grpc_shutdown()
 */
  __pyx_t_1 = ((__pyx_v_self->c_op.op == GRPC_OP_RECV_STATUS_ON_CLIENT) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":540
 *     # This means that we need to clean up after receive_status_on_client.
 *     if self.c_op.type == GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       gpr_free(self._received_status_details)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
    gpr_free(__pyx_v_self->_received_status_details);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":539
 *     # Python. The remaining one(s) are primitive fields filled in by GRPC core.
 *     # This means that we need to clean up after receive_status_on_client.
 *     if self.c_op.type == GRPC_OP_RECV_STATUS_ON_CLIENT:             # <<<<<<<<<<<<<<
 *       gpr_free(self._received_status_details)
 *     grpc_shutdown()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":541
 *     if self.c_op.type == GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       gpr_free(self._received_status_details)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 * def operation_send_initial_metadata(Metadata metadata, int flags):
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":535
 *     return False if self._received_cancelled == 0 else True
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     # We *almost* don't need to do anything; most of the objects are handled by
 *     # Python. The remaining one(s) are primitive fields filled in by GRPC core.
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pxd.pxi":117
 *   cdef size_t _received_status_details_capacity
 *   cdef int _received_cancelled
 *   cdef readonly bint is_valid             # <<<<<<<<<<<<<<
 *   cdef object references
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_8is_valid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_8is_valid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_9Operation_8is_valid___get__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_9Operation_8is_valid___get__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_valid); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operation.is_valid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":543
 *     grpc_shutdown()
 * 
 * def operation_send_initial_metadata(Metadata metadata, int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_INITIAL_METADATA
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_21operation_send_initial_metadata(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_21operation_send_initial_metadata = {"operation_send_initial_metadata", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_21operation_send_initial_metadata, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_21operation_send_initial_metadata(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata = 0;
  int __pyx_v_flags;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("operation_send_initial_metadata (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_metadata,&__pyx_n_s_flags,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_metadata)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("operation_send_initial_metadata", 1, 2, 2, 1); __PYX_ERR(0, 543, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "operation_send_initial_metadata") < 0)) __PYX_ERR(0, 543, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)values[0]);
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 543, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("operation_send_initial_metadata", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 543, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_send_initial_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_metadata), __pyx_ptype_4grpc_7_cython_6cygrpc_Metadata, 1, "metadata", 0))) __PYX_ERR(0, 543, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_20operation_send_initial_metadata(__pyx_self, __pyx_v_metadata, __pyx_v_flags);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_20operation_send_initial_metadata(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata, int __pyx_v_flags) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_op = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  grpc_metadata *__pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("operation_send_initial_metadata", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":544
 * 
 * def operation_send_initial_metadata(Metadata metadata, int flags):
 *   cdef Operation op = Operation()             # <<<<<<<<<<<<<<
 *   op.c_op.type = GRPC_OP_SEND_INITIAL_METADATA
 *   op.c_op.flags = flags
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operation), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":545
 * def operation_send_initial_metadata(Metadata metadata, int flags):
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_INITIAL_METADATA             # <<<<<<<<<<<<<<
 *   op.c_op.flags = flags
 *   op.c_op.data.send_initial_metadata.count = metadata.c_metadata_array.count
 */
  __pyx_v_op->c_op.op = GRPC_OP_SEND_INITIAL_METADATA;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":546
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_INITIAL_METADATA
 *   op.c_op.flags = flags             # <<<<<<<<<<<<<<
 *   op.c_op.data.send_initial_metadata.count = metadata.c_metadata_array.count
 *   op.c_op.data.send_initial_metadata.metadata = (
 */
  __pyx_v_op->c_op.flags = __pyx_v_flags;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":547
 *   op.c_op.type = GRPC_OP_SEND_INITIAL_METADATA
 *   op.c_op.flags = flags
 *   op.c_op.data.send_initial_metadata.count = metadata.c_metadata_array.count             # <<<<<<<<<<<<<<
 *   op.c_op.data.send_initial_metadata.metadata = (
 *       metadata.c_metadata_array.metadata)
 */
  __pyx_t_2 = __pyx_v_metadata->c_metadata_array.count;
  __pyx_v_op->c_op.data.send_initial_metadata.count = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":549
 *   op.c_op.data.send_initial_metadata.count = metadata.c_metadata_array.count
 *   op.c_op.data.send_initial_metadata.metadata = (
 *       metadata.c_metadata_array.metadata)             # <<<<<<<<<<<<<<
 *   op.references.append(metadata)
 *   op.is_valid = True
 */
  __pyx_t_3 = __pyx_v_metadata->c_metadata_array.metadata;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":548
 *   op.c_op.flags = flags
 *   op.c_op.data.send_initial_metadata.count = metadata.c_metadata_array.count
 *   op.c_op.data.send_initial_metadata.metadata = (             # <<<<<<<<<<<<<<
 *       metadata.c_metadata_array.metadata)
 *   op.references.append(metadata)
 */
  __pyx_v_op->c_op.data.send_initial_metadata.metadata = __pyx_t_3;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":550
 *   op.c_op.data.send_initial_metadata.metadata = (
 *       metadata.c_metadata_array.metadata)
 *   op.references.append(metadata)             # <<<<<<<<<<<<<<
 *   op.is_valid = True
 *   return op
 */
  __pyx_t_4 = __Pyx_PyObject_Append(__pyx_v_op->references, ((PyObject *)__pyx_v_metadata)); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 550, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":551
 *       metadata.c_metadata_array.metadata)
 *   op.references.append(metadata)
 *   op.is_valid = True             # <<<<<<<<<<<<<<
 *   return op
 * 
 */
  __pyx_v_op->is_valid = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":552
 *   op.references.append(metadata)
 *   op.is_valid = True
 *   return op             # <<<<<<<<<<<<<<
 * 
 * def operation_send_message(data, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = ((PyObject *)__pyx_v_op);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":543
 *     grpc_shutdown()
 * 
 * def operation_send_initial_metadata(Metadata metadata, int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_INITIAL_METADATA
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_send_initial_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":554
 *   return op
 * 
 * def operation_send_message(data, int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_MESSAGE
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_23operation_send_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_23operation_send_message = {"operation_send_message", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_23operation_send_message, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_23operation_send_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  int __pyx_v_flags;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("operation_send_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,&__pyx_n_s_flags,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("operation_send_message", 1, 2, 2, 1); __PYX_ERR(0, 554, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "operation_send_message") < 0)) __PYX_ERR(0, 554, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_data = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 554, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("operation_send_message", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 554, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_22operation_send_message(__pyx_self, __pyx_v_data, __pyx_v_flags);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_22operation_send_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data, int __pyx_v_flags) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_op = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *__pyx_v_byte_buffer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  grpc_byte_buffer *__pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("operation_send_message", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":555
 * 
 * def operation_send_message(data, int flags):
 *   cdef Operation op = Operation()             # <<<<<<<<<<<<<<
 *   op.c_op.type = GRPC_OP_SEND_MESSAGE
 *   op.c_op.flags = flags
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operation), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":556
 * def operation_send_message(data, int flags):
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_MESSAGE             # <<<<<<<<<<<<<<
 *   op.c_op.flags = flags
 *   byte_buffer = ByteBuffer(data)
 */
  __pyx_v_op->c_op.op = GRPC_OP_SEND_MESSAGE;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":557
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_MESSAGE
 *   op.c_op.flags = flags             # <<<<<<<<<<<<<<
 *   byte_buffer = ByteBuffer(data)
 *   op.c_op.data.send_message = byte_buffer.c_byte_buffer
 */
  __pyx_v_op->c_op.flags = __pyx_v_flags;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":558
 *   op.c_op.type = GRPC_OP_SEND_MESSAGE
 *   op.c_op.flags = flags
 *   byte_buffer = ByteBuffer(data)             # <<<<<<<<<<<<<<
 *   op.c_op.data.send_message = byte_buffer.c_byte_buffer
 *   op.references.append(byte_buffer)
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_data);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_ByteBuffer), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_byte_buffer = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":559
 *   op.c_op.flags = flags
 *   byte_buffer = ByteBuffer(data)
 *   op.c_op.data.send_message = byte_buffer.c_byte_buffer             # <<<<<<<<<<<<<<
 *   op.references.append(byte_buffer)
 *   op.is_valid = True
 */
  __pyx_t_3 = __pyx_v_byte_buffer->c_byte_buffer;
  __pyx_v_op->c_op.data.send_message = __pyx_t_3;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":560
 *   byte_buffer = ByteBuffer(data)
 *   op.c_op.data.send_message = byte_buffer.c_byte_buffer
 *   op.references.append(byte_buffer)             # <<<<<<<<<<<<<<
 *   op.is_valid = True
 *   return op
 */
  __pyx_t_4 = __Pyx_PyObject_Append(__pyx_v_op->references, ((PyObject *)__pyx_v_byte_buffer)); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 560, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":561
 *   op.c_op.data.send_message = byte_buffer.c_byte_buffer
 *   op.references.append(byte_buffer)
 *   op.is_valid = True             # <<<<<<<<<<<<<<
 *   return op
 * 
 */
  __pyx_v_op->is_valid = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":562
 *   op.references.append(byte_buffer)
 *   op.is_valid = True
 *   return op             # <<<<<<<<<<<<<<
 * 
 * def operation_send_close_from_client(int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = ((PyObject *)__pyx_v_op);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":554
 *   return op
 * 
 * def operation_send_message(data, int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_MESSAGE
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XDECREF((PyObject *)__pyx_v_byte_buffer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":564
 *   return op
 * 
 * def operation_send_close_from_client(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_25operation_send_close_from_client(PyObject *__pyx_self, PyObject *__pyx_arg_flags); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_25operation_send_close_from_client = {"operation_send_close_from_client", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_25operation_send_close_from_client, METH_O, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_25operation_send_close_from_client(PyObject *__pyx_self, PyObject *__pyx_arg_flags) {
  int __pyx_v_flags;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("operation_send_close_from_client (wrapper)", 0);
  assert(__pyx_arg_flags); {
    __pyx_v_flags = __Pyx_PyInt_As_int(__pyx_arg_flags); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_send_close_from_client", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_24operation_send_close_from_client(__pyx_self, ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_24operation_send_close_from_client(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_op = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("operation_send_close_from_client", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":565
 * 
 * def operation_send_close_from_client(int flags):
 *   cdef Operation op = Operation()             # <<<<<<<<<<<<<<
 *   op.c_op.type = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 *   op.c_op.flags = flags
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operation), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":566
 * def operation_send_close_from_client(int flags):
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_CLOSE_FROM_CLIENT             # <<<<<<<<<<<<<<
 *   op.c_op.flags = flags
 *   op.is_valid = True
 */
  __pyx_v_op->c_op.op = GRPC_OP_SEND_CLOSE_FROM_CLIENT;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":567
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 *   op.c_op.flags = flags             # <<<<<<<<<<<<<<
 *   op.is_valid = True
 *   return op
 */
  __pyx_v_op->c_op.flags = __pyx_v_flags;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":568
 *   op.c_op.type = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 *   op.c_op.flags = flags
 *   op.is_valid = True             # <<<<<<<<<<<<<<
 *   return op
 * 
 */
  __pyx_v_op->is_valid = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":569
 *   op.c_op.flags = flags
 *   op.is_valid = True
 *   return op             # <<<<<<<<<<<<<<
 * 
 * def operation_send_status_from_server(
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = ((PyObject *)__pyx_v_op);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":564
 *   return op
 * 
 * def operation_send_close_from_client(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_send_close_from_client", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":571
 *   return op
 * 
 * def operation_send_status_from_server(             # <<<<<<<<<<<<<<
 *     Metadata metadata, grpc_status_code code, bytes details, int flags):
 *   cdef Operation op = Operation()
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_27operation_send_status_from_server(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_27operation_send_status_from_server = {"operation_send_status_from_server", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_27operation_send_status_from_server, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_27operation_send_status_from_server(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata = 0;
  grpc_status_code __pyx_v_code;
  PyObject *__pyx_v_details = 0;
  int __pyx_v_flags;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("operation_send_status_from_server (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_metadata,&__pyx_n_s_code,&__pyx_n_s_details,&__pyx_n_s_flags,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_metadata)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_code)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("operation_send_status_from_server", 1, 4, 4, 1); __PYX_ERR(0, 571, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_details)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("operation_send_status_from_server", 1, 4, 4, 2); __PYX_ERR(0, 571, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("operation_send_status_from_server", 1, 4, 4, 3); __PYX_ERR(0, 571, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "operation_send_status_from_server") < 0)) __PYX_ERR(0, 571, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)values[0]);
    __pyx_v_code = ((grpc_status_code)__Pyx_PyInt_As_grpc_status_code(values[1])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L3_error)
    __pyx_v_details = ((PyObject*)values[2]);
    __pyx_v_flags = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("operation_send_status_from_server", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 571, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_send_status_from_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_metadata), __pyx_ptype_4grpc_7_cython_6cygrpc_Metadata, 1, "metadata", 0))) __PYX_ERR(0, 572, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_details), (&PyBytes_Type), 1, "details", 1))) __PYX_ERR(0, 572, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_26operation_send_status_from_server(__pyx_self, __pyx_v_metadata, __pyx_v_code, __pyx_v_details, __pyx_v_flags);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_26operation_send_status_from_server(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *__pyx_v_metadata, grpc_status_code __pyx_v_code, PyObject *__pyx_v_details, int __pyx_v_flags) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_op = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  grpc_metadata *__pyx_t_3;
  char const *__pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("operation_send_status_from_server", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":573
 * def operation_send_status_from_server(
 *     Metadata metadata, grpc_status_code code, bytes details, int flags):
 *   cdef Operation op = Operation()             # <<<<<<<<<<<<<<
 *   op.c_op.type = GRPC_OP_SEND_STATUS_FROM_SERVER
 *   op.c_op.flags = flags
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operation), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":574
 *     Metadata metadata, grpc_status_code code, bytes details, int flags):
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_STATUS_FROM_SERVER             # <<<<<<<<<<<<<<
 *   op.c_op.flags = flags
 *   op.c_op.data.send_status_from_server.trailing_metadata_count = (
 */
  __pyx_v_op->c_op.op = GRPC_OP_SEND_STATUS_FROM_SERVER;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":575
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_STATUS_FROM_SERVER
 *   op.c_op.flags = flags             # <<<<<<<<<<<<<<
 *   op.c_op.data.send_status_from_server.trailing_metadata_count = (
 *       metadata.c_metadata_array.count)
 */
  __pyx_v_op->c_op.flags = __pyx_v_flags;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":577
 *   op.c_op.flags = flags
 *   op.c_op.data.send_status_from_server.trailing_metadata_count = (
 *       metadata.c_metadata_array.count)             # <<<<<<<<<<<<<<
 *   op.c_op.data.send_status_from_server.trailing_metadata = (
 *       metadata.c_metadata_array.metadata)
 */
  __pyx_t_2 = __pyx_v_metadata->c_metadata_array.count;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":576
 *   op.c_op.type = GRPC_OP_SEND_STATUS_FROM_SERVER
 *   op.c_op.flags = flags
 *   op.c_op.data.send_status_from_server.trailing_metadata_count = (             # <<<<<<<<<<<<<<
 *       metadata.c_metadata_array.count)
 *   op.c_op.data.send_status_from_server.trailing_metadata = (
 */
  __pyx_v_op->c_op.data.send_status_from_server.trailing_metadata_count = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":579
 *       metadata.c_metadata_array.count)
 *   op.c_op.data.send_status_from_server.trailing_metadata = (
 *       metadata.c_metadata_array.metadata)             # <<<<<<<<<<<<<<
 *   op.c_op.data.send_status_from_server.status = code
 *   op.c_op.data.send_status_from_server.status_details = details
 */
  __pyx_t_3 = __pyx_v_metadata->c_metadata_array.metadata;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":578
 *   op.c_op.data.send_status_from_server.trailing_metadata_count = (
 *       metadata.c_metadata_array.count)
 *   op.c_op.data.send_status_from_server.trailing_metadata = (             # <<<<<<<<<<<<<<
 *       metadata.c_metadata_array.metadata)
 *   op.c_op.data.send_status_from_server.status = code
 */
  __pyx_v_op->c_op.data.send_status_from_server.trailing_metadata = __pyx_t_3;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":580
 *   op.c_op.data.send_status_from_server.trailing_metadata = (
 *       metadata.c_metadata_array.metadata)
 *   op.c_op.data.send_status_from_server.status = code             # <<<<<<<<<<<<<<
 *   op.c_op.data.send_status_from_server.status_details = details
 *   op.references.append(metadata)
 */
  __pyx_v_op->c_op.data.send_status_from_server.status = __pyx_v_code;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":581
 *       metadata.c_metadata_array.metadata)
 *   op.c_op.data.send_status_from_server.status = code
 *   op.c_op.data.send_status_from_server.status_details = details             # <<<<<<<<<<<<<<
 *   op.references.append(metadata)
 *   op.references.append(details)
 */
  __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_details); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L1_error)
  __pyx_v_op->c_op.data.send_status_from_server.status_details = __pyx_t_4;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":582
 *   op.c_op.data.send_status_from_server.status = code
 *   op.c_op.data.send_status_from_server.status_details = details
 *   op.references.append(metadata)             # <<<<<<<<<<<<<<
 *   op.references.append(details)
 *   op.is_valid = True
 */
  __pyx_t_5 = __Pyx_PyObject_Append(__pyx_v_op->references, ((PyObject *)__pyx_v_metadata)); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 582, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":583
 *   op.c_op.data.send_status_from_server.status_details = details
 *   op.references.append(metadata)
 *   op.references.append(details)             # <<<<<<<<<<<<<<
 *   op.is_valid = True
 *   return op
 */
  __pyx_t_5 = __Pyx_PyObject_Append(__pyx_v_op->references, __pyx_v_details); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 583, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":584
 *   op.references.append(metadata)
 *   op.references.append(details)
 *   op.is_valid = True             # <<<<<<<<<<<<<<
 *   return op
 * 
 */
  __pyx_v_op->is_valid = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":585
 *   op.references.append(details)
 *   op.is_valid = True
 *   return op             # <<<<<<<<<<<<<<
 * 
 * def operation_receive_initial_metadata(int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = ((PyObject *)__pyx_v_op);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":571
 *   return op
 * 
 * def operation_send_status_from_server(             # <<<<<<<<<<<<<<
 *     Metadata metadata, grpc_status_code code, bytes details, int flags):
 *   cdef Operation op = Operation()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_send_status_from_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":587
 *   return op
 * 
 * def operation_receive_initial_metadata(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_INITIAL_METADATA
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_29operation_receive_initial_metadata(PyObject *__pyx_self, PyObject *__pyx_arg_flags); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_29operation_receive_initial_metadata = {"operation_receive_initial_metadata", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_29operation_receive_initial_metadata, METH_O, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_29operation_receive_initial_metadata(PyObject *__pyx_self, PyObject *__pyx_arg_flags) {
  int __pyx_v_flags;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("operation_receive_initial_metadata (wrapper)", 0);
  assert(__pyx_arg_flags); {
    __pyx_v_flags = __Pyx_PyInt_As_int(__pyx_arg_flags); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 587, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_receive_initial_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_28operation_receive_initial_metadata(__pyx_self, ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_28operation_receive_initial_metadata(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_op = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("operation_receive_initial_metadata", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":588
 * 
 * def operation_receive_initial_metadata(int flags):
 *   cdef Operation op = Operation()             # <<<<<<<<<<<<<<
 *   op.c_op.type = GRPC_OP_RECV_INITIAL_METADATA
 *   op.c_op.flags = flags
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operation), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":589
 * def operation_receive_initial_metadata(int flags):
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_INITIAL_METADATA             # <<<<<<<<<<<<<<
 *   op.c_op.flags = flags
 *   op._received_metadata = Metadata([])
 */
  __pyx_v_op->c_op.op = GRPC_OP_RECV_INITIAL_METADATA;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":590
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_INITIAL_METADATA
 *   op.c_op.flags = flags             # <<<<<<<<<<<<<<
 *   op._received_metadata = Metadata([])
 *   op.c_op.data.receive_initial_metadata = (
 */
  __pyx_v_op->c_op.flags = __pyx_v_flags;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":591
 *   op.c_op.type = GRPC_OP_RECV_INITIAL_METADATA
 *   op.c_op.flags = flags
 *   op._received_metadata = Metadata([])             # <<<<<<<<<<<<<<
 *   op.c_op.data.receive_initial_metadata = (
 *       &op._received_metadata.c_metadata_array)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Metadata), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_op->_received_metadata);
  __Pyx_DECREF(((PyObject *)__pyx_v_op->_received_metadata));
  __pyx_v_op->_received_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":592
 *   op.c_op.flags = flags
 *   op._received_metadata = Metadata([])
 *   op.c_op.data.receive_initial_metadata = (             # <<<<<<<<<<<<<<
 *       &op._received_metadata.c_metadata_array)
 *   op.is_valid = True
 */
  __pyx_v_op->c_op.data.recv_initial_metadata = (&__pyx_v_op->_received_metadata->c_metadata_array);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":594
 *   op.c_op.data.receive_initial_metadata = (
 *       &op._received_metadata.c_metadata_array)
 *   op.is_valid = True             # <<<<<<<<<<<<<<
 *   return op
 * 
 */
  __pyx_v_op->is_valid = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":595
 *       &op._received_metadata.c_metadata_array)
 *   op.is_valid = True
 *   return op             # <<<<<<<<<<<<<<
 * 
 * def operation_receive_message(int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = ((PyObject *)__pyx_v_op);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":587
 *   return op
 * 
 * def operation_receive_initial_metadata(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_INITIAL_METADATA
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_receive_initial_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":597
 *   return op
 * 
 * def operation_receive_message(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_31operation_receive_message(PyObject *__pyx_self, PyObject *__pyx_arg_flags); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_31operation_receive_message = {"operation_receive_message", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_31operation_receive_message, METH_O, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_31operation_receive_message(PyObject *__pyx_self, PyObject *__pyx_arg_flags) {
  int __pyx_v_flags;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("operation_receive_message (wrapper)", 0);
  assert(__pyx_arg_flags); {
    __pyx_v_flags = __Pyx_PyInt_As_int(__pyx_arg_flags); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 597, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_receive_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_30operation_receive_message(__pyx_self, ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_30operation_receive_message(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_op = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("operation_receive_message", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":598
 * 
 * def operation_receive_message(int flags):
 *   cdef Operation op = Operation()             # <<<<<<<<<<<<<<
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE
 *   op.c_op.flags = flags
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operation), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":599
 * def operation_receive_message(int flags):
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE             # <<<<<<<<<<<<<<
 *   op.c_op.flags = flags
 *   op._received_message = ByteBuffer(None)
 */
  __pyx_v_op->c_op.op = GRPC_OP_RECV_MESSAGE;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":600
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE
 *   op.c_op.flags = flags             # <<<<<<<<<<<<<<
 *   op._received_message = ByteBuffer(None)
 *   # n.b. the c_op.data.receive_message field needs to be deleted by us,
 */
  __pyx_v_op->c_op.flags = __pyx_v_flags;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":601
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE
 *   op.c_op.flags = flags
 *   op._received_message = ByteBuffer(None)             # <<<<<<<<<<<<<<
 *   # n.b. the c_op.data.receive_message field needs to be deleted by us,
 *   # anyway, so we just let that be handled by the ByteBuffer() we allocated
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_ByteBuffer), __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_op->_received_message);
  __Pyx_DECREF(((PyObject *)__pyx_v_op->_received_message));
  __pyx_v_op->_received_message = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":605
 *   # anyway, so we just let that be handled by the ByteBuffer() we allocated
 *   # the line before.
 *   op.c_op.data.receive_message = &op._received_message.c_byte_buffer             # <<<<<<<<<<<<<<
 *   op.is_valid = True
 *   return op
 */
  __pyx_v_op->c_op.data.recv_message = (&__pyx_v_op->_received_message->c_byte_buffer);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":606
 *   # the line before.
 *   op.c_op.data.receive_message = &op._received_message.c_byte_buffer
 *   op.is_valid = True             # <<<<<<<<<<<<<<
 *   return op
 * 
 */
  __pyx_v_op->is_valid = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":607
 *   op.c_op.data.receive_message = &op._received_message.c_byte_buffer
 *   op.is_valid = True
 *   return op             # <<<<<<<<<<<<<<
 * 
 * def operation_receive_status_on_client(int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = ((PyObject *)__pyx_v_op);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":597
 *   return op
 * 
 * def operation_receive_message(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_receive_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":609
 *   return op
 * 
 * def operation_receive_status_on_client(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_STATUS_ON_CLIENT
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_33operation_receive_status_on_client(PyObject *__pyx_self, PyObject *__pyx_arg_flags); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_33operation_receive_status_on_client = {"operation_receive_status_on_client", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_33operation_receive_status_on_client, METH_O, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_33operation_receive_status_on_client(PyObject *__pyx_self, PyObject *__pyx_arg_flags) {
  int __pyx_v_flags;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("operation_receive_status_on_client (wrapper)", 0);
  assert(__pyx_arg_flags); {
    __pyx_v_flags = __Pyx_PyInt_As_int(__pyx_arg_flags); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 609, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_receive_status_on_client", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_32operation_receive_status_on_client(__pyx_self, ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_32operation_receive_status_on_client(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_op = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("operation_receive_status_on_client", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":610
 * 
 * def operation_receive_status_on_client(int flags):
 *   cdef Operation op = Operation()             # <<<<<<<<<<<<<<
 *   op.c_op.type = GRPC_OP_RECV_STATUS_ON_CLIENT
 *   op.c_op.flags = flags
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operation), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":611
 * def operation_receive_status_on_client(int flags):
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_STATUS_ON_CLIENT             # <<<<<<<<<<<<<<
 *   op.c_op.flags = flags
 *   op._received_metadata = Metadata([])
 */
  __pyx_v_op->c_op.op = GRPC_OP_RECV_STATUS_ON_CLIENT;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":612
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_STATUS_ON_CLIENT
 *   op.c_op.flags = flags             # <<<<<<<<<<<<<<
 *   op._received_metadata = Metadata([])
 *   op.c_op.data.receive_status_on_client.trailing_metadata = (
 */
  __pyx_v_op->c_op.flags = __pyx_v_flags;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":613
 *   op.c_op.type = GRPC_OP_RECV_STATUS_ON_CLIENT
 *   op.c_op.flags = flags
 *   op._received_metadata = Metadata([])             # <<<<<<<<<<<<<<
 *   op.c_op.data.receive_status_on_client.trailing_metadata = (
 *       &op._received_metadata.c_metadata_array)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Metadata), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_op->_received_metadata);
  __Pyx_DECREF(((PyObject *)__pyx_v_op->_received_metadata));
  __pyx_v_op->_received_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":614
 *   op.c_op.flags = flags
 *   op._received_metadata = Metadata([])
 *   op.c_op.data.receive_status_on_client.trailing_metadata = (             # <<<<<<<<<<<<<<
 *       &op._received_metadata.c_metadata_array)
 *   op.c_op.data.receive_status_on_client.status = (
 */
  __pyx_v_op->c_op.data.recv_status_on_client.trailing_metadata = (&__pyx_v_op->_received_metadata->c_metadata_array);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":616
 *   op.c_op.data.receive_status_on_client.trailing_metadata = (
 *       &op._received_metadata.c_metadata_array)
 *   op.c_op.data.receive_status_on_client.status = (             # <<<<<<<<<<<<<<
 *       &op._received_status_code)
 *   op.c_op.data.receive_status_on_client.status_details = (
 */
  __pyx_v_op->c_op.data.recv_status_on_client.status = (&__pyx_v_op->_received_status_code);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":618
 *   op.c_op.data.receive_status_on_client.status = (
 *       &op._received_status_code)
 *   op.c_op.data.receive_status_on_client.status_details = (             # <<<<<<<<<<<<<<
 *       &op._received_status_details)
 *   op.c_op.data.receive_status_on_client.status_details_capacity = (
 */
  __pyx_v_op->c_op.data.recv_status_on_client.status_details = (&__pyx_v_op->_received_status_details);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":620
 *   op.c_op.data.receive_status_on_client.status_details = (
 *       &op._received_status_details)
 *   op.c_op.data.receive_status_on_client.status_details_capacity = (             # <<<<<<<<<<<<<<
 *       &op._received_status_details_capacity)
 *   op.is_valid = True
 */
  __pyx_v_op->c_op.data.recv_status_on_client.status_details_capacity = (&__pyx_v_op->_received_status_details_capacity);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":622
 *   op.c_op.data.receive_status_on_client.status_details_capacity = (
 *       &op._received_status_details_capacity)
 *   op.is_valid = True             # <<<<<<<<<<<<<<
 *   return op
 * 
 */
  __pyx_v_op->is_valid = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":623
 *       &op._received_status_details_capacity)
 *   op.is_valid = True
 *   return op             # <<<<<<<<<<<<<<
 * 
 * def operation_receive_close_on_server(int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = ((PyObject *)__pyx_v_op);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":609
 *   return op
 * 
 * def operation_receive_status_on_client(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_STATUS_ON_CLIENT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_receive_status_on_client", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":625
 *   return op
 * 
 * def operation_receive_close_on_server(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_CLOSE_ON_SERVER
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_35operation_receive_close_on_server(PyObject *__pyx_self, PyObject *__pyx_arg_flags); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_35operation_receive_close_on_server = {"operation_receive_close_on_server", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_35operation_receive_close_on_server, METH_O, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_35operation_receive_close_on_server(PyObject *__pyx_self, PyObject *__pyx_arg_flags) {
  int __pyx_v_flags;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("operation_receive_close_on_server (wrapper)", 0);
  assert(__pyx_arg_flags); {
    __pyx_v_flags = __Pyx_PyInt_As_int(__pyx_arg_flags); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 625, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_receive_close_on_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_34operation_receive_close_on_server(__pyx_self, ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_34operation_receive_close_on_server(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_flags) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *__pyx_v_op = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("operation_receive_close_on_server", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":626
 * 
 * def operation_receive_close_on_server(int flags):
 *   cdef Operation op = Operation()             # <<<<<<<<<<<<<<
 *   op.c_op.type = GRPC_OP_RECV_CLOSE_ON_SERVER
 *   op.c_op.flags = flags
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operation), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":627
 * def operation_receive_close_on_server(int flags):
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_CLOSE_ON_SERVER             # <<<<<<<<<<<<<<
 *   op.c_op.flags = flags
 *   op.c_op.data.receive_close_on_server.cancelled = &op._received_cancelled
 */
  __pyx_v_op->c_op.op = GRPC_OP_RECV_CLOSE_ON_SERVER;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":628
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_CLOSE_ON_SERVER
 *   op.c_op.flags = flags             # <<<<<<<<<<<<<<
 *   op.c_op.data.receive_close_on_server.cancelled = &op._received_cancelled
 *   op.is_valid = True
 */
  __pyx_v_op->c_op.flags = __pyx_v_flags;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":629
 *   op.c_op.type = GRPC_OP_RECV_CLOSE_ON_SERVER
 *   op.c_op.flags = flags
 *   op.c_op.data.receive_close_on_server.cancelled = &op._received_cancelled             # <<<<<<<<<<<<<<
 *   op.is_valid = True
 *   return op
 */
  __pyx_v_op->c_op.data.recv_close_on_server.cancelled = (&__pyx_v_op->_received_cancelled);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":630
 *   op.c_op.flags = flags
 *   op.c_op.data.receive_close_on_server.cancelled = &op._received_cancelled
 *   op.is_valid = True             # <<<<<<<<<<<<<<
 *   return op
 * 
 */
  __pyx_v_op->is_valid = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":631
 *   op.c_op.data.receive_close_on_server.cancelled = &op._received_cancelled
 *   op.is_valid = True
 *   return op             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = ((PyObject *)__pyx_v_op);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":625
 *   return op
 * 
 * def operation_receive_close_on_server(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_CLOSE_ON_SERVER
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.operation_receive_close_on_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":639
 *   cdef Operations operations
 * 
 *   def __cinit__(self, Operations operations not None):             # <<<<<<<<<<<<<<
 *     self.i = 0
 *     self.operations = operations
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_operations = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_operations,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_operations)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 639, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 639, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc._OperationsIterator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_operations), __pyx_ptype_4grpc_7_cython_6cygrpc_Operations, 0, "operations", 0))) __PYX_ERR(0, 639, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *)__pyx_v_self), __pyx_v_operations);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_operations) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":640
 * 
 *   def __cinit__(self, Operations operations not None):
 *     self.i = 0             # <<<<<<<<<<<<<<
 *     self.operations = operations
 * 
 */
  __pyx_v_self->i = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":641
 *   def __cinit__(self, Operations operations not None):
 *     self.i = 0
 *     self.operations = operations             # <<<<<<<<<<<<<<
 * 
 *   def __iter__(self):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_operations));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_operations));
  __Pyx_GOTREF(__pyx_v_self->operations);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->operations));
  __pyx_v_self->operations = __pyx_v_operations;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":639
 *   cdef Operations operations
 * 
 *   def __cinit__(self, Operations operations not None):             # <<<<<<<<<<<<<<
 *     self.i = 0
 *     self.operations = operations
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":643
 *     self.operations = operations
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_3__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator_2__iter__(((struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator_2__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":644
 * 
 *   def __iter__(self):
 *     return self             # <<<<<<<<<<<<<<
 * 
 *   def __next__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":643
 *     self.operations = operations
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":646
 *     return self
 * 
 *   def __next__(self):             # <<<<<<<<<<<<<<
 *     if self.i < len(self.operations):
 *       result = self.operations[self.i]
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_5__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_5__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator_4__next__(((struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_19_OperationsIterator_4__next__(struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":647
 * 
 *   def __next__(self):
 *     if self.i < len(self.operations):             # <<<<<<<<<<<<<<
 *       result = self.operations[self.i]
 *       self.i = self.i + 1
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->operations);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((__pyx_v_self->i < __pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":648
 *   def __next__(self):
 *     if self.i < len(self.operations):
 *       result = self.operations[self.i]             # <<<<<<<<<<<<<<
 *       self.i = self.i + 1
 *       return result
 */
    __pyx_t_1 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->operations), __pyx_v_self->i, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_result = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":649
 *     if self.i < len(self.operations):
 *       result = self.operations[self.i]
 *       self.i = self.i + 1             # <<<<<<<<<<<<<<
 *       return result
 *     else:
 */
    __pyx_v_self->i = (__pyx_v_self->i + 1);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":650
 *       result = self.operations[self.i]
 *       self.i = self.i + 1
 *       return result             # <<<<<<<<<<<<<<
 *     else:
 *       raise StopIteration
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":647
 * 
 *   def __next__(self):
 *     if self.i < len(self.operations):             # <<<<<<<<<<<<<<
 *       result = self.operations[self.i]
 *       self.i = self.i + 1
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":652
 *       return result
 *     else:
 *       raise StopIteration             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
    __PYX_ERR(0, 652, __pyx_L1_error)
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":646
 *     return self
 * 
 *   def __next__(self):             # <<<<<<<<<<<<<<
 *     if self.i < len(self.operations):
 *       result = self.operations[self.i]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc._OperationsIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":657
 * cdef class Operations:
 * 
 *   def __cinit__(self, operations):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.operations = list(operations)  # normalize iterable
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_operations = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_operations,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_operations)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 657, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_operations = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 657, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operations.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10Operations___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)__pyx_v_self), __pyx_v_operations);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_10Operations___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self, PyObject *__pyx_v_operations) {
  PyObject *__pyx_v_operation = NULL;
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  grpc_op __pyx_t_8;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":658
 * 
 *   def __cinit__(self, operations):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     self.operations = list(operations)  # normalize iterable
 *     self.c_ops = NULL
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":659
 *   def __cinit__(self, operations):
 *     grpc_init()
 *     self.operations = list(operations)  # normalize iterable             # <<<<<<<<<<<<<<
 *     self.c_ops = NULL
 *     self.c_nops = 0
 */
  __pyx_t_1 = PySequence_List(__pyx_v_operations); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->operations);
  __Pyx_DECREF(__pyx_v_self->operations);
  __pyx_v_self->operations = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":660
 *     grpc_init()
 *     self.operations = list(operations)  # normalize iterable
 *     self.c_ops = NULL             # <<<<<<<<<<<<<<
 *     self.c_nops = 0
 *     for operation in self.operations:
 */
  __pyx_v_self->c_ops = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":661
 *     self.operations = list(operations)  # normalize iterable
 *     self.c_ops = NULL
 *     self.c_nops = 0             # <<<<<<<<<<<<<<
 *     for operation in self.operations:
 *       if not isinstance(operation, Operation):
 */
  __pyx_v_self->c_nops = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":662
 *     self.c_ops = NULL
 *     self.c_nops = 0
 *     for operation in self.operations:             # <<<<<<<<<<<<<<
 *       if not isinstance(operation, Operation):
 *         raise TypeError("expected operations to be iterable of Operation")
 */
  if (unlikely(__pyx_v_self->operations == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 662, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->operations; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 662, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_operation, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":663
 *     self.c_nops = 0
 *     for operation in self.operations:
 *       if not isinstance(operation, Operation):             # <<<<<<<<<<<<<<
 *         raise TypeError("expected operations to be iterable of Operation")
 *     self.c_nops = len(self.operations)
 */
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_operation, __pyx_ptype_4grpc_7_cython_6cygrpc_Operation); 
    __pyx_t_5 = ((!(__pyx_t_4 != 0)) != 0);
    if (__pyx_t_5) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":664
 *     for operation in self.operations:
 *       if not isinstance(operation, Operation):
 *         raise TypeError("expected operations to be iterable of Operation")             # <<<<<<<<<<<<<<
 *     self.c_nops = len(self.operations)
 *     with nogil:
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 664, __pyx_L1_error)

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":663
 *     self.c_nops = 0
 *     for operation in self.operations:
 *       if not isinstance(operation, Operation):             # <<<<<<<<<<<<<<
 *         raise TypeError("expected operations to be iterable of Operation")
 *     self.c_nops = len(self.operations)
 */
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":662
 *     self.c_ops = NULL
 *     self.c_nops = 0
 *     for operation in self.operations:             # <<<<<<<<<<<<<<
 *       if not isinstance(operation, Operation):
 *         raise TypeError("expected operations to be iterable of Operation")
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":665
 *       if not isinstance(operation, Operation):
 *         raise TypeError("expected operations to be iterable of Operation")
 *     self.c_nops = len(self.operations)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       self.c_ops = <grpc_op *>gpr_malloc(sizeof(grpc_op)*self.c_nops)
 */
  __pyx_t_1 = __pyx_v_self->operations;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 665, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->c_nops = __pyx_t_2;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":666
 *         raise TypeError("expected operations to be iterable of Operation")
 *     self.c_nops = len(self.operations)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_ops = <grpc_op *>gpr_malloc(sizeof(grpc_op)*self.c_nops)
 *     for i in range(self.c_nops):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":667
 *     self.c_nops = len(self.operations)
 *     with nogil:
 *       self.c_ops = <grpc_op *>gpr_malloc(sizeof(grpc_op)*self.c_nops)             # <<<<<<<<<<<<<<
 *     for i in range(self.c_nops):
 *       self.c_ops[i] = (<Operation>(self.operations[i])).c_op
 */
        __pyx_v_self->c_ops = ((grpc_op *)gpr_malloc(((sizeof(grpc_op)) * __pyx_v_self->c_nops)));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":666
 *         raise TypeError("expected operations to be iterable of Operation")
 *     self.c_nops = len(self.operations)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_ops = <grpc_op *>gpr_malloc(sizeof(grpc_op)*self.c_nops)
 *     for i in range(self.c_nops):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":668
 *     with nogil:
 *       self.c_ops = <grpc_op *>gpr_malloc(sizeof(grpc_op)*self.c_nops)
 *     for i in range(self.c_nops):             # <<<<<<<<<<<<<<
 *       self.c_ops[i] = (<Operation>(self.operations[i])).c_op
 * 
 */
  __pyx_t_6 = __pyx_v_self->c_nops;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":669
 *       self.c_ops = <grpc_op *>gpr_malloc(sizeof(grpc_op)*self.c_nops)
 *     for i in range(self.c_nops):
 *       self.c_ops[i] = (<Operation>(self.operations[i])).c_op             # <<<<<<<<<<<<<<
 * 
 *   def __len__(self):
 */
    if (unlikely(__pyx_v_self->operations == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 669, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_self->operations, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)__pyx_t_1)->c_op;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->c_ops[__pyx_v_i]) = __pyx_t_8;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":657
 * cdef class Operations:
 * 
 *   def __cinit__(self, operations):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     self.operations = list(operations)  # normalize iterable
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operations.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_operation);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":671
 *       self.c_ops[i] = (<Operation>(self.operations[i])).c_op
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     return self.c_nops
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_3__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10Operations_2__len__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_4grpc_7_cython_6cygrpc_10Operations_2__len__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":672
 * 
 *   def __len__(self):
 *     return self.c_nops             # <<<<<<<<<<<<<<
 * 
 *   def __getitem__(self, size_t i):
 */
  __pyx_r = __pyx_v_self->c_nops;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":671
 *       self.c_ops[i] = (<Operation>(self.operations[i])).c_op
 * 
 *   def __len__(self):             # <<<<<<<<<<<<<<
 *     return self.c_nops
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":674
 *     return self.c_nops
 * 
 *   def __getitem__(self, size_t i):             # <<<<<<<<<<<<<<
 *     # self.operations is never stale; it's only updated from this file
 *     return self.operations[i]
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10Operations_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10Operations_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_size_t(__pyx_arg_i); if (unlikely((__pyx_v_i == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operations.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10Operations_4__getitem__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)__pyx_v_self), ((size_t)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10Operations_4__getitem__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self, size_t __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":676
 *   def __getitem__(self, size_t i):
 *     # self.operations is never stale; it's only updated from this file
 *     return self.operations[i]             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->operations == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 676, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_self->operations, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":674
 *     return self.c_nops
 * 
 *   def __getitem__(self, size_t i):             # <<<<<<<<<<<<<<
 *     # self.operations is never stale; it's only updated from this file
 *     return self.operations[i]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operations.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":678
 *     return self.operations[i]
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       gpr_free(self.c_ops)
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_7__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_7__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_10Operations_6__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_10Operations_6__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":679
 * 
 *   def __dealloc__(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_free(self.c_ops)
 *     grpc_shutdown()
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":680
 *   def __dealloc__(self):
 *     with nogil:
 *       gpr_free(self.c_ops)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 * 
 */
        gpr_free(__pyx_v_self->c_ops);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":679
 * 
 *   def __dealloc__(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       gpr_free(self.c_ops)
 *     grpc_shutdown()
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":681
 *     with nogil:
 *       gpr_free(self.c_ops)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 * 
 *   def __iter__(self):
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":678
 *     return self.operations[i]
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       gpr_free(self.c_ops)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":683
 *     grpc_shutdown()
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return _OperationsIterator(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10Operations_9__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_10Operations_9__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_10Operations_8__iter__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_10Operations_8__iter__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":684
 * 
 *   def __iter__(self):
 *     return _OperationsIterator(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc__OperationsIterator), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":683
 *     grpc_shutdown()
 * 
 *   def __iter__(self):             # <<<<<<<<<<<<<<
 *     return _OperationsIterator(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Operations.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":689
 * cdef class CompressionOptions:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_compression_options_init(&self.c_options)
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":690
 * 
 *   def __cinit__(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_compression_options_init(&self.c_options)
 * 
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":691
 *   def __cinit__(self):
 *     with nogil:
 *       grpc_compression_options_init(&self.c_options)             # <<<<<<<<<<<<<<
 * 
 *   def enable_algorithm(self, grpc_compression_algorithm algorithm):
 */
        grpc_compression_options_init((&__pyx_v_self->c_options));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":690
 * 
 *   def __cinit__(self):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_compression_options_init(&self.c_options)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":689
 * cdef class CompressionOptions:
 * 
 *   def __cinit__(self):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_compression_options_init(&self.c_options)
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":693
 *       grpc_compression_options_init(&self.c_options)
 * 
 *   def enable_algorithm(self, grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_compression_options_enable_algorithm(&self.c_options, algorithm)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_3enable_algorithm(PyObject *__pyx_v_self, PyObject *__pyx_arg_algorithm); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_3enable_algorithm(PyObject *__pyx_v_self, PyObject *__pyx_arg_algorithm) {
  grpc_compression_algorithm __pyx_v_algorithm;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enable_algorithm (wrapper)", 0);
  assert(__pyx_arg_algorithm); {
    __pyx_v_algorithm = ((grpc_compression_algorithm)__Pyx_PyInt_As_grpc_compression_algorithm(__pyx_arg_algorithm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 693, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompressionOptions.enable_algorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_2enable_algorithm(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *)__pyx_v_self), ((grpc_compression_algorithm)__pyx_v_algorithm));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_2enable_algorithm(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self, grpc_compression_algorithm __pyx_v_algorithm) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enable_algorithm", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":694
 * 
 *   def enable_algorithm(self, grpc_compression_algorithm algorithm):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_compression_options_enable_algorithm(&self.c_options, algorithm)
 * 
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":695
 *   def enable_algorithm(self, grpc_compression_algorithm algorithm):
 *     with nogil:
 *       grpc_compression_options_enable_algorithm(&self.c_options, algorithm)             # <<<<<<<<<<<<<<
 * 
 *   def disable_algorithm(self, grpc_compression_algorithm algorithm):
 */
        grpc_compression_options_enable_algorithm((&__pyx_v_self->c_options), __pyx_v_algorithm);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":694
 * 
 *   def enable_algorithm(self, grpc_compression_algorithm algorithm):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_compression_options_enable_algorithm(&self.c_options, algorithm)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":693
 *       grpc_compression_options_init(&self.c_options)
 * 
 *   def enable_algorithm(self, grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_compression_options_enable_algorithm(&self.c_options, algorithm)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":697
 *       grpc_compression_options_enable_algorithm(&self.c_options, algorithm)
 * 
 *   def disable_algorithm(self, grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_compression_options_disable_algorithm(&self.c_options, algorithm)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_5disable_algorithm(PyObject *__pyx_v_self, PyObject *__pyx_arg_algorithm); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_5disable_algorithm(PyObject *__pyx_v_self, PyObject *__pyx_arg_algorithm) {
  grpc_compression_algorithm __pyx_v_algorithm;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("disable_algorithm (wrapper)", 0);
  assert(__pyx_arg_algorithm); {
    __pyx_v_algorithm = ((grpc_compression_algorithm)__Pyx_PyInt_As_grpc_compression_algorithm(__pyx_arg_algorithm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 697, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompressionOptions.disable_algorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_4disable_algorithm(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *)__pyx_v_self), ((grpc_compression_algorithm)__pyx_v_algorithm));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_4disable_algorithm(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self, grpc_compression_algorithm __pyx_v_algorithm) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("disable_algorithm", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":698
 * 
 *   def disable_algorithm(self, grpc_compression_algorithm algorithm):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_compression_options_disable_algorithm(&self.c_options, algorithm)
 * 
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":699
 *   def disable_algorithm(self, grpc_compression_algorithm algorithm):
 *     with nogil:
 *       grpc_compression_options_disable_algorithm(&self.c_options, algorithm)             # <<<<<<<<<<<<<<
 * 
 *   def is_algorithm_enabled(self, grpc_compression_algorithm algorithm):
 */
        grpc_compression_options_disable_algorithm((&__pyx_v_self->c_options), __pyx_v_algorithm);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":698
 * 
 *   def disable_algorithm(self, grpc_compression_algorithm algorithm):
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_compression_options_disable_algorithm(&self.c_options, algorithm)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":697
 *       grpc_compression_options_enable_algorithm(&self.c_options, algorithm)
 * 
 *   def disable_algorithm(self, grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_compression_options_disable_algorithm(&self.c_options, algorithm)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":701
 *       grpc_compression_options_disable_algorithm(&self.c_options, algorithm)
 * 
 *   def is_algorithm_enabled(self, grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *     cdef int result
 *     with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_7is_algorithm_enabled(PyObject *__pyx_v_self, PyObject *__pyx_arg_algorithm); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_7is_algorithm_enabled(PyObject *__pyx_v_self, PyObject *__pyx_arg_algorithm) {
  grpc_compression_algorithm __pyx_v_algorithm;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_algorithm_enabled (wrapper)", 0);
  assert(__pyx_arg_algorithm); {
    __pyx_v_algorithm = ((grpc_compression_algorithm)__Pyx_PyInt_As_grpc_compression_algorithm(__pyx_arg_algorithm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 701, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompressionOptions.is_algorithm_enabled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_6is_algorithm_enabled(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *)__pyx_v_self), ((grpc_compression_algorithm)__pyx_v_algorithm));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_6is_algorithm_enabled(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self, grpc_compression_algorithm __pyx_v_algorithm) {
  int __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("is_algorithm_enabled", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":703
 *   def is_algorithm_enabled(self, grpc_compression_algorithm algorithm):
 *     cdef int result
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_compression_options_is_algorithm_enabled(
 *           &self.c_options, algorithm)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":704
 *     cdef int result
 *     with nogil:
 *       result = grpc_compression_options_is_algorithm_enabled(             # <<<<<<<<<<<<<<
 *           &self.c_options, algorithm)
 *     return result
 */
        __pyx_v_result = grpc_compression_options_is_algorithm_enabled((&__pyx_v_self->c_options), __pyx_v_algorithm);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":703
 *   def is_algorithm_enabled(self, grpc_compression_algorithm algorithm):
 *     cdef int result
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_compression_options_is_algorithm_enabled(
 *           &self.c_options, algorithm)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":706
 *       result = grpc_compression_options_is_algorithm_enabled(
 *           &self.c_options, algorithm)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   def to_channel_arg(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":701
 *       grpc_compression_options_disable_algorithm(&self.c_options, algorithm)
 * 
 *   def is_algorithm_enabled(self, grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *     cdef int result
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompressionOptions.is_algorithm_enabled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":708
 *     return result
 * 
 *   def to_channel_arg(self):             # <<<<<<<<<<<<<<
 *     return ChannelArg(GRPC_COMPRESSION_CHANNEL_ENABLED_ALGORITHMS_BITSET,
 *                       self.c_options.enabled_algorithms_bitset)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_9to_channel_arg(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_9to_channel_arg(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_channel_arg (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_8to_channel_arg(((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_18CompressionOptions_8to_channel_arg(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("to_channel_arg", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":709
 * 
 *   def to_channel_arg(self):
 *     return ChannelArg(GRPC_COMPRESSION_CHANNEL_ENABLED_ALGORITHMS_BITSET,             # <<<<<<<<<<<<<<
 *                       self.c_options.enabled_algorithms_bitset)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(GRPC_COMPRESSION_CHANNEL_ENABLED_ALGORITHMS_BITSET); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":710
 *   def to_channel_arg(self):
 *     return ChannelArg(GRPC_COMPRESSION_CHANNEL_ENABLED_ALGORITHMS_BITSET,
 *                       self.c_options.enabled_algorithms_bitset)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->c_options.enabled_algorithms_bitset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":709
 * 
 *   def to_channel_arg(self):
 *     return ChannelArg(GRPC_COMPRESSION_CHANNEL_ENABLED_ALGORITHMS_BITSET,             # <<<<<<<<<<<<<<
 *                       self.c_options.enabled_algorithms_bitset)
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_ChannelArg), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":708
 *     return result
 * 
 *   def to_channel_arg(self):             # <<<<<<<<<<<<<<
 *     return ChannelArg(GRPC_COMPRESSION_CHANNEL_ENABLED_ALGORITHMS_BITSET,
 *                       self.c_options.enabled_algorithms_bitset)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("grpc._cython.cygrpc.CompressionOptions.to_channel_arg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":713
 * 
 * 
 * def compression_algorithm_name(grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *   cdef char* name
 *   with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_37compression_algorithm_name(PyObject *__pyx_self, PyObject *__pyx_arg_algorithm); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_37compression_algorithm_name = {"compression_algorithm_name", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_37compression_algorithm_name, METH_O, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_37compression_algorithm_name(PyObject *__pyx_self, PyObject *__pyx_arg_algorithm) {
  grpc_compression_algorithm __pyx_v_algorithm;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compression_algorithm_name (wrapper)", 0);
  assert(__pyx_arg_algorithm); {
    __pyx_v_algorithm = ((grpc_compression_algorithm)__Pyx_PyInt_As_grpc_compression_algorithm(__pyx_arg_algorithm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 713, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.compression_algorithm_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_36compression_algorithm_name(__pyx_self, ((grpc_compression_algorithm)__pyx_v_algorithm));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_36compression_algorithm_name(CYTHON_UNUSED PyObject *__pyx_self, grpc_compression_algorithm __pyx_v_algorithm) {
  char *__pyx_v_name;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("compression_algorithm_name", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":715
 * def compression_algorithm_name(grpc_compression_algorithm algorithm):
 *   cdef char* name
 *   with nogil:             # <<<<<<<<<<<<<<
 *     grpc_compression_algorithm_name(algorithm, &name)
 *   # Let Cython do the right thing with string casting
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":716
 *   cdef char* name
 *   with nogil:
 *     grpc_compression_algorithm_name(algorithm, &name)             # <<<<<<<<<<<<<<
 *   # Let Cython do the right thing with string casting
 *   return name
 */
        grpc_compression_algorithm_name(__pyx_v_algorithm, (&__pyx_v_name));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":715
 * def compression_algorithm_name(grpc_compression_algorithm algorithm):
 *   cdef char* name
 *   with nogil:             # <<<<<<<<<<<<<<
 *     grpc_compression_algorithm_name(algorithm, &name)
 *   # Let Cython do the right thing with string casting
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":718
 *     grpc_compression_algorithm_name(algorithm, &name)
 *   # Let Cython do the right thing with string casting
 *   return name             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":713
 * 
 * 
 * def compression_algorithm_name(grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *   cdef char* name
 *   with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.compression_algorithm_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":35
 * 
 * 
 * cdef grpc_ssl_roots_override_result ssl_roots_override_callback(             # <<<<<<<<<<<<<<
 *     char **pem_root_certs) with gil:
 *   temporary_pem_root_certs = pkg_resources.resource_string(
 */

static grpc_ssl_roots_override_result __pyx_f_4grpc_7_cython_6cygrpc_ssl_roots_override_callback(char **__pyx_v_pem_root_certs) {
  PyObject *__pyx_v_temporary_pem_root_certs = NULL;
  grpc_ssl_roots_override_result __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  char *__pyx_t_4;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("ssl_roots_override_callback", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":37
 * cdef grpc_ssl_roots_override_result ssl_roots_override_callback(
 *     char **pem_root_certs) with gil:
 *   temporary_pem_root_certs = pkg_resources.resource_string(             # <<<<<<<<<<<<<<
 *       'grpc._cython', '_credentials/roots.pem')
 *   pem_root_certs[0] = <char *>gpr_malloc(len(temporary_pem_root_certs) + 1)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_pkg_resources); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_resource_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_temporary_pem_root_certs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":39
 *   temporary_pem_root_certs = pkg_resources.resource_string(
 *       'grpc._cython', '_credentials/roots.pem')
 *   pem_root_certs[0] = <char *>gpr_malloc(len(temporary_pem_root_certs) + 1)             # <<<<<<<<<<<<<<
 *   memcpy(
 *       pem_root_certs[0], <char *>temporary_pem_root_certs,
 */
  __pyx_t_3 = PyObject_Length(__pyx_v_temporary_pem_root_certs); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(8, 39, __pyx_L1_error)
  (__pyx_v_pem_root_certs[0]) = ((char *)gpr_malloc((__pyx_t_3 + 1)));

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":41
 *   pem_root_certs[0] = <char *>gpr_malloc(len(temporary_pem_root_certs) + 1)
 *   memcpy(
 *       pem_root_certs[0], <char *>temporary_pem_root_certs,             # <<<<<<<<<<<<<<
 *       len(temporary_pem_root_certs))
 *   pem_root_certs[0][len(temporary_pem_root_certs)] = '\0'
 */
  __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_temporary_pem_root_certs); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(8, 41, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":42
 *   memcpy(
 *       pem_root_certs[0], <char *>temporary_pem_root_certs,
 *       len(temporary_pem_root_certs))             # <<<<<<<<<<<<<<
 *   pem_root_certs[0][len(temporary_pem_root_certs)] = '\0'
 *   return GRPC_SSL_ROOTS_OVERRIDE_OK
 */
  __pyx_t_3 = PyObject_Length(__pyx_v_temporary_pem_root_certs); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(8, 42, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":40
 *       'grpc._cython', '_credentials/roots.pem')
 *   pem_root_certs[0] = <char *>gpr_malloc(len(temporary_pem_root_certs) + 1)
 *   memcpy(             # <<<<<<<<<<<<<<
 *       pem_root_certs[0], <char *>temporary_pem_root_certs,
 *       len(temporary_pem_root_certs))
 */
  memcpy((__pyx_v_pem_root_certs[0]), ((char *)__pyx_t_4), __pyx_t_3);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":43
 *       pem_root_certs[0], <char *>temporary_pem_root_certs,
 *       len(temporary_pem_root_certs))
 *   pem_root_certs[0][len(temporary_pem_root_certs)] = '\0'             # <<<<<<<<<<<<<<
 *   return GRPC_SSL_ROOTS_OVERRIDE_OK
 */
  __pyx_t_3 = PyObject_Length(__pyx_v_temporary_pem_root_certs); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(8, 43, __pyx_L1_error)
  ((__pyx_v_pem_root_certs[0])[__pyx_t_3]) = '\x00';

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":44
 *       len(temporary_pem_root_certs))
 *   pem_root_certs[0][len(temporary_pem_root_certs)] = '\0'
 *   return GRPC_SSL_ROOTS_OVERRIDE_OK             # <<<<<<<<<<<<<<
 */
  __pyx_r = GRPC_SSL_ROOTS_OVERRIDE_OK;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":35
 * 
 * 
 * cdef grpc_ssl_roots_override_result ssl_roots_override_callback(             # <<<<<<<<<<<<<<
 *     char **pem_root_certs) with gil:
 *   temporary_pem_root_certs = pkg_resources.resource_string(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("grpc._cython.cygrpc.ssl_roots_override_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = (grpc_ssl_roots_override_result) 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_temporary_pem_root_certs);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":37
 * cdef class Server:
 * 
 *   def __cinit__(self, ChannelArgs arguments=None):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     cdef grpc_channel_args *c_arguments = NULL
 */

/* Python wrapper */
static int __pyx_pw_4grpc_7_cython_6cygrpc_6Server_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4grpc_7_cython_6cygrpc_6Server_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_arguments = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arguments,0};
    PyObject* values[1] = {0};
    values[0] = (PyObject *)((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_arguments);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(4, 37, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_arguments = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 37, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_arguments), __pyx_ptype_4grpc_7_cython_6cygrpc_ChannelArgs, 1, "arguments", 0))) __PYX_ERR(4, 37, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6Server___cinit__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self), __pyx_v_arguments);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4grpc_7_cython_6cygrpc_6Server___cinit__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *__pyx_v_arguments) {
  grpc_channel_args *__pyx_v_c_arguments;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":38
 * 
 *   def __cinit__(self, ChannelArgs arguments=None):
 *     grpc_init()             # <<<<<<<<<<<<<<
 *     cdef grpc_channel_args *c_arguments = NULL
 *     self.references = []
 */
  grpc_init();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":39
 *   def __cinit__(self, ChannelArgs arguments=None):
 *     grpc_init()
 *     cdef grpc_channel_args *c_arguments = NULL             # <<<<<<<<<<<<<<
 *     self.references = []
 *     self.registered_completion_queues = []
 */
  __pyx_v_c_arguments = NULL;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":40
 *     grpc_init()
 *     cdef grpc_channel_args *c_arguments = NULL
 *     self.references = []             # <<<<<<<<<<<<<<
 *     self.registered_completion_queues = []
 *     if arguments is not None:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->references);
  __Pyx_DECREF(__pyx_v_self->references);
  __pyx_v_self->references = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":41
 *     cdef grpc_channel_args *c_arguments = NULL
 *     self.references = []
 *     self.registered_completion_queues = []             # <<<<<<<<<<<<<<
 *     if arguments is not None:
 *       c_arguments = &arguments.c_args
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->registered_completion_queues);
  __Pyx_DECREF(__pyx_v_self->registered_completion_queues);
  __pyx_v_self->registered_completion_queues = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":42
 *     self.references = []
 *     self.registered_completion_queues = []
 *     if arguments is not None:             # <<<<<<<<<<<<<<
 *       c_arguments = &arguments.c_args
 *       self.references.append(arguments)
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_arguments) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":43
 *     self.registered_completion_queues = []
 *     if arguments is not None:
 *       c_arguments = &arguments.c_args             # <<<<<<<<<<<<<<
 *       self.references.append(arguments)
 *     with nogil:
 */
    __pyx_v_c_arguments = (&__pyx_v_arguments->c_args);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":44
 *     if arguments is not None:
 *       c_arguments = &arguments.c_args
 *       self.references.append(arguments)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       self.c_server = grpc_server_create(c_arguments, NULL)
 */
    if (unlikely(__pyx_v_self->references == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
      __PYX_ERR(4, 44, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_self->references, ((PyObject *)__pyx_v_arguments)); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(4, 44, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":42
 *     self.references = []
 *     self.registered_completion_queues = []
 *     if arguments is not None:             # <<<<<<<<<<<<<<
 *       c_arguments = &arguments.c_args
 *       self.references.append(arguments)
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":45
 *       c_arguments = &arguments.c_args
 *       self.references.append(arguments)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_server = grpc_server_create(c_arguments, NULL)
 *     self.is_started = False
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":46
 *       self.references.append(arguments)
 *     with nogil:
 *       self.c_server = grpc_server_create(c_arguments, NULL)             # <<<<<<<<<<<<<<
 *     self.is_started = False
 *     self.is_shutting_down = False
 */
        __pyx_v_self->c_server = grpc_server_create(__pyx_v_c_arguments, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":45
 *       c_arguments = &arguments.c_args
 *       self.references.append(arguments)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       self.c_server = grpc_server_create(c_arguments, NULL)
 *     self.is_started = False
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":47
 *     with nogil:
 *       self.c_server = grpc_server_create(c_arguments, NULL)
 *     self.is_started = False             # <<<<<<<<<<<<<<
 *     self.is_shutting_down = False
 *     self.is_shutdown = False
 */
  __pyx_v_self->is_started = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":48
 *       self.c_server = grpc_server_create(c_arguments, NULL)
 *     self.is_started = False
 *     self.is_shutting_down = False             # <<<<<<<<<<<<<<
 *     self.is_shutdown = False
 * 
 */
  __pyx_v_self->is_shutting_down = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":49
 *     self.is_started = False
 *     self.is_shutting_down = False
 *     self.is_shutdown = False             # <<<<<<<<<<<<<<
 * 
 *   def request_call(
 */
  __pyx_v_self->is_shutdown = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":37
 * cdef class Server:
 * 
 *   def __cinit__(self, ChannelArgs arguments=None):             # <<<<<<<<<<<<<<
 *     grpc_init()
 *     cdef grpc_channel_args *c_arguments = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":51
 *     self.is_shutdown = False
 * 
 *   def request_call(             # <<<<<<<<<<<<<<
 *       self, CompletionQueue call_queue not None,
 *       CompletionQueue server_queue not None, tag):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_3request_call(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_3request_call(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_call_queue = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_server_queue = 0;
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("request_call (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_call_queue,&__pyx_n_s_server_queue,&__pyx_n_s_tag,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_call_queue)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_server_queue)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("request_call", 1, 3, 3, 1); __PYX_ERR(4, 51, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tag)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("request_call", 1, 3, 3, 2); __PYX_ERR(4, 51, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "request_call") < 0)) __PYX_ERR(4, 51, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_call_queue = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)values[0]);
    __pyx_v_server_queue = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)values[1]);
    __pyx_v_tag = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("request_call", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 51, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.request_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_call_queue), __pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue, 0, "call_queue", 0))) __PYX_ERR(4, 52, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_server_queue), __pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue, 0, "server_queue", 0))) __PYX_ERR(4, 53, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6Server_2request_call(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self), __pyx_v_call_queue, __pyx_v_server_queue, __pyx_v_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_2request_call(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_call_queue, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_server_queue, PyObject *__pyx_v_tag) {
  grpc_call_error __pyx_v_result;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *__pyx_v_operation_tag = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("request_call", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":54
 *       self, CompletionQueue call_queue not None,
 *       CompletionQueue server_queue not None, tag):
 *     if not self.is_started or self.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise ValueError("server must be started and not shutting down")
 *     if server_queue not in self.registered_completion_queues:
 */
  __pyx_t_2 = ((!(__pyx_v_self->is_started != 0)) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->is_shutting_down != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":55
 *       CompletionQueue server_queue not None, tag):
 *     if not self.is_started or self.is_shutting_down:
 *       raise ValueError("server must be started and not shutting down")             # <<<<<<<<<<<<<<
 *     if server_queue not in self.registered_completion_queues:
 *       raise ValueError("server_queue must be a registered completion queue")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 55, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":54
 *       self, CompletionQueue call_queue not None,
 *       CompletionQueue server_queue not None, tag):
 *     if not self.is_started or self.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise ValueError("server must be started and not shutting down")
 *     if server_queue not in self.registered_completion_queues:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":56
 *     if not self.is_started or self.is_shutting_down:
 *       raise ValueError("server must be started and not shutting down")
 *     if server_queue not in self.registered_completion_queues:             # <<<<<<<<<<<<<<
 *       raise ValueError("server_queue must be a registered completion queue")
 *     cdef grpc_call_error result
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_v_server_queue), __pyx_v_self->registered_completion_queues, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(4, 56, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":57
 *       raise ValueError("server must be started and not shutting down")
 *     if server_queue not in self.registered_completion_queues:
 *       raise ValueError("server_queue must be a registered completion queue")             # <<<<<<<<<<<<<<
 *     cdef grpc_call_error result
 *     cdef OperationTag operation_tag = OperationTag(tag)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 57, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":56
 *     if not self.is_started or self.is_shutting_down:
 *       raise ValueError("server must be started and not shutting down")
 *     if server_queue not in self.registered_completion_queues:             # <<<<<<<<<<<<<<
 *       raise ValueError("server_queue must be a registered completion queue")
 *     cdef grpc_call_error result
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":59
 *       raise ValueError("server_queue must be a registered completion queue")
 *     cdef grpc_call_error result
 *     cdef OperationTag operation_tag = OperationTag(tag)             # <<<<<<<<<<<<<<
 *     operation_tag.operation_call = Call()
 *     operation_tag.request_call_details = CallDetails()
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_tag);
  __Pyx_GIVEREF(__pyx_v_tag);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_tag);
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_OperationTag), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_operation_tag = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":60
 *     cdef grpc_call_error result
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     operation_tag.operation_call = Call()             # <<<<<<<<<<<<<<
 *     operation_tag.request_call_details = CallDetails()
 *     operation_tag.request_metadata = Metadata([])
 */
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Call), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_operation_tag->operation_call);
  __Pyx_DECREF(((PyObject *)__pyx_v_operation_tag->operation_call));
  __pyx_v_operation_tag->operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":61
 *     cdef OperationTag operation_tag = OperationTag(tag)
 *     operation_tag.operation_call = Call()
 *     operation_tag.request_call_details = CallDetails()             # <<<<<<<<<<<<<<
 *     operation_tag.request_metadata = Metadata([])
 *     operation_tag.references.extend([self, call_queue, server_queue])
 */
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_CallDetails), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_operation_tag->request_call_details);
  __Pyx_DECREF(((PyObject *)__pyx_v_operation_tag->request_call_details));
  __pyx_v_operation_tag->request_call_details = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":62
 *     operation_tag.operation_call = Call()
 *     operation_tag.request_call_details = CallDetails()
 *     operation_tag.request_metadata = Metadata([])             # <<<<<<<<<<<<<<
 *     operation_tag.references.extend([self, call_queue, server_queue])
 *     operation_tag.is_new_request = True
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Metadata), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_operation_tag->request_metadata);
  __Pyx_DECREF(((PyObject *)__pyx_v_operation_tag->request_metadata));
  __pyx_v_operation_tag->request_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":63
 *     operation_tag.request_call_details = CallDetails()
 *     operation_tag.request_metadata = Metadata([])
 *     operation_tag.references.extend([self, call_queue, server_queue])             # <<<<<<<<<<<<<<
 *     operation_tag.is_new_request = True
 *     operation_tag.batch_operations = Operations([])
 */
  if (unlikely(__pyx_v_operation_tag->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "extend");
    __PYX_ERR(4, 63, __pyx_L1_error)
  }
  __pyx_t_4 = PyList_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyList_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(((PyObject *)__pyx_v_call_queue));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_call_queue));
  PyList_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_call_queue));
  __Pyx_INCREF(((PyObject *)__pyx_v_server_queue));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_server_queue));
  PyList_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_server_queue));
  __pyx_t_5 = __Pyx_PyList_Extend(__pyx_v_operation_tag->references, __pyx_t_4); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(4, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":64
 *     operation_tag.request_metadata = Metadata([])
 *     operation_tag.references.extend([self, call_queue, server_queue])
 *     operation_tag.is_new_request = True             # <<<<<<<<<<<<<<
 *     operation_tag.batch_operations = Operations([])
 *     cpython.Py_INCREF(operation_tag)
 */
  __pyx_v_operation_tag->is_new_request = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":65
 *     operation_tag.references.extend([self, call_queue, server_queue])
 *     operation_tag.is_new_request = True
 *     operation_tag.batch_operations = Operations([])             # <<<<<<<<<<<<<<
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Operations), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_operation_tag->batch_operations);
  __Pyx_DECREF(((PyObject *)__pyx_v_operation_tag->batch_operations));
  __pyx_v_operation_tag->batch_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":66
 *     operation_tag.is_new_request = True
 *     operation_tag.batch_operations = Operations([])
 *     cpython.Py_INCREF(operation_tag)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       result = grpc_server_request_call(
 */
  Py_INCREF(((PyObject *)__pyx_v_operation_tag));

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":67
 *     operation_tag.batch_operations = Operations([])
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_server_request_call(
 *           self.c_server, &operation_tag.operation_call.c_call,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":68
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:
 *       result = grpc_server_request_call(             # <<<<<<<<<<<<<<
 *           self.c_server, &operation_tag.operation_call.c_call,
 *           &operation_tag.request_call_details.c_details,
 */
        __pyx_v_result = grpc_server_request_call(__pyx_v_self->c_server, (&__pyx_v_operation_tag->operation_call->c_call), (&__pyx_v_operation_tag->request_call_details->c_details), (&__pyx_v_operation_tag->request_metadata->c_metadata_array), __pyx_v_call_queue->c_completion_queue, __pyx_v_server_queue->c_completion_queue, ((PyObject *)__pyx_v_operation_tag));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":67
 *     operation_tag.batch_operations = Operations([])
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       result = grpc_server_request_call(
 *           self.c_server, &operation_tag.operation_call.c_call,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":74
 *           call_queue.c_completion_queue, server_queue.c_completion_queue,
 *           <cpython.PyObject *>operation_tag)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   def register_completion_queue(
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_From_grpc_call_error(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":51
 *     self.is_shutdown = False
 * 
 *   def request_call(             # <<<<<<<<<<<<<<
 *       self, CompletionQueue call_queue not None,
 *       CompletionQueue server_queue not None, tag):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.request_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_operation_tag);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":76
 *     return result
 * 
 *   def register_completion_queue(             # <<<<<<<<<<<<<<
 *       self, CompletionQueue queue not None):
 *     if self.is_started:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_5register_completion_queue(PyObject *__pyx_v_self, PyObject *__pyx_v_queue); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_5register_completion_queue(PyObject *__pyx_v_self, PyObject *__pyx_v_queue) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("register_completion_queue (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_queue), __pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue, 0, "queue", 0))) __PYX_ERR(4, 77, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6Server_4register_completion_queue(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self), ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_queue));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_4register_completion_queue(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("register_completion_queue", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":78
 *   def register_completion_queue(
 *       self, CompletionQueue queue not None):
 *     if self.is_started:             # <<<<<<<<<<<<<<
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:
 */
  __pyx_t_1 = (__pyx_v_self->is_started != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":79
 *       self, CompletionQueue queue not None):
 *     if self.is_started:
 *       raise ValueError("cannot register completion queues after start")             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_server_register_completion_queue(
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 79, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":78
 *   def register_completion_queue(
 *       self, CompletionQueue queue not None):
 *     if self.is_started:             # <<<<<<<<<<<<<<
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":80
 *     if self.is_started:
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_server_register_completion_queue(
 *           self.c_server, queue.c_completion_queue, NULL)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":81
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:
 *       grpc_server_register_completion_queue(             # <<<<<<<<<<<<<<
 *           self.c_server, queue.c_completion_queue, NULL)
 *     self.registered_completion_queues.append(queue)
 */
        grpc_server_register_completion_queue(__pyx_v_self->c_server, __pyx_v_queue->c_completion_queue, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":80
 *     if self.is_started:
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_server_register_completion_queue(
 *           self.c_server, queue.c_completion_queue, NULL)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":83
 *       grpc_server_register_completion_queue(
 *           self.c_server, queue.c_completion_queue, NULL)
 *     self.registered_completion_queues.append(queue)             # <<<<<<<<<<<<<<
 * 
 *   def register_non_listening_completion_queue(
 */
  if (unlikely(__pyx_v_self->registered_completion_queues == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(4, 83, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_self->registered_completion_queues, ((PyObject *)__pyx_v_queue)); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(4, 83, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":76
 *     return result
 * 
 *   def register_completion_queue(             # <<<<<<<<<<<<<<
 *       self, CompletionQueue queue not None):
 *     if self.is_started:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.register_completion_queue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":85
 *     self.registered_completion_queues.append(queue)
 * 
 *   def register_non_listening_completion_queue(             # <<<<<<<<<<<<<<
 *       self, CompletionQueue queue not None):
 *     if self.is_started:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_7register_non_listening_completion_queue(PyObject *__pyx_v_self, PyObject *__pyx_v_queue); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_7register_non_listening_completion_queue(PyObject *__pyx_v_self, PyObject *__pyx_v_queue) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("register_non_listening_completion_queue (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_queue), __pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue, 0, "queue", 0))) __PYX_ERR(4, 86, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6Server_6register_non_listening_completion_queue(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self), ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_v_queue));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_6register_non_listening_completion_queue(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("register_non_listening_completion_queue", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":87
 *   def register_non_listening_completion_queue(
 *       self, CompletionQueue queue not None):
 *     if self.is_started:             # <<<<<<<<<<<<<<
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:
 */
  __pyx_t_1 = (__pyx_v_self->is_started != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":88
 *       self, CompletionQueue queue not None):
 *     if self.is_started:
 *       raise ValueError("cannot register completion queues after start")             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_server_register_non_listening_completion_queue(
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 88, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":87
 *   def register_non_listening_completion_queue(
 *       self, CompletionQueue queue not None):
 *     if self.is_started:             # <<<<<<<<<<<<<<
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":89
 *     if self.is_started:
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_server_register_non_listening_completion_queue(
 *           self.c_server, queue.c_completion_queue, NULL)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":90
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:
 *       grpc_server_register_non_listening_completion_queue(             # <<<<<<<<<<<<<<
 *           self.c_server, queue.c_completion_queue, NULL)
 *     self.registered_completion_queues.append(queue)
 */
        grpc_server_register_non_listening_completion_queue(__pyx_v_self->c_server, __pyx_v_queue->c_completion_queue, NULL);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":89
 *     if self.is_started:
 *       raise ValueError("cannot register completion queues after start")
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_server_register_non_listening_completion_queue(
 *           self.c_server, queue.c_completion_queue, NULL)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":92
 *       grpc_server_register_non_listening_completion_queue(
 *           self.c_server, queue.c_completion_queue, NULL)
 *     self.registered_completion_queues.append(queue)             # <<<<<<<<<<<<<<
 * 
 *   def start(self):
 */
  if (unlikely(__pyx_v_self->registered_completion_queues == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(4, 92, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_self->registered_completion_queues, ((PyObject *)__pyx_v_queue)); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(4, 92, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":85
 *     self.registered_completion_queues.append(queue)
 * 
 *   def register_non_listening_completion_queue(             # <<<<<<<<<<<<<<
 *       self, CompletionQueue queue not None):
 *     if self.is_started:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.register_non_listening_completion_queue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":94
 *     self.registered_completion_queues.append(queue)
 * 
 *   def start(self):             # <<<<<<<<<<<<<<
 *     if self.is_started:
 *       raise ValueError("the server has already started")
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_9start(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_9start(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6Server_8start(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_8start(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("start", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":95
 * 
 *   def start(self):
 *     if self.is_started:             # <<<<<<<<<<<<<<
 *       raise ValueError("the server has already started")
 *     self.backup_shutdown_queue = CompletionQueue()
 */
  __pyx_t_1 = (__pyx_v_self->is_started != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":96
 *   def start(self):
 *     if self.is_started:
 *       raise ValueError("the server has already started")             # <<<<<<<<<<<<<<
 *     self.backup_shutdown_queue = CompletionQueue()
 *     self.register_non_listening_completion_queue(self.backup_shutdown_queue)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 96, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":95
 * 
 *   def start(self):
 *     if self.is_started:             # <<<<<<<<<<<<<<
 *       raise ValueError("the server has already started")
 *     self.backup_shutdown_queue = CompletionQueue()
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":97
 *     if self.is_started:
 *       raise ValueError("the server has already started")
 *     self.backup_shutdown_queue = CompletionQueue()             # <<<<<<<<<<<<<<
 *     self.register_non_listening_completion_queue(self.backup_shutdown_queue)
 *     self.is_started = True
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->backup_shutdown_queue);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->backup_shutdown_queue));
  __pyx_v_self->backup_shutdown_queue = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":98
 *       raise ValueError("the server has already started")
 *     self.backup_shutdown_queue = CompletionQueue()
 *     self.register_non_listening_completion_queue(self.backup_shutdown_queue)             # <<<<<<<<<<<<<<
 *     self.is_started = True
 *     with nogil:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_register_non_listening_completio); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_self->backup_shutdown_queue)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    __Pyx_INCREF(((PyObject *)__pyx_v_self->backup_shutdown_queue));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->backup_shutdown_queue));
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_self->backup_shutdown_queue));
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":99
 *     self.backup_shutdown_queue = CompletionQueue()
 *     self.register_non_listening_completion_queue(self.backup_shutdown_queue)
 *     self.is_started = True             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_server_start(self.c_server)
 */
  __pyx_v_self->is_started = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":100
 *     self.register_non_listening_completion_queue(self.backup_shutdown_queue)
 *     self.is_started = True
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_server_start(self.c_server)
 *     # Ensure the core has gotten a chance to do the start-up work
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":101
 *     self.is_started = True
 *     with nogil:
 *       grpc_server_start(self.c_server)             # <<<<<<<<<<<<<<
 *     # Ensure the core has gotten a chance to do the start-up work
 *     self.backup_shutdown_queue.poll(Timespec(None))
 */
        grpc_server_start(__pyx_v_self->c_server);
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":100
 *     self.register_non_listening_completion_queue(self.backup_shutdown_queue)
 *     self.is_started = True
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_server_start(self.c_server)
 *     # Ensure the core has gotten a chance to do the start-up work
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":103
 *       grpc_server_start(self.c_server)
 *     # Ensure the core has gotten a chance to do the start-up work
 *     self.backup_shutdown_queue.poll(Timespec(None))             # <<<<<<<<<<<<<<
 * 
 *   def add_http2_port(self, bytes address,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->backup_shutdown_queue), __pyx_n_s_poll); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec), __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":94
 *     self.registered_completion_queues.append(queue)
 * 
 *   def start(self):             # <<<<<<<<<<<<<<
 *     if self.is_started:
 *       raise ValueError("the server has already started")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":105
 *     self.backup_shutdown_queue.poll(Timespec(None))
 * 
 *   def add_http2_port(self, bytes address,             # <<<<<<<<<<<<<<
 *                      ServerCredentials server_credentials=None):
 *     address = str_to_bytes(address)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_11add_http2_port(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_11add_http2_port(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_address = 0;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *__pyx_v_server_credentials = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_http2_port (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_address,&__pyx_n_s_server_credentials,0};
    PyObject* values[2] = {0,0};

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":106
 * 
 *   def add_http2_port(self, bytes address,
 *                      ServerCredentials server_credentials=None):             # <<<<<<<<<<<<<<
 *     address = str_to_bytes(address)
 *     self.references.append(address)
 */
    values[1] = (PyObject *)((struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_address)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_server_credentials);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add_http2_port") < 0)) __PYX_ERR(4, 105, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_address = ((PyObject*)values[0]);
    __pyx_v_server_credentials = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_http2_port", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 105, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.add_http2_port", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_address), (&PyBytes_Type), 1, "address", 1))) __PYX_ERR(4, 105, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_server_credentials), __pyx_ptype_4grpc_7_cython_6cygrpc_ServerCredentials, 1, "server_credentials", 0))) __PYX_ERR(4, 106, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6Server_10add_http2_port(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self), __pyx_v_address, __pyx_v_server_credentials);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":105
 *     self.backup_shutdown_queue.poll(Timespec(None))
 * 
 *   def add_http2_port(self, bytes address,             # <<<<<<<<<<<<<<
 *                      ServerCredentials server_credentials=None):
 *     address = str_to_bytes(address)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_10add_http2_port(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, PyObject *__pyx_v_address, struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *__pyx_v_server_credentials) {
  int __pyx_v_result;
  char *__pyx_v_address_c_string;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  char *__pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("add_http2_port", 0);
  __Pyx_INCREF(__pyx_v_address);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":107
 *   def add_http2_port(self, bytes address,
 *                      ServerCredentials server_credentials=None):
 *     address = str_to_bytes(address)             # <<<<<<<<<<<<<<
 *     self.references.append(address)
 *     cdef int result
 */
  __pyx_t_1 = __pyx_f_4grpc_7_cython_6cygrpc_str_to_bytes(__pyx_v_address); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_address, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":108
 *                      ServerCredentials server_credentials=None):
 *     address = str_to_bytes(address)
 *     self.references.append(address)             # <<<<<<<<<<<<<<
 *     cdef int result
 *     cdef char *address_c_string = address
 */
  if (unlikely(__pyx_v_self->references == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(4, 108, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_Append(__pyx_v_self->references, __pyx_v_address); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(4, 108, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":110
 *     self.references.append(address)
 *     cdef int result
 *     cdef char *address_c_string = address             # <<<<<<<<<<<<<<
 *     if server_credentials is not None:
 *       self.references.append(server_credentials)
 */
  __pyx_t_3 = __Pyx_PyObject_AsString(__pyx_v_address); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(4, 110, __pyx_L1_error)
  __pyx_v_address_c_string = __pyx_t_3;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":111
 *     cdef int result
 *     cdef char *address_c_string = address
 *     if server_credentials is not None:             # <<<<<<<<<<<<<<
 *       self.references.append(server_credentials)
 *       with nogil:
 */
  __pyx_t_4 = (((PyObject *)__pyx_v_server_credentials) != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":112
 *     cdef char *address_c_string = address
 *     if server_credentials is not None:
 *       self.references.append(server_credentials)             # <<<<<<<<<<<<<<
 *       with nogil:
 *         result = grpc_server_add_secure_http2_port(
 */
    if (unlikely(__pyx_v_self->references == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
      __PYX_ERR(4, 112, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyList_Append(__pyx_v_self->references, ((PyObject *)__pyx_v_server_credentials)); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(4, 112, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":113
 *     if server_credentials is not None:
 *       self.references.append(server_credentials)
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_server_add_secure_http2_port(
 *             self.c_server, address_c_string, server_credentials.c_credentials)
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":114
 *       self.references.append(server_credentials)
 *       with nogil:
 *         result = grpc_server_add_secure_http2_port(             # <<<<<<<<<<<<<<
 *             self.c_server, address_c_string, server_credentials.c_credentials)
 *     else:
 */
          __pyx_v_result = grpc_server_add_secure_http2_port(__pyx_v_self->c_server, __pyx_v_address_c_string, __pyx_v_server_credentials->c_credentials);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":113
 *     if server_credentials is not None:
 *       self.references.append(server_credentials)
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_server_add_secure_http2_port(
 *             self.c_server, address_c_string, server_credentials.c_credentials)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L6;
          }
          __pyx_L6:;
        }
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":111
 *     cdef int result
 *     cdef char *address_c_string = address
 *     if server_credentials is not None:             # <<<<<<<<<<<<<<
 *       self.references.append(server_credentials)
 *       with nogil:
 */
    goto __pyx_L3;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":117
 *             self.c_server, address_c_string, server_credentials.c_credentials)
 *     else:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_server_add_insecure_http2_port(self.c_server,
 *                                                      address_c_string)
 */
  /*else*/ {
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":118
 *     else:
 *       with nogil:
 *         result = grpc_server_add_insecure_http2_port(self.c_server,             # <<<<<<<<<<<<<<
 *                                                      address_c_string)
 *     return result
 */
          __pyx_v_result = grpc_server_add_insecure_http2_port(__pyx_v_self->c_server, __pyx_v_address_c_string);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":117
 *             self.c_server, address_c_string, server_credentials.c_credentials)
 *     else:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         result = grpc_server_add_insecure_http2_port(self.c_server,
 *                                                      address_c_string)
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }
  }
  __pyx_L3:;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":120
 *         result = grpc_server_add_insecure_http2_port(self.c_server,
 *                                                      address_c_string)
 *     return result             # <<<<<<<<<<<<<<
 * 
 *   cdef _c_shutdown(self, CompletionQueue queue, tag):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":105
 *     self.backup_shutdown_queue.poll(Timespec(None))
 * 
 *   def add_http2_port(self, bytes address,             # <<<<<<<<<<<<<<
 *                      ServerCredentials server_credentials=None):
 *     address = str_to_bytes(address)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.add_http2_port", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_address);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":122
 *     return result
 * 
 *   cdef _c_shutdown(self, CompletionQueue queue, tag):             # <<<<<<<<<<<<<<
 *     self.is_shutting_down = True
 *     operation_tag = OperationTag(tag)
 */

static PyObject *__pyx_f_4grpc_7_cython_6cygrpc_6Server__c_shutdown(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue, PyObject *__pyx_v_tag) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *__pyx_v_operation_tag = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_c_shutdown", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":123
 * 
 *   cdef _c_shutdown(self, CompletionQueue queue, tag):
 *     self.is_shutting_down = True             # <<<<<<<<<<<<<<
 *     operation_tag = OperationTag(tag)
 *     operation_tag.shutting_down_server = self
 */
  __pyx_v_self->is_shutting_down = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":124
 *   cdef _c_shutdown(self, CompletionQueue queue, tag):
 *     self.is_shutting_down = True
 *     operation_tag = OperationTag(tag)             # <<<<<<<<<<<<<<
 *     operation_tag.shutting_down_server = self
 *     cpython.Py_INCREF(operation_tag)
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_tag);
  __Pyx_GIVEREF(__pyx_v_tag);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_tag);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_OperationTag), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_operation_tag = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":125
 *     self.is_shutting_down = True
 *     operation_tag = OperationTag(tag)
 *     operation_tag.shutting_down_server = self             # <<<<<<<<<<<<<<
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_operation_tag->shutting_down_server);
  __Pyx_DECREF(((PyObject *)__pyx_v_operation_tag->shutting_down_server));
  __pyx_v_operation_tag->shutting_down_server = __pyx_v_self;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":126
 *     operation_tag = OperationTag(tag)
 *     operation_tag.shutting_down_server = self
 *     cpython.Py_INCREF(operation_tag)             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_server_shutdown_and_notify(
 */
  Py_INCREF(((PyObject *)__pyx_v_operation_tag));

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":127
 *     operation_tag.shutting_down_server = self
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_server_shutdown_and_notify(
 *           self.c_server, queue.c_completion_queue,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      #endif
      /*try:*/ {

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":128
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:
 *       grpc_server_shutdown_and_notify(             # <<<<<<<<<<<<<<
 *           self.c_server, queue.c_completion_queue,
 *           <cpython.PyObject *>operation_tag)
 */
        grpc_server_shutdown_and_notify(__pyx_v_self->c_server, __pyx_v_queue->c_completion_queue, ((PyObject *)__pyx_v_operation_tag));
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":127
 *     operation_tag.shutting_down_server = self
 *     cpython.Py_INCREF(operation_tag)
 *     with nogil:             # <<<<<<<<<<<<<<
 *       grpc_server_shutdown_and_notify(
 *           self.c_server, queue.c_completion_queue,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":122
 *     return result
 * 
 *   cdef _c_shutdown(self, CompletionQueue queue, tag):             # <<<<<<<<<<<<<<
 *     self.is_shutting_down = True
 *     operation_tag = OperationTag(tag)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server._c_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_operation_tag);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":132
 *           <cpython.PyObject *>operation_tag)
 * 
 *   def shutdown(self, CompletionQueue queue not None, tag):             # <<<<<<<<<<<<<<
 *     cdef OperationTag operation_tag
 *     if queue.is_shutting_down:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_13shutdown(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_13shutdown(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue = 0;
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shutdown (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_queue,&__pyx_n_s_tag,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_queue)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tag)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shutdown", 1, 2, 2, 1); __PYX_ERR(4, 132, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "shutdown") < 0)) __PYX_ERR(4, 132, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_queue = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)values[0]);
    __pyx_v_tag = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("shutdown", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 132, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_queue), __pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue, 0, "queue", 0))) __PYX_ERR(4, 132, __pyx_L1_error)
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6Server_12shutdown(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self), __pyx_v_queue, __pyx_v_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_12shutdown(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *__pyx_v_queue, PyObject *__pyx_v_tag) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("shutdown", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":134
 *   def shutdown(self, CompletionQueue queue not None, tag):
 *     cdef OperationTag operation_tag
 *     if queue.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise ValueError("queue must be live")
 *     elif not self.is_started:
 */
  __pyx_t_1 = (__pyx_v_queue->is_shutting_down != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":135
 *     cdef OperationTag operation_tag
 *     if queue.is_shutting_down:
 *       raise ValueError("queue must be live")             # <<<<<<<<<<<<<<
 *     elif not self.is_started:
 *       raise ValueError("the server hasn't started yet")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 135, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":134
 *   def shutdown(self, CompletionQueue queue not None, tag):
 *     cdef OperationTag operation_tag
 *     if queue.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise ValueError("queue must be live")
 *     elif not self.is_started:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":136
 *     if queue.is_shutting_down:
 *       raise ValueError("queue must be live")
 *     elif not self.is_started:             # <<<<<<<<<<<<<<
 *       raise ValueError("the server hasn't started yet")
 *     elif self.is_shutting_down:
 */
  __pyx_t_1 = ((!(__pyx_v_self->is_started != 0)) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":137
 *       raise ValueError("queue must be live")
 *     elif not self.is_started:
 *       raise ValueError("the server hasn't started yet")             # <<<<<<<<<<<<<<
 *     elif self.is_shutting_down:
 *       return
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 137, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":136
 *     if queue.is_shutting_down:
 *       raise ValueError("queue must be live")
 *     elif not self.is_started:             # <<<<<<<<<<<<<<
 *       raise ValueError("the server hasn't started yet")
 *     elif self.is_shutting_down:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":138
 *     elif not self.is_started:
 *       raise ValueError("the server hasn't started yet")
 *     elif self.is_shutting_down:             # <<<<<<<<<<<<<<
 *       return
 *     elif queue not in self.registered_completion_queues:
 */
  __pyx_t_1 = (__pyx_v_self->is_shutting_down != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":139
 *       raise ValueError("the server hasn't started yet")
 *     elif self.is_shutting_down:
 *       return             # <<<<<<<<<<<<<<
 *     elif queue not in self.registered_completion_queues:
 *       raise ValueError("expected registered completion queue")
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":138
 *     elif not self.is_started:
 *       raise ValueError("the server hasn't started yet")
 *     elif self.is_shutting_down:             # <<<<<<<<<<<<<<
 *       return
 *     elif queue not in self.registered_completion_queues:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":140
 *     elif self.is_shutting_down:
 *       return
 *     elif queue not in self.registered_completion_queues:             # <<<<<<<<<<<<<<
 *       raise ValueError("expected registered completion queue")
 *     else:
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_v_queue), __pyx_v_self->registered_completion_queues, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(4, 140, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (__pyx_t_3) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":141
 *       return
 *     elif queue not in self.registered_completion_queues:
 *       raise ValueError("expected registered completion queue")             # <<<<<<<<<<<<<<
 *     else:
 *       self._c_shutdown(queue, tag)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 141, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":140
 *     elif self.is_shutting_down:
 *       return
 *     elif queue not in self.registered_completion_queues:             # <<<<<<<<<<<<<<
 *       raise ValueError("expected registered completion queue")
 *     else:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":143
 *       raise ValueError("expected registered completion queue")
 *     else:
 *       self._c_shutdown(queue, tag)             # <<<<<<<<<<<<<<
 * 
 *   cdef notify_shutdown_complete(self):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self->__pyx_vtab)->_c_shutdown(__pyx_v_self, __pyx_v_queue, __pyx_v_tag); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":132
 *           <cpython.PyObject *>operation_tag)
 * 
 *   def shutdown(self, CompletionQueue queue not None, tag):             # <<<<<<<<<<<<<<
 *     cdef OperationTag operation_tag
 *     if queue.is_shutting_down:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":145
 *       self._c_shutdown(queue, tag)
 * 
 *   cdef notify_shutdown_complete(self):             # <<<<<<<<<<<<<<
 *     # called only by a completion queue on receiving our shutdown operation tag
 *     self.is_shutdown = True
 */

static PyObject *__pyx_f_4grpc_7_cython_6cygrpc_6Server_notify_shutdown_complete(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("notify_shutdown_complete", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":147
 *   cdef notify_shutdown_complete(self):
 *     # called only by a completion queue on receiving our shutdown operation tag
 *     self.is_shutdown = True             # <<<<<<<<<<<<<<
 * 
 *   def cancel_all_calls(self):
 */
  __pyx_v_self->is_shutdown = 1;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":145
 *       self._c_shutdown(queue, tag)
 * 
 *   cdef notify_shutdown_complete(self):             # <<<<<<<<<<<<<<
 *     # called only by a completion queue on receiving our shutdown operation tag
 *     self.is_shutdown = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":149
 *     self.is_shutdown = True
 * 
 *   def cancel_all_calls(self):             # <<<<<<<<<<<<<<
 *     if not self.is_shutting_down:
 *       raise RuntimeError("the server must be shutting down to cancel all calls")
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_15cancel_all_calls(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_6Server_15cancel_all_calls(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel_all_calls (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_6Server_14cancel_all_calls(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_6Server_14cancel_all_calls(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cancel_all_calls", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":150
 * 
 *   def cancel_all_calls(self):
 *     if not self.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise RuntimeError("the server must be shutting down to cancel all calls")
 *     elif self.is_shutdown:
 */
  __pyx_t_1 = ((!(__pyx_v_self->is_shutting_down != 0)) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":151
 *   def cancel_all_calls(self):
 *     if not self.is_shutting_down:
 *       raise RuntimeError("the server must be shutting down to cancel all calls")             # <<<<<<<<<<<<<<
 *     elif self.is_shutdown:
 *       return
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 151, __pyx_L1_error)

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":150
 * 
 *   def cancel_all_calls(self):
 *     if not self.is_shutting_down:             # <<<<<<<<<<<<<<
 *       raise RuntimeError("the server must be shutting down to cancel all calls")
 *     elif self.is_shutdown:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":152
 *     if not self.is_shutting_down:
 *       raise RuntimeError("the server must be shutting down to cancel all calls")
 *     elif self.is_shutdown:             # <<<<<<<<<<<<<<
 *       return
 *     else:
 */
  __pyx_t_1 = (__pyx_v_self->is_shutdown != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":153
 *       raise RuntimeError("the server must be shutting down to cancel all calls")
 *     elif self.is_shutdown:
 *       return             # <<<<<<<<<<<<<<
 *     else:
 *       with nogil:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":152
 *     if not self.is_shutting_down:
 *       raise RuntimeError("the server must be shutting down to cancel all calls")
 *     elif self.is_shutdown:             # <<<<<<<<<<<<<<
 *       return
 *     else:
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":155
 *       return
 *     else:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         grpc_server_cancel_all_calls(self.c_server)
 * 
 */
  /*else*/ {
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        #endif
        /*try:*/ {

          /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":156
 *     else:
 *       with nogil:
 *         grpc_server_cancel_all_calls(self.c_server)             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
          grpc_server_cancel_all_calls(__pyx_v_self->c_server);
        }

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":155
 *       return
 *     else:
 *       with nogil:             # <<<<<<<<<<<<<<
 *         grpc_server_cancel_all_calls(self.c_server)
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L6;
          }
          __pyx_L6:;
        }
    }
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":149
 *     self.is_shutdown = True
 * 
 *   def cancel_all_calls(self):             # <<<<<<<<<<<<<<
 *     if not self.is_shutting_down:
 *       raise RuntimeError("the server must be shutting down to cancel all calls")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("grpc._cython.cygrpc.Server.cancel_all_calls", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":158
 *         grpc_server_cancel_all_calls(self.c_server)
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_server != NULL:
 *       if not self.is_started:
 */

/* Python wrapper */
static void __pyx_pw_4grpc_7_cython_6cygrpc_6Server_17__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_4grpc_7_cython_6cygrpc_6Server_17__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_4grpc_7_cython_6cygrpc_6Server_16__dealloc__(((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_4grpc_7_cython_6cygrpc_6Server_16__dealloc__(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":159
 * 
 *   def __dealloc__(self):
 *     if self.c_server != NULL:             # <<<<<<<<<<<<<<
 *       if not self.is_started:
 *         pass
 */
  __pyx_t_1 = ((__pyx_v_self->c_server != NULL) != 0);
  if (__pyx_t_1) {

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":160
 *   def __dealloc__(self):
 *     if self.c_server != NULL:
 *       if not self.is_started:             # <<<<<<<<<<<<<<
 *         pass
 *       elif self.is_shutdown:
 */
    __pyx_t_1 = ((!(__pyx_v_self->is_started != 0)) != 0);
    if (__pyx_t_1) {
      goto __pyx_L4;
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":162
 *       if not self.is_started:
 *         pass
 *       elif self.is_shutdown:             # <<<<<<<<<<<<<<
 *         pass
 *       elif not self.is_shutting_down:
 */
    __pyx_t_1 = (__pyx_v_self->is_shutdown != 0);
    if (__pyx_t_1) {
      goto __pyx_L4;
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":164
 *       elif self.is_shutdown:
 *         pass
 *       elif not self.is_shutting_down:             # <<<<<<<<<<<<<<
 *         # the user didn't call shutdown - use our backup queue
 *         self._c_shutdown(self.backup_shutdown_queue, None)
 */
    __pyx_t_1 = ((!(__pyx_v_self->is_shutting_down != 0)) != 0);
    if (__pyx_t_1) {

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":166
 *       elif not self.is_shutting_down:
 *         # the user didn't call shutdown - use our backup queue
 *         self._c_shutdown(self.backup_shutdown_queue, None)             # <<<<<<<<<<<<<<
 *         # and now we wait
 *         while not self.is_shutdown:
 */
      __pyx_t_2 = ((PyObject *)__pyx_v_self->backup_shutdown_queue);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = ((struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_Server *)__pyx_v_self->__pyx_vtab)->_c_shutdown(__pyx_v_self, ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)__pyx_t_2), Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":168
 *         self._c_shutdown(self.backup_shutdown_queue, None)
 *         # and now we wait
 *         while not self.is_shutdown:             # <<<<<<<<<<<<<<
 *           self.backup_shutdown_queue.poll()
 *       else:
 */
      while (1) {
        __pyx_t_1 = ((!(__pyx_v_self->is_shutdown != 0)) != 0);
        if (!__pyx_t_1) break;

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":169
 *         # and now we wait
 *         while not self.is_shutdown:
 *           self.backup_shutdown_queue.poll()             # <<<<<<<<<<<<<<
 *       else:
 *         # We're in the process of shutting down, but have not shutdown; can't do
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->backup_shutdown_queue), __pyx_n_s_poll); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 169, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 169, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 169, __pyx_L1_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":164
 *       elif self.is_shutdown:
 *         pass
 *       elif not self.is_shutting_down:             # <<<<<<<<<<<<<<
 *         # the user didn't call shutdown - use our backup queue
 *         self._c_shutdown(self.backup_shutdown_queue, None)
 */
      goto __pyx_L4;
    }

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":173
 *         # We're in the process of shutting down, but have not shutdown; can't do
 *         # much but repeatedly release the GIL and wait
 *         while not self.is_shutdown:             # <<<<<<<<<<<<<<
 *           time.sleep(0)
 *       grpc_server_destroy(self.c_server)
 */
    /*else*/ {
      while (1) {
        __pyx_t_1 = ((!(__pyx_v_self->is_shutdown != 0)) != 0);
        if (!__pyx_t_1) break;

        /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":174
 *         # much but repeatedly release the GIL and wait
 *         while not self.is_shutdown:
 *           time.sleep(0)             # <<<<<<<<<<<<<<
 *       grpc_server_destroy(self.c_server)
 *     grpc_shutdown()
 */
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sleep); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }
    __pyx_L4:;

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":175
 *         while not self.is_shutdown:
 *           time.sleep(0)
 *       grpc_server_destroy(self.c_server)             # <<<<<<<<<<<<<<
 *     grpc_shutdown()
 */
    grpc_server_destroy(__pyx_v_self->c_server);

    /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":159
 * 
 *   def __dealloc__(self):
 *     if self.c_server != NULL:             # <<<<<<<<<<<<<<
 *       if not self.is_started:
 *         pass
 */
  }

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":176
 *           time.sleep(0)
 *       grpc_server_destroy(self.c_server)
 *     grpc_shutdown()             # <<<<<<<<<<<<<<
 */
  grpc_shutdown();

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":158
 *         grpc_server_cancel_all_calls(self.c_server)
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.c_server != NULL:
 *       if not self.is_started:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("grpc._cython.cygrpc.Server.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "grpc/_cython/cygrpc.pyx":57
 * 
 * 
 * def _initialize():             # <<<<<<<<<<<<<<
 *   grpc_set_ssl_roots_override_callback(
 *           <grpc_ssl_roots_override_callback>ssl_roots_override_callback)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_39_initialize(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4grpc_7_cython_6cygrpc_39_initialize = {"_initialize", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_39_initialize, METH_NOARGS, 0};
static PyObject *__pyx_pw_4grpc_7_cython_6cygrpc_39_initialize(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_initialize (wrapper)", 0);
  __pyx_r = __pyx_pf_4grpc_7_cython_6cygrpc_38_initialize(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4grpc_7_cython_6cygrpc_38_initialize(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_initialize", 0);

  /* "grpc/_cython/cygrpc.pyx":58
 * 
 * def _initialize():
 *   grpc_set_ssl_roots_override_callback(             # <<<<<<<<<<<<<<
 *           <grpc_ssl_roots_override_callback>ssl_roots_override_callback)
 * 
 */
  grpc_set_ssl_roots_override_callback(((grpc_ssl_roots_override_callback)__pyx_f_4grpc_7_cython_6cygrpc_ssl_roots_override_callback));

  /* "grpc/_cython/cygrpc.pyx":57
 * 
 * 
 * def _initialize():             # <<<<<<<<<<<<<<
 *   grpc_set_ssl_roots_override_callback(
 *           <grpc_ssl_roots_override_callback>ssl_roots_override_callback)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Call(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)o);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_4Call_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Call(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_4Call_15__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->references);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Call(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)o;
  if (p->references) {
    e = (*v)(p->references, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_Call(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)o;
  tmp = ((PyObject*)p->references);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_4Call_is_valid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_4Call_8is_valid_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Call[] = {
  {"_start_batch", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_4Call_3_start_batch, METH_VARARGS|METH_KEYWORDS, 0},
  {"start_client_batch", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_4Call_5start_client_batch, METH_VARARGS|METH_KEYWORDS, 0},
  {"start_server_batch", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_4Call_7start_server_batch, METH_VARARGS|METH_KEYWORDS, 0},
  {"cancel", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_4Call_9cancel, METH_VARARGS|METH_KEYWORDS, 0},
  {"set_credentials", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_4Call_11set_credentials, METH_O, 0},
  {"peer", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_4Call_13peer, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_Call[] = {
  {(char *)"is_valid", __pyx_getprop_4grpc_7_cython_6cygrpc_4Call_is_valid, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Call = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Call", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Call), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Call, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Call, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_Call, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Call, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_Call, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Call, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Channel(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)o);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Channel(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_7Channel_11__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->references);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Channel(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)o;
  if (p->references) {
    e = (*v)(p->references, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_Channel(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel *)o;
  tmp = ((PyObject*)p->references);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Channel[] = {
  {"create_call", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_3create_call, METH_VARARGS|METH_KEYWORDS, 0},
  {"check_connectivity_state", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_5check_connectivity_state, METH_O, 0},
  {"watch_connectivity_state", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_7watch_connectivity_state, METH_VARARGS|METH_KEYWORDS, 0},
  {"target", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_7Channel_9target, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Channel = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Channel", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Channel), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Channel, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Channel, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_Channel, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Channel, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Channel, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelCredentials(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)o);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ChannelCredentials(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->references);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_ChannelCredentials(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)o;
  if (p->references) {
    e = (*v)(p->references, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_ChannelCredentials(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials *)o;
  tmp = ((PyObject*)p->references);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_18ChannelCredentials_is_valid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_18ChannelCredentials_8is_valid_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_ChannelCredentials[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_ChannelCredentials[] = {
  {(char *)"is_valid", __pyx_getprop_4grpc_7_cython_6cygrpc_18ChannelCredentials_is_valid, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_ChannelCredentials = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.ChannelCredentials", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelCredentials), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ChannelCredentials, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_ChannelCredentials, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_ChannelCredentials, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_ChannelCredentials, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_ChannelCredentials, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelCredentials, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CallCredentials(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)o);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CallCredentials(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->references);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_CallCredentials(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)o;
  if (p->references) {
    e = (*v)(p->references, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_CallCredentials(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials *)o;
  tmp = ((PyObject*)p->references);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_15CallCredentials_is_valid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_15CallCredentials_8is_valid_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_CallCredentials[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_CallCredentials[] = {
  {(char *)"is_valid", __pyx_getprop_4grpc_7_cython_6cygrpc_15CallCredentials_is_valid, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_CallCredentials = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.CallCredentials", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallCredentials), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CallCredentials, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_CallCredentials, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_CallCredentials, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_CallCredentials, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_CallCredentials, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_CallCredentials, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ServerCredentials(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)o);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_17ServerCredentials_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ServerCredentials(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_17ServerCredentials_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->references);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_ServerCredentials(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)o;
  if (p->references) {
    e = (*v)(p->references, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_ServerCredentials(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials *)o;
  tmp = ((PyObject*)p->references);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_ServerCredentials[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_ServerCredentials = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.ServerCredentials", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_ServerCredentials), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ServerCredentials, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_ServerCredentials, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_ServerCredentials, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_ServerCredentials, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_ServerCredentials, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin __pyx_vtable_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin;

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)o);
  p->__pyx_vtab = __pyx_vtabptr_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin;
  p->plugin_callback = Py_None; Py_INCREF(Py_None);
  p->plugin_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->plugin_callback);
  Py_CLEAR(p->plugin_name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)o;
  if (p->plugin_callback) {
    e = (*v)(p->plugin_callback, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *)o;
  tmp = ((PyObject*)p->plugin_callback);
  p->plugin_callback = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.CredentialsMetadataPlugin", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_AuthMetadataContext(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_AuthMetadataContext(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_19AuthMetadataContext_service_url(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11service_url_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_19AuthMetadataContext_method_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_19AuthMetadataContext_11method_name_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_AuthMetadataContext[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_AuthMetadataContext[] = {
  {(char *)"service_url", __pyx_getprop_4grpc_7_cython_6cygrpc_19AuthMetadataContext_service_url, 0, (char *)0, 0},
  {(char *)"method_name", __pyx_getprop_4grpc_7_cython_6cygrpc_19AuthMetadataContext_method_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_AuthMetadataContext = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.AuthMetadataContext", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_AuthMetadataContext), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_AuthMetadataContext, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_AuthMetadataContext, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_AuthMetadataContext, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_AuthMetadataContext, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_CompletionQueue __pyx_vtable_4grpc_7_cython_6cygrpc_CompletionQueue;

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CompletionQueue(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)o);
  p->__pyx_vtab = __pyx_vtabptr_4grpc_7_cython_6cygrpc_CompletionQueue;
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CompletionQueue(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_9__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_CompletionQueue[] = {
  {"poll", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_3poll, METH_VARARGS|METH_KEYWORDS, 0},
  {"shutdown", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_5shutdown, METH_NOARGS, 0},
  {"clear", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_15CompletionQueue_7clear, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_CompletionQueue = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.CompletionQueue", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CompletionQueue, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_CompletionQueue, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_CompletionQueue, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Timespec(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Timespec(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_8Timespec_seconds(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_7seconds_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_8Timespec_nanoseconds(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_11nanoseconds_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Timespec[] = {
  {"infinite_future", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_5infinite_future, METH_VARARGS|METH_KEYWORDS, 0},
  {"infinite_past", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_7infinite_past, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_Timespec[] = {
  {(char *)"seconds", __pyx_getprop_4grpc_7_cython_6cygrpc_8Timespec_seconds, 0, (char *)0, 0},
  {(char *)"nanoseconds", __pyx_getprop_4grpc_7_cython_6cygrpc_8Timespec_nanoseconds, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number_Timespec = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  __pyx_pw_4grpc_7_cython_6cygrpc_8Timespec_3__float__, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Timespec = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Timespec", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Timespec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Timespec, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  &__pyx_tp_as_number_Timespec, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Timespec, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_Timespec, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Timespec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CallDetails(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CallDetails(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_11CallDetails_method(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_6method_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_11CallDetails_host(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_4host_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_11CallDetails_deadline(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_11CallDetails_8deadline_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_CallDetails[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_CallDetails[] = {
  {(char *)"method", __pyx_getprop_4grpc_7_cython_6cygrpc_11CallDetails_method, 0, (char *)0, 0},
  {(char *)"host", __pyx_getprop_4grpc_7_cython_6cygrpc_11CallDetails_host, 0, (char *)0, 0},
  {(char *)"deadline", __pyx_getprop_4grpc_7_cython_6cygrpc_11CallDetails_deadline, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_CallDetails = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.CallDetails", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CallDetails, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_CallDetails, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_CallDetails, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_CallDetails, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_OperationTag(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)o);
  p->user_tag = Py_None; Py_INCREF(Py_None);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->shutting_down_server = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)Py_None); Py_INCREF(Py_None);
  p->operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)Py_None); Py_INCREF(Py_None);
  p->request_call_details = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)Py_None); Py_INCREF(Py_None);
  p->request_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None); Py_INCREF(Py_None);
  p->batch_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_12OperationTag_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_OperationTag(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->user_tag);
  Py_CLEAR(p->references);
  Py_CLEAR(p->shutting_down_server);
  Py_CLEAR(p->operation_call);
  Py_CLEAR(p->request_call_details);
  Py_CLEAR(p->request_metadata);
  Py_CLEAR(p->batch_operations);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_OperationTag(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)o;
  if (p->user_tag) {
    e = (*v)(p->user_tag, a); if (e) return e;
  }
  if (p->references) {
    e = (*v)(p->references, a); if (e) return e;
  }
  if (p->shutting_down_server) {
    e = (*v)(((PyObject*)p->shutting_down_server), a); if (e) return e;
  }
  if (p->operation_call) {
    e = (*v)(((PyObject*)p->operation_call), a); if (e) return e;
  }
  if (p->request_call_details) {
    e = (*v)(((PyObject*)p->request_call_details), a); if (e) return e;
  }
  if (p->request_metadata) {
    e = (*v)(((PyObject*)p->request_metadata), a); if (e) return e;
  }
  if (p->batch_operations) {
    e = (*v)(((PyObject*)p->batch_operations), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_OperationTag(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag *)o;
  tmp = ((PyObject*)p->user_tag);
  p->user_tag = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->references);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->shutting_down_server);
  p->shutting_down_server = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->operation_call);
  p->operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->request_call_details);
  p->request_call_details = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->request_metadata);
  p->request_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->batch_operations);
  p->batch_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_OperationTag[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_OperationTag = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.OperationTag", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_OperationTag), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_OperationTag, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_OperationTag, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_OperationTag, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_OperationTag, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_OperationTag, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Event(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)o);
  p->tag = Py_None; Py_INCREF(Py_None);
  p->request_call_details = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)Py_None); Py_INCREF(Py_None);
  p->request_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None); Py_INCREF(Py_None);
  p->operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)Py_None); Py_INCREF(Py_None);
  p->batch_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_5Event_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Event(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->tag);
  Py_CLEAR(p->request_call_details);
  Py_CLEAR(p->request_metadata);
  Py_CLEAR(p->operation_call);
  Py_CLEAR(p->batch_operations);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Event(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)o;
  if (p->tag) {
    e = (*v)(p->tag, a); if (e) return e;
  }
  if (p->request_call_details) {
    e = (*v)(((PyObject*)p->request_call_details), a); if (e) return e;
  }
  if (p->request_metadata) {
    e = (*v)(((PyObject*)p->request_metadata), a); if (e) return e;
  }
  if (p->operation_call) {
    e = (*v)(((PyObject*)p->operation_call), a); if (e) return e;
  }
  if (p->batch_operations) {
    e = (*v)(((PyObject*)p->batch_operations), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_Event(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Event *)o;
  tmp = ((PyObject*)p->tag);
  p->tag = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->request_call_details);
  p->request_call_details = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CallDetails *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->request_metadata);
  p->request_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->operation_call);
  p->operation_call = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Call *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->batch_operations);
  p->batch_operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_5Event_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_5Event_4type_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_5Event_success(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_5Event_7success_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_5Event_tag(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_5Event_3tag_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_5Event_is_new_request(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_5Event_14is_new_request_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_5Event_request_call_details(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_5Event_20request_call_details_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_5Event_request_metadata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_5Event_16request_metadata_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_5Event_operation_call(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_5Event_14operation_call_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_5Event_batch_operations(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_5Event_16batch_operations_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Event[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_Event[] = {
  {(char *)"type", __pyx_getprop_4grpc_7_cython_6cygrpc_5Event_type, 0, (char *)0, 0},
  {(char *)"success", __pyx_getprop_4grpc_7_cython_6cygrpc_5Event_success, 0, (char *)0, 0},
  {(char *)"tag", __pyx_getprop_4grpc_7_cython_6cygrpc_5Event_tag, 0, (char *)0, 0},
  {(char *)"is_new_request", __pyx_getprop_4grpc_7_cython_6cygrpc_5Event_is_new_request, 0, (char *)0, 0},
  {(char *)"request_call_details", __pyx_getprop_4grpc_7_cython_6cygrpc_5Event_request_call_details, 0, (char *)0, 0},
  {(char *)"request_metadata", __pyx_getprop_4grpc_7_cython_6cygrpc_5Event_request_metadata, 0, (char *)0, 0},
  {(char *)"operation_call", __pyx_getprop_4grpc_7_cython_6cygrpc_5Event_operation_call, 0, (char *)0, 0},
  {(char *)"batch_operations", __pyx_getprop_4grpc_7_cython_6cygrpc_5Event_batch_operations, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Event = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Event", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Event), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Event, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Event, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_Event, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Event, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_Event, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Event, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ByteBuffer(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ByteBuffer(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_9__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_ByteBuffer[] = {
  {"bytes", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_3bytes, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_ByteBuffer = {
  __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_ByteBuffer = {
  __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_5__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_ByteBuffer = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.ByteBuffer", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ByteBuffer, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_ByteBuffer, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_ByteBuffer, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_4grpc_7_cython_6cygrpc_10ByteBuffer_7__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_ByteBuffer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_ByteBuffer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_SslPemKeyCertPair(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)o);
  p->private_key = Py_None; Py_INCREF(Py_None);
  p->certificate_chain = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_SslPemKeyCertPair(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->private_key);
  Py_CLEAR(p->certificate_chain);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_SslPemKeyCertPair(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)o;
  if (p->private_key) {
    e = (*v)(p->private_key, a); if (e) return e;
  }
  if (p->certificate_chain) {
    e = (*v)(p->certificate_chain, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_SslPemKeyCertPair(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair *)o;
  tmp = ((PyObject*)p->private_key);
  p->private_key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->certificate_chain);
  p->certificate_chain = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_private_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_11private_key_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_certificate_chain(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_17certificate_chain_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_SslPemKeyCertPair[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_SslPemKeyCertPair[] = {
  {(char *)"private_key", __pyx_getprop_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_private_key, 0, (char *)0, 0},
  {(char *)"certificate_chain", __pyx_getprop_4grpc_7_cython_6cygrpc_17SslPemKeyCertPair_certificate_chain, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_SslPemKeyCertPair = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.SslPemKeyCertPair", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_SslPemKeyCertPair), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_SslPemKeyCertPair, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_SslPemKeyCertPair, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_SslPemKeyCertPair, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_SslPemKeyCertPair, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_SslPemKeyCertPair, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_SslPemKeyCertPair, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelArg(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *)o);
  p->key = Py_None; Py_INCREF(Py_None);
  p->value = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ChannelArg(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->key);
  Py_CLEAR(p->value);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_ChannelArg(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *)o;
  if (p->key) {
    e = (*v)(p->key, a); if (e) return e;
  }
  if (p->value) {
    e = (*v)(p->value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_ChannelArg(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg *)o;
  tmp = ((PyObject*)p->key);
  p->key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->value);
  p->value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_10ChannelArg_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_3key_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_10ChannelArg_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_10ChannelArg_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_ChannelArg[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_ChannelArg[] = {
  {(char *)"key", __pyx_getprop_4grpc_7_cython_6cygrpc_10ChannelArg_key, 0, (char *)0, 0},
  {(char *)"value", __pyx_getprop_4grpc_7_cython_6cygrpc_10ChannelArg_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_ChannelArg = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.ChannelArg", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArg), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ChannelArg, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_ChannelArg, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_ChannelArg, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_ChannelArg, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_ChannelArg, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelArg, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelArgs(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)o);
  p->args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ChannelArgs(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->args);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_ChannelArgs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)o;
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_ChannelArgs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs *)o;
  tmp = ((PyObject*)p->args);
  p->args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_4grpc_7_cython_6cygrpc_ChannelArgs(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_ChannelArgs[] = {
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_ChannelArgs = {
  __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_4grpc_7_cython_6cygrpc_ChannelArgs, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_ChannelArgs = {
  __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_5__len__, /*mp_length*/
  __pyx_pw_4grpc_7_cython_6cygrpc_11ChannelArgs_7__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_ChannelArgs = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.ChannelArgs", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_ChannelArgs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_ChannelArgs, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_ChannelArgs, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_ChannelArgs, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_ChannelArgs, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_ChannelArgs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_ChannelArgs, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_ChannelArgs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Metadatum(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)o);
  p->_key = Py_None; Py_INCREF(Py_None);
  p->_value = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Metadatum(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_key);
  Py_CLEAR(p->_value);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Metadatum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)o;
  if (p->_key) {
    e = (*v)(p->_key, a); if (e) return e;
  }
  if (p->_value) {
    e = (*v)(p->_value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_Metadatum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum *)o;
  tmp = ((PyObject*)p->_key);
  p->_key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_value);
  p->_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_4grpc_7_cython_6cygrpc_Metadatum(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Metadatum_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_3key_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Metadatum_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Metadatum[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_Metadatum[] = {
  {(char *)"key", __pyx_getprop_4grpc_7_cython_6cygrpc_9Metadatum_key, 0, (char *)0, 0},
  {(char *)"value", __pyx_getprop_4grpc_7_cython_6cygrpc_9Metadatum_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Metadatum = {
  __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_4grpc_7_cython_6cygrpc_Metadatum, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Metadatum = {
  __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_3__len__, /*mp_length*/
  __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Metadatum = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Metadatum", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadatum), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Metadatum, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Metadatum, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Metadatum, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Metadatum, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_Metadatum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_4grpc_7_cython_6cygrpc_9Metadatum_7__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Metadatum, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_Metadatum, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Metadatum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Metadata(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)o);
  p->metadata = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Metadata(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->metadata);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Metadata(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)o;
  if (p->metadata) {
    e = (*v)(p->metadata, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_Metadata(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)o;
  tmp = ((PyObject*)p->metadata);
  p->metadata = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_4grpc_7_cython_6cygrpc_Metadata(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Metadata[] = {
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Metadata = {
  __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_4grpc_7_cython_6cygrpc_Metadata, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Metadata = {
  __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_5__len__, /*mp_length*/
  __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_7__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Metadata = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Metadata", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Metadata, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Metadata, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Metadata, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Metadata, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_Metadata, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_4grpc_7_cython_6cygrpc_8Metadata_9__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Metadata, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Metadata, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Operation(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)o);
  p->_received_message = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)Py_None); Py_INCREF(Py_None);
  p->_received_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None); Py_INCREF(Py_None);
  p->references = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_9Operation_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Operation(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_received_message);
  Py_CLEAR(p->_received_metadata);
  Py_CLEAR(p->references);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Operation(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)o;
  if (p->_received_message) {
    e = (*v)(((PyObject*)p->_received_message), a); if (e) return e;
  }
  if (p->_received_metadata) {
    e = (*v)(((PyObject*)p->_received_metadata), a); if (e) return e;
  }
  if (p->references) {
    e = (*v)(p->references, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_Operation(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation *)o;
  tmp = ((PyObject*)p->_received_message);
  p->_received_message = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_ByteBuffer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_received_metadata);
  p->_received_metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->references);
  p->references = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_4type_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_5flags_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_has_status(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_10has_status_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_message(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_16received_message_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_message_or_none(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_24received_message_or_none_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_metadata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_17received_metadata_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_metadata_or_none(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_25received_metadata_or_none_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_status_code(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_20received_status_code_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_status_code_or_none(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_28received_status_code_or_none_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_status_details(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_23received_status_details_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_status_details_or_none(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_31received_status_details_or_none_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_cancelled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_18received_cancelled_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_cancelled_or_none(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_26received_cancelled_or_none_1__get__(o);
}

static PyObject *__pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_is_valid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4grpc_7_cython_6cygrpc_9Operation_8is_valid_1__get__(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Operation[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4grpc_7_cython_6cygrpc_Operation[] = {
  {(char *)"type", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_type, 0, (char *)0, 0},
  {(char *)"flags", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_flags, 0, (char *)0, 0},
  {(char *)"has_status", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_has_status, 0, (char *)0, 0},
  {(char *)"received_message", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_message, 0, (char *)0, 0},
  {(char *)"received_message_or_none", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_message_or_none, 0, (char *)0, 0},
  {(char *)"received_metadata", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_metadata, 0, (char *)0, 0},
  {(char *)"received_metadata_or_none", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_metadata_or_none, 0, (char *)0, 0},
  {(char *)"received_status_code", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_status_code, 0, (char *)0, 0},
  {(char *)"received_status_code_or_none", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_status_code_or_none, 0, (char *)0, 0},
  {(char *)"received_status_details", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_status_details, 0, (char *)0, 0},
  {(char *)"received_status_details_or_none", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_status_details_or_none, 0, (char *)0, 0},
  {(char *)"received_cancelled", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_cancelled, 0, (char *)0, 0},
  {(char *)"received_cancelled_or_none", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_received_cancelled_or_none, 0, (char *)0, 0},
  {(char *)"is_valid", __pyx_getprop_4grpc_7_cython_6cygrpc_9Operation_is_valid, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Operation = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Operation", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operation), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Operation, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Operation, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_Operation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Operation, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4grpc_7_cython_6cygrpc_Operation, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Operation, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Operations(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)o);
  p->operations = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_10Operations_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Operations(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_7__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->operations);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Operations(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)o;
  if (p->operations) {
    e = (*v)(p->operations, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_Operations(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)o;
  tmp = ((PyObject*)p->operations);
  p->operations = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_4grpc_7_cython_6cygrpc_Operations(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Operations[] = {
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Operations = {
  __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_4grpc_7_cython_6cygrpc_Operations, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Operations = {
  __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_3__len__, /*mp_length*/
  __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Operations = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Operations", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Operations, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Operations, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Operations, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Operations, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_Operations, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_4grpc_7_cython_6cygrpc_10Operations_9__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Operations, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Operations, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_CompressionOptions(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CompressionOptions(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_CompressionOptions[] = {
  {"enable_algorithm", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_3enable_algorithm, METH_O, 0},
  {"disable_algorithm", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_5disable_algorithm, METH_O, 0},
  {"is_algorithm_enabled", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_7is_algorithm_enabled, METH_O, 0},
  {"to_channel_arg", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_18CompressionOptions_9to_channel_arg, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_CompressionOptions = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.CompressionOptions", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompressionOptions), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_CompressionOptions, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_CompressionOptions, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_CompressionOptions, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_4grpc_7_cython_6cygrpc_Server __pyx_vtable_4grpc_7_cython_6cygrpc_Server;

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc_Server(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)o);
  p->__pyx_vtab = __pyx_vtabptr_4grpc_7_cython_6cygrpc_Server;
  p->backup_shutdown_queue = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)Py_None); Py_INCREF(Py_None);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->registered_completion_queues = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_6Server_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Server(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_4grpc_7_cython_6cygrpc_6Server_17__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->backup_shutdown_queue);
  Py_CLEAR(p->references);
  Py_CLEAR(p->registered_completion_queues);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Server(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)o;
  if (p->backup_shutdown_queue) {
    e = (*v)(((PyObject*)p->backup_shutdown_queue), a); if (e) return e;
  }
  if (p->references) {
    e = (*v)(p->references, a); if (e) return e;
  }
  if (p->registered_completion_queues) {
    e = (*v)(p->registered_completion_queues, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc_Server(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *)o;
  tmp = ((PyObject*)p->backup_shutdown_queue);
  p->backup_shutdown_queue = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->references);
  p->references = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->registered_completion_queues);
  p->registered_completion_queues = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc_Server[] = {
  {"request_call", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_6Server_3request_call, METH_VARARGS|METH_KEYWORDS, 0},
  {"register_completion_queue", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_6Server_5register_completion_queue, METH_O, 0},
  {"register_non_listening_completion_queue", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_6Server_7register_non_listening_completion_queue, METH_O, 0},
  {"start", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_6Server_9start, METH_NOARGS, 0},
  {"add_http2_port", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_6Server_11add_http2_port, METH_VARARGS|METH_KEYWORDS, 0},
  {"shutdown", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_6Server_13shutdown, METH_VARARGS|METH_KEYWORDS, 0},
  {"cancel_all_calls", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_6Server_15cancel_all_calls, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc_Server = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.Server", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc_Server, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc_Server, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc_Server, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc_Server, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc_Server, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc__MetadataIterator(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *)o);
  p->metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc__MetadataIterator(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->metadata);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc__MetadataIterator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *)o;
  if (p->metadata) {
    e = (*v)(((PyObject*)p->metadata), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc__MetadataIterator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator *)o;
  tmp = ((PyObject*)p->metadata);
  p->metadata = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Metadata *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc__MetadataIterator[] = {
  {"__next__", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_5__next__, METH_NOARGS|METH_COEXIST, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc__MetadataIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc._MetadataIterator", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc__MetadataIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc__MetadataIterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc__MetadataIterator, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc__MetadataIterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_3__iter__, /*tp_iter*/
  __pyx_pw_4grpc_7_cython_6cygrpc_17_MetadataIterator_5__next__, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc__MetadataIterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc__MetadataIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc__OperationsIterator(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *)o);
  p->operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc__OperationsIterator(PyObject *o) {
  struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->operations);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4grpc_7_cython_6cygrpc__OperationsIterator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *)o;
  if (p->operations) {
    e = (*v)(((PyObject*)p->operations), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4grpc_7_cython_6cygrpc__OperationsIterator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *p = (struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator *)o;
  tmp = ((PyObject*)p->operations);
  p->operations = ((struct __pyx_obj_4grpc_7_cython_6cygrpc_Operations *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_4grpc_7_cython_6cygrpc__OperationsIterator[] = {
  {"__next__", (PyCFunction)__pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_5__next__, METH_NOARGS|METH_COEXIST, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc__OperationsIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc._OperationsIterator", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc__OperationsIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc__OperationsIterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4grpc_7_cython_6cygrpc__OperationsIterator, /*tp_traverse*/
  __pyx_tp_clear_4grpc_7_cython_6cygrpc__OperationsIterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_3__iter__, /*tp_iter*/
  __pyx_pw_4grpc_7_cython_6cygrpc_19_OperationsIterator_5__next__, /*tp_iternext*/
  __pyx_methods_4grpc_7_cython_6cygrpc__OperationsIterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc__OperationsIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata *__pyx_freelist_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata[8];
static int __pyx_freecount_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata = 0;

static PyObject *__pyx_tp_new_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata)))) {
    o = (PyObject*)__pyx_freelist_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata[--__pyx_freecount_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata];
    memset(o, 0, sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata)))) {
    __pyx_freelist_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata[__pyx_freecount_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata++] = ((struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata = {
  PyVarObject_HEAD_INIT(0, 0)
  "grpc._cython.cygrpc.__pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata", /*tp_name*/
  sizeof(struct __pyx_obj_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "cygrpc",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_CallError, __pyx_k_CallError, sizeof(__pyx_k_CallError), 0, 0, 1, 1},
  {&__pyx_n_s_ChannelArgKey, __pyx_k_ChannelArgKey, sizeof(__pyx_k_ChannelArgKey), 0, 0, 1, 1},
  {&__pyx_n_s_CompletionType, __pyx_k_CompletionType, sizeof(__pyx_k_CompletionType), 0, 0, 1, 1},
  {&__pyx_n_s_CompressionAlgorithm, __pyx_k_CompressionAlgorithm, sizeof(__pyx_k_CompressionAlgorithm), 0, 0, 1, 1},
  {&__pyx_n_s_CompressionLevel, __pyx_k_CompressionLevel, sizeof(__pyx_k_CompressionLevel), 0, 0, 1, 1},
  {&__pyx_n_s_ConnectivityState, __pyx_k_ConnectivityState, sizeof(__pyx_k_ConnectivityState), 0, 0, 1, 1},
  {&__pyx_kp_s_Expected_bytes_str_or_unicode_no, __pyx_k_Expected_bytes_str_or_unicode_no, sizeof(__pyx_k_Expected_bytes_str_or_unicode_no), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_int_or_bytes_got, __pyx_k_Expected_int_or_bytes_got, sizeof(__pyx_k_Expected_int_or_bytes_got), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_OperationType, __pyx_k_OperationType, sizeof(__pyx_k_OperationType), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_StatusCode, __pyx_k_StatusCode, sizeof(__pyx_k_StatusCode), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_WriteFlag, __pyx_k_WriteFlag, sizeof(__pyx_k_WriteFlag), 0, 0, 1, 1},
  {&__pyx_n_s_aborted, __pyx_k_aborted, sizeof(__pyx_k_aborted), 0, 0, 1, 1},
  {&__pyx_n_s_address, __pyx_k_address, sizeof(__pyx_k_address), 0, 0, 1, 1},
  {&__pyx_n_s_algorithm, __pyx_k_algorithm, sizeof(__pyx_k_algorithm), 0, 0, 1, 1},
  {&__pyx_n_s_already_accepted, __pyx_k_already_accepted, sizeof(__pyx_k_already_accepted), 0, 0, 1, 1},
  {&__pyx_n_s_already_exists, __pyx_k_already_exists, sizeof(__pyx_k_already_exists), 0, 0, 1, 1},
  {&__pyx_n_s_already_finished, __pyx_k_already_finished, sizeof(__pyx_k_already_finished), 0, 0, 1, 1},
  {&__pyx_n_s_already_invoked, __pyx_k_already_invoked, sizeof(__pyx_k_already_invoked), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_arguments, __pyx_k_arguments, sizeof(__pyx_k_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 0, 1, 1},
  {&__pyx_n_s_authority_selector, __pyx_k_authority_selector, sizeof(__pyx_k_authority_selector), 0, 0, 1, 1},
  {&__pyx_n_s_authority_selector_c_string, __pyx_k_authority_selector_c_string, sizeof(__pyx_k_authority_selector_c_string), 0, 0, 1, 1},
  {&__pyx_n_s_authorization_token, __pyx_k_authorization_token, sizeof(__pyx_k_authorization_token), 0, 0, 1, 1},
  {&__pyx_n_s_authorization_token_c_string, __pyx_k_authorization_token_c_string, sizeof(__pyx_k_authorization_token_c_string), 0, 0, 1, 1},
  {&__pyx_n_s_batch_operations, __pyx_k_batch_operations, sizeof(__pyx_k_batch_operations), 0, 0, 1, 1},
  {&__pyx_n_s_buffer_hint, __pyx_k_buffer_hint, sizeof(__pyx_k_buffer_hint), 0, 0, 1, 1},
  {&__pyx_n_s_byte_buffer, __pyx_k_byte_buffer, sizeof(__pyx_k_byte_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_c_pem_root_certificates, __pyx_k_c_pem_root_certificates, sizeof(__pyx_k_c_pem_root_certificates), 0, 0, 1, 1},
  {&__pyx_n_s_c_pem_root_certs, __pyx_k_c_pem_root_certs, sizeof(__pyx_k_c_pem_root_certs), 0, 0, 1, 1},
  {&__pyx_n_s_c_plugin, __pyx_k_c_plugin, sizeof(__pyx_k_c_plugin), 0, 0, 1, 1},
  {&__pyx_n_s_call_credentials_composite, __pyx_k_call_credentials_composite, sizeof(__pyx_k_call_credentials_composite), 0, 0, 1, 1},
  {&__pyx_n_s_call_credentials_google_compute, __pyx_k_call_credentials_google_compute, sizeof(__pyx_k_call_credentials_google_compute), 0, 0, 1, 1},
  {&__pyx_n_s_call_credentials_google_iam, __pyx_k_call_credentials_google_iam, sizeof(__pyx_k_call_credentials_google_iam), 0, 0, 1, 1},
  {&__pyx_n_s_call_credentials_google_refresh, __pyx_k_call_credentials_google_refresh, sizeof(__pyx_k_call_credentials_google_refresh), 0, 0, 1, 1},
  {&__pyx_n_s_call_credentials_metadata_plugin, __pyx_k_call_credentials_metadata_plugin, sizeof(__pyx_k_call_credentials_metadata_plugin), 0, 0, 1, 1},
  {&__pyx_n_s_call_credentials_service_account, __pyx_k_call_credentials_service_account, sizeof(__pyx_k_call_credentials_service_account), 0, 0, 1, 1},
  {&__pyx_n_s_call_queue, __pyx_k_call_queue, sizeof(__pyx_k_call_queue), 0, 0, 1, 1},
  {&__pyx_n_s_cancelled, __pyx_k_cancelled, sizeof(__pyx_k_cancelled), 0, 0, 1, 1},
  {&__pyx_kp_s_cannot_register_completion_queue, __pyx_k_cannot_register_completion_queue, sizeof(__pyx_k_cannot_register_completion_queue), 0, 0, 1, 0},
  {&__pyx_n_s_certificate_chain, __pyx_k_certificate_chain, sizeof(__pyx_k_certificate_chain), 0, 0, 1, 1},
  {&__pyx_n_s_channel_credentials, __pyx_k_channel_credentials, sizeof(__pyx_k_channel_credentials), 0, 0, 1, 1},
  {&__pyx_n_s_channel_credentials_composite, __pyx_k_channel_credentials_composite, sizeof(__pyx_k_channel_credentials_composite), 0, 0, 1, 1},
  {&__pyx_n_s_channel_credentials_google_defau, __pyx_k_channel_credentials_google_defau, sizeof(__pyx_k_channel_credentials_google_defau), 0, 0, 1, 1},
  {&__pyx_n_s_channel_credentials_ssl, __pyx_k_channel_credentials_ssl, sizeof(__pyx_k_channel_credentials_ssl), 0, 0, 1, 1},
  {&__pyx_n_s_code, __pyx_k_code, sizeof(__pyx_k_code), 0, 0, 1, 1},
  {&__pyx_n_s_compression_algorithm_name, __pyx_k_compression_algorithm_name, sizeof(__pyx_k_compression_algorithm_name), 0, 0, 1, 1},
  {&__pyx_n_s_connecting, __pyx_k_connecting, sizeof(__pyx_k_connecting), 0, 0, 1, 1},
  {&__pyx_n_s_credentials, __pyx_k_credentials, sizeof(__pyx_k_credentials), 0, 0, 1, 1},
  {&__pyx_n_s_credentials_1, __pyx_k_credentials_1, sizeof(__pyx_k_credentials_1), 0, 0, 1, 1},
  {&__pyx_n_s_credentials_2, __pyx_k_credentials_2, sizeof(__pyx_k_credentials_2), 0, 0, 1, 1},
  {&__pyx_kp_s_credentials_roots_pem, __pyx_k_credentials_roots_pem, sizeof(__pyx_k_credentials_roots_pem), 0, 0, 1, 0},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_data_loss, __pyx_k_data_loss, sizeof(__pyx_k_data_loss), 0, 0, 1, 1},
  {&__pyx_n_s_deadline, __pyx_k_deadline, sizeof(__pyx_k_deadline), 0, 0, 1, 1},
  {&__pyx_n_s_deadline_exceeded, __pyx_k_deadline_exceeded, sizeof(__pyx_k_deadline_exceeded), 0, 0, 1, 1},
  {&__pyx_n_s_default_authority, __pyx_k_default_authority, sizeof(__pyx_k_default_authority), 0, 0, 1, 1},
  {&__pyx_n_s_deflate, __pyx_k_deflate, sizeof(__pyx_k_deflate), 0, 0, 1, 1},
  {&__pyx_n_s_details, __pyx_k_details, sizeof(__pyx_k_details), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_enable_census, __pyx_k_enable_census, sizeof(__pyx_k_enable_census), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_error_code, __pyx_k_error_code, sizeof(__pyx_k_error_code), 0, 0, 1, 1},
  {&__pyx_n_s_error_details, __pyx_k_error_details, sizeof(__pyx_k_error_details), 0, 0, 1, 1},
  {&__pyx_kp_s_expected_callable_plugin_callbac, __pyx_k_expected_callable_plugin_callbac, sizeof(__pyx_k_expected_callable_plugin_callbac), 0, 0, 1, 0},
  {&__pyx_kp_s_expected_list_of_ChannelArg, __pyx_k_expected_list_of_ChannelArg, sizeof(__pyx_k_expected_list_of_ChannelArg), 0, 0, 1, 0},
  {&__pyx_kp_s_expected_list_of_Metadatum, __pyx_k_expected_list_of_Metadatum, sizeof(__pyx_k_expected_list_of_Metadatum), 0, 0, 1, 0},
  {&__pyx_kp_s_expected_operations_to_be_iterab, __pyx_k_expected_operations_to_be_iterab, sizeof(__pyx_k_expected_operations_to_be_iterab), 0, 0, 1, 0},
  {&__pyx_kp_s_expected_pem_key_cert_pairs_to_b, __pyx_k_expected_pem_key_cert_pairs_to_b, sizeof(__pyx_k_expected_pem_key_cert_pairs_to_b), 0, 0, 1, 0},
  {&__pyx_kp_s_expected_registered_completion_q, __pyx_k_expected_registered_completion_q, sizeof(__pyx_k_expected_registered_completion_q), 0, 0, 1, 0},
  {&__pyx_kp_s_expected_time_to_be_float_int_or, __pyx_k_expected_time_to_be_float_int_or, sizeof(__pyx_k_expected_time_to_be_float_int_or), 0, 0, 1, 0},
  {&__pyx_n_s_failed_precondition, __pyx_k_failed_precondition, sizeof(__pyx_k_failed_precondition), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_force_client_auth, __pyx_k_force_client_auth, sizeof(__pyx_k_force_client_auth), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_kp_s_grpc__cython, __pyx_k_grpc__cython, sizeof(__pyx_k_grpc__cython), 0, 0, 1, 0},
  {&__pyx_n_s_grpc__cython_cygrpc, __pyx_k_grpc__cython_cygrpc, sizeof(__pyx_k_grpc__cython_cygrpc), 0, 0, 1, 1},
  {&__pyx_n_s_gzip, __pyx_k_gzip, sizeof(__pyx_k_gzip), 0, 0, 1, 1},
  {&__pyx_n_s_high, __pyx_k_high, sizeof(__pyx_k_high), 0, 0, 1, 1},
  {&__pyx_n_s_host, __pyx_k_host, sizeof(__pyx_k_host), 0, 0, 1, 1},
  {&__pyx_n_s_http2_initial_sequence_number, __pyx_k_http2_initial_sequence_number, sizeof(__pyx_k_http2_initial_sequence_number), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_idle, __pyx_k_idle, sizeof(__pyx_k_idle), 0, 0, 1, 1},
  {&__pyx_kp_s_if_error_code_is_specified_so_mu, __pyx_k_if_error_code_is_specified_so_mu, sizeof(__pyx_k_if_error_code_is_specified_so_mu), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_kp_s_index_must_be_0_key_or_1_value, __pyx_k_index_must_be_0_key_or_1_value, sizeof(__pyx_k_index_must_be_0_key_or_1_value), 0, 0, 1, 0},
  {&__pyx_kp_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 0},
  {&__pyx_kp_s_inf_2, __pyx_k_inf_2, sizeof(__pyx_k_inf_2), 0, 0, 1, 0},
  {&__pyx_n_s_infinite_future, __pyx_k_infinite_future, sizeof(__pyx_k_infinite_future), 0, 0, 1, 1},
  {&__pyx_n_s_infinite_past, __pyx_k_infinite_past, sizeof(__pyx_k_infinite_past), 0, 0, 1, 1},
  {&__pyx_n_s_initialize, __pyx_k_initialize, sizeof(__pyx_k_initialize), 0, 0, 1, 1},
  {&__pyx_n_s_internal, __pyx_k_internal, sizeof(__pyx_k_internal), 0, 0, 1, 1},
  {&__pyx_n_s_invalid_argument, __pyx_k_invalid_argument, sizeof(__pyx_k_invalid_argument), 0, 0, 1, 1},
  {&__pyx_kp_s_invalid_call_object_cannot_be_us, __pyx_k_invalid_call_object_cannot_be_us, sizeof(__pyx_k_invalid_call_object_cannot_be_us), 0, 0, 1, 0},
  {&__pyx_n_s_invalid_flags, __pyx_k_invalid_flags, sizeof(__pyx_k_invalid_flags), 0, 0, 1, 1},
  {&__pyx_n_s_invalid_metadata, __pyx_k_invalid_metadata, sizeof(__pyx_k_invalid_metadata), 0, 0, 1, 1},
  {&__pyx_n_s_is_new_request, __pyx_k_is_new_request, sizeof(__pyx_k_is_new_request), 0, 0, 1, 1},
  {&__pyx_n_s_is_valid, __pyx_k_is_valid, sizeof(__pyx_k_is_valid), 0, 0, 1, 1},
  {&__pyx_n_s_json_key, __pyx_k_json_key, sizeof(__pyx_k_json_key), 0, 0, 1, 1},
  {&__pyx_n_s_json_key_c_string, __pyx_k_json_key_c_string, sizeof(__pyx_k_json_key_c_string), 0, 0, 1, 1},
  {&__pyx_n_s_json_refresh_token, __pyx_k_json_refresh_token, sizeof(__pyx_k_json_refresh_token), 0, 0, 1, 1},
  {&__pyx_n_s_json_refresh_token_c_string, __pyx_k_json_refresh_token_c_string, sizeof(__pyx_k_json_refresh_token_c_string), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_last_observed_state, __pyx_k_last_observed_state, sizeof(__pyx_k_last_observed_state), 0, 0, 1, 1},
  {&__pyx_n_s_low, __pyx_k_low, sizeof(__pyx_k_low), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max_concurrent_streams, __pyx_k_max_concurrent_streams, sizeof(__pyx_k_max_concurrent_streams), 0, 0, 1, 1},
  {&__pyx_n_s_max_message_length, __pyx_k_max_message_length, sizeof(__pyx_k_max_message_length), 0, 0, 1, 1},
  {&__pyx_n_s_medium, __pyx_k_medium, sizeof(__pyx_k_medium), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_metadata, __pyx_k_metadata, sizeof(__pyx_k_metadata), 0, 0, 1, 1},
  {&__pyx_n_s_method, __pyx_k_method, sizeof(__pyx_k_method), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_no_compress, __pyx_k_no_compress, sizeof(__pyx_k_no_compress), 0, 0, 1, 1},
  {&__pyx_n_s_none, __pyx_k_none, sizeof(__pyx_k_none), 0, 0, 1, 1},
  {&__pyx_n_s_not_found, __pyx_k_not_found, sizeof(__pyx_k_not_found), 0, 0, 1, 1},
  {&__pyx_n_s_not_invoked, __pyx_k_not_invoked, sizeof(__pyx_k_not_invoked), 0, 0, 1, 1},
  {&__pyx_n_s_not_on_client, __pyx_k_not_on_client, sizeof(__pyx_k_not_on_client), 0, 0, 1, 1},
  {&__pyx_n_s_not_on_server, __pyx_k_not_on_server, sizeof(__pyx_k_not_on_server), 0, 0, 1, 1},
  {&__pyx_n_s_ok, __pyx_k_ok, sizeof(__pyx_k_ok), 0, 0, 1, 1},
  {&__pyx_n_s_op, __pyx_k_op, sizeof(__pyx_k_op), 0, 0, 1, 1},
  {&__pyx_n_s_operation_call, __pyx_k_operation_call, sizeof(__pyx_k_operation_call), 0, 0, 1, 1},
  {&__pyx_n_s_operation_complete, __pyx_k_operation_complete, sizeof(__pyx_k_operation_complete), 0, 0, 1, 1},
  {&__pyx_n_s_operation_receive_close_on_serve, __pyx_k_operation_receive_close_on_serve, sizeof(__pyx_k_operation_receive_close_on_serve), 0, 0, 1, 1},
  {&__pyx_n_s_operation_receive_initial_metada, __pyx_k_operation_receive_initial_metada, sizeof(__pyx_k_operation_receive_initial_metada), 0, 0, 1, 1},
  {&__pyx_n_s_operation_receive_message, __pyx_k_operation_receive_message, sizeof(__pyx_k_operation_receive_message), 0, 0, 1, 1},
  {&__pyx_n_s_operation_receive_status_on_clie, __pyx_k_operation_receive_status_on_clie, sizeof(__pyx_k_operation_receive_status_on_clie), 0, 0, 1, 1},
  {&__pyx_n_s_operation_send_close_from_client, __pyx_k_operation_send_close_from_client, sizeof(__pyx_k_operation_send_close_from_client), 0, 0, 1, 1},
  {&__pyx_n_s_operation_send_initial_metadata, __pyx_k_operation_send_initial_metadata, sizeof(__pyx_k_operation_send_initial_metadata), 0, 0, 1, 1},
  {&__pyx_n_s_operation_send_message, __pyx_k_operation_send_message, sizeof(__pyx_k_operation_send_message), 0, 0, 1, 1},
  {&__pyx_n_s_operation_send_status_from_serve, __pyx_k_operation_send_status_from_serve, sizeof(__pyx_k_operation_send_status_from_serve), 0, 0, 1, 1},
  {&__pyx_n_s_operations, __pyx_k_operations, sizeof(__pyx_k_operations), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_os_path, __pyx_k_os_path, sizeof(__pyx_k_os_path), 0, 0, 1, 1},
  {&__pyx_n_s_out_of_range, __pyx_k_out_of_range, sizeof(__pyx_k_out_of_range), 0, 0, 1, 1},
  {&__pyx_n_s_pair, __pyx_k_pair, sizeof(__pyx_k_pair), 0, 0, 1, 1},
  {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
  {&__pyx_kp_s_passed_credentials_must_both_be, __pyx_k_passed_credentials_must_both_be, sizeof(__pyx_k_passed_credentials_must_both_be), 0, 0, 1, 0},
  {&__pyx_n_s_pem_key_cert_pairs, __pyx_k_pem_key_cert_pairs, sizeof(__pyx_k_pem_key_cert_pairs), 0, 0, 1, 1},
  {&__pyx_n_s_pem_root_certificates, __pyx_k_pem_root_certificates, sizeof(__pyx_k_pem_root_certificates), 0, 0, 1, 1},
  {&__pyx_n_s_pem_root_certs, __pyx_k_pem_root_certs, sizeof(__pyx_k_pem_root_certs), 0, 0, 1, 1},
  {&__pyx_n_s_permission_denied, __pyx_k_permission_denied, sizeof(__pyx_k_permission_denied), 0, 0, 1, 1},
  {&__pyx_n_s_pkg_resources, __pyx_k_pkg_resources, sizeof(__pyx_k_pkg_resources), 0, 0, 1, 1},
  {&__pyx_n_s_plugin, __pyx_k_plugin, sizeof(__pyx_k_plugin), 0, 0, 1, 1},
  {&__pyx_n_s_plugin_callback, __pyx_k_plugin_callback, sizeof(__pyx_k_plugin_callback), 0, 0, 1, 1},
  {&__pyx_n_s_plugin_get_metadata_locals_pytho, __pyx_k_plugin_get_metadata_locals_pytho, sizeof(__pyx_k_plugin_get_metadata_locals_pytho), 0, 0, 1, 1},
  {&__pyx_n_s_poll, __pyx_k_poll, sizeof(__pyx_k_poll), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_primary_user_agent_string, __pyx_k_primary_user_agent_string, sizeof(__pyx_k_primary_user_agent_string), 0, 0, 1, 1},
  {&__pyx_n_s_private_key, __pyx_k_private_key, sizeof(__pyx_k_private_key), 0, 0, 1, 1},
  {&__pyx_kp_s_private_var_folders_g4_3ppgfd8x, __pyx_k_private_var_folders_g4_3ppgfd8x, sizeof(__pyx_k_private_var_folders_g4_3ppgfd8x), 0, 0, 1, 0},
  {&__pyx_n_s_python_callback, __pyx_k_python_callback, sizeof(__pyx_k_python_callback), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_queue, __pyx_k_queue, sizeof(__pyx_k_queue), 0, 0, 1, 1},
  {&__pyx_kp_s_queue_must_be_live, __pyx_k_queue_must_be_live, sizeof(__pyx_k_queue_must_be_live), 0, 0, 1, 0},
  {&__pyx_kp_s_queue_must_be_shutting_down_to_b, __pyx_k_queue_must_be_shutting_down_to_b, sizeof(__pyx_k_queue_must_be_shutting_down_to_b), 0, 0, 1, 0},
  {&__pyx_kp_s_queue_must_not_be_shutting_down, __pyx_k_queue_must_not_be_shutting_down, sizeof(__pyx_k_queue_must_not_be_shutting_down), 0, 0, 1, 0},
  {&__pyx_n_s_queue_shutdown, __pyx_k_queue_shutdown, sizeof(__pyx_k_queue_shutdown), 0, 0, 1, 1},
  {&__pyx_n_s_queue_timeout, __pyx_k_queue_timeout, sizeof(__pyx_k_queue_timeout), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_ready, __pyx_k_ready, sizeof(__pyx_k_ready), 0, 0, 1, 1},
  {&__pyx_n_s_receive_close_on_server, __pyx_k_receive_close_on_server, sizeof(__pyx_k_receive_close_on_server), 0, 0, 1, 1},
  {&__pyx_n_s_receive_initial_metadata, __pyx_k_receive_initial_metadata, sizeof(__pyx_k_receive_initial_metadata), 0, 0, 1, 1},
  {&__pyx_n_s_receive_message, __pyx_k_receive_message, sizeof(__pyx_k_receive_message), 0, 0, 1, 1},
  {&__pyx_n_s_receive_status_on_client, __pyx_k_receive_status_on_client, sizeof(__pyx_k_receive_status_on_client), 0, 0, 1, 1},
  {&__pyx_n_s_register_non_listening_completio, __pyx_k_register_non_listening_completio, sizeof(__pyx_k_register_non_listening_completio), 0, 0, 1, 1},
  {&__pyx_n_s_request_call_details, __pyx_k_request_call_details, sizeof(__pyx_k_request_call_details), 0, 0, 1, 1},
  {&__pyx_n_s_request_metadata, __pyx_k_request_metadata, sizeof(__pyx_k_request_metadata), 0, 0, 1, 1},
  {&__pyx_n_s_resource_exhausted, __pyx_k_resource_exhausted, sizeof(__pyx_k_resource_exhausted), 0, 0, 1, 1},
  {&__pyx_n_s_resource_string, __pyx_k_resource_string, sizeof(__pyx_k_resource_string), 0, 0, 1, 1},
  {&__pyx_n_s_retain_self, __pyx_k_retain_self, sizeof(__pyx_k_retain_self), 0, 0, 1, 1},
  {&__pyx_n_s_secondary_user_agent_string, __pyx_k_secondary_user_agent_string, sizeof(__pyx_k_secondary_user_agent_string), 0, 0, 1, 1},
  {&__pyx_kp_s_self_must_be_an_operation_receiv, __pyx_k_self_must_be_an_operation_receiv, sizeof(__pyx_k_self_must_be_an_operation_receiv), 0, 0, 1, 0},
  {&__pyx_kp_s_self_must_be_an_operation_receiv_2, __pyx_k_self_must_be_an_operation_receiv_2, sizeof(__pyx_k_self_must_be_an_operation_receiv_2), 0, 0, 1, 0},
  {&__pyx_kp_s_self_must_be_an_operation_receiv_3, __pyx_k_self_must_be_an_operation_receiv_3, sizeof(__pyx_k_self_must_be_an_operation_receiv_3), 0, 0, 1, 0},
  {&__pyx_kp_s_self_must_be_an_operation_receiv_4, __pyx_k_self_must_be_an_operation_receiv_4, sizeof(__pyx_k_self_must_be_an_operation_receiv_4), 0, 0, 1, 0},
  {&__pyx_kp_s_self_must_be_an_operation_receiv_5, __pyx_k_self_must_be_an_operation_receiv_5, sizeof(__pyx_k_self_must_be_an_operation_receiv_5), 0, 0, 1, 0},
  {&__pyx_n_s_send_close_from_client, __pyx_k_send_close_from_client, sizeof(__pyx_k_send_close_from_client), 0, 0, 1, 1},
  {&__pyx_n_s_send_initial_metadata, __pyx_k_send_initial_metadata, sizeof(__pyx_k_send_initial_metadata), 0, 0, 1, 1},
  {&__pyx_n_s_send_message, __pyx_k_send_message, sizeof(__pyx_k_send_message), 0, 0, 1, 1},
  {&__pyx_n_s_send_status_from_server, __pyx_k_send_status_from_server, sizeof(__pyx_k_send_status_from_server), 0, 0, 1, 1},
  {&__pyx_n_s_server_credentials, __pyx_k_server_credentials, sizeof(__pyx_k_server_credentials), 0, 0, 1, 1},
  {&__pyx_n_s_server_credentials_ssl, __pyx_k_server_credentials_ssl, sizeof(__pyx_k_server_credentials_ssl), 0, 0, 1, 1},
  {&__pyx_kp_s_server_must_be_started_and_not_s, __pyx_k_server_must_be_started_and_not_s, sizeof(__pyx_k_server_must_be_started_and_not_s), 0, 0, 1, 0},
  {&__pyx_n_s_server_queue, __pyx_k_server_queue, sizeof(__pyx_k_server_queue), 0, 0, 1, 1},
  {&__pyx_kp_s_server_queue_must_be_a_registere, __pyx_k_server_queue_must_be_a_registere, sizeof(__pyx_k_server_queue_must_be_a_registere), 0, 0, 1, 0},
  {&__pyx_n_s_shutdown, __pyx_k_shutdown, sizeof(__pyx_k_shutdown), 0, 0, 1, 1},
  {&__pyx_n_s_sleep, __pyx_k_sleep, sizeof(__pyx_k_sleep), 0, 0, 1, 1},
  {&__pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_k_src_python_grpcio_grpc__cython, sizeof(__pyx_k_src_python_grpcio_grpc__cython), 0, 0, 1, 0},
  {&__pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_k_src_python_grpcio_grpc__cython_2, sizeof(__pyx_k_src_python_grpcio_grpc__cython_2), 0, 0, 1, 0},
  {&__pyx_n_s_ssl_pem_key_cert_pair, __pyx_k_ssl_pem_key_cert_pair, sizeof(__pyx_k_ssl_pem_key_cert_pair), 0, 0, 1, 1},
  {&__pyx_n_s_ssl_target_name_override, __pyx_k_ssl_target_name_override, sizeof(__pyx_k_ssl_target_name_override), 0, 0, 1, 1},
  {&__pyx_n_s_start_batch, __pyx_k_start_batch, sizeof(__pyx_k_start_batch), 0, 0, 1, 1},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_status, __pyx_k_status, sizeof(__pyx_k_status), 0, 0, 1, 1},
  {&__pyx_n_s_success, __pyx_k_success, sizeof(__pyx_k_success), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_tag, __pyx_k_tag, sizeof(__pyx_k_tag), 0, 0, 1, 1},
  {&__pyx_n_s_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_the_server_has_already_started, __pyx_k_the_server_has_already_started, sizeof(__pyx_k_the_server_has_already_started), 0, 0, 1, 0},
  {&__pyx_kp_s_the_server_hasn_t_started_yet, __pyx_k_the_server_hasn_t_started_yet, sizeof(__pyx_k_the_server_hasn_t_started_yet), 0, 0, 1, 0},
  {&__pyx_kp_s_the_server_must_be_shutting_down, __pyx_k_the_server_must_be_shutting_down, sizeof(__pyx_k_the_server_must_be_shutting_down), 0, 0, 1, 0},
  {&__pyx_n_s_threading, __pyx_k_threading, sizeof(__pyx_k_threading), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_token_lifetime, __pyx_k_token_lifetime, sizeof(__pyx_k_token_lifetime), 0, 0, 1, 1},
  {&__pyx_n_s_too_many_operations, __pyx_k_too_many_operations, sizeof(__pyx_k_too_many_operations), 0, 0, 1, 1},
  {&__pyx_n_s_transient_failure, __pyx_k_transient_failure, sizeof(__pyx_k_transient_failure), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_unauthenticated, __pyx_k_unauthenticated, sizeof(__pyx_k_unauthenticated), 0, 0, 1, 1},
  {&__pyx_n_s_unavailable, __pyx_k_unavailable, sizeof(__pyx_k_unavailable), 0, 0, 1, 1},
  {&__pyx_n_s_unimplemented, __pyx_k_unimplemented, sizeof(__pyx_k_unimplemented), 0, 0, 1, 1},
  {&__pyx_n_s_unknown, __pyx_k_unknown, sizeof(__pyx_k_unknown), 0, 0, 1, 1},
  {&__pyx_n_s_user_tag, __pyx_k_user_tag, sizeof(__pyx_k_user_tag), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 167, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 39, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(2, 43, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(3, 278, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 378, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 402, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(4, 151, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/grpc_string.pyx.pxi":37
 *     return s
 *   elif isinstance(s, unicode):
 *     return s.encode('ascii')             # <<<<<<<<<<<<<<
 *   else:
 *     raise TypeError('Expected bytes, str, or unicode, not {}'.format(type(s)))
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_n_s_ascii); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":43
 *   def _start_batch(self, operations, tag, retain_self):
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")             # <<<<<<<<<<<<<<
 *     cdef grpc_call_error result
 *     cdef Operations cy_operations = Operations(operations)
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_invalid_call_object_cannot_be_us); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":72
 *     details = str_to_bytes(details)
 *     if not self.is_valid:
 *       raise ValueError("invalid call object cannot be used from Python")             # <<<<<<<<<<<<<<
 *     if (details is None) != (error_code == GRPC_STATUS__DO_NOT_USE):
 *       raise ValueError("if error_code is specified, so must details "
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_invalid_call_object_cannot_be_us); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":74
 *       raise ValueError("invalid call object cannot be used from Python")
 *     if (details is None) != (error_code == GRPC_STATUS__DO_NOT_USE):
 *       raise ValueError("if error_code is specified, so must details "             # <<<<<<<<<<<<<<
 *                        "(and vice-versa)")
 *     cdef grpc_call_error result
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_if_error_code_is_specified_so_mu); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(2, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/channel.pyx.pxi":61
 *                   method, host, Timespec deadline not None):
 *     if queue.is_shutting_down:
 *       raise ValueError("queue must not be shutting down or shutdown")             # <<<<<<<<<<<<<<
 *     cdef char *method_c_string = method
 *     cdef char *host_c_string = NULL
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_queue_must_not_be_shutting_down); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(5, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":101
 *     grpc_init()
 *     if not callable(plugin_callback):
 *       raise ValueError('expected callable plugin_callback')             # <<<<<<<<<<<<<<
 *     self.plugin_callback = plugin_callback
 *     self.plugin_name = name
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_expected_callable_plugin_callbac); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(3, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":141
 *     void *state, grpc_auth_metadata_context context,
 *     grpc_credentials_plugin_metadata_cb cb, void *user_data) with gil:
 *   def python_callback(             # <<<<<<<<<<<<<<
 *       Metadata metadata, grpc_status_code status,
 *       bytes error_details):
 */
  __pyx_tuple__8 = PyTuple_Pack(3, __pyx_n_s_metadata, __pyx_n_s_status, __pyx_n_s_error_details); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(3, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_python_callback, 141, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(3, 141, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":183
 *     CallCredentials credentials_2 not None):
 *   if not credentials_1.is_valid or not credentials_2.is_valid:
 *     raise ValueError("passed credentials must both be valid")             # <<<<<<<<<<<<<<
 *   cdef ChannelCredentials credentials = ChannelCredentials()
 *   with nogil:
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_passed_credentials_must_both_be); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(3, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":196
 *     CallCredentials credentials_2 not None):
 *   if not credentials_1.is_valid or not credentials_2.is_valid:
 *     raise ValueError("passed credentials must both be valid")             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_passed_credentials_must_both_be); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(3, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":266
 *   for pair in pem_key_cert_pairs:
 *     if not isinstance(pair, SslPemKeyCertPair):
 *       raise TypeError("expected pem_key_cert_pairs to be sequence of "             # <<<<<<<<<<<<<<
 *                       "SslPemKeyCertPair")
 *   cdef ServerCredentials credentials = ServerCredentials()
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_expected_pem_key_cert_pairs_to_b); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(3, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":116
 *   def clear(self):
 *     if not self.is_shutting_down:
 *       raise ValueError('queue must be shutting down to be cleared')             # <<<<<<<<<<<<<<
 *     while self.poll().type != GRPC_QUEUE_SHUTDOWN:
 *       pass
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_queue_must_be_shutting_down_to_b); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(6, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":330
 *     for arg in self.args:
 *       if not isinstance(arg, ChannelArg):
 *         raise TypeError("expected list of ChannelArg")             # <<<<<<<<<<<<<<
 *     self.c_args.arguments_length = len(self.args)
 *     with nogil:
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_expected_list_of_ChannelArg); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":378
 *       return self.value
 *     else:
 *       raise IndexError("index must be 0 (key) or 1 (value)")             # <<<<<<<<<<<<<<
 * 
 *   def __iter__(self):
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_index_must_be_0_key_or_1_value); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":412
 *     for metadatum in metadata:
 *       if not isinstance(metadatum, Metadatum):
 *         raise TypeError("expected list of Metadatum")             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_metadata_array_init(&self.c_metadata_array)
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_expected_list_of_Metadatum); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":469
 *   def received_message(self):
 *     if self.c_op.type != GRPC_OP_RECV_MESSAGE:
 *       raise TypeError("self must be an operation receiving a message")             # <<<<<<<<<<<<<<
 *     return self._received_message
 * 
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_self_must_be_an_operation_receiv); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":482
 *     if (self.c_op.type != GRPC_OP_RECV_INITIAL_METADATA and
 *         self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT):
 *       raise TypeError("self must be an operation receiving metadata")             # <<<<<<<<<<<<<<
 *     return self._received_metadata
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_self_must_be_an_operation_receiv_2); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":495
 *   def received_status_code(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving a status code")             # <<<<<<<<<<<<<<
 *     return self._received_status_code
 * 
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_self_must_be_an_operation_receiv_3); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":507
 *   def received_status_details(self):
 *     if self.c_op.type != GRPC_OP_RECV_STATUS_ON_CLIENT:
 *       raise TypeError("self must be an operation receiving status details")             # <<<<<<<<<<<<<<
 *     if self._received_status_details:
 *       return self._received_status_details
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_self_must_be_an_operation_receiv_4); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":525
 *   def received_cancelled(self):
 *     if self.c_op.type != GRPC_OP_RECV_CLOSE_ON_SERVER:
 *       raise TypeError("self must be an operation receiving cancellation "             # <<<<<<<<<<<<<<
 *                       "information")
 *     return False if self._received_cancelled == 0 else True
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_self_must_be_an_operation_receiv_5); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":601
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE
 *   op.c_op.flags = flags
 *   op._received_message = ByteBuffer(None)             # <<<<<<<<<<<<<<
 *   # n.b. the c_op.data.receive_message field needs to be deleted by us,
 *   # anyway, so we just let that be handled by the ByteBuffer() we allocated
 */
  __pyx_tuple__22 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":664
 *     for operation in self.operations:
 *       if not isinstance(operation, Operation):
 *         raise TypeError("expected operations to be iterable of Operation")             # <<<<<<<<<<<<<<
 *     self.c_nops = len(self.operations)
 *     with nogil:
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_expected_operations_to_be_iterab); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":37
 * cdef grpc_ssl_roots_override_result ssl_roots_override_callback(
 *     char **pem_root_certs) with gil:
 *   temporary_pem_root_certs = pkg_resources.resource_string(             # <<<<<<<<<<<<<<
 *       'grpc._cython', '_credentials/roots.pem')
 *   pem_root_certs[0] = <char *>gpr_malloc(len(temporary_pem_root_certs) + 1)
 */
  __pyx_tuple__24 = PyTuple_Pack(2, __pyx_kp_s_grpc__cython, __pyx_kp_s_credentials_roots_pem); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(8, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":55
 *       CompletionQueue server_queue not None, tag):
 *     if not self.is_started or self.is_shutting_down:
 *       raise ValueError("server must be started and not shutting down")             # <<<<<<<<<<<<<<
 *     if server_queue not in self.registered_completion_queues:
 *       raise ValueError("server_queue must be a registered completion queue")
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_server_must_be_started_and_not_s); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(4, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":57
 *       raise ValueError("server must be started and not shutting down")
 *     if server_queue not in self.registered_completion_queues:
 *       raise ValueError("server_queue must be a registered completion queue")             # <<<<<<<<<<<<<<
 *     cdef grpc_call_error result
 *     cdef OperationTag operation_tag = OperationTag(tag)
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_server_queue_must_be_a_registere); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(4, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":79
 *       self, CompletionQueue queue not None):
 *     if self.is_started:
 *       raise ValueError("cannot register completion queues after start")             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_server_register_completion_queue(
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_cannot_register_completion_queue); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(4, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":88
 *       self, CompletionQueue queue not None):
 *     if self.is_started:
 *       raise ValueError("cannot register completion queues after start")             # <<<<<<<<<<<<<<
 *     with nogil:
 *       grpc_server_register_non_listening_completion_queue(
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_cannot_register_completion_queue); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(4, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":96
 *   def start(self):
 *     if self.is_started:
 *       raise ValueError("the server has already started")             # <<<<<<<<<<<<<<
 *     self.backup_shutdown_queue = CompletionQueue()
 *     self.register_non_listening_completion_queue(self.backup_shutdown_queue)
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_the_server_has_already_started); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(4, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":103
 *       grpc_server_start(self.c_server)
 *     # Ensure the core has gotten a chance to do the start-up work
 *     self.backup_shutdown_queue.poll(Timespec(None))             # <<<<<<<<<<<<<<
 * 
 *   def add_http2_port(self, bytes address,
 */
  __pyx_tuple__30 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(4, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":135
 *     cdef OperationTag operation_tag
 *     if queue.is_shutting_down:
 *       raise ValueError("queue must be live")             # <<<<<<<<<<<<<<
 *     elif not self.is_started:
 *       raise ValueError("the server hasn't started yet")
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_queue_must_be_live); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(4, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":137
 *       raise ValueError("queue must be live")
 *     elif not self.is_started:
 *       raise ValueError("the server hasn't started yet")             # <<<<<<<<<<<<<<
 *     elif self.is_shutting_down:
 *       return
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_the_server_hasn_t_started_yet); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(4, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":141
 *       return
 *     elif queue not in self.registered_completion_queues:
 *       raise ValueError("expected registered completion queue")             # <<<<<<<<<<<<<<
 *     else:
 *       self._c_shutdown(queue, tag)
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_expected_registered_completion_q); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(4, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":151
 *   def cancel_all_calls(self):
 *     if not self.is_shutting_down:
 *       raise RuntimeError("the server must be shutting down to cancel all calls")             # <<<<<<<<<<<<<<
 *     elif self.is_shutdown:
 *       return
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_the_server_must_be_shutting_down); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(4, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":174
 *         # much but repeatedly release the GIL and wait
 *         while not self.is_shutdown:
 *           time.sleep(0)             # <<<<<<<<<<<<<<
 *       grpc_server_destroy(self.c_server)
 *     grpc_shutdown()
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(4, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":154
 *   cpython.Py_DECREF(<CredentialsMetadataPlugin>state)
 * 
 * def channel_credentials_google_default():             # <<<<<<<<<<<<<<
 *   cdef ChannelCredentials credentials = ChannelCredentials();
 *   with nogil:
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_n_s_credentials); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(3, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(0, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_channel_credentials_google_defau, 154, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(3, 154, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":160
 *   return credentials
 * 
 * def channel_credentials_ssl(pem_root_certificates,             # <<<<<<<<<<<<<<
 *                             SslPemKeyCertPair ssl_pem_key_cert_pair):
 *   pem_root_certificates = str_to_bytes(pem_root_certificates)
 */
  __pyx_tuple__38 = PyTuple_Pack(4, __pyx_n_s_pem_root_certificates, __pyx_n_s_ssl_pem_key_cert_pair, __pyx_n_s_credentials, __pyx_n_s_c_pem_root_certificates); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(3, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_channel_credentials_ssl, 160, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(3, 160, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":179
 *   return credentials
 * 
 * def channel_credentials_composite(             # <<<<<<<<<<<<<<
 *     ChannelCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 */
  __pyx_tuple__40 = PyTuple_Pack(3, __pyx_n_s_credentials_1, __pyx_n_s_credentials_2, __pyx_n_s_credentials); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(3, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_channel_credentials_composite, 179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(3, 179, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":192
 *   return credentials
 * 
 * def call_credentials_composite(             # <<<<<<<<<<<<<<
 *     CallCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 */
  __pyx_tuple__42 = PyTuple_Pack(3, __pyx_n_s_credentials_1, __pyx_n_s_credentials_2, __pyx_n_s_credentials); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(3, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_call_credentials_composite, 192, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(3, 192, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":205
 *   return credentials
 * 
 * def call_credentials_google_compute_engine():             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_n_s_credentials); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(3, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(0, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_call_credentials_google_compute, 205, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(3, 205, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":212
 *   return credentials
 * 
 * def call_credentials_service_account_jwt_access(             # <<<<<<<<<<<<<<
 *     json_key, Timespec token_lifetime not None):
 *   json_key = str_to_bytes(json_key)
 */
  __pyx_tuple__46 = PyTuple_Pack(4, __pyx_n_s_json_key, __pyx_n_s_token_lifetime, __pyx_n_s_credentials, __pyx_n_s_json_key_c_string); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(3, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_call_credentials_service_account, 212, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(3, 212, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":224
 *   return credentials
 * 
 * def call_credentials_google_refresh_token(json_refresh_token):             # <<<<<<<<<<<<<<
 *   json_refresh_token = str_to_bytes(json_refresh_token)
 *   cdef CallCredentials credentials = CallCredentials()
 */
  __pyx_tuple__48 = PyTuple_Pack(3, __pyx_n_s_json_refresh_token, __pyx_n_s_credentials, __pyx_n_s_json_refresh_token_c_string); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(3, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_call_credentials_google_refresh, 224, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(3, 224, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":234
 *   return credentials
 * 
 * def call_credentials_google_iam(authorization_token, authority_selector):             # <<<<<<<<<<<<<<
 *   authorization_token = str_to_bytes(authorization_token)
 *   authority_selector = str_to_bytes(authority_selector)
 */
  __pyx_tuple__50 = PyTuple_Pack(5, __pyx_n_s_authorization_token, __pyx_n_s_authority_selector, __pyx_n_s_credentials, __pyx_n_s_authorization_token_c_string, __pyx_n_s_authority_selector_c_string); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(3, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_call_credentials_google_iam, 234, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(3, 234, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":247
 *   return credentials
 * 
 * def call_credentials_metadata_plugin(CredentialsMetadataPlugin plugin):             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()
 */
  __pyx_tuple__52 = PyTuple_Pack(3, __pyx_n_s_plugin, __pyx_n_s_credentials, __pyx_n_s_c_plugin); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(3, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_call_credentials_metadata_plugin, 247, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(3, 247, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":257
 *   return credentials
 * 
 * def server_credentials_ssl(pem_root_certs, pem_key_cert_pairs,             # <<<<<<<<<<<<<<
 *                            bint force_client_auth):
 *   pem_root_certs = str_to_bytes(pem_root_certs)
 */
  __pyx_tuple__54 = PyTuple_Pack(7, __pyx_n_s_pem_root_certs, __pyx_n_s_pem_key_cert_pairs, __pyx_n_s_force_client_auth, __pyx_n_s_c_pem_root_certs, __pyx_n_s_pair, __pyx_n_s_credentials, __pyx_n_s_i); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(3, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(3, 0, 7, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython, __pyx_n_s_server_credentials_ssl, 257, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(3, 257, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":168
 * 
 *   @staticmethod
 *   def infinite_future():             # <<<<<<<<<<<<<<
 *     return Timespec(float("+inf"))
 * 
 */
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_infinite_future, 168, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 168, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":172
 * 
 *   @staticmethod
 *   def infinite_past():             # <<<<<<<<<<<<<<
 *     return Timespec(float("-inf"))
 * 
 */
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_infinite_past, 172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 172, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":543
 *     grpc_shutdown()
 * 
 * def operation_send_initial_metadata(Metadata metadata, int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_INITIAL_METADATA
 */
  __pyx_tuple__58 = PyTuple_Pack(3, __pyx_n_s_metadata, __pyx_n_s_flags, __pyx_n_s_op); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_operation_send_initial_metadata, 543, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 543, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":554
 *   return op
 * 
 * def operation_send_message(data, int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_MESSAGE
 */
  __pyx_tuple__60 = PyTuple_Pack(4, __pyx_n_s_data, __pyx_n_s_flags, __pyx_n_s_op, __pyx_n_s_byte_buffer); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_operation_send_message, 554, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 554, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":564
 *   return op
 * 
 * def operation_send_close_from_client(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 */
  __pyx_tuple__62 = PyTuple_Pack(3, __pyx_n_s_flags, __pyx_n_s_flags, __pyx_n_s_op); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__62, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_operation_send_close_from_client, 564, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 564, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":571
 *   return op
 * 
 * def operation_send_status_from_server(             # <<<<<<<<<<<<<<
 *     Metadata metadata, grpc_status_code code, bytes details, int flags):
 *   cdef Operation op = Operation()
 */
  __pyx_tuple__64 = PyTuple_Pack(5, __pyx_n_s_metadata, __pyx_n_s_code, __pyx_n_s_details, __pyx_n_s_flags, __pyx_n_s_op); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_operation_send_status_from_serve, 571, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(0, 571, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":587
 *   return op
 * 
 * def operation_receive_initial_metadata(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_INITIAL_METADATA
 */
  __pyx_tuple__66 = PyTuple_Pack(3, __pyx_n_s_flags, __pyx_n_s_flags, __pyx_n_s_op); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_operation_receive_initial_metada, 587, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 587, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":597
 *   return op
 * 
 * def operation_receive_message(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE
 */
  __pyx_tuple__68 = PyTuple_Pack(3, __pyx_n_s_flags, __pyx_n_s_flags, __pyx_n_s_op); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_operation_receive_message, 597, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 597, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":609
 *   return op
 * 
 * def operation_receive_status_on_client(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_STATUS_ON_CLIENT
 */
  __pyx_tuple__70 = PyTuple_Pack(3, __pyx_n_s_flags, __pyx_n_s_flags, __pyx_n_s_op); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_operation_receive_status_on_clie, 609, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 609, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":625
 *   return op
 * 
 * def operation_receive_close_on_server(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_CLOSE_ON_SERVER
 */
  __pyx_tuple__72 = PyTuple_Pack(3, __pyx_n_s_flags, __pyx_n_s_flags, __pyx_n_s_op); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__72, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_operation_receive_close_on_serve, 625, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(0, 625, __pyx_L1_error)

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":713
 * 
 * 
 * def compression_algorithm_name(grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *   cdef char* name
 *   with nogil:
 */
  __pyx_tuple__74 = PyTuple_Pack(3, __pyx_n_s_algorithm, __pyx_n_s_algorithm, __pyx_n_s_name); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__74, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_python_grpcio_grpc__cython_2, __pyx_n_s_compression_algorithm_name, 713, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 713, __pyx_L1_error)

  /* "grpc/_cython/cygrpc.pyx":57
 * 
 * 
 * def _initialize():             # <<<<<<<<<<<<<<
 *   grpc_set_ssl_roots_override_callback(
 *           <grpc_ssl_roots_override_callback>ssl_roots_override_callback)
 */
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_private_var_folders_g4_3ppgfd8x, __pyx_n_s_initialize, 57, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(9, 57, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(9, 1, __pyx_L1_error);
  __pyx_float_1e9 = PyFloat_FromDouble(1e9); if (unlikely(!__pyx_float_1e9)) __PYX_ERR(9, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(9, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initcygrpc(void); /*proto*/
PyMODINIT_FUNC initcygrpc(void)
#else
PyMODINIT_FUNC PyInit_cygrpc(void); /*proto*/
PyMODINIT_FUNC PyInit_cygrpc(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_cygrpc(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(9, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(9, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(9, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cygrpc", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(9, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(9, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(9, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(9, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_grpc___cython__cygrpc) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(9, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "grpc._cython.cygrpc")) {
      if (unlikely(PyDict_SetItemString(modules, "grpc._cython.cygrpc", __pyx_m) < 0)) __PYX_ERR(9, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("plugin_get_metadata", (void (*)(void))__pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata, "void (void *, grpc_auth_metadata_context, grpc_credentials_plugin_metadata_cb, void *)") < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("plugin_destroy_c_plugin_state", (void (*)(void))__pyx_f_4grpc_7_cython_6cygrpc_plugin_destroy_c_plugin_state, "void (void *)") < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ssl_roots_override_callback", (void (*)(void))__pyx_f_4grpc_7_cython_6cygrpc_ssl_roots_override_callback, "grpc_ssl_roots_override_result (char **)") < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Call) < 0) __PYX_ERR(2, 33, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Call.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Call", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Call) < 0) __PYX_ERR(2, 33, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Call = &__pyx_type_4grpc_7_cython_6cygrpc_Call;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Channel) < 0) __PYX_ERR(5, 33, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Channel.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Channel", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Channel) < 0) __PYX_ERR(5, 33, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Channel = &__pyx_type_4grpc_7_cython_6cygrpc_Channel;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_ChannelCredentials) < 0) __PYX_ERR(3, 33, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_ChannelCredentials.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "ChannelCredentials", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_ChannelCredentials) < 0) __PYX_ERR(3, 33, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_ChannelCredentials = &__pyx_type_4grpc_7_cython_6cygrpc_ChannelCredentials;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_CallCredentials) < 0) __PYX_ERR(3, 54, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_CallCredentials.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "CallCredentials", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_CallCredentials) < 0) __PYX_ERR(3, 54, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_CallCredentials = &__pyx_type_4grpc_7_cython_6cygrpc_CallCredentials;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_ServerCredentials) < 0) __PYX_ERR(3, 73, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_ServerCredentials.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "ServerCredentials", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_ServerCredentials) < 0) __PYX_ERR(3, 73, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_ServerCredentials = &__pyx_type_4grpc_7_cython_6cygrpc_ServerCredentials;
  __pyx_vtabptr_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin = &__pyx_vtable_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin;
  __pyx_vtable_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin.make_c_plugin = (grpc_metadata_credentials_plugin (*)(struct __pyx_obj_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin *))__pyx_f_4grpc_7_cython_6cygrpc_25CredentialsMetadataPlugin_make_c_plugin;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin) < 0) __PYX_ERR(3, 86, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin.tp_dict, __pyx_vtabptr_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin) < 0) __PYX_ERR(3, 86, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "CredentialsMetadataPlugin", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin) < 0) __PYX_ERR(3, 86, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin = &__pyx_type_4grpc_7_cython_6cygrpc_CredentialsMetadataPlugin;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_AuthMetadataContext) < 0) __PYX_ERR(3, 119, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_AuthMetadataContext.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "AuthMetadataContext", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_AuthMetadataContext) < 0) __PYX_ERR(3, 119, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_AuthMetadataContext = &__pyx_type_4grpc_7_cython_6cygrpc_AuthMetadataContext;
  __pyx_vtabptr_4grpc_7_cython_6cygrpc_CompletionQueue = &__pyx_vtable_4grpc_7_cython_6cygrpc_CompletionQueue;
  __pyx_vtable_4grpc_7_cython_6cygrpc_CompletionQueue._interpret_event = (PyObject *(*)(struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *, grpc_event))__pyx_f_4grpc_7_cython_6cygrpc_15CompletionQueue__interpret_event;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_CompletionQueue) < 0) __PYX_ERR(6, 38, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_CompletionQueue.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_4grpc_7_cython_6cygrpc_CompletionQueue.tp_dict, __pyx_vtabptr_4grpc_7_cython_6cygrpc_CompletionQueue) < 0) __PYX_ERR(6, 38, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "CompletionQueue", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_CompletionQueue) < 0) __PYX_ERR(6, 38, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_CompletionQueue = &__pyx_type_4grpc_7_cython_6cygrpc_CompletionQueue;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Timespec) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Timespec.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Timespec", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Timespec) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Timespec = &__pyx_type_4grpc_7_cython_6cygrpc_Timespec;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_CallDetails) < 0) __PYX_ERR(0, 176, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_CallDetails.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "CallDetails", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_CallDetails) < 0) __PYX_ERR(0, 176, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_CallDetails = &__pyx_type_4grpc_7_cython_6cygrpc_CallDetails;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_OperationTag) < 0) __PYX_ERR(0, 209, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_OperationTag.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "OperationTag", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_OperationTag) < 0) __PYX_ERR(0, 209, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_OperationTag = &__pyx_type_4grpc_7_cython_6cygrpc_OperationTag;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Event) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Event.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Event", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Event) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Event = &__pyx_type_4grpc_7_cython_6cygrpc_Event;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_ByteBuffer) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_ByteBuffer.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "ByteBuffer", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_ByteBuffer) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_ByteBuffer = &__pyx_type_4grpc_7_cython_6cygrpc_ByteBuffer;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_SslPemKeyCertPair) < 0) __PYX_ERR(0, 297, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_SslPemKeyCertPair.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "SslPemKeyCertPair", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_SslPemKeyCertPair) < 0) __PYX_ERR(0, 297, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_SslPemKeyCertPair = &__pyx_type_4grpc_7_cython_6cygrpc_SslPemKeyCertPair;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_ChannelArg) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_ChannelArg.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "ChannelArg", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_ChannelArg) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_ChannelArg = &__pyx_type_4grpc_7_cython_6cygrpc_ChannelArg;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_ChannelArgs) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_ChannelArgs.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "ChannelArgs", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_ChannelArgs) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_ChannelArgs = &__pyx_type_4grpc_7_cython_6cygrpc_ChannelArgs;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Metadatum) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Metadatum.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Metadatum", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Metadatum) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Metadatum = &__pyx_type_4grpc_7_cython_6cygrpc_Metadatum;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Metadata) < 0) __PYX_ERR(0, 405, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Metadata.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Metadata", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Metadata) < 0) __PYX_ERR(0, 405, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Metadata = &__pyx_type_4grpc_7_cython_6cygrpc_Metadata;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Operation) < 0) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Operation.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Operation", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Operation) < 0) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Operation = &__pyx_type_4grpc_7_cython_6cygrpc_Operation;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Operations) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Operations.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Operations", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Operations) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Operations = &__pyx_type_4grpc_7_cython_6cygrpc_Operations;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_CompressionOptions) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_CompressionOptions.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "CompressionOptions", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_CompressionOptions) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_CompressionOptions = &__pyx_type_4grpc_7_cython_6cygrpc_CompressionOptions;
  __pyx_vtabptr_4grpc_7_cython_6cygrpc_Server = &__pyx_vtable_4grpc_7_cython_6cygrpc_Server;
  __pyx_vtable_4grpc_7_cython_6cygrpc_Server._c_shutdown = (PyObject *(*)(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *, struct __pyx_obj_4grpc_7_cython_6cygrpc_CompletionQueue *, PyObject *))__pyx_f_4grpc_7_cython_6cygrpc_6Server__c_shutdown;
  __pyx_vtable_4grpc_7_cython_6cygrpc_Server.notify_shutdown_complete = (PyObject *(*)(struct __pyx_obj_4grpc_7_cython_6cygrpc_Server *))__pyx_f_4grpc_7_cython_6cygrpc_6Server_notify_shutdown_complete;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc_Server) < 0) __PYX_ERR(4, 35, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc_Server.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_4grpc_7_cython_6cygrpc_Server.tp_dict, __pyx_vtabptr_4grpc_7_cython_6cygrpc_Server) < 0) __PYX_ERR(4, 35, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "Server", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc_Server) < 0) __PYX_ERR(4, 35, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc_Server = &__pyx_type_4grpc_7_cython_6cygrpc_Server;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc__MetadataIterator) < 0) __PYX_ERR(0, 384, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc__MetadataIterator.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "_MetadataIterator", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc__MetadataIterator) < 0) __PYX_ERR(0, 384, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc__MetadataIterator = &__pyx_type_4grpc_7_cython_6cygrpc__MetadataIterator;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc__OperationsIterator) < 0) __PYX_ERR(0, 634, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc__OperationsIterator.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "_OperationsIterator", (PyObject *)&__pyx_type_4grpc_7_cython_6cygrpc__OperationsIterator) < 0) __PYX_ERR(0, 634, __pyx_L1_error)
  __pyx_ptype_4grpc_7_cython_6cygrpc__OperationsIterator = &__pyx_type_4grpc_7_cython_6cygrpc__OperationsIterator;
  if (PyType_Ready(&__pyx_type_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata) < 0) __PYX_ERR(3, 138, __pyx_L1_error)
  __pyx_type_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata.tp_print = 0;
  __pyx_ptype_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata = &__pyx_type_4grpc_7_cython_6cygrpc___pyx_scope_struct____pyx_f_4grpc_7_cython_6cygrpc_plugin_get_metadata;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(10, 9, __pyx_L1_error)
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) __PYX_ERR(11, 8, __pyx_L1_error)
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) __PYX_ERR(12, 15, __pyx_L1_error)
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  #endif

  /* "grpc/_cython/cygrpc.pyx":32
 * cimport cpython
 * 
 * import pkg_resources             # <<<<<<<<<<<<<<
 * import os.path
 * import sys
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pkg_resources, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pkg_resources, __pyx_t_1) < 0) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "grpc/_cython/cygrpc.pyx":33
 * 
 * import pkg_resources
 * import os.path             # <<<<<<<<<<<<<<
 * import sys
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os_path, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(9, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "grpc/_cython/cygrpc.pyx":34
 * import pkg_resources
 * import os.path
 * import sys             # <<<<<<<<<<<<<<
 * 
 * # TODO(atash): figure out why the coverage tool gets confused about the Cython
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(9, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/call.pyx.pxi":68
 * 
 *   def cancel(
 *       self, grpc_status_code error_code=GRPC_STATUS__DO_NOT_USE,             # <<<<<<<<<<<<<<
 *       details=None):
 *     details = str_to_bytes(details)
 */
  __pyx_k__3 = GRPC_STATUS__DO_NOT_USE;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":154
 *   cpython.Py_DECREF(<CredentialsMetadataPlugin>state)
 * 
 * def channel_credentials_google_default():             # <<<<<<<<<<<<<<
 *   cdef ChannelCredentials credentials = ChannelCredentials();
 *   with nogil:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_1channel_credentials_google_default, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_channel_credentials_google_defau, __pyx_t_1) < 0) __PYX_ERR(3, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":160
 *   return credentials
 * 
 * def channel_credentials_ssl(pem_root_certificates,             # <<<<<<<<<<<<<<
 *                             SslPemKeyCertPair ssl_pem_key_cert_pair):
 *   pem_root_certificates = str_to_bytes(pem_root_certificates)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_3channel_credentials_ssl, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_channel_credentials_ssl, __pyx_t_1) < 0) __PYX_ERR(3, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":179
 *   return credentials
 * 
 * def channel_credentials_composite(             # <<<<<<<<<<<<<<
 *     ChannelCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_5channel_credentials_composite, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_channel_credentials_composite, __pyx_t_1) < 0) __PYX_ERR(3, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":192
 *   return credentials
 * 
 * def call_credentials_composite(             # <<<<<<<<<<<<<<
 *     CallCredentials credentials_1 not None,
 *     CallCredentials credentials_2 not None):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_7call_credentials_composite, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_call_credentials_composite, __pyx_t_1) < 0) __PYX_ERR(3, 192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":205
 *   return credentials
 * 
 * def call_credentials_google_compute_engine():             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   with nogil:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_9call_credentials_google_compute_engine, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_call_credentials_google_compute, __pyx_t_1) < 0) __PYX_ERR(3, 205, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":212
 *   return credentials
 * 
 * def call_credentials_service_account_jwt_access(             # <<<<<<<<<<<<<<
 *     json_key, Timespec token_lifetime not None):
 *   json_key = str_to_bytes(json_key)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_11call_credentials_service_account_jwt_access, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_call_credentials_service_account, __pyx_t_1) < 0) __PYX_ERR(3, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":224
 *   return credentials
 * 
 * def call_credentials_google_refresh_token(json_refresh_token):             # <<<<<<<<<<<<<<
 *   json_refresh_token = str_to_bytes(json_refresh_token)
 *   cdef CallCredentials credentials = CallCredentials()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_13call_credentials_google_refresh_token, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_call_credentials_google_refresh, __pyx_t_1) < 0) __PYX_ERR(3, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":234
 *   return credentials
 * 
 * def call_credentials_google_iam(authorization_token, authority_selector):             # <<<<<<<<<<<<<<
 *   authorization_token = str_to_bytes(authorization_token)
 *   authority_selector = str_to_bytes(authority_selector)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_15call_credentials_google_iam, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_call_credentials_google_iam, __pyx_t_1) < 0) __PYX_ERR(3, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":247
 *   return credentials
 * 
 * def call_credentials_metadata_plugin(CredentialsMetadataPlugin plugin):             # <<<<<<<<<<<<<<
 *   cdef CallCredentials credentials = CallCredentials()
 *   cdef grpc_metadata_credentials_plugin c_plugin = plugin.make_c_plugin()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_17call_credentials_metadata_plugin, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_call_credentials_metadata_plugin, __pyx_t_1) < 0) __PYX_ERR(3, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/credentials.pyx.pxi":257
 *   return credentials
 * 
 * def server_credentials_ssl(pem_root_certs, pem_key_cert_pairs,             # <<<<<<<<<<<<<<
 *                            bint force_client_auth):
 *   pem_root_certs = str_to_bytes(pem_root_certs)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_19server_credentials_ssl, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_server_credentials_ssl, __pyx_t_1) < 0) __PYX_ERR(3, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":32
 * cimport cpython
 * 
 * import threading             # <<<<<<<<<<<<<<
 * import time
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_threading, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_threading, __pyx_t_1) < 0) __PYX_ERR(6, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":33
 * 
 * import threading
 * import time             # <<<<<<<<<<<<<<
 * 
 * cdef int _INTERRUPT_CHECK_PERIOD_MS = 200
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_time, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_1) < 0) __PYX_ERR(6, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/completion_queue.pyx.pxi":35
 * import time
 * 
 * cdef int _INTERRUPT_CHECK_PERIOD_MS = 200             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_4grpc_7_cython_6cygrpc__INTERRUPT_CHECK_PERIOD_MS = 0xC8;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":31
 * 
 * 
 * class ConnectivityState:             # <<<<<<<<<<<<<<
 *   idle = GRPC_CHANNEL_IDLE
 *   connecting = GRPC_CHANNEL_CONNECTING
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_ConnectivityState, __pyx_n_s_ConnectivityState, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":32
 * 
 * class ConnectivityState:
 *   idle = GRPC_CHANNEL_IDLE             # <<<<<<<<<<<<<<
 *   connecting = GRPC_CHANNEL_CONNECTING
 *   ready = GRPC_CHANNEL_READY
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_connectivity_state(GRPC_CHANNEL_IDLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_idle, __pyx_t_2) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":33
 * class ConnectivityState:
 *   idle = GRPC_CHANNEL_IDLE
 *   connecting = GRPC_CHANNEL_CONNECTING             # <<<<<<<<<<<<<<
 *   ready = GRPC_CHANNEL_READY
 *   transient_failure = GRPC_CHANNEL_TRANSIENT_FAILURE
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_connectivity_state(GRPC_CHANNEL_CONNECTING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_connecting, __pyx_t_2) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":34
 *   idle = GRPC_CHANNEL_IDLE
 *   connecting = GRPC_CHANNEL_CONNECTING
 *   ready = GRPC_CHANNEL_READY             # <<<<<<<<<<<<<<
 *   transient_failure = GRPC_CHANNEL_TRANSIENT_FAILURE
 *   shutdown = GRPC_CHANNEL_SHUTDOWN
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_connectivity_state(GRPC_CHANNEL_READY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_ready, __pyx_t_2) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":35
 *   connecting = GRPC_CHANNEL_CONNECTING
 *   ready = GRPC_CHANNEL_READY
 *   transient_failure = GRPC_CHANNEL_TRANSIENT_FAILURE             # <<<<<<<<<<<<<<
 *   shutdown = GRPC_CHANNEL_SHUTDOWN
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_connectivity_state(GRPC_CHANNEL_TRANSIENT_FAILURE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_transient_failure, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":36
 *   ready = GRPC_CHANNEL_READY
 *   transient_failure = GRPC_CHANNEL_TRANSIENT_FAILURE
 *   shutdown = GRPC_CHANNEL_SHUTDOWN             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_connectivity_state(GRPC_CHANNEL_SHUTDOWN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_shutdown, __pyx_t_2) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":31
 * 
 * 
 * class ConnectivityState:             # <<<<<<<<<<<<<<
 *   idle = GRPC_CHANNEL_IDLE
 *   connecting = GRPC_CHANNEL_CONNECTING
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_ConnectivityState, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ConnectivityState, __pyx_t_2) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":39
 * 
 * 
 * class ChannelArgKey:             # <<<<<<<<<<<<<<
 *   enable_census = GRPC_ARG_ENABLE_CENSUS
 *   max_concurrent_streams = GRPC_ARG_MAX_CONCURRENT_STREAMS
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_ChannelArgKey, __pyx_n_s_ChannelArgKey, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":40
 * 
 * class ChannelArgKey:
 *   enable_census = GRPC_ARG_ENABLE_CENSUS             # <<<<<<<<<<<<<<
 *   max_concurrent_streams = GRPC_ARG_MAX_CONCURRENT_STREAMS
 *   max_message_length = GRPC_ARG_MAX_MESSAGE_LENGTH
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(GRPC_ARG_ENABLE_CENSUS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_enable_census, __pyx_t_2) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":41
 * class ChannelArgKey:
 *   enable_census = GRPC_ARG_ENABLE_CENSUS
 *   max_concurrent_streams = GRPC_ARG_MAX_CONCURRENT_STREAMS             # <<<<<<<<<<<<<<
 *   max_message_length = GRPC_ARG_MAX_MESSAGE_LENGTH
 *   http2_initial_sequence_number = GRPC_ARG_HTTP2_INITIAL_SEQUENCE_NUMBER
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(GRPC_ARG_MAX_CONCURRENT_STREAMS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_max_concurrent_streams, __pyx_t_2) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":42
 *   enable_census = GRPC_ARG_ENABLE_CENSUS
 *   max_concurrent_streams = GRPC_ARG_MAX_CONCURRENT_STREAMS
 *   max_message_length = GRPC_ARG_MAX_MESSAGE_LENGTH             # <<<<<<<<<<<<<<
 *   http2_initial_sequence_number = GRPC_ARG_HTTP2_INITIAL_SEQUENCE_NUMBER
 *   default_authority = GRPC_ARG_DEFAULT_AUTHORITY
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(GRPC_ARG_MAX_MESSAGE_LENGTH); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_max_message_length, __pyx_t_2) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":43
 *   max_concurrent_streams = GRPC_ARG_MAX_CONCURRENT_STREAMS
 *   max_message_length = GRPC_ARG_MAX_MESSAGE_LENGTH
 *   http2_initial_sequence_number = GRPC_ARG_HTTP2_INITIAL_SEQUENCE_NUMBER             # <<<<<<<<<<<<<<
 *   default_authority = GRPC_ARG_DEFAULT_AUTHORITY
 *   primary_user_agent_string = GRPC_ARG_PRIMARY_USER_AGENT_STRING
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(GRPC_ARG_HTTP2_INITIAL_SEQUENCE_NUMBER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_http2_initial_sequence_number, __pyx_t_2) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":44
 *   max_message_length = GRPC_ARG_MAX_MESSAGE_LENGTH
 *   http2_initial_sequence_number = GRPC_ARG_HTTP2_INITIAL_SEQUENCE_NUMBER
 *   default_authority = GRPC_ARG_DEFAULT_AUTHORITY             # <<<<<<<<<<<<<<
 *   primary_user_agent_string = GRPC_ARG_PRIMARY_USER_AGENT_STRING
 *   secondary_user_agent_string = GRPC_ARG_SECONDARY_USER_AGENT_STRING
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(GRPC_ARG_DEFAULT_AUTHORITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_default_authority, __pyx_t_2) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":45
 *   http2_initial_sequence_number = GRPC_ARG_HTTP2_INITIAL_SEQUENCE_NUMBER
 *   default_authority = GRPC_ARG_DEFAULT_AUTHORITY
 *   primary_user_agent_string = GRPC_ARG_PRIMARY_USER_AGENT_STRING             # <<<<<<<<<<<<<<
 *   secondary_user_agent_string = GRPC_ARG_SECONDARY_USER_AGENT_STRING
 *   ssl_target_name_override = GRPC_SSL_TARGET_NAME_OVERRIDE_ARG
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(GRPC_ARG_PRIMARY_USER_AGENT_STRING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_primary_user_agent_string, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":46
 *   default_authority = GRPC_ARG_DEFAULT_AUTHORITY
 *   primary_user_agent_string = GRPC_ARG_PRIMARY_USER_AGENT_STRING
 *   secondary_user_agent_string = GRPC_ARG_SECONDARY_USER_AGENT_STRING             # <<<<<<<<<<<<<<
 *   ssl_target_name_override = GRPC_SSL_TARGET_NAME_OVERRIDE_ARG
 * 
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(GRPC_ARG_SECONDARY_USER_AGENT_STRING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_secondary_user_agent_string, __pyx_t_2) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":47
 *   primary_user_agent_string = GRPC_ARG_PRIMARY_USER_AGENT_STRING
 *   secondary_user_agent_string = GRPC_ARG_SECONDARY_USER_AGENT_STRING
 *   ssl_target_name_override = GRPC_SSL_TARGET_NAME_OVERRIDE_ARG             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(GRPC_SSL_TARGET_NAME_OVERRIDE_ARG); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_ssl_target_name_override, __pyx_t_2) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":39
 * 
 * 
 * class ChannelArgKey:             # <<<<<<<<<<<<<<
 *   enable_census = GRPC_ARG_ENABLE_CENSUS
 *   max_concurrent_streams = GRPC_ARG_MAX_CONCURRENT_STREAMS
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_ChannelArgKey, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ChannelArgKey, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":50
 * 
 * 
 * class WriteFlag:             # <<<<<<<<<<<<<<
 *   buffer_hint = GRPC_WRITE_BUFFER_HINT
 *   no_compress = GRPC_WRITE_NO_COMPRESS
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_WriteFlag, __pyx_n_s_WriteFlag, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":51
 * 
 * class WriteFlag:
 *   buffer_hint = GRPC_WRITE_BUFFER_HINT             # <<<<<<<<<<<<<<
 *   no_compress = GRPC_WRITE_NO_COMPRESS
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(GRPC_WRITE_BUFFER_HINT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_buffer_hint, __pyx_t_2) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":52
 * class WriteFlag:
 *   buffer_hint = GRPC_WRITE_BUFFER_HINT
 *   no_compress = GRPC_WRITE_NO_COMPRESS             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(GRPC_WRITE_NO_COMPRESS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_no_compress, __pyx_t_2) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":50
 * 
 * 
 * class WriteFlag:             # <<<<<<<<<<<<<<
 *   buffer_hint = GRPC_WRITE_BUFFER_HINT
 *   no_compress = GRPC_WRITE_NO_COMPRESS
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_WriteFlag, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_WriteFlag, __pyx_t_2) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":55
 * 
 * 
 * class StatusCode:             # <<<<<<<<<<<<<<
 *   ok = GRPC_STATUS_OK
 *   cancelled = GRPC_STATUS_CANCELLED
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_StatusCode, __pyx_n_s_StatusCode, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":56
 * 
 * class StatusCode:
 *   ok = GRPC_STATUS_OK             # <<<<<<<<<<<<<<
 *   cancelled = GRPC_STATUS_CANCELLED
 *   unknown = GRPC_STATUS_UNKNOWN
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_OK); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_ok, __pyx_t_2) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":57
 * class StatusCode:
 *   ok = GRPC_STATUS_OK
 *   cancelled = GRPC_STATUS_CANCELLED             # <<<<<<<<<<<<<<
 *   unknown = GRPC_STATUS_UNKNOWN
 *   invalid_argument = GRPC_STATUS_INVALID_ARGUMENT
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_CANCELLED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_cancelled, __pyx_t_2) < 0) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":58
 *   ok = GRPC_STATUS_OK
 *   cancelled = GRPC_STATUS_CANCELLED
 *   unknown = GRPC_STATUS_UNKNOWN             # <<<<<<<<<<<<<<
 *   invalid_argument = GRPC_STATUS_INVALID_ARGUMENT
 *   deadline_exceeded = GRPC_STATUS_DEADLINE_EXCEEDED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_UNKNOWN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_unknown, __pyx_t_2) < 0) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":59
 *   cancelled = GRPC_STATUS_CANCELLED
 *   unknown = GRPC_STATUS_UNKNOWN
 *   invalid_argument = GRPC_STATUS_INVALID_ARGUMENT             # <<<<<<<<<<<<<<
 *   deadline_exceeded = GRPC_STATUS_DEADLINE_EXCEEDED
 *   not_found = GRPC_STATUS_NOT_FOUND
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_INVALID_ARGUMENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_invalid_argument, __pyx_t_2) < 0) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":60
 *   unknown = GRPC_STATUS_UNKNOWN
 *   invalid_argument = GRPC_STATUS_INVALID_ARGUMENT
 *   deadline_exceeded = GRPC_STATUS_DEADLINE_EXCEEDED             # <<<<<<<<<<<<<<
 *   not_found = GRPC_STATUS_NOT_FOUND
 *   already_exists = GRPC_STATUS_ALREADY_EXISTS
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_DEADLINE_EXCEEDED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_deadline_exceeded, __pyx_t_2) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":61
 *   invalid_argument = GRPC_STATUS_INVALID_ARGUMENT
 *   deadline_exceeded = GRPC_STATUS_DEADLINE_EXCEEDED
 *   not_found = GRPC_STATUS_NOT_FOUND             # <<<<<<<<<<<<<<
 *   already_exists = GRPC_STATUS_ALREADY_EXISTS
 *   permission_denied = GRPC_STATUS_PERMISSION_DENIED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_NOT_FOUND); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_not_found, __pyx_t_2) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":62
 *   deadline_exceeded = GRPC_STATUS_DEADLINE_EXCEEDED
 *   not_found = GRPC_STATUS_NOT_FOUND
 *   already_exists = GRPC_STATUS_ALREADY_EXISTS             # <<<<<<<<<<<<<<
 *   permission_denied = GRPC_STATUS_PERMISSION_DENIED
 *   unauthenticated = GRPC_STATUS_UNAUTHENTICATED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_ALREADY_EXISTS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_already_exists, __pyx_t_2) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":63
 *   not_found = GRPC_STATUS_NOT_FOUND
 *   already_exists = GRPC_STATUS_ALREADY_EXISTS
 *   permission_denied = GRPC_STATUS_PERMISSION_DENIED             # <<<<<<<<<<<<<<
 *   unauthenticated = GRPC_STATUS_UNAUTHENTICATED
 *   resource_exhausted = GRPC_STATUS_RESOURCE_EXHAUSTED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_PERMISSION_DENIED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_permission_denied, __pyx_t_2) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":64
 *   already_exists = GRPC_STATUS_ALREADY_EXISTS
 *   permission_denied = GRPC_STATUS_PERMISSION_DENIED
 *   unauthenticated = GRPC_STATUS_UNAUTHENTICATED             # <<<<<<<<<<<<<<
 *   resource_exhausted = GRPC_STATUS_RESOURCE_EXHAUSTED
 *   failed_precondition = GRPC_STATUS_FAILED_PRECONDITION
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_UNAUTHENTICATED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_unauthenticated, __pyx_t_2) < 0) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":65
 *   permission_denied = GRPC_STATUS_PERMISSION_DENIED
 *   unauthenticated = GRPC_STATUS_UNAUTHENTICATED
 *   resource_exhausted = GRPC_STATUS_RESOURCE_EXHAUSTED             # <<<<<<<<<<<<<<
 *   failed_precondition = GRPC_STATUS_FAILED_PRECONDITION
 *   aborted = GRPC_STATUS_ABORTED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_RESOURCE_EXHAUSTED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_resource_exhausted, __pyx_t_2) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":66
 *   unauthenticated = GRPC_STATUS_UNAUTHENTICATED
 *   resource_exhausted = GRPC_STATUS_RESOURCE_EXHAUSTED
 *   failed_precondition = GRPC_STATUS_FAILED_PRECONDITION             # <<<<<<<<<<<<<<
 *   aborted = GRPC_STATUS_ABORTED
 *   out_of_range = GRPC_STATUS_OUT_OF_RANGE
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_FAILED_PRECONDITION); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_failed_precondition, __pyx_t_2) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":67
 *   resource_exhausted = GRPC_STATUS_RESOURCE_EXHAUSTED
 *   failed_precondition = GRPC_STATUS_FAILED_PRECONDITION
 *   aborted = GRPC_STATUS_ABORTED             # <<<<<<<<<<<<<<
 *   out_of_range = GRPC_STATUS_OUT_OF_RANGE
 *   unimplemented = GRPC_STATUS_UNIMPLEMENTED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_ABORTED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_aborted, __pyx_t_2) < 0) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":68
 *   failed_precondition = GRPC_STATUS_FAILED_PRECONDITION
 *   aborted = GRPC_STATUS_ABORTED
 *   out_of_range = GRPC_STATUS_OUT_OF_RANGE             # <<<<<<<<<<<<<<
 *   unimplemented = GRPC_STATUS_UNIMPLEMENTED
 *   internal = GRPC_STATUS_INTERNAL
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_OUT_OF_RANGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_out_of_range, __pyx_t_2) < 0) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":69
 *   aborted = GRPC_STATUS_ABORTED
 *   out_of_range = GRPC_STATUS_OUT_OF_RANGE
 *   unimplemented = GRPC_STATUS_UNIMPLEMENTED             # <<<<<<<<<<<<<<
 *   internal = GRPC_STATUS_INTERNAL
 *   unavailable = GRPC_STATUS_UNAVAILABLE
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_UNIMPLEMENTED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_unimplemented, __pyx_t_2) < 0) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":70
 *   out_of_range = GRPC_STATUS_OUT_OF_RANGE
 *   unimplemented = GRPC_STATUS_UNIMPLEMENTED
 *   internal = GRPC_STATUS_INTERNAL             # <<<<<<<<<<<<<<
 *   unavailable = GRPC_STATUS_UNAVAILABLE
 *   data_loss = GRPC_STATUS_DATA_LOSS
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_INTERNAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_internal, __pyx_t_2) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":71
 *   unimplemented = GRPC_STATUS_UNIMPLEMENTED
 *   internal = GRPC_STATUS_INTERNAL
 *   unavailable = GRPC_STATUS_UNAVAILABLE             # <<<<<<<<<<<<<<
 *   data_loss = GRPC_STATUS_DATA_LOSS
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_UNAVAILABLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_unavailable, __pyx_t_2) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":72
 *   internal = GRPC_STATUS_INTERNAL
 *   unavailable = GRPC_STATUS_UNAVAILABLE
 *   data_loss = GRPC_STATUS_DATA_LOSS             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_status_code(GRPC_STATUS_DATA_LOSS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_data_loss, __pyx_t_2) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":55
 * 
 * 
 * class StatusCode:             # <<<<<<<<<<<<<<
 *   ok = GRPC_STATUS_OK
 *   cancelled = GRPC_STATUS_CANCELLED
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_StatusCode, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StatusCode, __pyx_t_2) < 0) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":75
 * 
 * 
 * class CallError:             # <<<<<<<<<<<<<<
 *   ok = GRPC_CALL_OK
 *   error = GRPC_CALL_ERROR
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_CallError, __pyx_n_s_CallError, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":76
 * 
 * class CallError:
 *   ok = GRPC_CALL_OK             # <<<<<<<<<<<<<<
 *   error = GRPC_CALL_ERROR
 *   not_on_server = GRPC_CALL_ERROR_NOT_ON_SERVER
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_OK); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_ok, __pyx_t_2) < 0) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":77
 * class CallError:
 *   ok = GRPC_CALL_OK
 *   error = GRPC_CALL_ERROR             # <<<<<<<<<<<<<<
 *   not_on_server = GRPC_CALL_ERROR_NOT_ON_SERVER
 *   not_on_client = GRPC_CALL_ERROR_NOT_ON_CLIENT
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_error, __pyx_t_2) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":78
 *   ok = GRPC_CALL_OK
 *   error = GRPC_CALL_ERROR
 *   not_on_server = GRPC_CALL_ERROR_NOT_ON_SERVER             # <<<<<<<<<<<<<<
 *   not_on_client = GRPC_CALL_ERROR_NOT_ON_CLIENT
 *   already_accepted = GRPC_CALL_ERROR_ALREADY_ACCEPTED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_NOT_ON_SERVER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_not_on_server, __pyx_t_2) < 0) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":79
 *   error = GRPC_CALL_ERROR
 *   not_on_server = GRPC_CALL_ERROR_NOT_ON_SERVER
 *   not_on_client = GRPC_CALL_ERROR_NOT_ON_CLIENT             # <<<<<<<<<<<<<<
 *   already_accepted = GRPC_CALL_ERROR_ALREADY_ACCEPTED
 *   already_invoked = GRPC_CALL_ERROR_ALREADY_INVOKED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_NOT_ON_CLIENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_not_on_client, __pyx_t_2) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":80
 *   not_on_server = GRPC_CALL_ERROR_NOT_ON_SERVER
 *   not_on_client = GRPC_CALL_ERROR_NOT_ON_CLIENT
 *   already_accepted = GRPC_CALL_ERROR_ALREADY_ACCEPTED             # <<<<<<<<<<<<<<
 *   already_invoked = GRPC_CALL_ERROR_ALREADY_INVOKED
 *   not_invoked = GRPC_CALL_ERROR_NOT_INVOKED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_ALREADY_ACCEPTED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_already_accepted, __pyx_t_2) < 0) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":81
 *   not_on_client = GRPC_CALL_ERROR_NOT_ON_CLIENT
 *   already_accepted = GRPC_CALL_ERROR_ALREADY_ACCEPTED
 *   already_invoked = GRPC_CALL_ERROR_ALREADY_INVOKED             # <<<<<<<<<<<<<<
 *   not_invoked = GRPC_CALL_ERROR_NOT_INVOKED
 *   already_finished = GRPC_CALL_ERROR_ALREADY_FINISHED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_ALREADY_INVOKED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_already_invoked, __pyx_t_2) < 0) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":82
 *   already_accepted = GRPC_CALL_ERROR_ALREADY_ACCEPTED
 *   already_invoked = GRPC_CALL_ERROR_ALREADY_INVOKED
 *   not_invoked = GRPC_CALL_ERROR_NOT_INVOKED             # <<<<<<<<<<<<<<
 *   already_finished = GRPC_CALL_ERROR_ALREADY_FINISHED
 *   too_many_operations = GRPC_CALL_ERROR_TOO_MANY_OPERATIONS
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_NOT_INVOKED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_not_invoked, __pyx_t_2) < 0) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":83
 *   already_invoked = GRPC_CALL_ERROR_ALREADY_INVOKED
 *   not_invoked = GRPC_CALL_ERROR_NOT_INVOKED
 *   already_finished = GRPC_CALL_ERROR_ALREADY_FINISHED             # <<<<<<<<<<<<<<
 *   too_many_operations = GRPC_CALL_ERROR_TOO_MANY_OPERATIONS
 *   invalid_flags = GRPC_CALL_ERROR_INVALID_FLAGS
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_ALREADY_FINISHED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_already_finished, __pyx_t_2) < 0) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":84
 *   not_invoked = GRPC_CALL_ERROR_NOT_INVOKED
 *   already_finished = GRPC_CALL_ERROR_ALREADY_FINISHED
 *   too_many_operations = GRPC_CALL_ERROR_TOO_MANY_OPERATIONS             # <<<<<<<<<<<<<<
 *   invalid_flags = GRPC_CALL_ERROR_INVALID_FLAGS
 *   invalid_metadata = GRPC_CALL_ERROR_INVALID_METADATA
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_TOO_MANY_OPERATIONS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_too_many_operations, __pyx_t_2) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":85
 *   already_finished = GRPC_CALL_ERROR_ALREADY_FINISHED
 *   too_many_operations = GRPC_CALL_ERROR_TOO_MANY_OPERATIONS
 *   invalid_flags = GRPC_CALL_ERROR_INVALID_FLAGS             # <<<<<<<<<<<<<<
 *   invalid_metadata = GRPC_CALL_ERROR_INVALID_METADATA
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_INVALID_FLAGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_invalid_flags, __pyx_t_2) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":86
 *   too_many_operations = GRPC_CALL_ERROR_TOO_MANY_OPERATIONS
 *   invalid_flags = GRPC_CALL_ERROR_INVALID_FLAGS
 *   invalid_metadata = GRPC_CALL_ERROR_INVALID_METADATA             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_call_error(GRPC_CALL_ERROR_INVALID_METADATA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_invalid_metadata, __pyx_t_2) < 0) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":75
 * 
 * 
 * class CallError:             # <<<<<<<<<<<<<<
 *   ok = GRPC_CALL_OK
 *   error = GRPC_CALL_ERROR
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_CallError, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CallError, __pyx_t_2) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":89
 * 
 * 
 * class CompletionType:             # <<<<<<<<<<<<<<
 *   queue_shutdown = GRPC_QUEUE_SHUTDOWN
 *   queue_timeout = GRPC_QUEUE_TIMEOUT
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_CompletionType, __pyx_n_s_CompletionType, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":90
 * 
 * class CompletionType:
 *   queue_shutdown = GRPC_QUEUE_SHUTDOWN             # <<<<<<<<<<<<<<
 *   queue_timeout = GRPC_QUEUE_TIMEOUT
 *   operation_complete = GRPC_OP_COMPLETE
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_completion_type(GRPC_QUEUE_SHUTDOWN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_queue_shutdown, __pyx_t_2) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":91
 * class CompletionType:
 *   queue_shutdown = GRPC_QUEUE_SHUTDOWN
 *   queue_timeout = GRPC_QUEUE_TIMEOUT             # <<<<<<<<<<<<<<
 *   operation_complete = GRPC_OP_COMPLETE
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_completion_type(GRPC_QUEUE_TIMEOUT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_queue_timeout, __pyx_t_2) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":92
 *   queue_shutdown = GRPC_QUEUE_SHUTDOWN
 *   queue_timeout = GRPC_QUEUE_TIMEOUT
 *   operation_complete = GRPC_OP_COMPLETE             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_completion_type(GRPC_OP_COMPLETE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_operation_complete, __pyx_t_2) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":89
 * 
 * 
 * class CompletionType:             # <<<<<<<<<<<<<<
 *   queue_shutdown = GRPC_QUEUE_SHUTDOWN
 *   queue_timeout = GRPC_QUEUE_TIMEOUT
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_CompletionType, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CompletionType, __pyx_t_2) < 0) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":95
 * 
 * 
 * class OperationType:             # <<<<<<<<<<<<<<
 *   send_initial_metadata = GRPC_OP_SEND_INITIAL_METADATA
 *   send_message = GRPC_OP_SEND_MESSAGE
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_OperationType, __pyx_n_s_OperationType, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":96
 * 
 * class OperationType:
 *   send_initial_metadata = GRPC_OP_SEND_INITIAL_METADATA             # <<<<<<<<<<<<<<
 *   send_message = GRPC_OP_SEND_MESSAGE
 *   send_close_from_client = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_op_type(GRPC_OP_SEND_INITIAL_METADATA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_send_initial_metadata, __pyx_t_2) < 0) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":97
 * class OperationType:
 *   send_initial_metadata = GRPC_OP_SEND_INITIAL_METADATA
 *   send_message = GRPC_OP_SEND_MESSAGE             # <<<<<<<<<<<<<<
 *   send_close_from_client = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 *   send_status_from_server = GRPC_OP_SEND_STATUS_FROM_SERVER
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_op_type(GRPC_OP_SEND_MESSAGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_send_message, __pyx_t_2) < 0) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":98
 *   send_initial_metadata = GRPC_OP_SEND_INITIAL_METADATA
 *   send_message = GRPC_OP_SEND_MESSAGE
 *   send_close_from_client = GRPC_OP_SEND_CLOSE_FROM_CLIENT             # <<<<<<<<<<<<<<
 *   send_status_from_server = GRPC_OP_SEND_STATUS_FROM_SERVER
 *   receive_initial_metadata = GRPC_OP_RECV_INITIAL_METADATA
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_op_type(GRPC_OP_SEND_CLOSE_FROM_CLIENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_send_close_from_client, __pyx_t_2) < 0) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":99
 *   send_message = GRPC_OP_SEND_MESSAGE
 *   send_close_from_client = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 *   send_status_from_server = GRPC_OP_SEND_STATUS_FROM_SERVER             # <<<<<<<<<<<<<<
 *   receive_initial_metadata = GRPC_OP_RECV_INITIAL_METADATA
 *   receive_message = GRPC_OP_RECV_MESSAGE
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_op_type(GRPC_OP_SEND_STATUS_FROM_SERVER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_send_status_from_server, __pyx_t_2) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":100
 *   send_close_from_client = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 *   send_status_from_server = GRPC_OP_SEND_STATUS_FROM_SERVER
 *   receive_initial_metadata = GRPC_OP_RECV_INITIAL_METADATA             # <<<<<<<<<<<<<<
 *   receive_message = GRPC_OP_RECV_MESSAGE
 *   receive_status_on_client = GRPC_OP_RECV_STATUS_ON_CLIENT
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_op_type(GRPC_OP_RECV_INITIAL_METADATA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_receive_initial_metadata, __pyx_t_2) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":101
 *   send_status_from_server = GRPC_OP_SEND_STATUS_FROM_SERVER
 *   receive_initial_metadata = GRPC_OP_RECV_INITIAL_METADATA
 *   receive_message = GRPC_OP_RECV_MESSAGE             # <<<<<<<<<<<<<<
 *   receive_status_on_client = GRPC_OP_RECV_STATUS_ON_CLIENT
 *   receive_close_on_server = GRPC_OP_RECV_CLOSE_ON_SERVER
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_op_type(GRPC_OP_RECV_MESSAGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_receive_message, __pyx_t_2) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":102
 *   receive_initial_metadata = GRPC_OP_RECV_INITIAL_METADATA
 *   receive_message = GRPC_OP_RECV_MESSAGE
 *   receive_status_on_client = GRPC_OP_RECV_STATUS_ON_CLIENT             # <<<<<<<<<<<<<<
 *   receive_close_on_server = GRPC_OP_RECV_CLOSE_ON_SERVER
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_op_type(GRPC_OP_RECV_STATUS_ON_CLIENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_receive_status_on_client, __pyx_t_2) < 0) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":103
 *   receive_message = GRPC_OP_RECV_MESSAGE
 *   receive_status_on_client = GRPC_OP_RECV_STATUS_ON_CLIENT
 *   receive_close_on_server = GRPC_OP_RECV_CLOSE_ON_SERVER             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_op_type(GRPC_OP_RECV_CLOSE_ON_SERVER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_receive_close_on_server, __pyx_t_2) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":95
 * 
 * 
 * class OperationType:             # <<<<<<<<<<<<<<
 *   send_initial_metadata = GRPC_OP_SEND_INITIAL_METADATA
 *   send_message = GRPC_OP_SEND_MESSAGE
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_OperationType, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OperationType, __pyx_t_2) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":106
 * 
 * 
 * class CompressionAlgorithm:             # <<<<<<<<<<<<<<
 *   none = GRPC_COMPRESS_NONE
 *   deflate = GRPC_COMPRESS_DEFLATE
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_CompressionAlgorithm, __pyx_n_s_CompressionAlgorithm, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":107
 * 
 * class CompressionAlgorithm:
 *   none = GRPC_COMPRESS_NONE             # <<<<<<<<<<<<<<
 *   deflate = GRPC_COMPRESS_DEFLATE
 *   gzip = GRPC_COMPRESS_GZIP
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_compression_algorithm(GRPC_COMPRESS_NONE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_none, __pyx_t_2) < 0) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":108
 * class CompressionAlgorithm:
 *   none = GRPC_COMPRESS_NONE
 *   deflate = GRPC_COMPRESS_DEFLATE             # <<<<<<<<<<<<<<
 *   gzip = GRPC_COMPRESS_GZIP
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_compression_algorithm(GRPC_COMPRESS_DEFLATE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_deflate, __pyx_t_2) < 0) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":109
 *   none = GRPC_COMPRESS_NONE
 *   deflate = GRPC_COMPRESS_DEFLATE
 *   gzip = GRPC_COMPRESS_GZIP             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_compression_algorithm(GRPC_COMPRESS_GZIP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_gzip, __pyx_t_2) < 0) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":106
 * 
 * 
 * class CompressionAlgorithm:             # <<<<<<<<<<<<<<
 *   none = GRPC_COMPRESS_NONE
 *   deflate = GRPC_COMPRESS_DEFLATE
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_CompressionAlgorithm, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CompressionAlgorithm, __pyx_t_2) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":112
 * 
 * 
 * class CompressionLevel:             # <<<<<<<<<<<<<<
 *   none = GRPC_COMPRESS_LEVEL_NONE
 *   low = GRPC_COMPRESS_LEVEL_LOW
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_CompressionLevel, __pyx_n_s_CompressionLevel, (PyObject *) NULL, __pyx_n_s_grpc__cython_cygrpc, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":113
 * 
 * class CompressionLevel:
 *   none = GRPC_COMPRESS_LEVEL_NONE             # <<<<<<<<<<<<<<
 *   low = GRPC_COMPRESS_LEVEL_LOW
 *   medium = GRPC_COMPRESS_LEVEL_MED
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_compression_level(GRPC_COMPRESS_LEVEL_NONE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_none, __pyx_t_2) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":114
 * class CompressionLevel:
 *   none = GRPC_COMPRESS_LEVEL_NONE
 *   low = GRPC_COMPRESS_LEVEL_LOW             # <<<<<<<<<<<<<<
 *   medium = GRPC_COMPRESS_LEVEL_MED
 *   high = GRPC_COMPRESS_LEVEL_HIGH
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_compression_level(GRPC_COMPRESS_LEVEL_LOW); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_low, __pyx_t_2) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":115
 *   none = GRPC_COMPRESS_LEVEL_NONE
 *   low = GRPC_COMPRESS_LEVEL_LOW
 *   medium = GRPC_COMPRESS_LEVEL_MED             # <<<<<<<<<<<<<<
 *   high = GRPC_COMPRESS_LEVEL_HIGH
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_compression_level(GRPC_COMPRESS_LEVEL_MED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_medium, __pyx_t_2) < 0) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":116
 *   low = GRPC_COMPRESS_LEVEL_LOW
 *   medium = GRPC_COMPRESS_LEVEL_MED
 *   high = GRPC_COMPRESS_LEVEL_HIGH             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_grpc_compression_level(GRPC_COMPRESS_LEVEL_HIGH); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_1, __pyx_n_s_high, __pyx_t_2) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":112
 * 
 * 
 * class CompressionLevel:             # <<<<<<<<<<<<<<
 *   none = GRPC_COMPRESS_LEVEL_NONE
 *   low = GRPC_COMPRESS_LEVEL_LOW
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_CompressionLevel, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CompressionLevel, __pyx_t_2) < 0) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":168
 * 
 *   @staticmethod
 *   def infinite_future():             # <<<<<<<<<<<<<<
 *     return Timespec(float("+inf"))
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_8Timespec_5infinite_future, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":167
 *     return <double>real_time.seconds + <double>real_time.nanoseconds / 1e9
 * 
 *   @staticmethod             # <<<<<<<<<<<<<<
 *   def infinite_future():
 *     return Timespec(float("+inf"))
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec->tp_dict, __pyx_n_s_infinite_future, __pyx_t_1) < 0) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":168
 * 
 *   @staticmethod
 *   def infinite_future():             # <<<<<<<<<<<<<<
 *     return Timespec(float("+inf"))
 * 
 */
  __pyx_t_1 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec, __pyx_n_s_infinite_future); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":167
 *     return <double>real_time.seconds + <double>real_time.nanoseconds / 1e9
 * 
 *   @staticmethod             # <<<<<<<<<<<<<<
 *   def infinite_future():
 *     return Timespec(float("+inf"))
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec->tp_dict, __pyx_n_s_infinite_future, __pyx_t_1) < 0) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":172
 * 
 *   @staticmethod
 *   def infinite_past():             # <<<<<<<<<<<<<<
 *     return Timespec(float("-inf"))
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_8Timespec_7infinite_past, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":171
 *     return Timespec(float("+inf"))
 * 
 *   @staticmethod             # <<<<<<<<<<<<<<
 *   def infinite_past():
 *     return Timespec(float("-inf"))
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec->tp_dict, __pyx_n_s_infinite_past, __pyx_t_1) < 0) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":172
 * 
 *   @staticmethod
 *   def infinite_past():             # <<<<<<<<<<<<<<
 *     return Timespec(float("-inf"))
 * 
 */
  __pyx_t_1 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec, __pyx_n_s_infinite_past); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":171
 *     return Timespec(float("+inf"))
 * 
 *   @staticmethod             # <<<<<<<<<<<<<<
 *   def infinite_past():
 *     return Timespec(float("-inf"))
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec->tp_dict, __pyx_n_s_infinite_past, __pyx_t_1) < 0) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_4grpc_7_cython_6cygrpc_Timespec);

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":543
 *     grpc_shutdown()
 * 
 * def operation_send_initial_metadata(Metadata metadata, int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_INITIAL_METADATA
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_21operation_send_initial_metadata, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operation_send_initial_metadata, __pyx_t_1) < 0) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":554
 *   return op
 * 
 * def operation_send_message(data, int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_MESSAGE
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_23operation_send_message, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operation_send_message, __pyx_t_1) < 0) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":564
 *   return op
 * 
 * def operation_send_close_from_client(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_SEND_CLOSE_FROM_CLIENT
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_25operation_send_close_from_client, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operation_send_close_from_client, __pyx_t_1) < 0) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":571
 *   return op
 * 
 * def operation_send_status_from_server(             # <<<<<<<<<<<<<<
 *     Metadata metadata, grpc_status_code code, bytes details, int flags):
 *   cdef Operation op = Operation()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_27operation_send_status_from_server, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operation_send_status_from_serve, __pyx_t_1) < 0) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":587
 *   return op
 * 
 * def operation_receive_initial_metadata(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_INITIAL_METADATA
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_29operation_receive_initial_metadata, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operation_receive_initial_metada, __pyx_t_1) < 0) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":597
 *   return op
 * 
 * def operation_receive_message(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_MESSAGE
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_31operation_receive_message, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operation_receive_message, __pyx_t_1) < 0) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":609
 *   return op
 * 
 * def operation_receive_status_on_client(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_STATUS_ON_CLIENT
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_33operation_receive_status_on_client, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operation_receive_status_on_clie, __pyx_t_1) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":625
 *   return op
 * 
 * def operation_receive_close_on_server(int flags):             # <<<<<<<<<<<<<<
 *   cdef Operation op = Operation()
 *   op.c_op.type = GRPC_OP_RECV_CLOSE_ON_SERVER
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_35operation_receive_close_on_server, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operation_receive_close_on_serve, __pyx_t_1) < 0) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/records.pyx.pxi":713
 * 
 * 
 * def compression_algorithm_name(grpc_compression_algorithm algorithm):             # <<<<<<<<<<<<<<
 *   cdef char* name
 *   with nogil:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_37compression_algorithm_name, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compression_algorithm_name, __pyx_t_1) < 0) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/security.pyx.pxi":32
 * from libc.string cimport memcpy
 * 
 * import pkg_resources             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pkg_resources, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pkg_resources, __pyx_t_1) < 0) __PYX_ERR(8, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/python/grpcio/grpc/_cython/_cygrpc/server.pyx.pxi":32
 * cimport cpython
 * 
 * import time             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_time, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_1) < 0) __PYX_ERR(4, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "grpc/_cython/cygrpc.pyx":57
 * 
 * 
 * def _initialize():             # <<<<<<<<<<<<<<
 *   grpc_set_ssl_roots_override_callback(
 *           <grpc_ssl_roots_override_callback>ssl_roots_override_callback)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4grpc_7_cython_6cygrpc_39_initialize, NULL, __pyx_n_s_grpc__cython_cygrpc); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_initialize, __pyx_t_1) < 0) __PYX_ERR(9, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "grpc/_cython/cygrpc.pyx":62
 * 
 * 
 * _initialize()             # <<<<<<<<<<<<<<
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_initialize); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 62, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "grpc/_cython/cygrpc.pyx":1
 * # Copyright 2015, Google Inc.             # <<<<<<<<<<<<<<
 * # All rights reserved.
 * #
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(9, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init grpc._cython.cygrpc", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init grpc._cython.cygrpc");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseArgTupleInvalid */
    static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
    static CYTHON_INLINE int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_CheckExact(key)) && unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseDoubleKeywords */
    static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
    static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
    static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* WriteUnraisableException */
    static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* FetchCommonType */
    static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
    static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_COMPILING_IN_CPYTHON
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
#if CYTHON_COMPILING_IN_PYPY
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    PyObject *self = f->m_self;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
#else
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
	return PyCFunction_Call(func, arg, kw);
}
#endif
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_Call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
#if !CYTHON_COMPILING_IN_PYPY
    __pyx_CyFunctionType_type.tp_call = PyCFunction_Call;
#endif
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* GetItemInt */
        static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallNoArg */
        #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectCallMethod1 */
          static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto bad;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            args = PyTuple_New(2);
            if (unlikely(!args)) goto bad;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            Py_DECREF(method); method = NULL;
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
bad:
    Py_XDECREF(method);
    return result;
}

/* append */
          static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* GetModuleGlobalName */
          static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* SetVTable */
            static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* Import */
            static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* CalculateMetaclass */
            static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
            static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* GetNameInClass */
            static PyObject *__Pyx_GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result)
        result = __Pyx_GetModuleGlobalName(name);
    return result;
}

/* CodeObjectCache */
            static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
            #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_completion_type(grpc_completion_type value) {
    const grpc_completion_type neg_one = (grpc_completion_type) -1, const_zero = (grpc_completion_type) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(grpc_completion_type) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_completion_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(grpc_completion_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(grpc_completion_type) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_completion_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(grpc_completion_type),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
            #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_connectivity_state(grpc_connectivity_state value) {
    const grpc_connectivity_state neg_one = (grpc_connectivity_state) -1, const_zero = (grpc_connectivity_state) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(grpc_connectivity_state) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_connectivity_state) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(grpc_connectivity_state) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(grpc_connectivity_state) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_connectivity_state) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(grpc_connectivity_state),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_status_code(grpc_status_code value) {
    const grpc_status_code neg_one = (grpc_status_code) -1, const_zero = (grpc_status_code) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(grpc_status_code) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_status_code) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(grpc_status_code) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(grpc_status_code) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_status_code) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(grpc_status_code),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_call_error(grpc_call_error value) {
    const grpc_call_error neg_one = (grpc_call_error) -1, const_zero = (grpc_call_error) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(grpc_call_error) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_call_error) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(grpc_call_error) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(grpc_call_error) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_call_error) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(grpc_call_error),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_op_type(grpc_op_type value) {
    const grpc_op_type neg_one = (grpc_op_type) -1, const_zero = (grpc_op_type) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(grpc_op_type) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_op_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(grpc_op_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(grpc_op_type) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_op_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(grpc_op_type),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_compression_algorithm(grpc_compression_algorithm value) {
    const grpc_compression_algorithm neg_one = (grpc_compression_algorithm) -1, const_zero = (grpc_compression_algorithm) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(grpc_compression_algorithm) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_compression_algorithm) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(grpc_compression_algorithm) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(grpc_compression_algorithm) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_compression_algorithm) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(grpc_compression_algorithm),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_grpc_compression_level(grpc_compression_level value) {
    const grpc_compression_level neg_one = (grpc_compression_level) -1, const_zero = (grpc_compression_level) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(grpc_compression_level) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_compression_level) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(grpc_compression_level) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(grpc_compression_level) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(grpc_compression_level) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(grpc_compression_level),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
            static CYTHON_INLINE grpc_status_code __Pyx_PyInt_As_grpc_status_code(PyObject *x) {
    const grpc_status_code neg_one = (grpc_status_code) -1, const_zero = (grpc_status_code) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(grpc_status_code) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(grpc_status_code, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (grpc_status_code) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (grpc_status_code) 0;
                case  1: __PYX_VERIFY_RETURN_INT(grpc_status_code, digit, digits[0])
                case 2:
                    if (8 * sizeof(grpc_status_code) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) >= 2 * PyLong_SHIFT) {
                            return (grpc_status_code) (((((grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(grpc_status_code) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) >= 3 * PyLong_SHIFT) {
                            return (grpc_status_code) (((((((grpc_status_code)digits[2]) << PyLong_SHIFT) | (grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(grpc_status_code) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) >= 4 * PyLong_SHIFT) {
                            return (grpc_status_code) (((((((((grpc_status_code)digits[3]) << PyLong_SHIFT) | (grpc_status_code)digits[2]) << PyLong_SHIFT) | (grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (grpc_status_code) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(grpc_status_code) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_status_code, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(grpc_status_code) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_status_code, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (grpc_status_code) 0;
                case -1: __PYX_VERIFY_RETURN_INT(grpc_status_code, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(grpc_status_code,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(grpc_status_code) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) - 1 > 2 * PyLong_SHIFT) {
                            return (grpc_status_code) (((grpc_status_code)-1)*(((((grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(grpc_status_code) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) - 1 > 2 * PyLong_SHIFT) {
                            return (grpc_status_code) ((((((grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(grpc_status_code) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) - 1 > 3 * PyLong_SHIFT) {
                            return (grpc_status_code) (((grpc_status_code)-1)*(((((((grpc_status_code)digits[2]) << PyLong_SHIFT) | (grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(grpc_status_code) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) - 1 > 3 * PyLong_SHIFT) {
                            return (grpc_status_code) ((((((((grpc_status_code)digits[2]) << PyLong_SHIFT) | (grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(grpc_status_code) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) - 1 > 4 * PyLong_SHIFT) {
                            return (grpc_status_code) (((grpc_status_code)-1)*(((((((((grpc_status_code)digits[3]) << PyLong_SHIFT) | (grpc_status_code)digits[2]) << PyLong_SHIFT) | (grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(grpc_status_code) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_status_code, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_status_code) - 1 > 4 * PyLong_SHIFT) {
                            return (grpc_status_code) ((((((((((grpc_status_code)digits[3]) << PyLong_SHIFT) | (grpc_status_code)digits[2]) << PyLong_SHIFT) | (grpc_status_code)digits[1]) << PyLong_SHIFT) | (grpc_status_code)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(grpc_status_code) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_status_code, long, PyLong_AsLong(x))
            } else if (sizeof(grpc_status_code) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_status_code, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            grpc_status_code val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (grpc_status_code) -1;
        }
    } else {
        grpc_status_code val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (grpc_status_code) -1;
        val = __Pyx_PyInt_As_grpc_status_code(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to grpc_status_code");
    return (grpc_status_code) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to grpc_status_code");
    return (grpc_status_code) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE grpc_connectivity_state __Pyx_PyInt_As_grpc_connectivity_state(PyObject *x) {
    const grpc_connectivity_state neg_one = (grpc_connectivity_state) -1, const_zero = (grpc_connectivity_state) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(grpc_connectivity_state) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (grpc_connectivity_state) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (grpc_connectivity_state) 0;
                case  1: __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, digit, digits[0])
                case 2:
                    if (8 * sizeof(grpc_connectivity_state) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) >= 2 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) (((((grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(grpc_connectivity_state) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) >= 3 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) (((((((grpc_connectivity_state)digits[2]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(grpc_connectivity_state) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) >= 4 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) (((((((((grpc_connectivity_state)digits[3]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[2]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (grpc_connectivity_state) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(grpc_connectivity_state) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_connectivity_state, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(grpc_connectivity_state) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_connectivity_state, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (grpc_connectivity_state) 0;
                case -1: __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(grpc_connectivity_state,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(grpc_connectivity_state) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) - 1 > 2 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) (((grpc_connectivity_state)-1)*(((((grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(grpc_connectivity_state) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) - 1 > 2 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) ((((((grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(grpc_connectivity_state) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) - 1 > 3 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) (((grpc_connectivity_state)-1)*(((((((grpc_connectivity_state)digits[2]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(grpc_connectivity_state) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) - 1 > 3 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) ((((((((grpc_connectivity_state)digits[2]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(grpc_connectivity_state) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) - 1 > 4 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) (((grpc_connectivity_state)-1)*(((((((((grpc_connectivity_state)digits[3]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[2]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(grpc_connectivity_state) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_connectivity_state, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_connectivity_state) - 1 > 4 * PyLong_SHIFT) {
                            return (grpc_connectivity_state) ((((((((((grpc_connectivity_state)digits[3]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[2]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[1]) << PyLong_SHIFT) | (grpc_connectivity_state)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(grpc_connectivity_state) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_connectivity_state, long, PyLong_AsLong(x))
            } else if (sizeof(grpc_connectivity_state) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_connectivity_state, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            grpc_connectivity_state val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (grpc_connectivity_state) -1;
        }
    } else {
        grpc_connectivity_state val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (grpc_connectivity_state) -1;
        val = __Pyx_PyInt_As_grpc_connectivity_state(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to grpc_connectivity_state");
    return (grpc_connectivity_state) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to grpc_connectivity_state");
    return (grpc_connectivity_state) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE grpc_completion_type __Pyx_PyInt_As_grpc_completion_type(PyObject *x) {
    const grpc_completion_type neg_one = (grpc_completion_type) -1, const_zero = (grpc_completion_type) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(grpc_completion_type) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(grpc_completion_type, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (grpc_completion_type) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (grpc_completion_type) 0;
                case  1: __PYX_VERIFY_RETURN_INT(grpc_completion_type, digit, digits[0])
                case 2:
                    if (8 * sizeof(grpc_completion_type) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) >= 2 * PyLong_SHIFT) {
                            return (grpc_completion_type) (((((grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(grpc_completion_type) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) >= 3 * PyLong_SHIFT) {
                            return (grpc_completion_type) (((((((grpc_completion_type)digits[2]) << PyLong_SHIFT) | (grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(grpc_completion_type) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) >= 4 * PyLong_SHIFT) {
                            return (grpc_completion_type) (((((((((grpc_completion_type)digits[3]) << PyLong_SHIFT) | (grpc_completion_type)digits[2]) << PyLong_SHIFT) | (grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (grpc_completion_type) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(grpc_completion_type) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_completion_type, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(grpc_completion_type) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_completion_type, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (grpc_completion_type) 0;
                case -1: __PYX_VERIFY_RETURN_INT(grpc_completion_type, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(grpc_completion_type,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(grpc_completion_type) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) - 1 > 2 * PyLong_SHIFT) {
                            return (grpc_completion_type) (((grpc_completion_type)-1)*(((((grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(grpc_completion_type) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) - 1 > 2 * PyLong_SHIFT) {
                            return (grpc_completion_type) ((((((grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(grpc_completion_type) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) - 1 > 3 * PyLong_SHIFT) {
                            return (grpc_completion_type) (((grpc_completion_type)-1)*(((((((grpc_completion_type)digits[2]) << PyLong_SHIFT) | (grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(grpc_completion_type) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) - 1 > 3 * PyLong_SHIFT) {
                            return (grpc_completion_type) ((((((((grpc_completion_type)digits[2]) << PyLong_SHIFT) | (grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(grpc_completion_type) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) - 1 > 4 * PyLong_SHIFT) {
                            return (grpc_completion_type) (((grpc_completion_type)-1)*(((((((((grpc_completion_type)digits[3]) << PyLong_SHIFT) | (grpc_completion_type)digits[2]) << PyLong_SHIFT) | (grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(grpc_completion_type) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_completion_type, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_completion_type) - 1 > 4 * PyLong_SHIFT) {
                            return (grpc_completion_type) ((((((((((grpc_completion_type)digits[3]) << PyLong_SHIFT) | (grpc_completion_type)digits[2]) << PyLong_SHIFT) | (grpc_completion_type)digits[1]) << PyLong_SHIFT) | (grpc_completion_type)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(grpc_completion_type) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_completion_type, long, PyLong_AsLong(x))
            } else if (sizeof(grpc_completion_type) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_completion_type, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            grpc_completion_type val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (grpc_completion_type) -1;
        }
    } else {
        grpc_completion_type val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (grpc_completion_type) -1;
        val = __Pyx_PyInt_As_grpc_completion_type(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to grpc_completion_type");
    return (grpc_completion_type) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to grpc_completion_type");
    return (grpc_completion_type) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE grpc_compression_algorithm __Pyx_PyInt_As_grpc_compression_algorithm(PyObject *x) {
    const grpc_compression_algorithm neg_one = (grpc_compression_algorithm) -1, const_zero = (grpc_compression_algorithm) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(grpc_compression_algorithm) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (grpc_compression_algorithm) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (grpc_compression_algorithm) 0;
                case  1: __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, digit, digits[0])
                case 2:
                    if (8 * sizeof(grpc_compression_algorithm) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) >= 2 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) (((((grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(grpc_compression_algorithm) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) >= 3 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) (((((((grpc_compression_algorithm)digits[2]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(grpc_compression_algorithm) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) >= 4 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) (((((((((grpc_compression_algorithm)digits[3]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[2]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (grpc_compression_algorithm) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(grpc_compression_algorithm) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_compression_algorithm, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(grpc_compression_algorithm) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_compression_algorithm, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (grpc_compression_algorithm) 0;
                case -1: __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(grpc_compression_algorithm) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) - 1 > 2 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) (((grpc_compression_algorithm)-1)*(((((grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(grpc_compression_algorithm) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) - 1 > 2 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) ((((((grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(grpc_compression_algorithm) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) - 1 > 3 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) (((grpc_compression_algorithm)-1)*(((((((grpc_compression_algorithm)digits[2]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(grpc_compression_algorithm) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) - 1 > 3 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) ((((((((grpc_compression_algorithm)digits[2]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(grpc_compression_algorithm) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) - 1 > 4 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) (((grpc_compression_algorithm)-1)*(((((((((grpc_compression_algorithm)digits[3]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[2]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(grpc_compression_algorithm) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(grpc_compression_algorithm, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(grpc_compression_algorithm) - 1 > 4 * PyLong_SHIFT) {
                            return (grpc_compression_algorithm) ((((((((((grpc_compression_algorithm)digits[3]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[2]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[1]) << PyLong_SHIFT) | (grpc_compression_algorithm)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(grpc_compression_algorithm) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_compression_algorithm, long, PyLong_AsLong(x))
            } else if (sizeof(grpc_compression_algorithm) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(grpc_compression_algorithm, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            grpc_compression_algorithm val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (grpc_compression_algorithm) -1;
        }
    } else {
        grpc_compression_algorithm val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (grpc_compression_algorithm) -1;
        val = __Pyx_PyInt_As_grpc_compression_algorithm(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to grpc_compression_algorithm");
    return (grpc_compression_algorithm) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to grpc_compression_algorithm");
    return (grpc_compression_algorithm) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
            static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
            static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
            #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
            #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* InitStrings */
            static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
